Package: merTools
Title: Tools for Analyzing Mixed Effect Regression Models
Version: 1.0.0
Authors@R: c(
    person(c("Jared", "E."), "Knowles", email = "jared@civilytics.com", role = c("aut", "cre")),
    person("Carl", "Frederick", email="carlbfrederick@gmail.com", role = c("aut")),
    person("Alex", "Whitworth", email="whitworth.alex@gmail.com", role = c("ctb")))
Description: Provides methods for extracting results from mixed-effect model
    objects fit with the 'lme4' package. Allows construction of prediction intervals
    efficiently from large scale linear and generalized linear mixed-effects models.
    This method draws from the simulation framework used in the Gelman and Hill (2007) textbook:
    Data Analysis Using Regression and Multilevel/Hierarchical Models.
Depends: R (>= 3.0.2), lme4 (>= 1.1-11), methods
Suggests: testthat, knitr, rmarkdown, shiny, nlme, future.apply,
        rstanarm, Amelia, DT
Imports: arm, dplyr, mvtnorm, reformulas, foreach, abind, ggplot2,
        blme, broom.mixed, Matrix
License: GPL (>= 2)
LazyData: true
VignetteBuilder: knitr
RoxygenNote: 7.3.3
Encoding: UTF-8
URL: https://jknowles.github.io/merTools/,
        https://github.com/jknowles/merTools
BugReports: https://github.com/jknowles/merTools/issues
Config/testthat/edition: 3
Config/testthat/parallel: false
NeedsCompilation: no
Packaged: 2026-05-30 20:25:08 UTC; jared
Author: Jared E. Knowles [aut, cre],
  Carl Frederick [aut],
  Alex Whitworth [ctb]
Maintainer: Jared E. Knowles <jared@civilytics.com>
Repository: CRAN
Date/Publication: 2026-05-31 05:11:49 UTC
Built: R 4.5.2; ; 2026-05-31 07:27:55 UTC; unix

Package: merTools
Title: Tools for Analyzing Mixed Effect Regression Models
Version: 1.0.0
Authors@R: c(
    person(c("Jared", "E."), "Knowles", email = "jared@civilytics.com", role = c("aut", "cre")),
    person("Carl", "Frederick", email="carlbfrederick@gmail.com", role = c("aut")),
    person("Alex", "Whitworth", email="whitworth.alex@gmail.com", role = c("ctb")))
Description: Provides methods for extracting results from mixed-effect model
    objects fit with the 'lme4' package. Allows construction of prediction intervals
    efficiently from large scale linear and generalized linear mixed-effects models.
    This method draws from the simulation framework used in the Gelman and Hill (2007) textbook:
    Data Analysis Using Regression and Multilevel/Hierarchical Models.
Depends: R (>= 3.0.2), lme4 (>= 1.1-11), methods
Suggests: testthat, knitr, rmarkdown, shiny, nlme, future.apply,
        rstanarm, Amelia, DT
Imports: arm, dplyr, mvtnorm, reformulas, foreach, abind, ggplot2,
        blme, broom.mixed, Matrix
License: GPL (>= 2)
LazyData: true
VignetteBuilder: knitr
RoxygenNote: 7.3.3
Encoding: UTF-8
URL: https://jknowles.github.io/merTools/,
        https://github.com/jknowles/merTools
BugReports: https://github.com/jknowles/merTools/issues
Config/testthat/edition: 3
Config/testthat/parallel: false
NeedsCompilation: no
Packaged: 2026-05-30 20:25:08 UTC; jared
Author: Jared E. Knowles [aut, cre],
  Carl Frederick [aut],
  Alex Whitworth [ctb]
Maintainer: Jared E. Knowles <jared@civilytics.com>
Repository: CRAN
Date/Publication: 2026-05-31 05:11:49 UTC
Built: R 4.6.0; ; 2026-05-31 08:25:06 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(VarCorr,merModList)
S3method(draw,merMod)
S3method(fastdisp,merMod)
S3method(fastdisp,merModList)
S3method(fixef,merModList)
S3method(print,merModList)
S3method(print,summary.merModList)
S3method(ranef,merModList)
S3method(summary,merModList)
export(FEsim)
export(ICC)
export(REcorrExtract)
export(REextract)
export(REimpact)
export(REmargins)
export(REquantile)
export(REsdExtract)
export(REsim)
export(RMSE.merMod)
export(averageObs)
export(bglmerModList)
export(blmerModList)
export(draw)
export(expectedRank)
export(fastdisp)
export(findFormFuns)
export(glmerModList)
export(lmerModList)
export(modelFixedEff)
export(modelInfo)
export(modelRandEffStats)
export(plotFEsim)
export(plotREimpact)
export(plotREsim)
export(predictInterval)
export(randomObs)
export(shinyMer)
export(subBoot)
export(superFactor)
export(thetaExtract)
export(wiggle)
import(dplyr)
import(ggplot2)
import(lme4)
importFrom(abind,abind)
importFrom(arm,fround)
importFrom(arm,pfround)
importFrom(arm,sigma.hat)
importFrom(arm,sim)
importFrom(blme,bglmer)
importFrom(blme,blmer)
importFrom(broom.mixed,tidy)
importFrom(foreach,"%dopar%")
importFrom(foreach,foreach)
importFrom(methods,as)
importFrom(methods,is)
importFrom(mvtnorm,rmvnorm)
importFrom(stats,AIC)
importFrom(stats,as.formula)
importFrom(stats,delete.response)
importFrom(stats,df)
importFrom(stats,formula)
importFrom(stats,getCall)
importFrom(stats,logLik)
importFrom(stats,median)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,na.omit)
importFrom(stats,na.pass)
importFrom(stats,pnorm)
importFrom(stats,pt)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,rbinom)
importFrom(stats,reformulate)
importFrom(stats,reshape)
importFrom(stats,residuals)
importFrom(stats,rgamma)
importFrom(stats,rnorm)
importFrom(stats,rpois)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,terms)
importFrom(stats,vcov)
importFrom(stats,weighted.mean)
importFrom(utils,packageVersion)

[![CRAN_Status_Badge](http://www.r-pkg.org/badges/version/merTools)](https://cran.r-project.org/package=merTools)
[![Downloads](http://cranlogs.r-pkg.org/badges/merTools)](https://cran.r-project.org/package=merTools)
[![Downloads](http://cranlogs.r-pkg.org/badges/grand-total/merTools)](https://cran.r-project.org/package=merTools)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# merTools

A package for getting the most out of large multilevel models in R

by Jared E. Knowles and Carl Frederick

Working with generalized linear mixed models (GLMM) and linear mixed
models (LMM) has become increasingly easy with advances in the `lme4`
package. As we have found ourselves using these models more and more
within our work, we, the authors, have developed a set of tools for
simplifying and speeding up common tasks for interacting with `merMod`
objects from `lme4`. This package provides those tools.

## Installation

``` r
# development version
library(devtools)
install_github("jknowles/merTools")

# CRAN version
install.packages("merTools")
```

## Recent Updates

### merTools 1.0.0 (May 2026): Long-Term Support release

- This release marks `merTools` as **feature complete**. The package now
  enters **maintenance mode**: future releases will focus on bug fixes,
  dependency and CRAN compatibility, and documentation rather than new
  features.
- It also resolves the last of the open issues: a correctness fix for
  `predictInterval()` on nested random effects (#124), a repaired and
  extended `shinyMer()` (#32, \#78), the new `plotREimpact()` plot (#84,
  \#85), and refreshed documentation (#116, \#136, \#137). See `NEWS.md`
  for the full list.

### merTools 0.6.4 (January 2026)

- Maintenance release to merge @DavisVaughan changes to accommodate
  upstream changes in `vctrs` package impacting `dplyr::bind_rows()`
  usage in `REsim` (#133)

### merTools 0.6.3 (September 2025)

- Maintenance release to fix crossreference issues with function
  documentation

### merTools 0.6.2 (Early 2024)

- Maintenance release to fix minor issues with function documentation
- Fix \#130 by avoiding conflict with `vcov` in the `merDeriv` package
- Upgrade package test infrastructure to 3e testthat specification

### merTools 0.6.1 (Spring 2023)

- Maintenance release to keep package listed on CRAN
- Fix a small bug where parallel code path is run twice (#126)
- Update plotting functions to avoid deprecated `aes_string()` calls
  (#127)
- Fix (#115) in description
- Speed up PI using @bbolker pull request (#120)
- Updated package maintainer contact information

### merTools 0.5.0

#### New Features

- `subBoot` now works with `glmerMod` objects as well
- `reMargins` a new function that allows the user to marginalize the
  prediction over breaks in the distribution of random effect
  distributions, see `?reMargins` and the new `reMargins` vignette
  (closes \#73)

#### Bug fixes

- Fixed an issue where known convergence errors were issuing warnings
  and causing the test suite to not work
- Fixed an issue where models with a random slope, no intercept, and no
  fixed term were unable to be predicted (#101)
- Fixed an issue with shinyMer not working with substantive fixed
  effects (#93)

### merTools 0.4.1

#### New Features

- Standard errors reported by `merModList` functions now apply the Rubin
  correction for multiple imputation

#### Bug fixes

- Contribution by Alex Whitworth (@alexWhitworth) adding error checking
  to plotting functions

## Shiny App and Demo

The easiest way to demo the features of this application is to use the
bundled Shiny application which launches a number of the metrics here to
aide in exploring the model. To do this:

    library(merTools)
    m1 <- lmer(y ~ service + lectage + studage + (1|d) + (1|s), data=InstEval)
    shinyMer(m1, simData = InstEval[1:100, ]) # just try the first 100 rows of data

![](man/figures/README-predPanel.png)

On the first tab, the function presents the prediction intervals for the
data selected by user which are calculated using the `predictInterval`
function within the package. This function calculates prediction
intervals quickly by sampling from the simulated distribution of the
fixed effect and random effect terms and combining these simulated
estimates to produce a distribution of predictions for each observation.
This allows prediction intervals to be generated from very large models
where the use of `bootMer` would not be feasible computationally.

![](man/figures/README-effPanel.png)

On the next tab the distribution of the fixed effect and group-level
effects is depicted on confidence interval plots. These are useful for
diagnostics and provide a way to inspect the relative magnitudes of
various parameters. This tab makes use of four related functions in
`merTools`: `FEsim`, `plotFEsim`, `REsim` and `plotREsim` which are
available to be used on their own as well.

![](man/figures/README-substPanel.png)

On the third tab are some convenient ways to show the influence or
magnitude of effects by leveraging the power of `predictInterval`. For
each case, up to 12, in the selected data type, the user can view the
impact of changing either one of the fixed effect or one of the grouping
level terms. Using the `REimpact` function, each case is simulated with
the model’s prediction if all else was held equal, but the observation
was moved through the distribution of the fixed effect or the random
effect term. This is plotted on the scale of the dependent variable,
which allows the user to compare the magnitude of effects across
variables, and also between models on the same data.

## Predicting

Standard prediction looks like so.

``` r
predict(m1, newdata = InstEval[1:10, ])
#>        1        2        3        4        5        6        7        8 
#> 3.146337 3.165212 3.398499 3.114249 3.320686 3.252670 4.180897 3.845219 
#>        9       10 
#> 3.779337 3.331013
```

With `predictInterval` we obtain predictions that are more like the
standard objects produced by `lm` and `glm`:

``` r
predictInterval(m1, newdata = InstEval[1:10, ], n.sims = 500, level = 0.9,
                stat = 'median')
#>         fit      upr      lwr
#> 1  3.161371 5.103826 1.139883
#> 2  3.232387 5.009761 1.332832
#> 3  3.445031 5.291794 1.499974
#> 4  3.013080 5.244352 1.104054
#> 5  3.291172 5.228658 1.150961
#> 6  3.311819 5.336952 1.060889
#> 7  4.165035 5.971790 2.084237
#> 8  3.830382 5.578582 1.791503
#> 9  3.801973 5.845741 1.924566
#> 10 3.358777 5.288322 1.581321
```

Note that `predictInterval` is slower because it is computing
simulations. It can also return all of the simulated `yhat` values as an
attribute to the predict object itself.

`predictInterval` uses the `sim` function from the `arm` package heavily
to draw the distributions of the parameters of the model. It then
combines these simulated values to create a distribution of the `yhat`
for each observation.

### Inspecting the Prediction Components

We can also explore the components of the prediction interval by asking
`predictInterval` to return specific components of the prediction
interval.

``` r
predictInterval(m1, newdata = InstEval[1:10, ], n.sims = 200, level = 0.9,
                stat = 'median', which = "all")
#>      effect          fit      upr        lwr obs
#> 1  combined  3.083593245 5.062599  0.9674464   1
#> 2  combined  3.235036697 5.212885  1.1708064   2
#> 3  combined  3.342847853 5.371818  1.2686879   3
#> 4  combined  3.077033326 5.251489  1.3107275   4
#> 5  combined  3.316029839 5.337122  1.4697923   5
#> 6  combined  3.443738022 5.172142  0.9068009   6
#> 7  combined  4.112843798 6.535431  2.1650584   7
#> 8  combined  3.816085776 5.974373  1.8064576   8
#> 9  combined  3.792304602 5.609345  1.8844468   9
#> 10 combined  3.333815423 5.402147  1.5005593  10
#> 11        s  0.231536444 1.790168 -1.7006576   1
#> 12        s  0.134265072 1.957335 -1.7931547   2
#> 13        s  0.218560304 2.021749 -1.7836963   3
#> 14        s -0.099702591 2.029545 -1.8110851   4
#> 15        s -0.045505671 1.693573 -2.1095589   5
#> 16        s -0.087337382 1.716194 -2.0527377   6
#> 17        s  0.376170634 2.109191 -1.8125491   7
#> 18        s  0.364157813 2.329252 -1.4930631   8
#> 19        s  0.415591035 2.240454 -1.4130724   9
#> 20        s  0.356965891 2.283956 -1.7016778  10
#> 21        d  0.008946287 1.700966 -2.0450912   1
#> 22        d -0.095226122 1.741526 -2.0126363   2
#> 23        d -0.171781743 2.068428 -2.2161950   3
#> 24        d -0.225090139 1.962778 -2.2287493   4
#> 25        d  0.242943045 2.185860 -1.5681800   5
#> 26        d -0.083381572 1.958019 -2.1527131   6
#> 27        d  0.505729236 2.272959 -1.0985370   7
#> 28        d  0.223848219 1.892653 -1.4133758   8
#> 29        d  0.232838207 2.161019 -1.6092655   9
#> 30        d -0.411865471 1.744540 -2.4933411  10
#> 31    fixed  3.102540070 5.098888  1.2745417   1
#> 32    fixed  3.071049868 5.045066  1.1160584   2
#> 33    fixed  3.325102518 5.242442  1.1231595   3
#> 34    fixed  3.185105756 4.766187  1.1267020   4
#> 35    fixed  3.392547857 4.902974  1.3989885   5
#> 36    fixed  3.213560141 4.949743  1.2420229   6
#> 37    fixed  3.197190288 5.109854  1.1668196   7
#> 38    fixed  3.305944109 5.132413  1.5010800   8
#> 39    fixed  3.395006368 5.327011  1.7248772   9
#> 40    fixed  3.415382668 5.052743  1.1706732  10
```

This can lead to some useful plotting:

``` r
library(ggplot2)
plotdf <- predictInterval(m1, newdata = InstEval[1:10, ], n.sims = 2000,
                          level = 0.9, stat = 'median', which = "all",
                          include.resid.var = FALSE)
plotdfb <- predictInterval(m1, newdata = InstEval[1:10, ], n.sims = 2000,
                          level = 0.9, stat = 'median', which = "all",
                          include.resid.var = TRUE)

plotdf <- dplyr::bind_rows(plotdf, plotdfb, .id = "residVar")
plotdf$residVar <- ifelse(plotdf$residVar == 1, "No Model Variance",
                          "Model Variance")

ggplot(plotdf, aes(x = obs, y = fit, ymin = lwr, ymax = upr)) +
  geom_pointrange() +
  geom_hline(yintercept = 0, color = I("red"), size = 1.1) +
  scale_x_continuous(breaks = c(1, 10)) +
  facet_grid(residVar~effect) + theme_bw()
#> Warning: Using `size` aesthetic for lines was deprecated in ggplot2 3.4.0.
#> ℹ Please use `linewidth` instead.
#> This warning is displayed once per session.
#> Call `lifecycle::last_lifecycle_warnings()` to see where this warning was
#> generated.
```

![](man/figures/README_unnamed-chunk-7-1.png)<!-- -->

We can also investigate the makeup of the prediction for each
observation.

``` r
ggplot(plotdf[plotdf$obs < 6,],
       aes(x = effect, y = fit, ymin = lwr, ymax = upr)) +
  geom_pointrange() +
  geom_hline(yintercept = 0, color = I("red"), size = 1.1) +
  facet_grid(residVar~obs) + theme_bw()
```

![](man/figures/README_unnamed-chunk-8-1.png)<!-- -->

## Plotting

`merTools` also provides functionality for inspecting `merMod` objects
visually. The easiest are getting the posterior distributions of both
fixed and random effect parameters.

``` r
feSims <- FEsim(m1, n.sims = 100)
head(feSims)
#>          term        mean      median         sd
#> 1 (Intercept)  3.22480254  3.22496233 0.01614456
#> 2    service1 -0.06978571 -0.07023666 0.01257453
#> 3   lectage.L -0.18553409 -0.18582412 0.01558176
#> 4   lectage.Q  0.02485585  0.02585718 0.01197095
#> 5   lectage.C -0.02587443 -0.02424027 0.01327372
#> 6   lectage^4 -0.02253842 -0.02115560 0.01434107
```

And we can also plot this:

``` r
plotFEsim(FEsim(m1, n.sims = 100), level = 0.9, stat = 'median', intercept = FALSE)
```

![](man/figures/README_FEsimPlot-1.png)<!-- -->

We can also quickly make caterpillar plots for the random-effect terms:

``` r
reSims <- REsim(m1, n.sims = 100)
head(reSims)
#>   groupFctr groupID        term         mean      median        sd
#> 1         s       1 (Intercept)  0.148244481 0.170819888 0.3386784
#> 2         s       2 (Intercept) -0.006549032 0.009520074 0.3202212
#> 3         s       3 (Intercept)  0.264735849 0.268038770 0.2765385
#> 4         s       4 (Intercept)  0.259182685 0.242682150 0.3018633
#> 5         s       5 (Intercept)  0.020047123 0.019007888 0.3068737
#> 6         s       6 (Intercept)  0.081449654 0.078937547 0.2519452
```

``` r
plotREsim(REsim(m1, n.sims = 100), stat = 'median', sd = TRUE)
```

![](man/figures/README_reSimplot-1.png)<!-- -->

Note that `plotREsim` highlights group levels that have a simulated
distribution that does not overlap 0 – these appear darker. The lighter
bars represent grouping levels that are not distinguishable from 0 in
the data.

Sometimes the random effects can be hard to interpret and not all of
them are meaningfully different from zero. To help with this `merTools`
provides the `expectedRank` function, which provides the percentile
ranks for the observed groups in the random effect distribution taking
into account both the magnitude and uncertainty of the estimated effect
for each group.

``` r
ranks <- expectedRank(m1, groupFctr = "d")
head(ranks)
#>   groupFctr groupLevel      term   estimate  std.error       ER pctER
#> 2         d          1 Intercept  0.3944919 0.08665152 835.3005    74
#> 3         d          6 Intercept -0.4428949 0.03901988 239.5363    21
#> 4         d          7 Intercept  0.6562681 0.03717200 997.3569    88
#> 5         d          8 Intercept -0.6430680 0.02210017 138.3445    12
#> 6         d         12 Intercept  0.1902940 0.04024063 702.3410    62
#> 7         d         13 Intercept  0.2497464 0.03216255 750.0174    66
```

A nice features `expectedRank` is that you can return the expected rank
for all factors simultaneously and use them:

``` r
ranks <- expectedRank(m1)
head(ranks)
#>   groupFctr groupLevel      term    estimate  std.error       ER pctER
#> 2         s          1 Intercept  0.16732800 0.08165665 1931.570    65
#> 3         s          2 Intercept -0.04409538 0.09234250 1368.160    46
#> 4         s          3 Intercept  0.30382219 0.05204082 2309.941    78
#> 5         s          4 Intercept  0.24756175 0.06641699 2151.828    72
#> 6         s          5 Intercept  0.05232329 0.08174130 1627.693    55
#> 7         s          6 Intercept  0.10191653 0.06648394 1772.548    60

ggplot(ranks, aes(x = term, y = estimate)) +
  geom_violin(fill = "gray50") + facet_wrap(~groupFctr) +
  theme_bw()
```

![](man/figures/README_unnamed-chunk-12-1.png)<!-- -->

## Effect Simulation

It can still be difficult to interpret the results of LMM and GLMM
models, especially the relative influence of varying parameters on the
predicted outcome. This is where the `REimpact` and the `wiggle`
functions in `merTools` can be handy.

``` r
impSim <- REimpact(m1, InstEval[7, ], groupFctr = "d", breaks = 5,
                   n.sims = 300, level = 0.9)
#> Warning: executing %dopar% sequentially: no parallel backend registered
impSim
#>   case bin   AvgFit     AvgFitSE nobs
#> 1    1   1 2.780092 3.023753e-04  193
#> 2    1   2 3.253589 6.367631e-05  240
#> 3    1   3 3.545126 5.249834e-05  254
#> 4    1   4 3.829405 6.408711e-05  265
#> 5    1   5 4.232709 1.919177e-04  176
```

The result of `REimpact` shows the change in the `yhat` as the case we
supplied to `newdata` is moved from the first to the fifth quintile in
terms of the magnitude of the group factor coefficient. We can see here
that the individual professor effect has a strong impact on the outcome
variable. The new `plotREimpact()` function visualizes this directly:

``` r
plotREimpact(impSim)
```

![](man/figures/README_reImpactplot-1.png)<!-- -->

Here the standard error is a bit different – it is the weighted standard
error of the mean effect within the bin. It does not take into account
the variability within the effects of each observation in the bin –
accounting for this variation will be a future addition to `merTools`.
\### Comparing grouping factors with `plotREimpact()`

New in merTools 1.0.0, `plotREimpact()` plots `REimpact()` output
directly and can overlay a *named list* of results on a single chart.
This makes it easy to compare how strongly different grouping factors
move the predicted outcome for the same case – here, the instructor
(`d`) and student (`s`) effects from the model above:

``` r
s_impSim <- REimpact(m1, InstEval[7, ], groupFctr = "s", breaks = 5,
                     n.sims = 300, level = 0.9)
plotREimpact(list("Instructor (d)" = impSim, "Student (s)" = s_impSim))
```

![](man/figures/README_reImpactCompare-1.png)<!-- -->

## Explore Substantive Impacts

Another feature of `merTools` is the ability to easily generate
hypothetical scenarios to explore the predicted outcomes of a `merMod`
object and understand what the model is saying in terms of the outcome
variable.

Let’s take the case where we want to explore the impact of a model with
an interaction term between a category and a continuous predictor.
First, we fit a model with interactions:

``` r
data(VerbAgg)
fmVA <- glmer(r2 ~ (Anger + Gender + btype + situ)^2 +
           (1|id) + (1|item), family = binomial,
           data = VerbAgg)
#> Warning in checkConv(attr(opt, "derivs"), opt$par, ctrl = control$checkConv, : Model failed to converge with max|grad| = 0.0729926 (tol = 0.002, component 1)
#>   See ?lme4::convergence and ?lme4::troubleshooting.
```

Now we prep the data using the `draw` function in `merTools`. Here we
draw the average observation from the model frame. We then `wiggle` the
data by expanding the dataframe to include the same observation repeated
but with different values of the variable specified by the `var`
parameter. Here, we expand the dataset to all values of `btype`, `situ`,
and `Anger` subsequently.

``` r
# Select the average case
newData <- draw(fmVA, type = "average")
newData <- wiggle(newData, varlist = "btype",
                  valueslist = list(unique(VerbAgg$btype)))
newData <- wiggle(newData, var = "situ",
                  valueslist = list(unique(VerbAgg$situ)))
newData <- wiggle(newData, var = "Anger",
                  valueslist = list(unique(VerbAgg$Anger)))
head(newData, 10)
#>    r2 Anger Gender btype  situ  id        item
#> 1   N    20      F curse other 149 S3WantCurse
#> 2   N    20      F scold other 149 S3WantCurse
#> 3   N    20      F shout other 149 S3WantCurse
#> 4   N    20      F curse  self 149 S3WantCurse
#> 5   N    20      F scold  self 149 S3WantCurse
#> 6   N    20      F shout  self 149 S3WantCurse
#> 7   N    11      F curse other 149 S3WantCurse
#> 8   N    11      F scold other 149 S3WantCurse
#> 9   N    11      F shout other 149 S3WantCurse
#> 10  N    11      F curse  self 149 S3WantCurse
```

The next step is familiar – we simply pass this new dataset to
`predictInterval` in order to generate predictions for these
counterfactuals. Then we plot the predicted values against the
continuous variable, `Anger`, and facet and group on the two categorical
variables `situ` and `btype` respectively.

``` r
plotdf <- predictInterval(fmVA, newdata = newData, type = "probability",
            stat = "median", n.sims = 1000)
#> Warning: For binomial GLMMs, include.resid.var = TRUE simulates from the
#> conditional binomial distribution (n-trial binomial simulation).
#> This is the theoretically correct approach.
#> To get predictions without residual variance, set include.resid.var = FALSE.
plotdf <- cbind(plotdf, newData)

ggplot(plotdf, aes(y = fit, x = Anger, color = btype, group = btype)) +
  geom_point() + geom_smooth(aes(color = btype), method = "lm") +
  facet_wrap(~situ) + theme_bw() +
  labs(y = "Predicted Probability")
#> `geom_smooth()` using formula = 'y ~ x'
```

![](man/figures/README_substImpactPredict-1.png)<!-- -->

## Cross-version numeric regression checks (for contributors)

`predictInterval()` is stochastic, so the usual unit-test tolerances are
not a reliable way to confirm that a refactor of the simulation
internals has preserved numeric behavior for users who rely on a fixed
seed. To address this, the package ships a standalone regression harness
that pins a canonical set of inputs (LMM and GLMM models, various
`which`, `level`, `stat`, `ignore.fixed.terms`,
`fix.intercept.variance`, and single-row-newdata cases) and dumps
`predictInterval()` output to an RDS bundle for two package versions so
they can be diffed bit-for-bit.

The script lives at `tests/comparisons/predictInterval-regression.R` and
is NOT part of `R CMD check`. It has two modes:

    # Generate an output bundle for one package version
    Rscript tests/comparisons/predictInterval-regression.R harness \
            <pkg_path> <output.rds>

    # Diff two bundles
    Rscript tests/comparisons/predictInterval-regression.R diff \
            <a.rds> <b.rds>

Typical workflow — comparing the current checkout against
`origin/master`:

    git worktree add /tmp/mT-old origin/master
    Rscript tests/comparisons/predictInterval-regression.R harness /tmp/mT-old /tmp/old.rds
    Rscript tests/comparisons/predictInterval-regression.R harness .          /tmp/new.rds
    Rscript tests/comparisons/predictInterval-regression.R diff  /tmp/old.rds /tmp/new.rds
    git worktree remove /tmp/mT-old

The diff output categorizes every case as `IDENTICAL`, `numeric-equal`
(\< 1e-6), or various drift tiers. Any LMM case showing more than
`numeric-equal` indicates that the refactor changed behavior for a
user-supplied seed and should be investigated before the change is
merged. Known-intentional numeric differences (for example, the GLMM
`include.resid.var = TRUE` binomial-residual simulation fix introduced
in 0.9.0) will show up only in the two `glmm_bin_with_resid_*` cases.

Run this whenever touching `R/merPredict.R`,
`R/predictInterval_helpers.R`, or the simulation helpers they call.

## Marginalizing Random Effects

``` r
# get cases
case_idx <- sample(1:nrow(VerbAgg), 10)
mfx <- REmargins(fmVA, newdata = VerbAgg[case_idx,], breaks = 4, groupFctr = "item",
                 type = "probability")
#> Warning: For binomial GLMMs, include.resid.var = TRUE simulates from the
#> conditional binomial distribution (n-trial binomial simulation).
#> This is the theoretically correct approach.
#> To get predictions without residual variance, set include.resid.var = FALSE.

ggplot(mfx, aes(y = fit_combined, x = breaks, group = case)) +
  geom_point() + geom_line() +
  theme_bw() +
  scale_y_continuous(breaks = 1:10/10, limits = c(0, 1)) +
  coord_cartesian(expand = FALSE) +
  labs(x = "Quartile of item random effect Intercept for term 'item'",
       y = "Predicted Probability",
       title = "Simulated Effect of Item Intercept on Predicted Probability for 10 Random Cases")
```

![](man/figures/README_unnamed-chunk-14-1.png)<!-- -->

[![CRAN_Status_Badge](http://www.r-pkg.org/badges/version/merTools)](https://cran.r-project.org/package=merTools)
[![Downloads](http://cranlogs.r-pkg.org/badges/merTools)](https://cran.r-project.org/package=merTools)
[![Downloads](http://cranlogs.r-pkg.org/badges/grand-total/merTools)](https://cran.r-project.org/package=merTools)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# merTools

A package for getting the most out of large multilevel models in R

by Jared E. Knowles and Carl Frederick

Working with generalized linear mixed models (GLMM) and linear mixed
models (LMM) has become increasingly easy with advances in the `lme4`
package. As we have found ourselves using these models more and more
within our work, we, the authors, have developed a set of tools for
simplifying and speeding up common tasks for interacting with `merMod`
objects from `lme4`. This package provides those tools.

## Installation

``` r
# development version
library(devtools)
install_github("jknowles/merTools")

# CRAN version
install.packages("merTools")
```

## Recent Updates

## merTools 0.6.2 (Early 2024)

- Maintenance release to fix minor issues with function documentation
- Fix \#130 by avoiding conflict with `vcov` in the `merDeriv` package
- Upgrade package test infrastructure to 3e testthat specification

## merTools 0.6.1 (Spring 2023)

- Maintenance release to keep package listed on CRAN
- Fix a small bug where parallel code path is run twice (#126)
- Update plotting functions to avoid deprecated `aes_string()` calls
  (#127)
- Fix (#115) in description
- Speed up PI using @bbolker pull request (#120)
- Updated package maintainer contact information

### merTools 0.5.0

#### New Features

- `subBoot` now works with `glmerMod` objects as well
- `reMargins` a new function that allows the user to marginalize the
  prediction over breaks in the distribution of random effect
  distributions, see `?reMargins` and the new `reMargins` vignette
  (closes \#73)

#### Bug fixes

- Fixed an issue where known convergence errors were issuing warnings
  and causing the test suite to not work
- Fixed an issue where models with a random slope, no intercept, and no
  fixed term were unable to be predicted (#101)
- Fixed an issue with shinyMer not working with substantive fixed
  effects (#93)

### merTools 0.4.1

#### New Features

- Standard errors reported by `merModList` functions now apply the Rubin
  correction for multiple imputation

#### Bug fixes

- Contribution by Alex Whitworth (@alexWhitworth) adding error checking
  to plotting functions

## Shiny App and Demo

The easiest way to demo the features of this application is to use the
bundled Shiny application which launches a number of the metrics here to
aide in exploring the model. To do this:

    library(merTools)
    m1 <- lmer(y ~ service + lectage + studage + (1|d) + (1|s), data=InstEval)
    shinyMer(m1, simData = InstEval[1:100, ]) # just try the first 100 rows of data

![](man/figures/README-predPanel.png)

On the first tab, the function presents the prediction intervals for the
data selected by user which are calculated using the `predictInterval`
function within the package. This function calculates prediction
intervals quickly by sampling from the simulated distribution of the
fixed effect and random effect terms and combining these simulated
estimates to produce a distribution of predictions for each observation.
This allows prediction intervals to be generated from very large models
where the use of `bootMer` would not be feasible computationally.

![](man/figures/README-effPanel.png)

On the next tab the distribution of the fixed effect and group-level
effects is depicted on confidence interval plots. These are useful for
diagnostics and provide a way to inspect the relative magnitudes of
various parameters. This tab makes use of four related functions in
`merTools`: `FEsim`, `plotFEsim`, `REsim` and `plotREsim` which are
available to be used on their own as well.

![](man/figures/README-substPanel.png)

On the third tab are some convenient ways to show the influence or
magnitude of effects by leveraging the power of `predictInterval`. For
each case, up to 12, in the selected data type, the user can view the
impact of changing either one of the fixed effect or one of the grouping
level terms. Using the `REimpact` function, each case is simulated with
the model’s prediction if all else was held equal, but the observation
was moved through the distribution of the fixed effect or the random
effect term. This is plotted on the scale of the dependent variable,
which allows the user to compare the magnitude of effects across
variables, and also between models on the same data.

## Predicting

Standard prediction looks like so.

``` r
predict(m1, newdata = InstEval[1:10, ])
#>        1        2        3        4        5        6        7        8 
#> 3.146337 3.165212 3.398499 3.114249 3.320686 3.252670 4.180897 3.845219 
#>        9       10 
#> 3.779337 3.331013
```

With `predictInterval` we obtain predictions that are more like the
standard objects produced by `lm` and `glm`:

``` r
predictInterval(m1, newdata = InstEval[1:10, ], n.sims = 500, level = 0.9, 
                stat = 'median')
#>         fit      upr       lwr
#> 1  3.215698 5.302545 1.4367495
#> 2  3.155941 5.327796 1.2210140
#> 3  3.374129 5.287901 1.4875231
#> 4  3.101672 5.183841 0.9248584
#> 5  3.299367 5.298370 1.3287058
#> 6  3.147238 5.368311 1.1132248
#> 7  4.155194 6.273147 2.2167207
#> 8  3.873493 5.705669 1.9152401
#> 9  3.740978 5.737517 2.0454222
#> 10 3.291242 5.297614 1.2375007
```

Note that `predictInterval` is slower because it is computing
simulations. It can also return all of the simulated `yhat` values as an
attribute to the predict object itself.

`predictInterval` uses the `sim` function from the `arm` package heavily
to draw the distributions of the parameters of the model. It then
combines these simulated values to create a distribution of the `yhat`
for each observation.

### Inspecting the Prediction Components

We can also explore the components of the prediction interval by asking
`predictInterval` to return specific components of the prediction
interval.

``` r
predictInterval(m1, newdata = InstEval[1:10, ], n.sims = 200, level = 0.9, 
                stat = 'median', which = "all")
#>      effect         fit      upr       lwr obs
#> 1  combined  3.35554348 5.217964  1.615782   1
#> 2  combined  3.21487934 5.327824  1.114338   2
#> 3  combined  3.44493242 5.474256  1.809136   3
#> 4  combined  3.24123655 4.838427  1.272174   4
#> 5  combined  3.20539661 5.367651  1.068128   5
#> 6  combined  3.54335144 5.481756  1.585809   6
#> 7  combined  4.23212790 6.267669  2.284923   7
#> 8  combined  4.05055116 5.684968  1.931558   8
#> 9  combined  3.84266853 5.492163  2.091312   9
#> 10 combined  3.24121727 5.183680  1.196101  10
#> 11        s -0.02342248 1.948494 -1.691035   1
#> 12        s  0.04148408 2.091467 -1.782386   2
#> 13        s  0.04477028 2.087629 -2.144621   3
#> 14        s  0.26160482 2.114509 -1.733429   4
#> 15        s -0.10803386 1.714535 -1.982283   5
#> 16        s -0.04962613 1.916212 -1.909187   6
#> 17        s  0.24916111 2.001528 -1.628554   7
#> 18        s  0.19640074 2.070513 -1.473660   8
#> 19        s  0.27031215 2.119763 -1.643120   9
#> 20        s  0.13772544 2.313012 -1.855489  10
#> 21        d -0.32196201 1.357316 -2.397083   1
#> 22        d -0.29691477 1.422141 -2.662141   2
#> 23        d  0.24828667 1.782181 -1.987563   3
#> 24        d -0.37893052 1.471225 -2.350781   4
#> 25        d  0.02142086 2.172075 -2.148417   5
#> 26        d  0.07926221 2.003462 -1.677765   6
#> 27        d  0.76480967 2.767889 -1.274501   7
#> 28        d  0.08757337 2.374201 -1.958689   8
#> 29        d  0.25289032 2.083732 -1.376630   9
#> 30        d -0.17775160 1.601744 -2.115104  10
#> 31    fixed  3.16750528 5.010517  1.371678   1
#> 32    fixed  3.21493166 5.246672  1.074857   2
#> 33    fixed  3.36233628 5.581696  1.474776   3
#> 34    fixed  3.17926915 5.107315  1.621278   4
#> 35    fixed  3.16562882 5.136197  1.156010   5
#> 36    fixed  3.15944014 5.114967  1.506315   6
#> 37    fixed  3.32101367 5.149819  1.407884   7
#> 38    fixed  3.34020282 5.189215  1.651446   8
#> 39    fixed  3.17901802 5.000429  1.132874   9
#> 40    fixed  3.41100236 5.207451  1.555844  10
```

This can lead to some useful plotting:

``` r
library(ggplot2)
#> Warning: package 'ggplot2' was built under R version 4.3.2
plotdf <- predictInterval(m1, newdata = InstEval[1:10, ], n.sims = 2000, 
                          level = 0.9, stat = 'median', which = "all", 
                          include.resid.var = FALSE)
plotdfb <- predictInterval(m1, newdata = InstEval[1:10, ], n.sims = 2000, 
                          level = 0.9, stat = 'median', which = "all", 
                          include.resid.var = TRUE)

plotdf <- dplyr::bind_rows(plotdf, plotdfb, .id = "residVar")
plotdf$residVar <- ifelse(plotdf$residVar == 1, "No Model Variance", 
                          "Model Variance")

ggplot(plotdf, aes(x = obs, y = fit, ymin = lwr, ymax = upr)) + 
  geom_pointrange() +
  geom_hline(yintercept = 0, color = I("red"), size = 1.1) +
  scale_x_continuous(breaks = c(1, 10)) +
  facet_grid(residVar~effect) + theme_bw()
#> Warning: Using `size` aesthetic for lines was deprecated in ggplot2 3.4.0.
#> ℹ Please use `linewidth` instead.
#> This warning is displayed once every 8 hours.
#> Call `lifecycle::last_lifecycle_warnings()` to see where this warning was
#> generated.
```

![](man/figures/README_unnamed-chunk-7-1.png)<!-- -->

We can also investigate the makeup of the prediction for each
observation.

``` r
ggplot(plotdf[plotdf$obs < 6,], 
       aes(x = effect, y = fit, ymin = lwr, ymax = upr)) + 
  geom_pointrange() +
  geom_hline(yintercept = 0, color = I("red"), size = 1.1) +
  facet_grid(residVar~obs) + theme_bw()
```

![](man/figures/README_unnamed-chunk-8-1.png)<!-- -->

## Plotting

`merTools` also provides functionality for inspecting `merMod` objects
visually. The easiest are getting the posterior distributions of both
fixed and random effect parameters.

``` r
feSims <- FEsim(m1, n.sims = 100)
head(feSims)
#>          term        mean      median         sd
#> 1 (Intercept)  3.22469489  3.22427807 0.01767782
#> 2    service1 -0.07112136 -0.07190156 0.01382577
#> 3   lectage.L -0.18681442 -0.18833173 0.01617233
#> 4   lectage.Q  0.02208290  0.02252175 0.01259911
#> 5   lectage.C -0.02656846 -0.02507924 0.01099316
#> 6   lectage^4 -0.02165334 -0.02073531 0.01355675
```

And we can also plot this:

``` r
plotFEsim(FEsim(m1, n.sims = 100), level = 0.9, stat = 'median', intercept = FALSE)
```

![](man/figures/README_FEsimPlot-1.png)<!-- -->

We can also quickly make caterpillar plots for the random-effect terms:

``` r
reSims <- REsim(m1, n.sims = 100)
head(reSims)
#>   groupFctr groupID        term        mean      median        sd
#> 1         s       1 (Intercept)  0.21300834  0.22465501 0.3266393
#> 2         s       2 (Intercept) -0.04278168 -0.08474934 0.3164631
#> 3         s       3 (Intercept)  0.41998629  0.42229608 0.2865010
#> 4         s       4 (Intercept)  0.28818926  0.28746523 0.3019460
#> 5         s       5 (Intercept)  0.05833502  0.03342223 0.3539464
#> 6         s       6 (Intercept)  0.08108947  0.08478943 0.2049882
```

``` r
plotREsim(REsim(m1, n.sims = 100), stat = 'median', sd = TRUE)
```

![](man/figures/README_reSimplot-1.png)<!-- -->

Note that `plotREsim` highlights group levels that have a simulated
distribution that does not overlap 0 – these appear darker. The lighter
bars represent grouping levels that are not distinguishable from 0 in
the data.

Sometimes the random effects can be hard to interpret and not all of
them are meaningfully different from zero. To help with this `merTools`
provides the `expectedRank` function, which provides the percentile
ranks for the observed groups in the random effect distribution taking
into account both the magnitude and uncertainty of the estimated effect
for each group.

``` r
ranks <- expectedRank(m1, groupFctr = "d")
head(ranks)
#>   groupFctr groupLevel      term   estimate  std.error       ER pctER
#> 2         d          1 Intercept  0.3944919 0.08665152 835.3005    74
#> 3         d          6 Intercept -0.4428949 0.03901988 239.5363    21
#> 4         d          7 Intercept  0.6562681 0.03717200 997.3569    88
#> 5         d          8 Intercept -0.6430680 0.02210017 138.3445    12
#> 6         d         12 Intercept  0.1902940 0.04024063 702.3410    62
#> 7         d         13 Intercept  0.2497464 0.03216255 750.0174    66
```

A nice features `expectedRank` is that you can return the expected rank
for all factors simultaneously and use them:

``` r
ranks <- expectedRank(m1)
head(ranks)
#>   groupFctr groupLevel      term    estimate  std.error       ER pctER
#> 2         s          1 Intercept  0.16732800 0.08165665 1931.570    65
#> 3         s          2 Intercept -0.04409538 0.09234250 1368.160    46
#> 4         s          3 Intercept  0.30382219 0.05204082 2309.941    78
#> 5         s          4 Intercept  0.24756175 0.06641699 2151.828    72
#> 6         s          5 Intercept  0.05232329 0.08174130 1627.693    55
#> 7         s          6 Intercept  0.10191653 0.06648394 1772.548    60

ggplot(ranks, aes(x = term, y = estimate)) + 
  geom_violin(fill = "gray50") + facet_wrap(~groupFctr) +
  theme_bw()
```

![](man/figures/README_unnamed-chunk-12-1.png)<!-- -->

## Effect Simulation

It can still be difficult to interpret the results of LMM and GLMM
models, especially the relative influence of varying parameters on the
predicted outcome. This is where the `REimpact` and the `wiggle`
functions in `merTools` can be handy.

``` r
impSim <- REimpact(m1, InstEval[7, ], groupFctr = "d", breaks = 5, 
                   n.sims = 300, level = 0.9)
#> Warning: executing %dopar% sequentially: no parallel backend registered
impSim
#>   case bin   AvgFit     AvgFitSE nobs
#> 1    1   1 2.775244 2.889334e-04  193
#> 2    1   2 3.251612 5.378474e-05  240
#> 3    1   3 3.535261 5.204943e-05  254
#> 4    1   4 3.834013 6.310989e-05  265
#> 5    1   5 4.220246 1.859937e-04  176
```

The result of `REimpact` shows the change in the `yhat` as the case we
supplied to `newdata` is moved from the first to the fifth quintile in
terms of the magnitude of the group factor coefficient. We can see here
that the individual professor effect has a strong impact on the outcome
variable. This can be shown graphically as well:

``` r
ggplot(impSim, aes(x = factor(bin), y = AvgFit, ymin = AvgFit - 1.96*AvgFitSE, 
                   ymax = AvgFit + 1.96*AvgFitSE)) + 
  geom_pointrange() + theme_bw() + labs(x = "Bin of `d` term", y = "Predicted Fit")
```

![](man/figures/README_reImpactplot-1.png)<!-- -->

Here the standard error is a bit different – it is the weighted standard
error of the mean effect within the bin. It does not take into account
the variability within the effects of each observation in the bin –
accounting for this variation will be a future addition to `merTools`.

## Explore Substantive Impacts

Another feature of `merTools` is the ability to easily generate
hypothetical scenarios to explore the predicted outcomes of a `merMod`
object and understand what the model is saying in terms of the outcome
variable.

Let’s take the case where we want to explore the impact of a model with
an interaction term between a category and a continuous predictor.
First, we fit a model with interactions:

``` r
data(VerbAgg)
fmVA <- glmer(r2 ~ (Anger + Gender + btype + situ)^2 +
           (1|id) + (1|item), family = binomial, 
           data = VerbAgg)
#> Warning in checkConv(attr(opt, "derivs"), opt$par, ctrl = control$checkConv, :
#> Model failed to converge with max|grad| = 0.0543724 (tol = 0.002, component 1)
```

Now we prep the data using the `draw` function in `merTools`. Here we
draw the average observation from the model frame. We then `wiggle` the
data by expanding the dataframe to include the same observation repeated
but with different values of the variable specified by the `var`
parameter. Here, we expand the dataset to all values of `btype`, `situ`,
and `Anger` subsequently.

``` r
# Select the average case
newData <- draw(fmVA, type = "average")
newData <- wiggle(newData, varlist = "btype", 
                  valueslist = list(unique(VerbAgg$btype)))
newData <- wiggle(newData, var = "situ", 
                  valueslist = list(unique(VerbAgg$situ)))
newData <- wiggle(newData, var = "Anger", 
                  valueslist = list(unique(VerbAgg$Anger)))
head(newData, 10)
#>    r2 Anger Gender btype  situ id        item
#> 1   N    20      F curse other  5 S3WantCurse
#> 2   N    20      F scold other  5 S3WantCurse
#> 3   N    20      F shout other  5 S3WantCurse
#> 4   N    20      F curse  self  5 S3WantCurse
#> 5   N    20      F scold  self  5 S3WantCurse
#> 6   N    20      F shout  self  5 S3WantCurse
#> 7   N    11      F curse other  5 S3WantCurse
#> 8   N    11      F scold other  5 S3WantCurse
#> 9   N    11      F shout other  5 S3WantCurse
#> 10  N    11      F curse  self  5 S3WantCurse
```

The next step is familiar – we simply pass this new dataset to
`predictInterval` in order to generate predictions for these
counterfactuals. Then we plot the predicted values against the
continuous variable, `Anger`, and facet and group on the two categorical
variables `situ` and `btype` respectively.

``` r
plotdf <- predictInterval(fmVA, newdata = newData, type = "probability", 
            stat = "median", n.sims = 1000)
plotdf <- cbind(plotdf, newData)

ggplot(plotdf, aes(y = fit, x = Anger, color = btype, group = btype)) + 
  geom_point() + geom_smooth(aes(color = btype), method = "lm") + 
  facet_wrap(~situ) + theme_bw() +
  labs(y = "Predicted Probability")
#> `geom_smooth()` using formula = 'y ~ x'
```

![](man/figures/README_substImpactPredict-1.png)<!-- -->

## Marginalizing Random Effects

``` r
# get cases
case_idx <- sample(1:nrow(VerbAgg), 10)
mfx <- REmargins(fmVA, newdata = VerbAgg[case_idx,], breaks = 4, groupFctr = "item", 
                 type = "probability")

ggplot(mfx, aes(y = fit_combined, x = breaks, group = case)) + 
  geom_point() + geom_line() + 
  theme_bw() + 
  scale_y_continuous(breaks = 1:10/10, limits = c(0, 1)) +
  coord_cartesian(expand = FALSE) +
  labs(x = "Quartile of item random effect Intercept for term 'item'", 
       y = "Predicted Probability", 
       title = "Simulated Effect of Item Intercept on Predicted Probability for 10 Random Cases")
```

![](man/figures/README_unnamed-chunk-14-1.png)<!-- -->

# merTools test suite layers

The tests in this directory are organized into three conceptual layers. Each
layer answers a different question. Knowing which layer a test belongs to
determines how brittle it is allowed to be, how fast it must run, and what
platforms it can run on.

When you write a new test, pick the layer that matches the question you are
actually trying to answer. Mixing layers in one test (e.g., bolting a tight
Monte Carlo bias check onto a structural compatibility test) is the most
common cause of flaky tests in this package — don't do it.

---

## Layer 1 — Structural invariants

> *"Is the output still a valid `predictInterval()` result?"*

File: **`test-predictInterval-invariants.R`**

Tests properties that MUST hold for every valid output regardless of seed,
simulation noise, platform, model type, or option combination. Derived from
the function's contract, not from specific numbers.

Examples: output is a `data.frame`, has the expected column names, `upr >=
fit >= lwr` row-wise, `nrow(output) == nrow(newdata)`, no `NA`/`NaN`/`Inf`,
same seed produces identical output, higher `level` gives wider intervals,
`type = "probability"` values are in `[0, 1]`.

**Runs on every CI platform.** No `skip_on_os`. No `skip_on_ci`. Seed- and
platform-independent.

When adding a new invariant: phrase it as "for any valid input, X must hold"
and verify it is derivable from the function's documentation or mathematics.
Do not snapshot numeric output here — that's Layer 2.

---

## Layer 2 — Pinned numeric regression

> *"Did my edit silently change the numbers users see?"*

File: **`test-predictInterval-snapshot.R`**
Committed contract: **`_snaps/predictInterval-snapshot.md`**

Uses `testthat::expect_snapshot_value(style = "json2", tolerance = 1e-6)` to
pin the exact `predictInterval()` output for a canonical set of LMM and
GLMM inputs. Any code change that alters these outputs produces a
reviewable diff in the committed `.md` file, not a random-tolerance failure.

**Runs on Linux only.** macOS-ARM64 and Windows produce bit-level drift from
different BLAS/LAPACK that would pollute the snapshot. Linux is the
canonical platform for the committed contract. Cross-platform equivalence
is covered manually by Layer 3.

When adding a new model to Layer 2: first check that the design matrix is
NOT rank-deficient or near-singular. `mvtnorm::rmvnorm(method = "chol")` on
such a matrix produces output that differs across BLAS implementations by
enough to blow past any reasonable snapshot tolerance, even on the same OS
across R versions. Rank-deficient models belong in Layer 1 and Layer 3 only.

To intentionally accept a changed snapshot:

    testthat::snapshot_accept("predictInterval-snapshot")

Review the resulting `.md` diff carefully before committing.

---

## Layer 3 — Statistical validation and offline regression

> *"Is the algorithm statistically correct (unbiased, well-covered)?
>   And does my refactor preserve numeric behavior across versions?"*

There are two mechanisms for Layer 3 content:

### In-suite, tagged with `skip_on_ci()`

Tests that require large `n.sims`, simulate new data, check coverage or
unbiasedness with tight tolerances, or exercise large datasets. These
still run automatically during local `devtools::test()` but skip on CI,
where they would be slow and flaky by nature.

Examples in this directory:

- `test-predict1.R` — coverage and bias checks on simulated LMM/GLMM data
- `test-predict2.R` — tight-tolerance median-PI accuracy checks
- `test-predict3.R` — rank-deficient model accuracy
- `test-predict4.R` — parallel execution performance/consistency
- `test-predictInterval-helpers.R` (one block) — large `InstEval` fit

When adding a new statistical check, tag it with `skip_on_ci()` and add a
comment explaining what statistical property it is validating.

### Cross-version regression harness (manual)

File: **`tests/comparisons/predictInterval-regression.R`**

A two-mode script (`harness` + `diff`) that pins a canonical set of
`predictInterval()` inputs and serializes their outputs as RDS, enabling
bit-for-bit comparison between two package versions. Run manually when
touching simulation internals to confirm a refactor did not change
numeric behavior for users with a fixed seed.

See the script header or the top-level `README.md` for the full workflow.

---

## Other test files in this directory

The remaining files test **specific model types and behaviors** that are
not numerical regressions or universal invariants:

- `test-predict.R` — edge-case model specifications (nested effects, cross-
  level interactions, no fixed intercept). Structural compatibility only.
- `test-predictInterval-accuracy.R` — reproducibility and dimension checks
  for a range of options.
- `test-predictInterval-helpers.R` — white-box unit tests for internal
  helper functions (`simulate_residual_variance`, `simulate_fixed_effects`,
  etc.).
- `test-seeds.R` — asserts that `seed =` arguments behave correctly across
  all user-facing stochastic functions.
- `test-merData.R`, `test-merExtract.R`, `test-merModList.R`,
  `test-expectedRank.R`, `test-REmargins.R`, `test-substEff.R`, `test-helpers.R`,
  `test-subboot.R` — domain-specific tests for the non-`predictInterval`
  surface of `merTools`.

---

## Conventions

- **Seed**: Use `seed = 11213` (or `set.seed(11213)`) unless a different
  seed is required to demonstrate a specific property (e.g., `test-seeds.R`
  pairs a second seed to assert divergence).
- **`skip_on_cran()`**: Use on any test that takes more than ~1 second.
  CRAN auto-checks are time-sensitive.
- **`skip_on_ci()`**: Use on Layer 3 content per above.
- **`skip_on_os()`**: Use only in `test-predictInterval-snapshot.R` to
  scope Layer 2 to Linux.

## Pinned RNG

`helper-seed.R` sets `RNGversion("4.1.0")` and an explicit `RNGkind()` so
that the same seed produces the same stream across R-oldrel, R-release,
and R-devel on CI. Without this, default RNG tweaks across R versions can
silently change test outputs.

# ShinyMer
shinyUI(fluidPage(
  shiny::titlePanel("Explore your merMod interactively"),
  shiny::sidebarLayout(
    shiny::sidebarPanel(
      shiny::radioButtons("newdataType",
                          "Simulated data scenario",
                          choices = df.choices,
                          selected = NULL),
      # Optional subset of the model frame to draw a random/average case from
      # (issue #32). Only relevant for the "Random Obs" / "Average Obs" draws.
      conditionalPanel(
        condition = "input.newdataType == 'rand' || input.newdataType == 'mean'",
        shiny::selectInput("subsetVar",
                           "Restrict draw to a subset (optional):",
                           choices = subset.choices,
                           selected = ""),
        shiny::uiOutput("subsetVal_ui")
      ),
      shiny::numericInput("n.sims",
                          label = "Simulations (Max=1,000)",
                          value = 100,
                          min = 1,
                          max = 1000),
      conditionalPanel(condition = "input.conditionedPanels==3",
                       helpText("Here you can compare impact of
                                changing input variables on the
                                outcome variable for selected cases."),
                       selectInput("group", "Group Factor:",
                                   choices = names(ranef(merMod)),
                                   selected = NULL),
                       uiOutput("re.ui"),
                       sliderInput("nbin", "Effect Bins", min = 3,
                                   max = 10, value = 4, step = 1),
                       helpText("And modify your fixed effects"),
                       selectInput("fixef", "Fixed Effect:",
                                   choices = all.vars(nobars(formula(merMod)))[-1],
                                   selected = NULL)
                       ),
      shiny::numericInput("alpha",
                          label = "Credible Interval (%)",
                          value = 95,
                          min = 0,
                          max = 100),
      shiny::radioButtons("stat",
                          "Measure of central tendency",
                          choices = c("Median" = "median", "Mean" = "mean"),
                          selected = NULL),
      shiny::radioButtons("predMetric",
                          "Prediction metric",
                          choices = c("Linear Predictor" = "linear.prediction",
                                    "Probability" = "probability"),
                          selected = NULL),
      shiny::checkboxInput("resid.var",
                           label = "Include Residual Variation",
                           value = TRUE)
    ),
    shiny::mainPanel(
      shiny::tabsetPanel(type = "tabs",
                         shiny::tabPanel("Prediction uncertainty",
                                         shiny::h3("Prediction Intervals:"),
                                         plotOutput("predPlot"),
                                         shiny::h3("All Predictions"),
                                         if ("DT" %in% rownames(installed.packages())) {
                                            DT::DTOutput("dt") } else {
                                            tableOutput("dt")
                                            },
                                         shiny::downloadButton("downloadData", "Download predict interval data")
                                         , value = 1
                         ),
                         shiny::tabPanel("Parameters",
                                         shiny::h3("Original call"),
                                         textOutput("call"),
                                         shiny::h3("Fixed Effects"),
                                         plotOutput("feplot"),
                                         shiny::h3("Group Effects"),
                                         plotOutput("replot"), value = 2
                         ),
                         shiny::tabPanel("Substantive Effect",
                                         shiny::h3("Effect Sizes"),
                                         plotOutput("gPlot"),
                                         shiny::h3("Fixef Effect Impact"),
                                         plotOutput("wigglePlot"), value = 3
                         ),
                         shiny::tabPanel("Model Summary",
                                         shiny::h3("Model overview"),
                                         tableOutput("modelInfo"),
                                         shiny::h3("Original call"),
                                         verbatimTextOutput("summaryCall"),
                                         shiny::h3("Fixed effects"),
                                         tableOutput("fixefTab"),
                                         shiny::h3("Random effects (variances & correlations)"),
                                         verbatimTextOutput("vcTab"),
                                         shiny::h3("Grouping factors"),
                                         tableOutput("ngrpsTab"), value = 4
                         ), id = "conditionedPanels"
      )
    )
)
  ))

#SERVER----
server = function(input, output, session){
  output$text1 <- renderText({
    paste("You have selected", input$stat)
  })

  # Optional subset of the model frame for the Random/Average draws (issue #32).
  # Returns a named list suitable for draw(..., varList = ) or NULL when the user
  # has not chosen a variable to restrict on.
  subsetVarList <- reactive({
    if (is.null(input$subsetVar) || identical(input$subsetVar, "") ||
        is.null(input$subsetVal) || identical(input$subsetVal, "")) {
      return(NULL)
    }
    stats::setNames(list(input$subsetVal), input$subsetVar)
  })

  # Value picker for the chosen subset variable, populated from the model frame.
  output$subsetVal_ui <- renderUI({
    if (is.null(input$subsetVar) || identical(input$subsetVar, "")) {
      return(NULL)
    }
    vals <- sort(unique(as.character(merMod@frame[, input$subsetVar])))
    selectInput("subsetVal",
                paste0("Value of ", input$subsetVar, ":"),
                choices = vals,
                selected = vals[1])
  })

  # Resolve the requested scenario into a data.frame of cases to predict from.
  drawData <- function() {
    switch(input$newdataType,
           "orig" = merMod@frame,
           "mean" = draw(merMod, type = "average", varList = subsetVarList()),
           "rand" = draw(merMod, type = "random", varList = subsetVarList()),
           "user" = newdata)
  }

  predInput <- reactive({
    data <- drawData()
    cbind(predictInterval(merMod, newdata = data, level = input$alpha/100,
                          type = input$predMetric,
                          include.resid.var = input$resid.var,
                          n.sims = input$n.sims, stat = input$stat), data)
  })

  if ("DT" %in% rownames(installed.packages())) {
    output$dt <- DT::renderDT({
      predInput()
    })
  } else {
    output$dt <- renderTable({
      predInput()
    })
  }

  output$downloadData <- shiny::downloadHandler(
    filename = "predictIntervalResults.csv",
    content = function(file) {
      write.csv(shiny::isolate(predInput()), file)
    }
  )

  output$predPlot <- renderPlot({
    data <- predInput()
    data$x <- factor(seq(1:nrow(data)))
    ggplot(data, aes(x = x, y = fit, ymin = lwr, ymax = upr)) +
      geom_pointrange() +
      theme_bw() + theme(axis.text.x = element_blank(),
                         panel.grid.major.x = element_blank(),
                         panel.grid.minor.x = element_blank(),
                         axis.ticks.x = element_blank())
  })

  feData <- reactive({
    data <- FEsim(merMod, n.sims = input$n.sims)
    return(data)
  })

  output$feplot <- renderPlot({
    plotdf <- feData()
    scale <- input$alpha/100
    vartmp <- input$stat
    plotFEsim(plotdf, level = scale, stat = vartmp, sd = TRUE,
              intercept = FALSE)
  })

  reData <- reactive({
    data <- REsim(merMod, n.sims = input$n.sims)
    return(data)
  })

  output$replot <- renderPlot({
    plotdf <- reData()
    scale <- input$alpha/100
    vartmp <- input$stat
    plotREsim(plotdf, level = scale, stat = vartmp, sd = TRUE)
  })

  output$call <- renderPrint({
    merMod@call
  })

  # ---- Model Summary tab (issue #78) ----
  output$modelInfo <- renderTable({
    merTools::modelInfo(merMod)
  }, digits = 3)

  output$summaryCall <- renderPrint({
    print(merMod@call)
  })

  output$fixefTab <- renderTable({
    fe <- lme4::fixef(merMod)
    data.frame(term = names(fe), estimate = as.numeric(fe),
               check.names = FALSE)
  }, digits = 4)

  output$vcTab <- renderPrint({
    print(lme4::VarCorr(merMod), comp = c("Variance", "Std.Dev."))
  })

  output$ngrpsTab <- renderTable({
    ng <- lme4::ngrps(merMod)
    data.frame(grouping_factor = names(ng), n_levels = as.integer(ng),
               check.names = FALSE)
  })

  reEffInput <- reactive({
    data <- drawData()
    if(nrow(data) > 12){
      warning("Too much data selected, only using top 12 rows.")
      data <- data[1:12, ]
    }
    return(data)
  })

  groupData <- reactive({
    plotdf <- REimpact(merMod, newdata = reEffInput(),
                       groupFctr = input$group,
                       term = input$term,
                       level = input$alpha/100,
                       breaks = input$nbin,
                       type = input$predMetric,
                       include.resid.var = input$resid.var,
                       n.sims = input$n.sims, stat = input$stat)
    # Two-sided interval half-width from the requested level.
    z <- qnorm(1 - (1 - input$alpha/100) / 2)
    plotdf$upr <- plotdf$AvgFit + z * plotdf$AvgFitSE
    plotdf$lwr <- plotdf$AvgFit - z * plotdf$AvgFitSE
    plotdf$bin <- factor(plotdf$bin)
    return(plotdf)
  })

  output$gPlot <- renderPlot({
    ggplot(groupData(), aes(x = bin, y = AvgFit, ymin = lwr, ymax = upr)) +
      geom_pointrange() + facet_wrap(~case) +
      theme_bw() + labs(x = "Bin", y = "Value of DV",
                        title = "Impact of grouping term for selected case")
  })

  wiggleData <- reactive({
    base <- reEffInput()
    valLookup <- unique(merMod@frame[, input$fixef])
    if (is.numeric(valLookup)) {
      newvals <- seq(min(valLookup), max(valLookup), length.out = 20)
    } else if (length(valLookup) < 50) {
      newvals <- valLookup
    } else {
      newvals <- sample(valLookup, 50)
    }
    plotdf <- wiggle(base, input$fixef, values = list(newvals))
    plotdf <- cbind(plotdf, predictInterval(merMod, newdata = plotdf,
                                            type = input$predMetric,
                                            level = input$alpha/100,
                                            include.resid.var = input$resid.var,
                                            n.sims = input$n.sims, stat = input$stat))
    plotdf$X <- plotdf[, input$fixef]
    # wiggle() stacks the base data once per value, so each base observation
    # (the "case" we facet on) repeats across the value blocks.
    plotdf$case <- rep(seq_len(nrow(base)), times = length(newvals))
    return(plotdf)
  })

  output$re.ui <- renderUI({
    choices <- names(ranef(merMod)[[input$group]])
    selectInput("term", "Group Term:",
                choices = choices,
                selected = choices[1])
  })

  output$wigglePlot <- renderPlot({
    ggplot(wiggleData(), aes(x = X, y = fit, ymin = lwr,
                             ymax = upr)) +
      geom_pointrange() + facet_wrap(~case) +
      theme_bw() + labs(y = "Simulated Value of DV",
                        title = "Impact of selected fixed effect for
                        selected cases.")
  })

  }

# # Not working currently
#
# # Diagramming a model
# library(DiagrammeR)
#
#
# # Get the 'nycflights13' package if not already installed
# # install.packages('nycflights13')
#
# # Get the 'lubridate' package if not already installed
# # install.packages('lubridate')
#
# # Get the latest build of the 'DiagrammeR' package from GitHub
# devtools::install_github('rich-iannone/DiagrammeR')
#
# library("nycflights13")
# library("lubridate")
# library("DiagrammeR")
# library("pipeR")
#
# # Choose a day from 2013 for NYC flight data
# # (You can choose any Julian day, it's interesting to see results for different days)
# day_of_year <- 10
#
# # Get a data frame of complete cases (e.g., flights have departure and arrival times)
# nycflights13 <-
#   nycflights13::flights[which(complete.cases(nycflights13::flights) == TRUE), ]
#
# # Generate a POSIXct vector of dates using the 'ISOdatetime' function
# # Columns 1, 2, and 3 are year, month, and day columns
# # Column 4 is a 4-digit combination of hours (00-23) and minutes (00-59)
# date_time <-
#   data.frame("date_time" =
#                ISOdatetime(year = nycflights13[,1],
#                            month = nycflights13[,2],
#                            day = nycflights13[,3],
#                            hour = gsub("[0-9][0-9]$", "", nycflights13[,4]),
#                            min = gsub(".*([0-9][0-9])$", "\\1", nycflights13[,4]),
#                            sec = 0, tz = "GMT"))
#
# # Add the POSIXct vector 'date_time' to the 'nycflights13' data frame
# nycflights13 <- cbind(date_time, nycflights13)
#
# # Select flights only from the specified day of the year 2013
# nycflights13_day <-
#   subset(nycflights13,
#          date_time >= ymd('2013-01-01', tz = "GMT") + days(day_of_year - 1) &
#            date_time < ymd('2013-01-01', tz = "GMT") + days(day_of_year))
#
# # Create the 'nodes' data frame where at least one column is named "nodes" or "node_id"
# # Column 12 is the 3-letter code for the airport departing from
# # Column 13 is for the airport arriving to
# # (Option: change df to 'nycflights13_day' and only airports used for the day will be included)
# nodes_df <- create_nodes(nodes = unique(c(nycflights13[,12],
#                                           nycflights13[,13])),
#                          label = FALSE)
#
# # The 'edges' data frame must have columns named 'edge_from' and 'edge_to'
# # The color attribute is determined with an 'ifelse' statement, where
# # column 8 is the minutes early (negative values) or minutes late (positive values)
# # for the flight arrival
# edges_df <- create_edges(edge_from = nycflights13_day[,12],
#                          edge_to = nycflights13_day[,13],
#                          color = ifelse(nycflights13_day[,8] < 0,
#                                         "green", "red"))
#
# # Set the graph diagram's default attributes for...
#
# # ...nodes
# node_attrs <- c("style = filled", "fillcolor = lightblue",
#                 "color = gray", "shape = circle", "fontname = Helvetica",
#                 "width = 1")
#
# # ...edges
# edge_attrs <- c("arrowhead = dot")
#
# # ...and the graph itself
# graph_attrs <- c("layout = circo",
#                  "overlap = false",
#                  "fixedsize = true",
#                  "ranksep = 3",
#                  "outputorder = edgesfirst")
#
# # Generate the graph diagram in the RStudio Viewer.
# # The green lines show flights that weren't late (red indicates late arrivals)
# # This graph is for a single day of flights, airports that are unconnected on a
# # given day may be destinations on another day
# create_graph(nodes_df = nodes_df, edges_df = edges_df,
#              graph_attrs = graph_attrs, node_attrs = node_attrs,
#              edge_attrs = edge_attrs, directed = TRUE) %>>%
#   render_graph(width = 1200, height = 800)

# ----------------------------------------------------------------------------
# brms vs. lme4 + merTools::predictInterval(): prediction intervals compared
# ----------------------------------------------------------------------------
#
# Since merTools was written, brms has become a de-facto standard for fully
# Bayesian multilevel modeling. This script puts the fast, simulation-based
# intervals from `predictInterval()` next to the "gold standard" posterior
# intervals from brms across three studies:
#
#   STUDY 1  Gaussian random-slopes model (hsb: math achievement)
#            -- point estimates, interval agreement, out-of-sample coverage,
#               speed, and behaviour for entirely new groups (unseen schools).
#   STUDY 2  The same model WITHOUT the contextual fixed effect (meanses)
#            -- shows that predictInterval's new-group intervals get too narrow
#               (and the gap to brms widens) once a fixed effect no longer
#               absorbs the between-school variance.
#   STUDY 3  Binomial GLMM (grouseticks) compared on the probability scale
#            -- predictInterval(type = "probability") vs brms posterior_epred,
#               plus a calibration check.
#
# Requires: merTools (>= 1.0.0), lme4, brms, ggplot2.
#
# Output and the (cached) brms fits go to `MERTOOLS_BENCH_DIR` if set, else a
# per-session temp directory. Each brms fit is cached, so only the first run
# pays the Stan compile + sampling cost.
# ----------------------------------------------------------------------------

if (!requireNamespace("brms", quietly = TRUE)) {
  stop("This example needs the 'brms' package. install.packages('brms')")
}
suppressPackageStartupMessages({
  library(merTools)
  library(lme4)
  library(brms)
  library(ggplot2)
})

SEED       <- 11213
N_SIMS     <- 1000
NOMINAL    <- c(0.50, 0.80, 0.90, 0.95)
N_NEW_GRPS <- 6
TEST_FRAC  <- 0.20
BRM_ITER   <- 4000
BRM_WARMUP <- 1000

outdir <- Sys.getenv("MERTOOLS_BENCH_DIR", unset = "")
if (!nzchar(outdir)) outdir <- file.path(tempdir(), "brms_vs_predictInterval")
dir.create(outdir, showWarnings = FALSE, recursive = TRUE)
message("Output directory: ", outdir)
set.seed(SEED)

## ===========================================================================
## Shared helpers
## ===========================================================================

# Standardised brms fit with on-disk caching.
fit_brm <- function(form, data, family, cache) {
  brm(form, data = data, family = family, chains = 4, iter = BRM_ITER,
      warmup = BRM_WARMUP, seed = SEED, refresh = 0, silent = 2,
      control = list(adapt_delta = 0.95),
      file = file.path(outdir, cache), file_refit = "on_change")
}

# (N_SIMS x nrow(newdata)) matrix of predictive draws from predictInterval(),
# oriented like brms::posterior_predict() (draws x obs).
mt_draws <- function(fit, newdata, type = "linear.prediction",
                     include.resid.var = TRUE, new.levels = "zero") {
  pi <- suppressWarnings(predictInterval(
    fit, newdata = newdata, level = 0.95, n.sims = N_SIMS, seed = SEED,
    type = type, include.resid.var = include.resid.var,
    new.levels = new.levels, returnSims = TRUE))
  sims <- attr(pi, "sim.results")
  if (nrow(sims) == nrow(newdata)) sims <- t(sims)
  sims
}

qband <- function(draws, level) {
  a <- (1 - level) / 2
  data.frame(fit = apply(draws, 2, median),
             lwr = apply(draws, 2, function(x) unname(quantile(x, a))),
             upr = apply(draws, 2, function(x) unname(quantile(x, 1 - a))))
}

coverage <- function(draws, y, levels) {
  vapply(levels, function(p) {
    b <- qband(draws, p); mean(y >= b$lwr & y <= b$upr)
  }, numeric(1))
}

# Full report for one Gaussian study. Returns tables + draws (for figures).
report_gaussian <- function(label, m_lme, m_brm, test_seen, test_new) {
  cat("\n##########  ", label, "  ##########\n", sep = "")
  t_mt <- system.time({
    mt_seen <- mt_draws(m_lme, test_seen)
    mt_new  <- mt_draws(m_lme, test_new)
  })
  t_pp <- system.time({
    pp_seen <- posterior_predict(m_brm, newdata = test_seen, allow_new_levels = TRUE)
    pp_new  <- posterior_predict(m_brm, newdata = test_new, allow_new_levels = TRUE)
  })

  # 1. point estimates (conditional means)
  lme_mean <- predict(m_lme, newdata = test_seen)
  brm_mean <- colMeans(posterior_epred(m_brm, newdata = test_seen,
                                       allow_new_levels = TRUE))
  pe <- lme_mean - brm_mean
  cat(sprintf("\n[1] Point estimates  cor=%.4f  mean|diff|=%.4f  RMSE=%.4f  (y SD=%.2f)\n",
              cor(lme_mean, brm_mean), mean(abs(pe)), sqrt(mean(pe^2)),
              sd(test_seen[[all.vars(formula(m_lme))[1]]])))

  # 2. interval agreement (90%)
  b_mt <- qband(mt_seen, 0.90); b_brm <- qband(pp_seen, 0.90)
  cat(sprintf("[2] 90%% PI width  merTools=%.2f brms=%.2f  cor(lwr/upr)=%.3f/%.3f  mean|endpt diff|=%.3f\n",
              median(b_mt$upr - b_mt$lwr), median(b_brm$upr - b_brm$lwr),
              cor(b_mt$lwr, b_brm$lwr), cor(b_mt$upr, b_brm$upr),
              mean(abs(c(b_mt$lwr - b_brm$lwr, b_mt$upr - b_brm$upr)))))

  # 3. coverage (seen) and 4. new groups
  y_seen <- test_seen[[all.vars(formula(m_lme))[1]]]
  y_new  <- test_new[[all.vars(formula(m_lme))[1]]]
  cov_seen <- data.frame(nominal = NOMINAL,
                         merTools = coverage(mt_seen, y_seen, NOMINAL),
                         brms     = coverage(pp_seen, y_seen, NOMINAL))
  cat("[3] Coverage, held-out students in SEEN schools:\n")
  print(cov_seen, row.names = FALSE, digits = 3)

  # 4. New groups. The two packages each offer two conventions for an unseen
  #    level; the comparison is fair only when they match:
  #      predictInterval new.levels="zero" ~ brms re_formula=NA       (drop effect)
  #      predictInterval new.levels="draw" ~ brms allow_new_levels    (sample one)
  sd_school <- attr(lme4::VarCorr(m_lme)[[1]], "stddev")[["(Intercept)"]]
  mt_drawn <- mt_draws(m_lme, test_new, new.levels = "draw")
  pp_pop   <- posterior_predict(m_brm, newdata = test_new, re_formula = NA)
  w90 <- function(d) median(qband(d, 0.90)$upr - qband(d, 0.90)$lwr)
  cat(sprintf("[4] New groups (90%% interval): group SD=%.2f, residual SD=%.2f\n",
              sd_school, sigma(m_lme)))
  print(data.frame(
    unseen_group   = c("drop effect", "sample effect"),
    merTools_width = c(w90(mt_new),  w90(mt_drawn)),
    brms_width     = c(w90(pp_pop),  w90(pp_new)),
    merTools_cov90 = c(coverage(mt_new, y_new, 0.90),  coverage(mt_drawn, y_new, 0.90)),
    brms_cov90     = c(coverage(pp_pop, y_new, 0.90),  coverage(pp_new, y_new, 0.90))),
    row.names = FALSE, digits = 3)
  # coverage curves use the matched "sample" pairing (both draw a new effect)
  cov_new <- data.frame(nominal = NOMINAL,
                        merTools = coverage(mt_drawn, y_new, NOMINAL),
                        brms     = coverage(pp_new,   y_new, NOMINAL))

  invisible(list(point = data.frame(merTools = lme_mean, brms = brm_mean),
                 b_mt = b_mt, b_brm = b_brm,
                 cov_seen = cov_seen, cov_new = cov_new,
                 sd_school = sd_school,
                 timing = c(mt = t_mt[["elapsed"]], pp = t_pp[["elapsed"]])))
}

## ===========================================================================
## Data: hsb, with a train / test-seen / test-new split
## ===========================================================================
data(hsb, package = "merTools")
hsb$schid <- factor(hsb$schid)
new_schools <- sample(levels(hsb$schid), N_NEW_GRPS)
hsb$.set <- "train"
hsb$.set[hsb$schid %in% new_schools] <- "test_new"
seen_rows <- which(hsb$.set == "train")
hsb$.set[sample(seen_rows, round(TEST_FRAC * length(seen_rows)))] <- "test_seen"
train     <- droplevels(hsb[hsb$.set == "train", ])
test_seen <- hsb[hsb$.set == "test_seen" & hsb$schid %in% levels(train$schid), ]
test_new  <- hsb[hsb$.set == "test_new", ]
cat(sprintf("\nhsb split -> train %d / test_seen %d / test_new %d (%d new schools)\n",
            nrow(train), nrow(test_seen), nrow(test_new), length(new_schools)))

## ===========================================================================
## STUDY 1 -- Gaussian random slopes, WITH contextual fixed effect (meanses)
## ===========================================================================
f1 <- mathach ~ ses + meanses + (ses | schid)
t_lme1 <- system.time(m_lme1 <- lmer(f1, data = train))
brm1_cached <- file.exists(file.path(outdir, "brms_hsb_meanses.rds"))
t_brm1 <- system.time(m_brm1 <- fit_brm(f1, train, gaussian(), "brms_hsb_meanses"))
res1 <- report_gaussian("STUDY 1: ses + meanses + (ses|schid)",
                        m_lme1, m_brm1, test_seen, test_new)
cat("\n=== STUDY 1 timing (wall-clock) ===\n")
timing <- data.frame(
  step = c("lme4 fit", "predictInterval (seen+new)",
           "brms fit (compile+sample)", "posterior_predict (seen+new)"),
  seconds = c(t_lme1[["elapsed"]], res1$timing[["mt"]],
              t_brm1[["elapsed"]], res1$timing[["pp"]]))
print(timing, row.names = FALSE, digits = 3)
mt_total  <- t_lme1[["elapsed"]] + res1$timing[["mt"]]
brm_total <- t_brm1[["elapsed"]] + res1$timing[["pp"]]
if (brm1_cached) {
  cat("(brms fit was loaded from cache; delete it to time a fresh compile + sample.)\n")
} else {
  cat(sprintf("end-to-end: merTools %.1fs (lme4 + predictInterval) vs brms %.1fs  (~%.0fx faster)\n",
              mt_total, brm_total, brm_total / mt_total))
}

## ===========================================================================
## STUDY 2 -- same model WITHOUT meanses: the new-group contrast
## ===========================================================================
f2 <- mathach ~ ses + (ses | schid)
m_lme2 <- lmer(f2, data = train)
m_brm2 <- fit_brm(f2, train, gaussian(), "brms_hsb_nomeanses")
res2 <- report_gaussian("STUDY 2: ses + (ses|schid)  [no meanses]",
                        m_lme2, m_brm2, test_seen, test_new)

cat("\n=== Unseen-group width: drop vs sample the group effect ===\n")
w90 <- function(d) median(qband(d, 0.90)$upr - qband(d, 0.90)$lwr)
contrast <- data.frame(
  model = c("with meanses", "without meanses"),
  school_SD  = c(res1$sd_school, res2$sd_school),
  width_zero = c(w90(mt_draws(m_lme1, test_new, new.levels = "zero")),
                 w90(mt_draws(m_lme2, test_new, new.levels = "zero"))),
  width_draw = c(w90(mt_draws(m_lme1, test_new, new.levels = "draw")),
                 w90(mt_draws(m_lme2, test_new, new.levels = "draw"))))
contrast$diff <- contrast$width_draw - contrast$width_zero
print(contrast, row.names = FALSE, digits = 3)
cat("Under matched conventions merTools and brms agree for unseen groups (see\n",
    "each study's [4] table). What meanses changes is how far the two\n",
    "conventions sit apart: when the contextual fixed effect absorbs the\n",
    "between-school variance, dropping vs sampling the school effect barely\n",
    "matters; without it the school SD roughly doubles and the choice moves the\n",
    "interval noticeably.\n", sep = "")

## ===========================================================================
## STUDY 3 -- Binomial GLMM (grouseticks), compared on the probability scale
## ===========================================================================
data(grouseticks, package = "lme4")
grouseticks$TICKS_BIN <- as.integer(grouseticks$TICKS >= 1)
grouseticks$cHEIGHT   <- as.numeric(scale(grouseticks$HEIGHT))
gk <- grouseticks
gk$.set <- ifelse(runif(nrow(gk)) < TEST_FRAC, "test", "train")
gk_tr <- droplevels(gk[gk$.set == "train", ])
gk_te <- gk[gk$.set == "test" & gk$BROOD %in% levels(gk_tr$BROOD)
            & gk$LOCATION %in% unique(gk_tr$LOCATION), ]

gf <- TICKS_BIN ~ YEAR + cHEIGHT + (1 | BROOD) + (1 | LOCATION)
g_lme <- glmer(gf, family = binomial, data = gk_tr,
               control = glmerControl(optimizer = "bobyqa"))
g_brm <- fit_brm(gf, gk_tr, bernoulli(), "brms_grouseticks")

# probability draws (no Bernoulli noise): predictInterval prob vs brms epred
mt_pr <- mt_draws(g_lme, gk_te, type = "probability", include.resid.var = FALSE)
if (max(mt_pr) > 1 || min(mt_pr) < 0) mt_pr <- plogis(mt_pr)  # guard: ensure prob scale
brm_pr <- posterior_epred(g_brm, newdata = gk_te, allow_new_levels = TRUE)

cat("\n##########  STUDY 3: grouseticks binomial GLMM (probability scale)  ##########\n")
mt_p <- apply(mt_pr, 2, median); brm_p <- apply(brm_pr, 2, median)
cat(sprintf("[1] Predicted probabilities  cor=%.4f  mean|diff|=%.4f\n",
            cor(mt_p, brm_p), mean(abs(mt_p - brm_p))))
bp_mt <- qband(mt_pr, 0.90); bp_brm <- qband(brm_pr, 0.90)
cat(sprintf("[2] 90%% interval on P(tick)  cor(lwr/upr)=%.3f/%.3f  mean|endpt diff|=%.4f\n",
            cor(bp_mt$lwr, bp_brm$lwr), cor(bp_mt$upr, bp_brm$upr),
            mean(abs(c(bp_mt$lwr - bp_brm$lwr, bp_mt$upr - bp_brm$upr)))))
# calibration: bin predicted prob, compare to observed frequency
calib <- function(p, y, k = 5) {
  b <- cut(p, quantile(p, seq(0, 1, length.out = k + 1)), include.lowest = TRUE)
  data.frame(bin = seq_len(nlevels(b)),
             pred = tapply(p, b, mean), obs = tapply(y, b, mean),
             n = as.integer(table(b)))
}
cat("[3] Calibration (predicted vs observed P(tick) by quintile):\n")
cal_mt  <- calib(mt_p,  gk_te$TICKS_BIN); cal_mt$method  <- "merTools"
cal_brm <- calib(brm_p, gk_te$TICKS_BIN); cal_brm$method <- "brms"
print(rbind(cal_mt, cal_brm)[, c("method", "bin", "pred", "obs", "n")],
      row.names = FALSE, digits = 3)

## ===========================================================================
## Figures
## ===========================================================================
ggsave2 <- function(p, nm) ggsave(file.path(outdir, paste0("brms_compare_", nm, ".png")),
                                  p, width = 7, height = 4.5, dpi = 120)

# point estimates (study 1)
ggsave2(ggplot(res1$point, aes(merTools, brms)) +
  geom_abline(slope = 1, intercept = 0, linetype = 2, color = "grey50") +
  geom_point(alpha = 0.3, size = 0.8, color = "#1B9E77") + coord_equal() +
  labs(title = "Point estimates are essentially identical",
       subtitle = "Conditional mean per held-out student (STUDY 1)",
       x = "lme4 predict()", y = "brms posterior_epred()") +
  theme_minimal(base_size = 12), "point")

# coverage curves, both Gaussian studies x (seen/new)
cov_all <- do.call(rbind, list(
  transform(res1$cov_seen, study = "with meanses",    grp = "seen"),
  transform(res1$cov_new,  study = "with meanses",    grp = "new"),
  transform(res2$cov_seen, study = "without meanses", grp = "seen"),
  transform(res2$cov_new,  study = "without meanses", grp = "new")))
cov_long <- rbind(
  data.frame(cov_all[c("nominal", "study", "grp")], empirical = cov_all$merTools, method = "merTools"),
  data.frame(cov_all[c("nominal", "study", "grp")], empirical = cov_all$brms,     method = "brms"))
ggsave2(ggplot(cov_long, aes(nominal, empirical, color = method)) +
  geom_abline(slope = 1, intercept = 0, linetype = 2, color = "grey50") +
  geom_line() + geom_point() + facet_grid(study ~ grp) +
  coord_equal(xlim = c(0.4, 1), ylim = c(0.4, 1)) +
  labs(title = "Calibration: nominal vs. empirical coverage",
       x = "nominal level", y = "empirical coverage") +
  theme_minimal(base_size = 12) + theme(legend.position = "top"), "cover")

# GLMM probability agreement
ggsave2(ggplot(data.frame(merTools = mt_p, brms = brm_p), aes(merTools, brms)) +
  geom_abline(slope = 1, intercept = 0, linetype = 2, color = "grey50") +
  geom_point(alpha = 0.4, color = "#7570B3") + coord_equal(xlim = c(0,1), ylim = c(0,1)) +
  labs(title = "GLMM: predicted P(tick) agrees with brms",
       subtitle = "grouseticks binomial model, held-out observations",
       x = "merTools predictInterval(type='probability')",
       y = "brms posterior_epred()") +
  theme_minimal(base_size = 12), "glmm")

cat("\nFigures written to ", outdir, "\n", sep = "")

# Import variables passed in from shinyMer()
library(ggplot2)
library(shiny)

merMod <<- .shinyMerPar$merMod
if (is.null(.shinyMerPar$simData)) {

} else {
  newdata <<- .shinyMerPar$simData
}

if (!exists("newdata")) {
  df.choices <- c("Model Frame"   = "orig",
                  "Random Obs"    = "rand",
                  "Average Obs"   = "mean")
} else {
  df.choices <- c("User Supplied" = "user",
                  "Model Frame"   = "orig",
                  "Random Obs"    = "rand",
                  "Average Obs"   = "mean")
}

# Variables in the model frame that the user can restrict a random/average draw
# to (issue #32). Used to build the optional subset controls in the sidebar.
subset.choices <- c("(use all data)" = "", names(merMod@frame))

# #This file compares the prediction interval from the "Correct" bootmer method
# #to our quick and dirty method to see how they differ.
#
# #I renamed some things to ease my typing
#
# #Prep R####
#  #rm(list=ls())
#  library(lme4)
#  library(arm)
#  library(mvtnorm)
#  library(dplyr)
#  library(tidyr)
#  library(ggplot2)
#  library(knitr)
#  library(RPushbullet)
#
#  set.seed(51315)
#  data(InstEval)
#  data(sleepstudy)
#  data(VerbAgg)
#
#  #Cannonical Models for testing purposes
#  ##1) Sleepstudy eg - lmer with random slope and random intercept
#    m1.form <- Reaction ~ Days + (Days | Subject)
#    m1.df <- sleepstudy
#    m1.new.df <- m1.df
#  ##2) Verbal Aggression eg - lmer (could to glmer(logit)) with 2 levels
#    m2.form <- r2 ~ (Anger + Gender + btype + situ)^2 + (1|id) + (1|item)
#    m2.df <- VerbAgg
#    m2.new.df <- m2.df
#  ##3) Verbal Aggression eg - glmer(logit) with 2 levels
#    m3.form <- r2 ~ (Anger + Gender + btype + situ)^2 + (1|id) + (1|item)
#    m3.df <- VerbAgg
#    m3.new.df <- m3.df
#  ##4) Instructer Evaluation et - lmer cross-classified, with two different random slopes
#    m4.form <- y ~ studage + lectage + service + dept + (studage|s) + (lectage|d)
#    m4.df <- InstEval
#    m4.new.df <- m4.df
#
# #Step 1: Our method ####
# #  predictInterval <- function(model, newdata, level = 0.95, nsim=1000, stat="median", predict.type="link"){
# #    #Depends
# #    require(mvtnorm)
# #    require(lme4)
# #    #Prep Output
# #    outs <- newdata
# #
# #    #Sort out all the levels
# #    reTerms <- names(ngrps(model))
# #    n.reTerms = length(reTerms)
# #
# #    ##The following 3 lines produce a matrix of linear predictors created from the fixed coefs
# #    betaSim <- rmvnorm(nsim, mean = fixef(model), sigma = as.matrix(vcov(model)))
# #    newdata.modelMatrix <- lFormula(formula = model@call, data=newdata)$X
# #    fixed.xb <- newdata.modelMatrix %*% t(betaSim)
# #
# #    ##Draw from random effects distributions for each level and merge onto newdata
# #    reSim <- NULL
# #    for (j in seq_along(reTerms)) {
# #      group=reTerms[j]
# #      reMeans <- array(ranef(model)[[group]])
# #      reMatrix <- attr(ranef(model, condVar=TRUE)[[group]], which = "postVar")
# #      reSim[[group]] <- data.frame(rownames(reMeans), matrix(NA, nrow=nrow(reMeans), ncol=nsim))
# #      colnames(reSim[[group]]) <- c(group, paste("sim", 1:nsim, sep=""))
# #      for (k in 1:nrow(reMeans)) {
# #        lvl = rownames(reMeans)[k]
# #        reSim[[group]][k,2:ncol(reSim[[group]])] <- rmvnorm(nsim, mean=as.matrix(reMeans[k,]), sigma=as.matrix(reMatrix[,,k]))
# #      }
# #      cnames <- colnames(reSim[[group]])
# #      reSim[[group]] <- merge(newdata, reSim[[group]], by=group, all.x=TRUE)
# #      reSim[[group]] <- as.matrix(reSim[[group]][,setdiff(cnames, group)])
# #    }
# #
# #    #Calculate yhat as sum of components
# #    yhat <- fixed.xb + apply(simplify2array(reSim), c(1,2), sum)
# #
# #    #Output prediction intervals
# #    if (stat=="median") {
# #      outs$fit <- apply(yhat,1,function(x) as.numeric(quantile(x, .5)))
# #    }
# #    if (stat=="mean") {
# #      outs$fit <- apply(yhat,1,mean)
# #    }
# #    outs$upr <- apply(yhat,1,function(x) as.numeric(quantile(x, 1 - ((1-level)/2))))
# #    outs$lwr <- apply(yhat,1,function(x) as.numeric(quantile(x, ((1-level)/2))))
# #    if (predict.type=="response") {
# #      outs$fit <- model@resp$family$linkinv(outs$fit)
# #      outs$upr <- model@resp$family$linkinv(outs$upr)
# #      outr$lwr <- model@resp$family$linkinc(outs$lwr)
# #    }
# #    #Close it out
# #    return(outs)
# #  }
#
# #Step 2: Unit Test Function####
#  predictInterval.test <- function(model.form, model.df, model.type="lmer",
#                                   predict.df, predict.type="link", stat="median",
#                                   idvar=NULL, nSims=1000, ...) {
#    require(lme4); require(dplyr); require(tidyr); require(ggplot2);
#    ##Estimate model
#    if (model.type=="lmer") {
#      modelEstimation.time <- system.time(
#        m1  <- lmer(model.form, data=model.df, ...)
#      )
#    }
#    if (model.type=="glmer") {
#      modelEstimation.time <- system.time(
#        m1  <- glmer(model.form, data=model.df, ...)
#      )
#    }
#    if (model.type=="blmer") {
#      modelEstimation.time <- system.time(
#        m1  <- blmer(model.form, data=model.df, ...)
#      )
#    }
#    if (model.type=="bglmer") {
#      modelEstimation.time <- system.time(
#        m1  <- bglmer(model.form, data=model.df, ...)
#      )
#    }
#    ##If it does not have one, add unique identifier to predict.df
#    if (is.null(idvar)) {
#      predict.df$.newID <- paste("newID", rownames(predict.df), sep="")
#    }
#    ##Functions for bootMer() and objects
#    ####Return predicted values from bootstrap
#    mySumm <- function(.) {
#      predict(., newdata=predict.df, re.form=NULL, type=predict.type)
#    }
#    ####Collapse bootstrap into median, 95% PI
#    sumBoot <- function(merBoot, stat) {
#      if (stat=="median") {
#         fit = apply(merBoot$t, 2, function(x) as.numeric(quantile(x, probs=.5, na.rm=TRUE)))
#      }
#      if (stat=="mean") {
#         fit = apply(merBoot$t, 2, function(x) mean(x, na.rm=TRUE))
#      }
#      return(
#        data.frame(merBoot$data,
#                   fit = fit,
#                   lwr = apply(merBoot$t, 2, function(x) as.numeric(quantile(x, probs=.025, na.rm=TRUE))),
#                   upr = apply(merBoot$t, 2, function(x) as.numeric(quantile(x, probs=.975, na.rm=TRUE))),
#                   colNAs = apply(merBoot$t, 2, function(x) sum(is.na(x)))
#        )
#      )
#    }
#    ##Bootstrap
#    ####Method 1: parametric, re-estimate BLUPS
#    cat("\n Bootstrap Method 1")
#    boot1.time <- system.time(
#      boot1 <- bootMer(m1, mySumm, nsim=nSims,
#                          use.u=FALSE, type="parametric",
#                          .progress = "txt", PBargs=list(style=3))
#    )
#    ####Method 2: parametric, conditional on estimated BLUPS
#    cat("\n Bootstrap Method 2")
#    boot2.time <- system.time(
#      boot2 <- bootMer(m1, mySumm, nsim=nSims,
#                          use.u=TRUE, type="parametric",
#                          .progress = "txt", PBargs=list(style=3))
#    )
#    ####Method 3: semiparametric (draw from resid), conditional on estimated BLUPS
#    cat("\n Bootstrap Method 3")
#    boot3.time <- system.time(
#      boot3 <- bootMer(m1, mySumm, nsim=nSims,
#                          use.u=TRUE, type="semiparametric",
#                          .progress = "txt", PBargs=list(style=3))
#    )
#    ##Our Method
#    kf.time <- system.time(
#      kf.method <- predictInterval(m1, predict.df)
#    )
#    ##Compare Times
#    compare.time <- rbind(modelEstimation.time, kf.time, boot1.time, boot2.time, boot3.time)
#
#    ##Summarize and compare results
#    boot1.sum <- sumBoot(boot1, stat=stat)
#    boot2.sum <- sumBoot(boot2, stat=stat)
#    boot3.sum <- sumBoot(boot3, stat=stat)
#
#    eval <- merge(predict.df, boot1.sum) %>%
#      rename(boot1.fit=fit, boot1.lwr=lwr, boot1.upr=upr)
#    eval <- merge(eval, boot2.sum) %>%
#      rename(boot2.fit=fit, boot2.lwr=lwr, boot2.upr=upr)
#    eval <- merge(eval, boot3.sum) %>%
#      rename(boot3.fit=fit, boot3.lwr=lwr, boot3.upr=upr)
#    eval <- merge(eval, kf.method) %>%
#      rename(KF.fit=fit, KF.lwr=lwr, KF.upr=upr)
#
#    #Check if nrow(eval) still equals nrow(predict.df) because it should
#    if (nrow(eval)!=nrow(predict.df)) {
#      stop("Something happened when merging bootstrap summaries together ...")
#    }
#
#    ##Add lmer yhats (predict.merMod) on there
#    eval$with.u <- predict(m1, newdata=predict.df, re.form=NULL, type=predict.type)
#    eval$no.u <- predict(m1, newdata=predict.df, re.form=NA, type=predict.type)
#
#    ##Create summary statistics
#    piCoveragePct <- function(ref.upr, ref.lwr, new.upr, new.lwr) {
#      pct <- ifelse(ref.upr < new.lwr | ref.lwr > new.upr, 0,
#             ifelse(ref.upr < new.upr & ref.lwr > new.lwr, 1,
#             ifelse(ref.upr > new.upr & ref.lwr < new.lwr, (new.upr-new.lwr)/(ref.upr-ref.lwr),
#             ifelse(ref.upr < new.upr, (ref.upr-new.lwr)/(ref.upr-ref.lwr),
#             ifelse(ref.lwr > new.lwr, (new.upr-ref.lwr)/(ref.upr-ref.lwr), NA)))))
#      return(pct)
#    }
#    ###Pct of other PI covered by KF PI
#    eval$boot1.coverage <- piCoveragePct(eval$boot1.upr, eval$boot1.lwr, eval$KF.upr, eval$KF.lwr)
#    eval$boot2.coverage <- piCoveragePct(eval$boot2.upr, eval$boot2.lwr, eval$KF.upr, eval$KF.lwr)
#    eval$boot3.coverage <- piCoveragePct(eval$boot3.upr, eval$boot3.lwr, eval$KF.upr, eval$KF.lwr)
#    ###Does KF PI contain point estimates
#    eval$contain.boot1  <- piCoveragePct(eval$boot1.fit, eval$boot1.fit, eval$KF.upr, eval$KF.lwr)
#    eval$contain.boot2  <- piCoveragePct(eval$boot2.fit, eval$boot2.fit, eval$KF.upr, eval$KF.lwr)
#    eval$contain.boot3  <- piCoveragePct(eval$boot3.fit, eval$boot3.fit, eval$KF.upr, eval$KF.lwr)
#    eval$contain.with.u <- piCoveragePct(eval$with.u   , eval$with.u   , eval$KF.upr, eval$KF.lwr)
#    eval$contain.no.u   <- piCoveragePct(eval$no.u     , eval$no.u     , eval$KF.upr, eval$KF.lwr)
#    ###Point estimate "bias"
#    eval$distance.boot1  <- eval$KF.fit - eval$boot1.fit
#    eval$distance.boot2  <- eval$KF.fit - eval$boot2.fit
#    eval$distance.boot3  <- eval$KF.fit - eval$boot3.fit
#    eval$distance.with.u <- eval$KF.fit - eval$with.u
#    eval$distance.no.u   <- eval$KF.fit - eval$no.u
#
#    ##Close it out
#    return(
#      list(
#        compareTimes=compare.time,
#        bootstraps = list(boot1, boot2, boot3),
#        kf.method = kf.method,
#        model=m1,
#        evalData = eval
#        )
#      )
#  }
#
# #Step 2b: Post processing predictInterval.test()
#  PIE.graphics <- function(eval.df, response, grouping.factors, seed=314) {
#    require(dplyr); require(tidyr); require(ggplot2); require(grid); require(gridExtra);
#    ####Data prep
#    set.seed(seed)
#    eval$random <- runif(nrow(eval))
#    Eval <- eval %>%
#      mutate(
#        random = row_number(random)
#      ) %>%
#      gather(var, value, starts_with("boot"), starts_with("KF")) %>%
#      mutate(
#        simMethod = sub("[[:punct:]][0-9A-Za-z]*", "", var),
#        stat = sub("[0-9A-Za-z]*[[:punct:]]", "", var)
#      ) %>%
#      select(-var) %>%
#      spread(stat, value) %>%
#      group_by(.newID) %>%
#      arrange(simMethod) %>%
#      mutate(
#        x=row_number(simMethod),
#        x=(x-mean(x))/10
#      ) %>%
#      group_by(simMethod) %>%
#      mutate(
#        x=row_number(random)+x
#      )
#
#    ####Direct Comparison Plot
#    Eval.small <- arrange(Eval, random) %>% filter(random<=30) %>% arrange(x)
#    p1 <- ggplot(aes(y=fit,x=x, color=simMethod), data=Eval.small) +
#            geom_point(size=I(3)) +
#            geom_linerange(aes(ymax=upr, ymin=lwr), size=I(1)) +
#            geom_point(shape="with.u", color="black", size=I(4),
#                       data=summarize(group_by(Eval.small, .newID), x=mean(x), fit=mean(with.u))) +
#            geom_point(shape="no.u", color="black", size=I(4),
#                       data=summarize(group_by(Eval.small, .newID), x=mean(x), fit=mean(no.u))) +
#            theme_bw() +
#            labs(x="Index", y="Prediction Interval", title="95% Prediction interval by method for 30 random obs") +
#            scale_x_discrete(breaks=1:30, labels=Eval.small$.newID[Eval.small$simMethod=="KF"]) +
#            theme(axis.text.x = element_text(angle=90))
#
#    #####Distribution of fitted values
#    mean.statment <- paste("mean(",response,")", sep="")
#    p2 <- ggplot(aes(x=fit), data=Eval) +
#            geom_density(aes(color=simMethod)) +
#            geom_vline(x=mean(unlist(Eval[,response]))) +
#            geom_density(data=summarize_(
#                                group_by_(Eval, grouping.factors),
#                                fit = mean.statment), color="black") +
#            theme_bw() +
#            labs(x = "Estimate", y="Density",
#                 title="Average point estimates by prediction type \n (vertical black line is sample grand mean, \n black density is distribution of sample group means)")
#
#    #####Bar Graph of KF PI containing other point estimates
#    p3 <- eval %>% select(starts_with("contain.")) %>%
#      gather(simMethod, value) %>%
#      mutate(
#        simMethod = sub("[[:alpha:]]*.", "", simMethod)
#      ) %>%
#      group_by(simMethod) %>%
#      summarize(
#        value=100*mean(value)
#      ) %>%
#      qplot(value, x=simMethod, fill=simMethod, data=., geom="bar", position="dodge", stat="identity") +
#        labs(y="Percent", title="Percent of observations in which our P.I. contains other point estimates") +
#        theme_bw()
#
#    ##Summarizing Bias
#    SD <- sd(eval[,response])
#    bias.data <- eval %>% select(starts_with("distance.")) %>%
#      gather(simMethod, value) %>%
#      mutate(
#        simMethod = sub("[[:alpha:]]*.", "", simMethod),
#        value = value/SD,
#        ymax = max(density(value)$y),
#        xmin = min(value)
#      ) %>%
#      group_by(simMethod) %>%
#      summarize(
#        mean.Distance = round(mean(value),4),
#        mad.Distance = round(mad(value),2),
#        ymax = max(ymax),
#        xmin = min(xmin)
#      )
#    bias.xmin <- min(bias.data$xmin, na.rm=TRUE)
#    bias.xmax <- 0.2 * bias.xmin
#    bias.ymax <- max(bias.data$ymax, na.rm=TRUE)
#    bias.ymin <- .5 * bias.ymax
#    bias.data <- bias.data[,1:3]
#
#    p4a <- eval %>% select(starts_with("distance.")) %>%
#      gather(simMethod, value) %>%
#      mutate(
#        simMethod = sub("[[:alpha:]]*.", "", simMethod),
#        value = value/SD
#      ) %>%
#      qplot(x=value, color=simMethod, data=., geom="density") +
#        labs(x="Standard deviations of response variable") +
#        theme_bw()
#
#    p4b <- tableGrob(bias.data, show.rownames = FALSE)
#    p4 <- arrangeGrob(p4a, p4b, ncol=1, main="Distribution of distance from KF point estimates to other point estimates")
#
#
#    ##Summarizing Bias
#    coverage.data <- eval %>%
#      select(ends_with(".coverage")) %>%
#      gather(simMethod, value) %>%
#      mutate(
#        simMethod = gsub(".coverage","", simMethod, fixed=TRUE),
#        TotalCoverage=value==1,
#        Cov.90to100 = value>0.9 & value < 1,
#        Cov.80to90 = value>0.8 & value <= 0.9,
#        Cov.50to80 = value>0.5 & value <= 0.8,
#        Cov.0to50 = value>0 & value <= 0.5,
#        ZeroCoverage= value==0
#      ) %>%
#      group_by(simMethod) %>%
#      summarize(
#        TotalCoverage= round(100*sum(TotalCoverage)/n(),1),
#        Cover.90to100 =  round(100*sum(Cov.90to100)/n(),1),
#        Cover.80to90 = round(100*sum(Cov.80to90)/n(),1),
#        Cover.50to80 = round(100*sum(Cov.50to80)/n(),1),
#        Cover.0to50 = round(100*sum(Cov.0to50)/n(),1),
#        ZeroCoverage = round(100*sum(ZeroCoverage)/n(),1)
#      )
#    p5a <- eval %>%
#      select(ends_with(".coverage")) %>%
#      gather(simMethod, value) %>%
#      mutate(
#        simMethod = gsub(".coverage","", simMethod, fixed=TRUE)
#      ) %>%
#      ggplot(data=., aes(x=100*value, fill=simMethod)) +
#        geom_bar(aes(y=300*(..count..)/sum(..count..)), binwidth=5) +
#        facet_wrap(~simMethod, ncol=3) +
#        labs(y="Percent of Observations", x="Coverage Percentage") +
#        theme_bw()
#    p5b <- tableGrob(coverage.data, show.rownames = FALSE)
#    p5 <- arrangeGrob(p5a, p5b, ncol=1, main="Distribution of PI coverage percentages")
#
#    ##Wrap-up
#    return(
#      list(
#        CompareRandomObs=p1,
#        FitDistributions=p2,
#        PointEstimateCoverage=p3,
#        BiasSummary=p4,
#        CoverageSummary=p5
#        )
#      )
#  }
# # PIE.graphics(cannonical.1$evalData, response="Reaction", grouping.factors = "Subject")
#
#
# #Step 3: Summarize and Compare Results####
#  #debug(predictInterval.test)
#  cannonical.1 <-  predictInterval.test(model.form = m1.form, model.df = m1.df, predict.df = m1.new.df, nSims=2500)
#  pbPost("note", title="Finished Cannonical Model 1", body=paste(kable(cannonical.1$compareTimes[,1:3]), collapse ="\n"))
#  cannonical.2 <-  predictInterval.test(model.form = m2.form, model.df = m2.df, predict.df = m2.new.df, nSims=2500)
#  pbPost("note", title="Finished Cannonical Model 2", body=paste(kable(cannonical.2$compareTimes[,1:3]), collapse ="\n"))
#  cannonical.4 <-  predictInterval.test(model.form = m4.form, model.df = m4.df, predict.df = m4.new.df, nSims=2500)
#  pbPost("note", title="Finished Cannonical Model 4", body=paste(kable(cannonical.4$compareTimes[,1:3]), collapse ="\n"))
#  cannonical.3 <-  predictInterval.test(model.form = m3.form, model.df = m3.df, model.type="glmer",
#                                        predict.df=m3.new.df, predict.type="response", nSims=2500)
#  pbPost("note", title="Finished Cannonical Model 3", body=paste(kable(cannonical.3$compareTimes[,1:3]), collapse ="\n"))
#
#
#  # save.image()
#
# #Step 4: checking math with a fine-toothed comb ####
#  ##Pull pieces to create toy example
#  KF <- cannonical.1$kf.method
#  model <- cannonical.1$model
#  nsim=5

# Global variables
utils::globalVariables(c(".shinyMerPar", "sig", "sigma", "Lind", "group",
                         "est", "mean_est", "est_ss", "within_var", "between_var",
                         "statistic"))

#' @importFrom methods as is
#' @importFrom stats AIC as.formula formula logLik median model.matrix na.omit
#' pnorm qnorm quantile residuals rgamma rnorm sd vcov weighted.mean delete.response
#' model.frame na.pass reformulate runif terms getCall
#' @importFrom utils packageVersion
zzz <- function(){
  # Nothing

}


#' Title
#'
#' @param object a merMod object
#' @param correlation optional p value
#' @param use.hessian logical
#' @param ... additional arguments to pass through
#'
#' @return a summary of the object
sum.mm <- function(object,
                           correlation = (p <= getOption("lme4.summary.cor.max")),
                           use.hessian = NULL,
                           ...)
{
  if (length(list(...)) > 0) {
    ## FIXME: need testing code
    warning("additional arguments ignored")
  }
  ## se.calc:
  hess.avail <- (!is.null(h <- object@optinfo$derivs$Hessian) &&
                   nrow(h) > length(getME(object,"theta")))
  if (is.null(use.hessian)) use.hessian <- hess.avail
  if (use.hessian && !hess.avail)
    stop("'use.hessian=TRUE' specified, but Hessian is unavailable")
  resp <- object@resp
  devC <- object@devcomp
  dd <- devC$dims
  ## cmp <- devC$cmp
  useSc <- as.logical(dd[["useSc"]])
  sig <- sigma(object)
  ## REML <- isREML(object)

  famL <- famlink(resp = resp)
  p <- length(coefs <- fixef(object))

  vc <- as.matrix(vcov(object, use.hessian = use.hessian))
  stdError <- sqrt(diag(vc))
  coefs <- cbind("Estimate" = coefs,
                 "Std. Error" = stdError)
  if (p > 0) {
    coefs <- cbind(coefs, (cf3 <- coefs[,1]/coefs[,2]), deparse.level = 0)
    colnames(coefs)[3] <- paste(if(useSc) "t" else "z", "value")
    if (isGLMM(object)) # FIXME: if "t" above, cannot have "z" here
      coefs <- cbind(coefs, "Pr(>|z|)" =
                       2*pnorm(abs(cf3), lower.tail = FALSE))
  }

  llAIC <- llikAIC(object)
  ## FIXME: You can't count on object@re@flist,
  ##	      nor compute VarCorr() unless is(re, "reTrms"):
  varcor <- VarCorr(object)
  # use S3 class for now
  structure(list(methTitle = methTitle(dd),
                 objClass = class(object),
                 devcomp = devC,
                 isLmer = is(resp, "lmerResp"), useScale = useSc,
                 logLik = llAIC[["logLik"]],
                 family = famL$family, link = famL$link,
                 ngrps = ngrps(object),
                 coefficients = coefs, sigma = sig,
                 vcov = vcov(object, correlation = correlation, sigm = sig),
                 varcor = varcor, # and use formatVC(.) for printing.
                 AICtab = llAIC[["AICtab"]], call = object@call,
                 residuals = residuals(object,"pearson",scaled = TRUE),
                 fitMsgs = fetch.merMod.msgs(object),
                 optinfo = object@optinfo
  ), class = "summary.merMod")
}

#' Find link function family
#'
#' @param object a merMod object
#' @param resp the response vector
#'
#' @return the link function and family
famlink <- function(object, resp = object@resp) {
  if(is(resp, "glmResp"))
    resp$family[c("family", "link")]
  else list(family = NULL, link = NULL)
}


##' Extract all warning msgs from a merMod object
##'
##' @param x a merMod object
fetch.merMod.msgs <- function(x) {
  ## currently only those found with 'X' :
  aX <- attributes(x@pp$X)
  wmsgs <- grep("^msg", names(aX))
  if(any(has.msg <- nchar(Xwmsgs <- unlist(aX[wmsgs])) > 0))
    Xwmsgs[has.msg]
  else
    character()
}


##' Extract all warning msgs from a merMod object
##' @param type check a fixed or random effect
##' @inheritParams plotREsim
plot_sim_error_chks <- function(type= c("FE", "RE"), level = 0.95,
                                stat = c("mean", "median"),
                                sd = TRUE, sigmaScale = NULL,
                                oddsRatio = FALSE, labs = FALSE, facet= TRUE) {

  if (level <= 0 | level >= 1) stop("level must be specified as a numeric in (0,1).")
  stat <- match.arg(stat, several.ok= FALSE)
  if (!is.logical(sd)) stop("sd must be a logical expression.")
  if (!is.null(sigmaScale) && !is.logical(sigmaScale)) stop("sigmaScale must be a logical expression.")
  if (!is.logical(oddsRatio)) stop("oddsRatio must be a logical expression.")
  if (!is.logical(labs)) stop("labs must be a logical expression.")
  if (!is.logical(facet)) {
    if(any(c(!is.list(facet), is.null(names(facet)),
             names(facet) != c("groupFctr", "term"))))
      stop("facet must be either a logical expression or a named list.")
  }
}

# Imputation functions
#' Extract model information from a merMod
#'
#' @param object a merMod object
#'
#' @return Simple summary information about the object, number
#' of observations, number of grouping terms, AIC, and residual standard deviation
#' @export
#' @examples
#' \donttest{
#' sim_list <- replicate(n = 10,
#'         expr = sleepstudy[sample(row.names(sleepstudy), 180),],
#'         simplify=FALSE)
#' fml <- "Reaction ~ Days + (Days | Subject)"
#' mod <- lmerModList(fml, data = sim_list)
#' modelInfo(mod[[1]])
#' lapply(mod, modelInfo)
#' }
modelInfo <- function(object){
  ngrps <- lapply(object@flist, function(x) length(levels(x)))
  out <- data.frame("n.obs" = getME(object, "devcomp")$dims["n"],
                    "n.lvls" = length(ngrps),
                    "AIC" = AIC(object),
                    "sigma" = sigma(object))
  row.names(out) <- NULL
  return(out)
}

# Functions to extract standard deviation of random effects from model
#' Extract the standard deviation of the random effects from a merMod object
#'
#' @param model an object that inherits from class merMod
#'
#' @return a numeric vector for standard deviations of the random effects
#' @export
#' @examples
#' fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
#' REsdExtract(fm1)
REsdExtract <- function(model){
  out <- unlist(lapply(VarCorr(model), attr, "stddev"))
  return(out)
}

#' Extract the correlations between the slopes and the intercepts from a model
#'
#' @param model an object that inherits from class merMod
#'
#' @return a numeric vector of the correlations among the effects
#' @export
#' @examples
#' fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
#' REcorrExtract(fm1)
REcorrExtract <- function(model){
  out <- unlist(lapply(VarCorr(model), attr, "corre"))
  return(min(unique(out)))
}


#' Extract data.frame of random effect statistics from merMod List
#'
#' @param modList a list of multilevel models
#'
#' @return a data.frame
#' @import dplyr
#' @importFrom broom.mixed tidy
#' @export
#' @examples
#' \donttest{
#' sim_list <- replicate(n = 10,
#'         expr = sleepstudy[sample(row.names(sleepstudy), 180),],
#'         simplify=FALSE)
#' fml <- "Reaction ~ Days + (Days | Subject)"
#' mod <- lmerModList(fml, data = sim_list)
#' modelRandEffStats(mod)
#' }
modelRandEffStats <- function(modList){
  effList <- lapply(modList, tidy, effects = "ran_pars")
  effList <- do.call(rbind, effList)
  out <- effList %>% group_by(term, group) %>%
    summarize(est = mean(estimate),
              std.error = sd(estimate)) %>%
    rename(estimate = est)
  return(as.data.frame(out))
}

#' Extract averaged fixed effect parameters across a list of merMod objects
#'
#' @param modList an object of class merModList
#' @param ... additional arguments to pass to \code{\link{tidy}}
#'
#' @return a data.frame of the averaged fixed effect parameters
#' @details The Rubin correction for combining estimates and standard errors from
#' Rubin (1987) is applied to adjust for the within and between imputation variances.
#' @export
#' @importFrom broom.mixed tidy
#' @import dplyr
#' @examples
#' \donttest{
#' sim_list <- replicate(n = 10,
#'         expr = sleepstudy[sample(row.names(sleepstudy), 180),],
#'         simplify=FALSE)
#' fml <- "Reaction ~ Days + (Days | Subject)"
#' mod <- lmerModList(fml, data = sim_list)
#' modelFixedEff(mod)
#' }
modelFixedEff <- function(modList, ...){
  fixEst <- lapply(modList, tidy, effects = "fixed", ...)
  fixEst <- do.call(rbind, fixEst)
  # Collapse
  # Rubin correction, get length of list
  ml <- length(modList)
  # Get between and within imputation variance, apply total correction
  # Calculate degree of freedom correction
  rubin <- fixEst %>% group_by(term) %>%
    mutate(mean_est = mean(estimate)) %>%
    mutate(est_ss = (estimate - mean_est)^2) %>%
    summarize(estimate = mean(estimate),
              within_var = mean(std.error), # compute within imputation variance
              between_var =  mean(est_ss)) %>% # estimate the between imputation variance
    mutate(std.error = within_var + ((1 + 1/ml)*between_var),
           df = (ml-1)* (1 + within_var/((1 + 1/ml)*between_var))^2) # apply rubins total variance correction
  # add fallback
  if (any((((1 + 1/ml)*rubin$between_var)^2) < 0.000000001)) {
    warning("Between imputation variance is very small, are imputation sets too similar?")
  }
  # DEPRECATED method
  # out <- fixEst %>% dplyr::group_by(term) %>%
  #         dplyr::summarize(estimate = mean(estimate),
  #              std.error = mean(std.error))
  rubin$statistic <- rubin$estimate / rubin$std.error
  rubin <- rubin %>% dplyr::select(term, estimate, std.error, statistic, df)
  return(as.data.frame(rubin))
}


#' Extract fixed-effects estimates for a merModList
#'
#' @inheritParams lme4::fixef
#' @return a named, numeric vector of fixed-effects estimates.
#' @details Extract the estimates of the fixed-effects parameters from a list of
#' fitted \code{merMod} models. Takes the mean of the individual \code{fixef}
#' objects for each of the component models in the \code{merModList}.
#' @export
#' @examples
#' \donttest{
#' sim_list <- replicate(n = 10,
#'         expr = sleepstudy[sample(row.names(sleepstudy), 180),],
#'         simplify=FALSE)
#' fml <- "Reaction ~ Days + (Days | Subject)"
#' mod <- lmerModList(fml, data = sim_list)
#' fixef(mod)
#' }
fixef.merModList <- function(object, add.dropped = FALSE, ...){
  Reduce(`+`, lapply(object, fixef)) / length(object)
}

#' Extract random-effects estimates for a merModList
#'
#' @inheritParams lme4::ranef
#' @return a named, numeric vector of random-effects estimates.
#' @details Extract the estimates of the random-effects parameters from a list of
#' fitted \code{merMod} models. Takes the mean of the individual \code{ranef}
#' objects for each of the component models in the \code{merModList}.
#' @export
#' @examples
#' \donttest{
#' sim_list <- replicate(n = 10,
#'         expr = sleepstudy[sample(row.names(sleepstudy), 180),],
#'         simplify=FALSE)
#' fml <- "Reaction ~ Days + (Days | Subject)"
#' mod <- lmerModList(fml, data = sim_list)
#' ranef(mod)
#' }
ranef.merModList <- function(object, ...){
  levels <- getME(object[[1]], "n_rfacs")
  re <- vector(length = levels, mode = "list")
  for(i in seq_along(1:levels)){
    # <- Reduce(`+`, lapply(object, ranef)[i]) / length(object)
  re[i]  <- lapply(Reduce(`+`, lapply(object, ranef)[1]), function(x) x/length(object))
  }
  names(re) <- names(ranef(object[[1]]))
  return(re)
}


#' Extract the variances and correlations for random effects from a merMod list
#' @inheritParams lme4::VarCorr
#' @param rdig the number of digits to round to, integer
#' @return a list with two elements "stddev" and "correlation" for the standard
#' deviations and correlations averaged across models in the list
#' @export
#' @import lme4
#' @examples
#' sim_list <- replicate(n = 10,
#'         expr = sleepstudy[sample(row.names(sleepstudy), 180),],
#'         simplify=FALSE)
#' fml <- "Reaction ~ Days + (Days | Subject)"
#' mod <- lmerModList(fml, data = sim_list)
#' VarCorr(mod)
VarCorr.merModList <- function(x, sigma = 1, rdig = 3L, ...){
  modList <- x
  ngrps <- length(VarCorr(modList[[1]]))
  errorList <- vector(mode = 'list', length = ngrps)
  corrList <- vector(mode = 'list', length = ngrps)
  for(i in 1:ngrps){
    subList <- lapply(modList, function(x) VarCorr(x)[[i]])
    if(all(dim(subList[[1]]) == c(1, 1))){
      subList <- mean(sqrt(unlist(subList)))
      errorList[[i]] <- subList
      names(errorList) <- "Intercept"
      corrList[[i]] <- matrix(1)
      dimnames(corrList[[i]]) <- list("(Intercept)", "(Intercept)")
    } else {
      errorList[[i]] <- apply(simplify2array(lapply(subList, attr, "stddev")),
                              1, mean)
      corrList[[i]] <- apply(simplify2array(lapply(subList, attr, "corre")),
                             1:2,mean)
    }
  }
  for(i in 1:length(errorList)){
    if(is.null(names(errorList[[i]]))){
      names(errorList[[i]]) <- "(Intercept)"
    }
  }
  for(i in 1:length(corrList)){
    if(is.null(names(corrList[[i]])) & is.null(dim(corrList[[i]]))){
      names(corrList[[i]]) <- "(Intercept)"
    }
  }
  names(errorList) <- names(ranef(modList[[1]]))
  names(corrList) <- names(ranef(modList[[1]]))
  return(list("stddev" = errorList, "correlation" = corrList))
}

utils::globalVariables(c("term", "estimate","std.error"))
#' Print the results of a merMod list
#'
#' @param object a modList of class merModList
#' @param ... additional arguments
#'
#' @return summary content printed to console
#' @export
#' @examples
#' sim_list <- replicate(n = 10,
#'         expr = sleepstudy[sample(row.names(sleepstudy), 180),],
#'         simplify=FALSE)
#' fml <- "Reaction ~ Days + (Days | Subject)"
#' mod <- lmerModList(fml, data = sim_list)
#' print(mod)
summary.merModList <- function(object, ...){
  modList <- object
  args <- eval(substitute(alist(...)))
  if("digits" %in% names(args)){
    digits <- args$digits
  } else{
    digits <- 3
  }
  summ.ml <- list()
  summ.ml$len <- length(modList)
  summ.ml$form <- modList[[1]]@call
  summ.ml$method <- summary(modList[[1]])$methTitle
  summ.ml$family <- summary(modList[[1]])$family
  summ.ml$fe <- modelFixedEff(modList)
  dimnames(summ.ml$fe)[[1]] <- summ.ml$fe$term
  # pfround(summ.ml$fe[, -1], digits)
  summ.ml$ngrps <- length(VarCorr(modList[[1]]))
  summ.ml$errorList <- VarCorr(modList)$stddev
  summ.ml$corrList <- VarCorr(modList)$correlation

  # lapply(errorList, pfround, digits)
  summ.ml$residError <- mean(unlist(lapply(modList, function(x) attr(VarCorr(x), "sc"))))
  summ.ml$ngrps <- lapply(modList[[1]]@flist, function(x) length(levels(x)))
  summ.ml$modn <- getME(modList[[1]], "devcomp")$dims["n"]
  summ.ml$mAIC <- mean(unlist(lapply(modList, AIC)))
  summ.ml$moDsigma.hat <- mean(unlist(lapply(modList, sigma)))
  class(summ.ml) <- "summary.merModList"
  return(summ.ml)
}

#' Summarize a merMod list
#'
#' @param x a modList of class merModList
#' @param ... additional arguments
#'
#' @return a summary object of model information
#' @export
#' @examples
#' \donttest{
#' sim_list <- replicate(n = 10,
#'         expr = sleepstudy[sample(row.names(sleepstudy), 180),],
#'         simplify=FALSE)
#' fml <- "Reaction ~ Days + (Days | Subject)"
#' mod <- lmerModList(fml, data = sim_list)
#' summary(mod)
#' }
print.merModList <- function(x, ...){
  out <- lapply(x, sum.mm)
  # class(out) <- "summary.merModList"
  return(out)
}

#' Print the summary of a merMod list
#'
#' @param x a summary of amerModList object
#' @param ... additional arguments
#'
#' @return summary content printed to console
#' @export
print.summary.merModList <- function(x, ...){
  summ.ml <- x
  args <- eval(substitute(alist(...)))
  if("digits" %in% names(args)){
    digits <- args$digits
  } else{
    digits <- 3
  }
  print(summ.ml$method)
  cat("Model family: ", summ.ml$family)
  cat("\n")
  print(summ.ml$form)
  cat("\nFixed Effects:\n")
  pfround(summ.ml$fe[, -1], digits)
  cat("\nRandom Effects:\n")
  cat("\nError Term Standard Deviations by Level:\n")
  for(i in 1:length(summ.ml$errorList)){
    cat("\n")
    cat(names(summ.ml$errorList[i]))
    cat("\n")
    if(is.null(names(summ.ml$errorList[[i]]))){
      names(summ.ml$errorList[[i]]) <- "(Intercept)"
    }
    pfround(summ.ml$errorList[[i]], digits = digits)
    cat("\n")
  }
  # lapply(errorList, pfround, digits)
  cat("\nError Term Correlations:\n")
  for(i in 1:length(summ.ml$corrList)){
    cat("\n")
    cat(names(summ.ml$corrList[i]))
    cat("\n")
    if(is.null(names(summ.ml$corrList[[i]]))){
      names(summ.ml$corrList[[i]]) <- "(Intercept)"
    }
    pfround(summ.ml$corrList[[i]], digits = digits)
    cat("\n")
  }
  # lapply(corrList, pfround, digits)
  cat("\nResidual Error =", fround(summ.ml$residError,
                                   digits), "\n")
  cat("\n---Groups\n")
  cat(sprintf("number of obs: %d, groups: ", summ.ml$modn))
  cat(paste(paste(names(summ.ml$ngrps), summ.ml$ngrps, sep = ", "),
            collapse = "; "))
  cat("\n")
  cat("\nModel Fit Stats")
  cat(sprintf("\nAIC = %g", round(summ.ml$mAIC, 1)))
  cat("\nResidual standard deviation =", fround(summ.ml$moDsigma.hat,
                                                digits), "\n")
}

#' Apply a multilevel model to a list of data frames
#'
#' @param formula a formula to pass through compatible with merMod
#' @param data a list object with each element being a data.frame
#' @param parallel logical, should the models be run in parallel? Default FALSE. If so,
#' the `future_lapply` function from the `future.apply` package is used. See
#' details.
#' @param ... additional arguments to pass to the estimating function
#' @rdname merModList
#'
#' @details Parallel computing is provided by the `futures` package, and its
#' extension the `future.apply` package to provide the `future_lapply` function
#' for easy parallel computations on lists. To use this package, simply register
#' a parallel backend using the `plan()` function from `futures` - an example
#' is to use `plan(multisession)`
#'
#' @return a list of fitted merMod objects of class merModList
#' @export
#' @examples
#' \donttest{
#' sim_list <- replicate(n = 10,
#'         expr = sleepstudy[sample(row.names(sleepstudy), 180),],
#'         simplify=FALSE)
#' fml <- "Reaction ~ Days + (Days | Subject)"
#' mod <- lmerModList(fml, data = sim_list)
#' summary(mod)
#' }
lmerModList <- function(formula, data, parallel = FALSE, ...){
  if (parallel) {
    if (requireNamespace("future.apply", quietly=TRUE)) {
      ml <- future.apply::future_lapply(data, function(d) lmer(formula, data = d, ...))
    } else {
      warning("Parallel set but future.apply not available. Running sequentially.")
      ml <- lapply(data, function(d) lmer(formula, data = d, ...))
    }
  } else {
    ml <- lapply(data, function(d) lmer(formula, data = d, ...))
  }

  class(ml) <- "merModList"
  return(ml)
}

#' Apply a Bayesian multilevel model to a list of data frames
#'
#' @rdname merModList
#' @return a merModList
#' @importFrom blme blmer
#' @export
blmerModList <- function(formula, data, parallel = FALSE, ...){
  if(parallel) {
    if (requireNamespace("future.apply", quietly=TRUE)) {
      ml <- future.apply::future_lapply(data, function(d) blmer(formula, data = d, ...))
    }
    warning("Parallel set but future.apply not available. Running sequentially.")
    ml <- lapply(data, function(d) blmer(formula, data = d, ...))
  } else {
    ml <- lapply(data, function(d) blmer(formula, data = d, ...))
  }
  class(ml) <- "merModList"
  return(ml)
}

#' Apply a generalized linear multilevel model to a list of data frames
#'
#' @rdname merModList
#' @return a merModList
#' @export
glmerModList <- function(formula, data, parallel = FALSE, ...){
  if(parallel) {
    if (requireNamespace("future.apply", quietly=TRUE)) {
      ml <- future.apply::future_lapply(data, function(d) glmer(formula, data = d, ...))
    }
    warning("Parallel set but future.apply not available. Running sequentially.")
    ml <- lapply(data, function(d) glmer(formula, data = d, ...))
  } else {
    ml <- lapply(data, function(d) glmer(formula, data = d, ...))
  }
  class(ml) <- "merModList"
  return(ml)
}

#' Apply a Bayesian generalized linear multilevel model to a list of data frames
#'
#' @rdname merModList
#' @return a merModList
#' @importFrom blme bglmer
#' @export
bglmerModList <- function(formula, data, parallel = FALSE, ...){
  if(parallel) {
    if (requireNamespace("future.apply", quietly=TRUE)) {
      ml <- future.apply::future_lapply(data, function(d) bglmer(formula, data = d, ...))
    }
    warning("Parallel set but future.apply not available. Running sequentially.")
    ml <- lapply(data, function(d) bglmer(formula, data = d, ...))
  } else {
    ml <- lapply(data, function(d) bglmer(formula, data = d, ...))
  }
  class(ml) <- "merModList"
  return(ml)
}

# test merModList functions

#Do merModList objects get built and work----

old_warn <- getOption("warn")
options(warn = -1)
set.seed(432422)

test_that("simple cases work", {
  skip_on_cran()
  library(blme)
  d <- expand.grid(fac1=LETTERS[1:5], grp=factor(1:10),
                   obs=1:100)
  split <- sample(x = LETTERS[9:15], size = nrow(d), replace=TRUE)
  suppressMessages({
    d$y <- simulate(~fac1+(1|grp),family = gaussian,
                    newdata=d,
                    newparams=list(beta=c(2,1,3,4,7), theta=c(.25),
                                   sigma = c(.23)))[[1]]
  })

  out <- split(d, split)
  rm(split)
  # TODO change tolerances
  g1 <- lmerModList(formula = y~fac1+(1|grp), data=out,
                    control= lmerControl(check.conv.grad = .makeCC("warning", tol= 2e-3)))
  expect_s3_class(g1, "merModList")
  g2 <- blmerModList(formula = y~fac1+(1|grp), data=out,
                     control= lmerControl(check.conv.grad = .makeCC("warning", tol= 2e-3)))
  expect_s3_class(g2, "merModList")
  expect_false(class(g1[[1]]) == class(g2[[1]]))

  split <- sample(x = LETTERS[1:20], size = nrow(InstEval), replace=TRUE)
  out <- split(InstEval, split)
  rm(split)
  suppressMessages({
    g1 <- lmerModList(formula = y ~ lectage + studage + (1|d) + (1|dept),
                      data=out,
                      control= lmerControl(check.conv.grad = .makeCC("warning", tol = 1e-2)))
  })
  expect_s3_class(g1, "merModList")

})

test_that("print methods work for merModList", {
  skip_on_cran()
  d <- expand.grid(fac1=LETTERS[1:5], grp=factor(1:10),
                   obs=1:100)
  split <- sample(x = LETTERS[9:15], size = nrow(d), replace=TRUE)
  suppressMessages({
    d$y <- simulate(~fac1+(1|grp),family = gaussian,
                    newdata=d,
                    newparams=list(beta=c(2,1,3,4,7), theta=c(.25),
                                   sigma = c(.23)))[[1]]
  })

  out <- split(d, split)
  rm(split);
  g1 <- lmerModList(formula = y~fac1+(1|grp), data=out,
                    control= lmerControl(check.conv.grad = .makeCC("warning", tol= 1e-2)));
  {sink("NUL"); zz <- print(g1);
    sink()}
  expect_type(zz, "list")
  zz <- summary(g1)
  expect_s3_class(zz, "summary.merModList")

})

# Numerical accuracy of merModList print method----

test_that("print method for merModList works in general case", {
  skip_on_cran()
  data(grouseticks)
  grouseticks$HEIGHT <- scale(grouseticks$HEIGHT)
  grouseticks <- merge(grouseticks, grouseticks_agg[, 1:3], by = "BROOD")
  grouseticks$TICKS_BIN <- ifelse(grouseticks$TICKS >=1, 1, 0)
  form <- TICKS_BIN ~ HEIGHT +(1 + HEIGHT|BROOD) + (1|YEAR)
  modDat <- vector(5, mode="list")
  for(i in 1:length(modDat)){
    modDat[[i]] <- grouseticks[sample(x=1:nrow(grouseticks), size = nrow(grouseticks),
                                      replace=FALSE),]
  }

  g1 <- glmerModList(formula = form,
                    data = modDat, family = "binomial",
                    control = glmerControl(optimizer="bobyqa",
                                           optCtrl=list(maxfun = 1e6),
                                           check.conv.grad = .makeCC("warning", tol= 1e-2)))
  g1T <- glmer(form, family = "binomial", data = grouseticks,
               control = glmerControl(optimizer="bobyqa",
                                      optCtrl=list(maxfun = 1e6),
                                      check.conv.grad = .makeCC("warning", tol= 1e-2)))

  expect_equal(VarCorr(g1)$stddev$BROOD, attr(VarCorr(g1T)$BROOD, "stddev"),
               tolerance = 0.0001)
  expect_equal(VarCorr(g1)$stddev$YEAR, attr(VarCorr(g1T)$YEAR, "stddev"),
               tolerance = 0.0001)
  expect_equal(VarCorr(g1)$correlation$BROOD, attr(VarCorr(g1T)$BROOD, "corre"),
               tolerance = 0.0001)
  expect_equal(VarCorr(g1)$correlation$YEAR, attr(VarCorr(g1T)$YEAR, "corre"),
               tolerance = 0.0001)

  form <- TICKS_BIN ~ HEIGHT +(1|BROOD)
  g1 <- glmerModList(formula = form,
                     data = modDat, family = "binomial",
                     control = glmerControl(optimizer="bobyqa",
                                            optCtrl=list(maxfun = 1e6),
                                            check.conv.grad = .makeCC("warning", tol= 1e-2)))
  g1T <- glmer(form, family = "binomial", data = grouseticks,
               control = glmerControl(optimizer="bobyqa",
                                      optCtrl=list(maxfun = 1e6),
                                      check.conv.grad = .makeCC("warning", tol= 1e-2)))

  expect_equal(VarCorr(g1)$stddev$BROOD, attr(VarCorr(g1T)$BROOD, "stddev"),
               tolerance = 0.0001)
  expect_equal(VarCorr(g1)$stddev$YEAR, attr(VarCorr(g1T)$YEAR, "stddev"),
               tolerance = 0.0001)
  expect_equal(VarCorr(g1)$correlation$BROOD, attr(VarCorr(g1T)$BROOD, "corre"),
               tolerance = 0.0001)
  expect_equal(VarCorr(g1)$correlation$YEAR, attr(VarCorr(g1T)$YEAR, "corre"),
               tolerance = 0.0001)

})

#ICC function----

test_that("ICC function works", {
  skip_on_cran()
  ICC1 <- ICC(outcome = "Reaction", group = "Subject", data = sleepstudy)
  expect_type(ICC1, "double")
  expect_equal(ICC1, 0.3948896, tol = .001)
})

options(warn= old_warn)

set.seed(11213)

test_that("fastdisp.merMod returns a list invisibly with expected components", {
  skip_on_cran()
  m1 <- lmer(Reaction ~ Days + (1 | Subject), data = sleepstudy)
  result <- fastdisp(m1)
  expect_type(result, "list")
  expect_true(all(c("call", "coef", "se", "ngrps", "AIC", "n") %in% names(result)))
  expect_type(result$coef, "double")
  expect_type(result$se, "double")
  expect_true(is.numeric(result$AIC))
  expect_true(is.numeric(result$n))
})

test_that("fastdisp.merMod return value can be captured and is non-null", {
  skip_on_cran()
  m1 <- lmer(Reaction ~ Days + (1 | Subject), data = sleepstudy)
  result <- fastdisp(m1)
  expect_false(is.null(result))
  expect_equal(names(result$coef), c("(Intercept)", "Days"))
})

test_that("fastdisp.merModList returns a list with expected components", {
  skip_on_cran()
  sim_list <- replicate(
    n = 3,
    expr = sleepstudy[sample(row.names(sleepstudy), 180), ],
    simplify = FALSE
  )
  ml <- lmerModList("Reaction ~ Days + (1 | Subject)", data = sim_list)
  result <- fastdisp(ml)
  expect_type(result, "list")
  expect_true(all(c("call", "coef", "se", "ngrps", "AIC", "n") %in% names(result)))
  expect_false(is.null(result))
})

#' @title Plot the results of a simulation of the random effects
#' @name plotREsim
#' @description Plot the simulated random effects on a ggplot2 chart. Points that
#' are distinguishable from zero (i.e. the confidence band based on \code{level}
#' does not cross the red line) are highlighted. Currently, the plots are ordered
#' according to the grouping factor.
#' @param data a data.frame generated by \code{\link{REsim}} with simulations of
#' the random effects of a \code{\link{merMod}}
#' @param level the width of the confidence interval
#' @param stat a character value indicating the variable name in data of the
#' midpoint of the estimated interval, e.g. "mean" or "median"
#' @param sd a logical indicating whether or not to plot error bars around
#' the estimates (default is TRUE). Calculates the width of the error bars
#' based on \code{level} and the variable named "sd" in \code{data}
#' @param sigmaScale a numeric value to divide the estimate and the standard
#' deviation by in the case of doing an effect size calculation
#' @param oddsRatio logical, should the parameters be converted to odds ratios
#' before plotting
#' @param labs logical, include the labels of the groups on the x-axis
#' @param facet Accepts either logical (\code{TRUE}) or \code{list} to specify which
#' random effects to plot. If \code{TRUE}, facets by both \code{groupFctr} and \code{term}.
#' If \code{list} selects the panel specified by the named elements of the list
#' @return a ggplot2 plot of the coefficient effects
#' @examples
#' \donttest{
#'  fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
#'  (p1 <- plotREsim(REsim(fm1)))
#'  #Plot just the random effects for the Days slope
#'  (p2 <- plotREsim(REsim(fm1), facet= list(groupFctr= "Subject", term= "Days")))
#'  }
#' @export
#' @import ggplot2
plotREsim <- function(data, level = 0.95, stat = "median", sd = TRUE,
                      sigmaScale = NULL, oddsRatio = FALSE, labs = FALSE,
                      facet= TRUE){
  # error checking
  plot_sim_error_chks(type= "RE", level= level, stat= stat, sd= sd, sigmaScale= sigmaScale,
                      oddsRatio= oddsRatio, labs= labs, facet= facet)
  # check for faceting
  facet_logical <- is.logical(facet)
  if (!facet_logical) {
    data <- data[data$groupFctr == facet[[1]] & data$term == facet[[2]], ]
  }

  if(!missing(sigmaScale)){
    data[, "sd"] <- data[, "sd"] / sigmaScale
    data[, stat] <- data[, stat] / sigmaScale
  }
  data[, "sd"] <- data[, "sd"] * qnorm(1-((1-level)/2))
  data[, "ymax"] <- data[, stat] + data[, "sd"]
  data[, "ymin"] <- data[, stat] - data[, "sd"]
  data[, "sig"] <- data[, "ymin"] > 0 | data[, "ymax"] < 0
  hlineInt <- 0
  if(oddsRatio == TRUE){
    data[, "ymax"] <- exp(data[, "ymax"])
    data[, stat] <- exp(data[, stat])
    data[, "ymin"] <- exp(data[, "ymin"])
    hlineInt <- 1
  }
  data <- data[order(data[,"groupFctr"], data[,"term"], data[,stat]),]
  rownames(data) <- 1:nrow(data)
  data[,"xvar"] <- factor(paste(data$groupFctr, data$groupID, sep=""),
                          levels=unique(paste(data$groupFctr,data$groupID, sep="")),
                          ordered=TRUE)
  if(labs == TRUE){
    xlabs.tmp <- element_text(face = "bold", angle=90, vjust=.5)
  } else {
    data[,"xvar"] <- as.numeric(data[,"xvar"])
    xlabs.tmp <- element_blank()
  }

  p <- ggplot(data, aes(x = .data[["xvar"]], y = .data[[stat]], ymax = .data[["ymax"]], ymin = .data[["ymin"]])) +
         geom_hline(yintercept = hlineInt, color = I("red"), linewidth = I(1.1)) +
         geom_point(color="gray75", alpha=1/(nrow(data)^.33), size=I(0.5)) +
         geom_point(data=subset(data, sig==TRUE), size=I(3)) +
         labs(x = "Group", y = "Effect Range", title = "Effect Ranges") +
         theme_bw() +
         theme(panel.grid.major = element_blank(),
               panel.grid.minor = element_blank(),
               axis.text.x = xlabs.tmp,
               axis.ticks.x = element_blank())
  if (sd) {
    p <- p +
      geom_pointrange(alpha = 1/(nrow(data)^.33)) +
      geom_pointrange(data=subset(data, sig==TRUE), alpha = 0.25)
  }
  # check facet
  if (facet_logical) {
    return(p + facet_grid(term ~ groupFctr, scales = "free_x"))
  } else {
    return(p)
  }
}

#' @title Plot the results of a simulation of the fixed effects
#' @name plotFEsim
#' @description Plot the simulated fixed effects on a ggplot2 chart
#' @param data a data.frame generated by \code{\link{FEsim}} with simulations of
#' the fixed effects of a \code{\link{merMod}}
#' @param level the width of the confidence interval
#' @param stat a character value indicating the variable name in data of the
#' midpoint of the estimated interval, e.g. "mean" or "median"
#' @param sd logical, indicating whether or not to plot error bars around
#' the estimates (default is TRUE). Calculates the width of the error bars
#' based on \code{level} and the variable named "sd" in \code{data}
#' @param intercept logical, should the intercept be included, default is FALSE
#' @param sigmaScale a numeric value to divide the estimate and the standard
#' deviation by in the case of doing an effect size calculation
#' @param oddsRatio logical, should the parameters be converted to odds ratios
#' before plotting
#' @return a ggplot2 plot of the coefficient effects
#' @examples
#'  fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
#'  (p1 <- plotFEsim(FEsim(fm1)))
#' @export
#' @import ggplot2
plotFEsim <- function(data, level=0.95, stat = "median", sd = TRUE,
                      intercept = FALSE, sigmaScale = NULL, oddsRatio = FALSE){
  # error checking
  plot_sim_error_chks(type= "FE", level= level, stat= stat, sd= sd, sigmaScale= sigmaScale,
                      oddsRatio= oddsRatio, labs= TRUE, facet= TRUE)

  if(!missing(sigmaScale)){
    data[, "sd"] <- data[, "sd"] / sigmaScale
    data[, stat] <- data[, stat] / sigmaScale
  }
  if(intercept == FALSE){
    data <- data[data$term != "(Intercept)", ]
  }
  data[, "sd"] <- data[, "sd"] * qnorm(1-((1-level)/2))
  data[, "ymax"] <- data[, stat] + data[, "sd"]
  data[, "ymin"] <- data[, stat] - data[, "sd"]
  hlineInt <- 0
  if(oddsRatio == TRUE){
    data[, "ymax"] <- exp(data[, "ymax"])
    data[, stat] <- exp(data[, stat])
    data[, "ymin"] <- exp(data[, "ymin"])
    hlineInt <- 1
  }
  xvar <- "term"
  data$term <- as.character(data$term)
  data$term <- factor(data$term , levels = data[order(data[, stat]), 1])
  p <- ggplot(aes(x = .data[[xvar]], y = .data[[stat]], ymax = .data[["ymax"]], ymin = .data[["ymin"]]), data = data) +
    geom_hline(yintercept = hlineInt, color = I("red")) +
    geom_point(size=I(3)) +
    coord_flip() +
    theme_bw()
  if (sd) {
    p <- p + geom_errorbar(width = 0.2)
  }
  p
}

# Test plotting functions

# Plot functions return gg objects? ----

test_that("Prediction interval plots work for simple linear example", {
  skip_on_cran()
  d <- expand.grid(fac1=LETTERS[1:5], grp=factor(1:10),
                   obs=1:100)
  suppressMessages({
    d$y <- simulate(~fac1+(1|grp),family = gaussian,
                    newdata=d,
                    newparams=list(beta=c(2,1,3,4,7), theta=c(.25),
                                   sigma = c(.23)))[[1]]
  })

  subD <- d[sample(row.names(d), 1000),]

  g1 <- lmer(y~fac1+(1|grp), data=subD)
  FE1 <- FEsim(g1)
  p1 <- plotFEsim(FE1)
  expect_s3_class(p1, "gg")
  p1 <- plotREsim(REsim(g1))
  expect_s3_class(p1, "gg")

})

#' merTools: Provides methods for extracting and exploring results from merMod
#' objects in the lme4 package.
#'
#' The merTools package contains convenience tools for extracting useful
#' information from and exploring the implications of merMod objects created by
#' the lme4 package.  These convenience functions are especially useful for
#' merMod objects that take a long time to estimate due to their complexity or
#' because they are estimated on very large samples.
#'
#' See the vignettes for usage examples
#'
#' @section merMod extraction/utility functions:
#'
#' \itemize{
#'   \item \code{\link{fastdisp}}
#'   \item \code{\link{superFactor}}
#'   \item \code{\link{REextract}}
#'   \item \code{\link{REsim}}
#'   \item \code{\link{FEsim}}
#'   \item \code{\link{RMSE.merMod}}
#'   \item \code{\link{thetaExtract}}
#'   \item \code{\link{REquantile}}
#' }
#'
#' @section merMod exploration functions:
#'
#' \itemize{
#'   \item \code{\link{plotREsim}}
#'   \item \code{\link{plotFEsim}}
#'   \item \code{\link{draw}}
#'   \item \code{\link{wiggle}}
#'   \item \code{\link{subBoot}}
#'   \item \code{\link{predictInterval}}
#'   \item \code{\link{expectedRank}}
#'   \item \code{\link{REimpact}}
#'   \item \code{\link{shinyMer}}
#' }
#'

#' @keywords internal
"_PACKAGE"

#' A subset of data from the 1982 High School and Beyond survey used as examples for HLM software
#' @description A key example dataset used for examples in the HLM software manual.
#' Included here for use in replicating HLM analyses in R.
#' @format A data frame with 7,185 observations on the following 8 variables.
#' \describe{
#'  \item{\code{schid}}{a numeric vector, 160 unique values}
#'  \item{\code{mathach}}{a numeric vector for the performance on a standardized math assessment}
#'  \item{\code{female}}{a numeric vector coded 0 for male and 1 for female}
#'  \item{\code{ses}}{a numeric measure of student socio-economic status}
#'  \item{\code{minority}}{a numeric vector coded 0 for white and 1 for non-white students}
#'  \item{\code{schtype}}{a numeric vector coded 0 for public and 1 for private schools}
#'  \item{\code{meanses}}{a numeric, the average SES for each school in the data set}
#'  \item{\code{size}}{a numeric for the number of students in the school}
#' }
#' @details The data file used for this presentation is a subsample from the
#' 1982 High School and Beyond Survey and is used extensively in
#' Hierarchical Linear Models by Raudenbush and Bryk. It consists of 7,185 students
#' nested in 160 schools.
#' @source Data made available by UCLA Institute for Digital Research and Education
#' (IDRE) online: \url{https://stats.oarc.ucla.edu/other/hlm/hlm-mlm/introduction-to-multilevel-modeling-using-hlm/}
#' @references Stephen W. Raudenbush and Anthony S. Bryk (2002). Hierarchical
#' Linear Models: Applications and Data Analysis Methods (2nd ed.). SAGE.
#' @examples
#' data(hsb)
#' head(hsb)
"hsb"

# Parallel functions

#' Set up parallel environment
#'
#' @return Nothing
setup_parallel <- function() {
  if (!requireNamespace("foreach", quietly = TRUE)) {
    # EXCLUDE COVERAGE START
    stop("foreach package required for parallel future operation",
         call. = FALSE)
    # EXCLUDE COVERAGE END
  }
  if (foreach::getDoParWorkers() == 1) {
    # EXCLUDE COVERAGE START
    warning("No parallel backend registered", call. = TRUE)
    # EXCLUDE COVERAGE END
  }
}

# if (.parallel) {
#   setup_parallel()
#
#   i <- seq_len(n)
#   fe_call <- as.call(c(list(quote(foreach::foreach), i = i), .paropts))
#   fe <- eval(fe_call)
#
#   result <- foreach::`%dopar%`(fe, do.ply(i))
# } else {
#   result <- loop_apply(n, do.ply)
# }

#' Calculate the weighted mean of fitted values for various levels of
#' random effect terms.
#'
#' \code{REimpact} calculates the average predicted value for each row of a
#' new data frame across the distribution of \code{\link{expectedRank}} for a
#' merMod object. This allows the user to make meaningful comparisons about the
#' influence of random effect terms on the scale of the response variable,
#' for user-defined inputs, and accounting for the variability in grouping terms.
#'
#' The function predicts the response at every level in the random effect term
#' specified by the user. Then, the expected rank of each group level is binned
#' to the number of bins specified by the user. Finally, a weighted mean of the
#' fitted value for all observations in each bin of the expected ranks is
#' calculated using the inverse of the variance as the weight -- so that less
#' precise estimates are downweighted in the calculation of the mean for the bin.
#' Finally, a standard error for the bin mean is calculated.
#'
#' @param merMod An object of class merMod
#'
#' @param newdata a data frame of observations to calculate group-level differences
#' for
#'
#' @param groupFctr The name of the grouping factor over which the random
#'   coefficient of interest varies.  This is the variable to the right of the
#'   pipe, \code{|}, in the [g]lmer formula. This parameter is optional, if not
#'   specified, it will perform the calculation for the first effect listed
#'   by \code{ranef}.
#'
#' @param term The name of the random coefficient of interest. This is the
#'   variable to the left of the pipe, \code{|}, in the [g]lmer formula. Partial
#'   matching is attempted on the intercept term so the following character
#'   strings will all return rankings based on the intercept (\emph{provided that
#'   they do not match the name of another random coefficient for that factor}):
#'   \code{c("(Intercept)", "Int", "intercep", ...)}.
#'
#' @param breaks an integer representing the number of bins to divide the group
#' effects into, the default is 3; alternatively it can specify breaks from 0-100
#' for how to cut the expected rank distribution
#'
#' @param ... additional arguments to pass to \code{\link{predictInterval}}
#'
#' @return A data.frame with all unique combinations of the number of cases, rows
#' in the newdata element, and number of bins:
#'   \describe{
#'     \item{case}{The row number of the observation from newdata.}
#'     \item{bin}{The ranking bin for the expected rank, the higher the bin number,
#'     the greater the expected rank of the groups in that bin.}
#'     \item{AvgFitWght}{The weighted mean of the fitted values for case i in bin k}
#'     \item{AvgFitWghtSE}{The standard deviation of the mean of the fitted values
#'     for case i in bin k.}
#'     \item{nobs}{The number of group effects contained in that bin.}
#'   }
#'
#' @details This function uses the formula for variance of a weighted mean
#' recommended by Cochran (1977).
#'
#' @references
#' Gatz, DF and Smith, L. The Standard Error of a Weighted Mean Concentration.
#' I. Bootstrapping vs other methods. \emph{Atmospheric Environment}.
#' 1995;11(2)1185-1193. Available at
#' \url{https://www.sciencedirect.com/science/article/pii/135223109400210C}
#'
#' Cochran, WG. 1977. Sampling Techniques (3rd Edition). Wiley, New York.
#'
#' @seealso \code{\link{expectedRank}}, \code{\link{predictInterval}}
#'
#' @examples
#' \donttest{
#' #For a one-level random intercept model
#' m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
#' m1.er <- REimpact(m1, newdata = sleepstudy[1, ], breaks = 2)
#' #For a one-level random intercept model with multiple random terms
#' m2 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
#' #ranked by the random slope on Days
#' m2.er1 <- REimpact(m2,  newdata = sleepstudy[1, ],
#'            groupFctr = "Subject", term="Days")
#' #ranked by the random intercept
#' m2.er2 <- REimpact(m2, newdata = sleepstudy[1, ],
#'              groupFctr = "Subject", term="int")
#'
#' # You can also pass additional arguments to predictInterval through REimpact
#' g1 <- lmer(y ~ lectage + studage + (1|d) + (1|s), data=InstEval)
#' zed <- REimpact(g1, newdata = InstEval[9:12, ], groupFctr = "d", n.sims = 50,
#'                 include.resid.var = TRUE)
#' zed2 <- REimpact(g1, newdata = InstEval[9:12, ], groupFctr = "s", n.sims = 50,
#'                  include.resid.var = TRUE)
#' zed3 <- REimpact(g1, newdata = InstEval[9:12, ], groupFctr = "d", breaks = 5,
#'                 n.sims = 50, include.resid.var = TRUE)
#' }
#' @export
REimpact <- function(merMod, newdata, groupFctr=NULL, term = NULL, breaks = 3, ...){
  if(missing(groupFctr)){
    groupFctr <- names(ranef(merMod))[1]
  }
  lvls <- unique(merMod@frame[, groupFctr])
  zed <- as.data.frame(lapply(newdata, rep, length(lvls)))
  zed[, groupFctr] <- rep(lvls, each = nrow(newdata))
  zed[, "case"] <- rep(seq(1, nrow(newdata)), times = length(lvls))
  outs1 <- cbind(zed, predictInterval(merMod, newdata = zed, ...))
  outs1$var <- outs1$upr - outs1$lwr
  outs1$lwr <- NULL; outs1$upr <- NULL
  ranks <- expectedRank(merMod, groupFctr = groupFctr, term = term)
  ranks <- ranks[, c(2, 7)]
  outs1 <- merge(ranks, outs1, by.x = "groupLevel", by.y = groupFctr); rm(ranks)

  weighted.var.se <- function(x, w, na.rm=FALSE)
    #  Computes the variance of a weighted mean following Cochran 1977 definition
  {
    if (na.rm) { w <- w[i <- !is.na(x)]; x <- x[i] }
    n = length(w)
    xWbar = weighted.mean(x,w,na.rm=na.rm)
    wbar = mean(w)
    out = n/((n-1)*sum(w)^2)*(sum((w*x-wbar*xWbar)^2)-2*xWbar*sum((w-wbar)*(w*x-wbar*xWbar))+xWbar^2*sum((w-wbar)^2))
    return(out)
  }
  # bin pctER somehow
  outs1$bin <- cut(outs1$pctER, breaks = breaks, labels = FALSE,
                   include.lowest = TRUE)
  bySum <- function(x){
    AvgFit <- weighted.mean(x$fit, 1/x$var)
    AvgFitSE <- weighted.var.se(x$fit, 1/x$var)
    nobs <- length(x$fit)
    return(c(AvgFit, AvgFitSE, nobs))
  }
  outs1 <- outs1[order(outs1$case, outs1$bin),]

  wMeans <- by(outs1, INDICES = list(outs1$case, outs1$bin), bySum)
  ids <- expand.grid(unique(outs1$case), unique(outs1$bin))
  wMeans <- cbind(ids, do.call(rbind, wMeans))
  names(wMeans) <- c("case", "bin", "AvgFit", "AvgFitSE", "nobs")
  return(wMeans)
}

# Test substantive effects
library(lme4)
set.seed(157)

# Test all user parameters for REimpact----
#context("Test all user parameters for REimpact")

test_that("REimpact parameters are respected", {
  skip_on_cran()
  g1 <- lmer(y ~ lectage + studage + (1|d) + (1|s), data=InstEval)
  #Warning is about %dopar% call in predictInterval
  zed <- suppressWarnings(REimpact(g1, newdata = InstEval[9:12, ], groupFctr = "d", n.sims = 50,
                  include.resid.var = TRUE))
  expect_identical(names(zed), c("case", "bin", "AvgFit", "AvgFitSE", "nobs"))
  zed2 <- REimpact(g1, newdata = InstEval[9:12, ], groupFctr = "s", n.sims = 50,
                  include.resid.var = TRUE)
  expect_equal(nrow(zed), 3 * nrow(InstEval[9:12, ]))
  expect_false(all(zed$AvgFit == zed2$AvgFit))
  expect_false(all(zed$AvgFitSE == zed2$AvgFitSE))
  expect_identical(names(zed2), c("case", "bin", "AvgFit", "AvgFitSE", "nobs"))
  zed <- REimpact(g1, newdata = InstEval[9:12, ], groupFctr = "d", breaks = 5,
                  n.sims = 50, include.resid.var = TRUE)
  expect_equal(nrow(zed), 5 * nrow(InstEval[9:12, ]))
})

test_that("REimpact respects passed values for predictInterval", {
  skip_on_cran()
  skip_on_travis()
  d <- expand.grid(fac1=LETTERS[1:5], grp=factor(1:30),
                   obs=1:100)
  suppressMessages({
    d$y <- simulate(~fac1+(1|grp),family = gaussian,
                    newdata=d,
                    newparams=list(beta=c(2,1,3,4,7), theta=c(.25),
                                   sigma = c(.23)))[[1]]
  })

  subD <- d[sample(row.names(d), 1000),]

  g1 <- lmer(y ~ fac1 + (1|grp), data=subD)
  zed <- REimpact(g1, newdata = subD[23:25, ], groupFctr = "grp", breaks = 5,
                  include.resid.var = FALSE, n.sims = 100, level = 0.8)
  zed2 <- REimpact(g1, newdata = subD[23:25, ], groupFctr = "grp", breaks = 5,
                   n.sims = 500, include.resid.var = TRUE, level = 0.99)
  # expect_true(all(zed2$AvgFitSE > zed$AvgFitSE))
  expect_true(!all(zed2$AvgFit > zed$AvgFit))
  expect_identical(names(zed), c("case", "bin", "AvgFit", "AvgFitSE", "nobs"))
  expect_identical(names(zed2), c("case", "bin", "AvgFit", "AvgFitSE", "nobs"))

})

# Test for slopes, intercepts, and combinations----
#context("Test for slopes, intercepts, and combinations")

test_that("Multiple terms can be accessed", {
  skip_on_cran()
  data(grouseticks)
  grouseticks$HEIGHT <- scale(grouseticks$HEIGHT)
  grouseticks <- merge(grouseticks, grouseticks_agg[, 1:3], by = "BROOD")
  grouseticks$TICKS_BIN <- ifelse(grouseticks$TICKS >=1, 1, 0)
  # GLMER 3 level + slope
  form <- TICKS_BIN ~ YEAR + HEIGHT + (1 + HEIGHT|BROOD) + (1|LOCATION) + (1|INDEX)
  suppressMessages({
    glmer3LevSlope  <- glmer(form, family="binomial",data=grouseticks,
                             control = glmerControl(optimizer="bobyqa",
                                                    optCtrl=list(maxfun = 1e5)))
  })

  # This is the same issue of zero mean zero variance in the predict interval call
  zed1 <- suppressWarnings(REimpact(glmer3LevSlope, newdata = grouseticks[5, ], groupFctr = "BROOD",
                  term = "HEIGHT", n.sims = 500,
                  include.resid.var = FALSE, breaks = 4, type = "probability"))
  # This is the same issue of zero mean zero variance in the predict interval call
  zed2 <- suppressWarnings(REimpact(glmer3LevSlope, newdata = grouseticks[5, ], groupFctr = "BROOD",
                   term = "Intercept",
                  n.sims = 500,
                  include.resid.var = FALSE, breaks = 4, type = "probability"))
  # This is the same issue of zero mean zero variance in the predict interval call
  zed4 <- suppressWarnings(REimpact(glmer3LevSlope, newdata = grouseticks[5, ], groupFctr = "LOCATION",
                   n.sims = 500,
                   include.resid.var = FALSE, breaks = 4))

  expect_true(all(zed4$AvgFit < zed2$AvgFit))
  expect_true(all(zed4$AvgFit < zed1$AvgFit))
  expect_false(identical(zed1, zed2))
  expect_false(identical(zed1, zed2))
  # No longer an error after revision 0.2.3
  # expect_error(zed3 <- suppressWarnings(REimpact(glmer3LevSlope, newdata = grouseticks[5, ], groupFctr = "BROOD",
  #                  n.sims = 500,
  #                  include.resid.var = FALSE, breaks = 4)), "Must specify which")
  # Don't think we need this ... it throws an subsetting error
  expect_error(zed5 <- suppressWarnings(REimpact(glmer3LevSlope, newdata = grouseticks[5, ], groupFctr = "LOCATION",
                   term = "HEIGHT",
                   n.sims = 500,
                   include.resid.var = FALSE, breaks = 4)), "undefined columns selected")

})

# Custom breaks----
#context("Custom breaks")

test_that("Custom breakpoints can be set", {
  skip_on_cran()
  g1 <- lmer(y ~ lectage + studage + (1|d) + (1|s), data=InstEval)
  zed <- REimpact(g1, newdata = InstEval[9, ], breaks = c(0, 10, 50, 90, 100),
                  groupFctr = "d", n.sims = 50,
                  include.resid.var = TRUE)
  zed2 <- REimpact(g1, newdata = InstEval[9, ], breaks = c(1, 20, 40, 60, 80, 100),
                   groupFctr = "d", n.sims = 50,
                   include.resid.var = TRUE)
  zed3 <- REimpact(g1, newdata = InstEval[9, ], breaks = 5,
                   groupFctr = "d", n.sims = 50,
                   include.resid.var = TRUE)
  expect_false(nrow(zed) == nrow(zed2))
  expect_gt(sd(zed$nobs), sd(zed2$nobs))
  expect_gt(mean(zed$nobs), mean(zed2$nobs))
  expect_equal(zed3$nobs, zed2$nobs, tolerance = .05)
})

# Tests for plotREimpact() -- overlaying one or more REimpact() results (#84)

test_that("plotREimpact returns a ggplot for a single REimpact result", {
  skip_on_cran()
  set.seed(11213)
  m1 <- lme4::lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
  imp <- suppressWarnings(
    REimpact(m1, newdata = sleepstudy[1, ], groupFctr = "Subject",
             breaks = 4, n.sims = 100)
  )
  p <- plotREimpact(imp)
  expect_s3_class(p, "ggplot")
  expect_silent(ggplot2::ggplot_build(p))
})

test_that("plotREimpact overlays a named list of REimpact results (#84)", {
  skip_on_cran()
  set.seed(11213)
  m1 <- lme4::lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
  a <- suppressWarnings(
    REimpact(m1, newdata = sleepstudy[1, ], groupFctr = "Subject",
             breaks = 4, n.sims = 100)
  )
  b <- suppressWarnings(
    REimpact(m1, newdata = sleepstudy[2, ], groupFctr = "Subject",
             breaks = 4, n.sims = 100)
  )
  p <- plotREimpact(list("Case A" = a, "Case B" = b))
  expect_s3_class(p, "ggplot")
  built <- ggplot2::ggplot_build(p)
  # The two series must both be represented in the built plot data.
  expect_true(any(grepl("colour|color", names(built$data[[length(built$data)]]))))
})

test_that("plotREimpact honors facet and level arguments", {
  skip_on_cran()
  set.seed(11213)
  m1 <- lme4::lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
  imp <- suppressWarnings(
    REimpact(m1, newdata = sleepstudy[1:2, ], groupFctr = "Subject",
             breaks = 3, n.sims = 100)
  )
  p_wide <- plotREimpact(imp, level = 0.99)
  p_narrow <- plotREimpact(imp, level = 0.80)
  d_wide <- ggplot2::ggplot_build(p_wide)$data[[1]]
  d_narrow <- ggplot2::ggplot_build(p_narrow)$data[[1]]
  # A wider confidence level produces wider-or-equal intervals.
  expect_true(all((d_wide$ymax - d_wide$ymin) >= (d_narrow$ymax - d_narrow$ymin) - 1e-8))
  expect_s3_class(plotREimpact(imp, facet = FALSE), "ggplot")
})

test_that("plotREimpact validates its inputs", {
  expect_error(plotREimpact(data.frame(a = 1)), "REimpact")
  expect_error(plotREimpact(42), "data.frame")
  set.seed(11213)
  m1 <- lme4::lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
  imp <- suppressWarnings(
    REimpact(m1, newdata = sleepstudy[1, ], groupFctr = "Subject",
             breaks = 3, n.sims = 100)
  )
  expect_error(plotREimpact(imp, level = 1.5), "between 0 and 1")
  expect_error(plotREimpact(imp, level = 0), "between 0 and 1")
})

#' Calculate the predicted value for each observation across the distribution
#' of the random effect terms.
#'
#' \code{REmargins} calculates the average predicted value for each row of a
#' new data frame across the distribution of \code{\link{expectedRank}} for a
#' merMod object. This allows the user to make meaningful comparisons about the
#' influence of random effect terms on the scale of the response variable,
#' for user-defined inputs, and accounting for the variability in grouping terms.
#'
#' The function simulates the
#'
#' The function predicts the response at every level in the random effect term
#' specified by the user. Then, the expected rank of each group level is binned
#' to the number of bins specified by the user. Finally, a weighted mean of the
#' fitted value for all observations in each bin of the expected ranks is
#' calculated using the inverse of the variance as the weight -- so that less
#' precise estimates are downweighted in the calculation of the mean for the bin.
#' Finally, a standard error for the bin mean is calculated.
#'
#' @param merMod An object of class merMod
#'
#' @param newdata a data frame of observations to calculate group-level differences
#' for
#'
#' @param groupFctr The name of the grouping factor over which the random
#'   coefficient of interest varies.  This is the variable to the right of the
#'   pipe, \code{|}, in the [g]lmer formula. This parameter is optional, if not
#'   specified, it will perform the calculation for the first effect listed
#'   by \code{ranef}. If the length is > 1 then the combined effect of all
#'   listed groups will calculated and marginalized over co-occurences of those
#'   groups if desired.
#'
#' @param term The name of the random coefficient of interest. This is the
#'   variable to the left of the pipe, \code{|}, in the [g]lmer formula. Partial
#'   matching is attempted on the intercept term so the following character
#'   strings will all return rankings based on the intercept (\emph{provided that
#'   they do not match the name of another random coefficient for that factor}):
#'   \code{c("(Intercept)", "Int", "intercep", ...)}.
#'
#' @param breaks an integer representing the number of bins to divide the group
#' effects into, the default is 3.
#' @param .parallel, logical should parallel computation be used, default is TRUE
#'
#' @param ... additional arguments to pass to \code{\link{predictInterval}}
#'
#' @return A data.frame with all unique combinations of the number of cases, rows
#' in the newdata element:
#'   \describe{
#'     \item{...}{The columns of the original data taken from \code{newdata}}
#'     \item{case}{The row number of the observation from newdata. Each row in newdata will be
#'     repeated for all unique levels of the grouping_var, term, and breaks.}
#'     \item{grouping_var}{The grouping variable the random effect is being marginalized over.}
#'     \item{term}{The term for the grouping variable the random effect is being marginalized over.}
#'     \item{breaks}{The ntile of the effect size for \code{grouping_var} and \code{term}}
#'     \item{original_group_level}{The original grouping value for this \code{case}}
#'     \item{fit_combined}{The predicted value from \code{predictInterval} for this case simulated
#'     at the Nth ntile of the expected rank distribution of \code{grouping_var} and \code{term}}
#'     \item{upr_combined}{The upper bound of the predicted value.}
#'     \item{lwr_combined}{The lower bound of the predicted value.}
#'     \item{fit_XX}{For each grouping term in newdata the predicted value is decomposed into its
#'     fit components via predictInterval and these are all returned here}
#'     \item{upr_XX}{The upper bound for the effect of each grouping term}
#'     \item{lwr_XX}{The lower bound for the effect of each grouping term}
#'     \item{fit_fixed}{The predicted fit with all the grouping terms set to 0 (average)}
#'     \item{upr_fixed}{The upper bound fit with all the grouping terms set to 0 (average)}
#'     \item{lwr_fixed}{The lower bound fit with all the grouping terms set to 0 (average)}
#'   }
#'
#'
#' @references
#' Gatz, DF and Smith, L. The Standard Error of a Weighted Mean Concentration.
#' I. Bootstrapping vs other methods. \emph{Atmospheric Environment}.
#' 1995;11(2)1185-1193. Available at
#' \url{https://www.sciencedirect.com/science/article/pii/135223109400210C}
#'
#' Cochran, WG. 1977. Sampling Techniques (3rd Edition). Wiley, New York.
#'
#' @seealso \code{\link{expectedRank}}, \code{\link{predictInterval}}
#' @importFrom stats reshape
#' @examples
#' \donttest{
#' fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
#' mfx <- REmargins(merMod = fm1, newdata = sleepstudy[1:10,])
#'
#' # You can also pass additional arguments to predictInterval through REimpact
#'  g1 <- lmer(y ~ lectage + studage + (1|d) + (1|s), data=InstEval)
#'  margin_df <- REmargins(g1, newdata = InstEval[20:25, ], groupFctr = c("s"),
#'                         breaks = 4)
#'  margin_df <- REmargins(g1, newdata = InstEval[20:25, ], groupFctr = c("d"),
#'                          breaks = 3)
#' }
#' @export
REmargins <- function(merMod, newdata = NULL, groupFctr = NULL, term = NULL, breaks = 4,
                      .parallel = FALSE, ...){
  # Validate inputs
  if (is.null(groupFctr)) {
    # If the user doesn't tell us which term to use, we take the first term
    groupFctr <- names(ranef(merMod))[1]
  }
  if (is.null(newdata)) {
    # If the user doesn't give us data, we take the whole dataset
    # TODO - how does performance scale to a large number of observations?
    newdata <- merMod@frame
  }

  # If the user doesn't tell us what term to use, we take all the terms
  if (is.null(term)) {
    term <- names(ranef(merMod)[[groupFctr]])
    # Sub out intercept
    term[term == "(Intercept)"] <- "Intercept"
  }

  # This is a rough way to break the ER distribution into quantiles
  brks <- ceiling(seq(1, 100, by = 100/breaks))
  # Fallback so we always take a 99th percentile effect (for the maximum)
  if (!99 %in% brks) {
    brks <- c(brks, 99)
  }
  # Inputs are validated - now we get the effect distribution
  # Generate the expected rank distribution
  ER_DF <- expectedRank(merMod, groupFctr = groupFctr, term = term)
  # With many effects there is a lot of duplication - drop duplicated pctER
  ER_DF <- ER_DF[!duplicated(ER_DF[, c("groupFctr", "term", "pctER")]), ]

  # Now we create a data frame to capture the factor levels of each groupFctr that
  # correspond to the right break in the expectedRank distribution of the random
  # effect grouping factor and term
  par_df <- expand.grid("grouping_var" = groupFctr, "term" = term, "breaks" = 1:breaks)

  # Keep only factor levels that have effects at the margins
  # Need to match closest value here
  # Find N closest values
  # Drop duplicates
    # For each combination build an index of candidate rows/effect levels
  # Then choose the level that has the most precise estimate within a
  # tolerance of the effect size
  for (trm in term) {
    for (i in seq_along(brks)) {
      # Compute each terms distance from the break
      rank_dist <- abs(ER_DF$pctER[ER_DF$term == trm] - brks[i])
      # Get the index for the rank that minimizes the distance
      # TODO - how to break ties here?
      tmp <- which(rank_dist %in% rank_dist[order(rank_dist)][1])
      # Store the result in the par_df object
      par_df$groupLevel[par_df$breaks == i & par_df$term == trm] <- ER_DF$groupLevel[tmp]
    }
  }
  # Get ready to expand the data
  sim_data <- as.data.frame(lapply(newdata, rep, each = nrow(par_df)))
  # sim_data now repeats each row of newdata by the number of rows in par_df
  # case labels the rows with an integer for later mapping
  sim_data$case <- rep(1:nrow(newdata), each = nrow(par_df))
  sim_data <- cbind(sim_data, par_df)
  sim_data$original_group_level <- sim_data[, groupFctr]
  sim_data[, groupFctr] <- sim_data$groupLevel
  sim_data$groupLevel <- NULL
  #

  # Maybe strongly recommend parallel here?
  if (.parallel & requireNamespace("foreach", quietly = TRUE)) {
    # TODO use future here
      setup_parallel()
        out <- predictInterval(merMod, newdata = sim_data,
                               which = "all", ...)
        out_w <- stats::reshape(out, direction = "wide",
                                idvar = "obs", timevar = "effect",
                                v.names = c("fit", "upr", "lwr"), sep = "_")
        out_w$obs <- NULL
        sim_data <- cbind(sim_data, out_w)
    } else if ( .parallel & !requireNamespace("foreach", quietly = TRUE)) {
      warning("foreach package is unavailable, parallel computing not available")
    } else {
      out <- predictInterval(merMod, newdata = sim_data,
                             which = "all", ...)
      out_w <- stats::reshape(out, direction = "wide",
                       idvar = "obs", timevar = "effect",
                       v.names = c("fit", "upr", "lwr"), sep = "_")
      out_w$obs <- NULL
      sim_data <- cbind(sim_data, out_w)
    }
  # Case is the number of the row in newdata
  # obs is the variance among the selected random effects to marginalize over
  # So we want to collapse by case if we can

  return(sim_data)
}

# Test REmargins

test_that("Text marginalized effects object has the correct dimensions", {
  skip_on_travis()
  skip_on_cran()
  set.seed(51315)
  fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
  #

  #
  mfx <- REmargins(merMod = fm1, newdata = sleepstudy[1:10,])
  #
  # mfx has a row for each unique combo of row in newdata, breaks, grouping_var, and term
  expect_equal(nrow(mfx), 10 * length(unique(mfx$breaks)) * length(unique(mfx$grouping_var)) *
                 length(unique(mfx$term)))
  expect_equal(ncol(mfx), 17)


})

# ggplot(out_w) + aes(x = obs, y = fit_Subject) +
#   geom_line() +
#   facet_wrap(~case)

#Helpers

# Function to take only rows that form distinct levels of factors
# Need to figure out how to build a model matrix better.
trimModelFrame <- function(data){
  # Identify numerics
  nums <- sapply(data, is.numeric)
  vars <- names(nums[!nums == TRUE])
  dataList <- vector(mode = "list", length = length(vars))
  names(dataList) <- vars
    for(i in vars){
      dataList[[i]] <- data[!duplicated(data[, i]), ,drop=FALSE]
    }
    newdat <- do.call(rbind, dataList)
    newdat <- newdat[!duplicated(newdat),]
    return(newdat)
}

# FROM LME4
residDF.merMod <- function(object) {
  npar <- length(object@beta) + length(object@theta) +
    object@devcomp[["dims"]][["useSc"]]
  nobs <- nrow(object@frame)
  ## TODO: how do we feel about counting the scale parameter ???
  return(nobs - npar)
}


# from ARM as.matrix.VarCorr
easyVarCorr <- function (varc, useScale, digits){
  # VarCorr function for lmer objects, altered as follows:
  #   1.  specify rounding
  #   2.  print statement at end is removed
  #   3.  reMat is returned
  #   4.  last line kept in reMat even when there's no error term
  sc <- attr(varc, "sc")[[1]]
  if(is.na(sc)) sc <- 1
  #                  recorr <- lapply(varc, function(el) el@factors$correlation)
  recorr <- lapply(varc, function(el) attr(el, "correlation"))
  #reStdDev <- c(lapply(recorr, slot, "sd"), list(Residual = sc))
  reStdDev <- c(lapply(varc, function(el) attr(el, "stddev")), list(Residual = sc))
  reLens <- unlist(c(lapply(reStdDev, length)))
  reMat <- array('', c(sum(reLens), 4),
                 list(rep('', sum(reLens)),
                      c("Groups", "Name", "Variance", "Std.Dev.")))
  reMat[1+cumsum(reLens)-reLens, 1] <- names(reLens)
  reMat[,2] <- c(unlist(lapply(reStdDev, names)), "")
  #                  reMat[,3] <- format(unlist(reStdDev)^2, digits = digits)
  #                  reMat[,4] <- format(unlist(reStdDev), digits = digits)
  reMat[,3] <- fround(unlist(reStdDev)^2, digits)
  reMat[,4] <- fround(unlist(reStdDev), digits)
  if (any(reLens > 1)) {
    maxlen <- max(reLens)
    corr <-
      do.call("rbind",
              lapply(recorr,
                     function(x, maxlen) {
                       x <- as(x, "matrix")
                       #   cc <- format(round(x, 3), nsmall = 3)
                       cc <- fround (x, digits)
                       cc[!lower.tri(cc)] <- ""
                       nr <- dim(cc)[1]
                       if (nr >= maxlen) return(cc)
                       cbind(cc, matrix("", nr, maxlen-nr))
                     }, maxlen))
    colnames(corr) <- c("Corr", rep("", maxlen - 1))
    reMat <- cbind(reMat, rbind(corr, rep("", ncol(corr))))
  }
  #    if (!useScale) reMat <- reMat[-nrow(reMat),]
  if (useScale<0) reMat[nrow(reMat),] <- c ("No residual sd", rep("",ncol(reMat)-1))
  return (reMat)
}

#' Count the number of random effect terms
#' @source From lme4 package
#' @keywords internal
reTermCount <- function(model){
  sum(unlist(lapply(as.list(VarCorr(model)), function(x) sqrt(length(x)))))
}

#' Get names of random effect terms in a model object
#' @param model a merMod object with random effect terms
#' @return a data.frame with rows for each term with columns naming the grouping
#' term and the effect type
#' @keywords internal
reTermNames <- function(model){
  tmp <- NA
  for(i in 1:length(names(ngrps(model)))){
    cons <- names(ngrps(model))[i]
    vars <- paste(cons, unlist(dimnames(VarCorr(model)[[i]])[1]), sep = "-")
    tmp <- c(tmp, vars)
  }
  tmp <- na.omit(tmp)
  tmp <- t(as.data.frame(strsplit(tmp, "-")))
  row.names(tmp) <- NULL
  colnames(tmp) <- c("group", "effect")
  tmp <- as.data.frame(tmp)
  tmp$group <- as.character(tmp$group)
  tmp$effect <- as.character(tmp$effect)
  return(tmp)
}

#' Clean formula
#' @description a function to modify the formula for a merMod object to create
#' a model matrix with all predictor terms in both the group level and fixed
#' effect level
#' @param model a merMod object from lme4
#' @return a formula object
#' @keywords internal
formulaBuild <- function(model){
  slopeFX <- setdiff(all.vars(model@call$formula), names(ngrps(model)))
  missVar <- setdiff(slopeFX, all.vars(nobars(model@call$formula)))
  newForm <- nobars(model@call$formula)
  if(length(missVar > 0)){
    newForm <- paste(Reduce(paste, deparse(newForm)), paste(missVar, collapse = " +"), sep = " + ")
  }
  newForm <- as.formula(newForm)
  return(newForm)
}

##' Random Effects formula only
##' @param f a model formula
##' @param response logical, should the result include the response
##' @return a formula
##' @keywords internal
reOnly <- function(f,response=FALSE) {
  response <- if (response && length(f)==3) f[[2]] else NULL
  reformulate(paste0("(", vapply(findbars(f), safeDeparse, ""), ")"),
              response=response)
}

safeDeparse <- function(x, collapse=" ") paste(deparse(x, 500L), collapse=collapse)

#' Build model matrix
#' @description a function to create a model matrix with all predictor terms in
#' both the group level and fixed effect level
#' @param model a merMod object from lme4
#' @param newdata a data frame to construct the matrix from
#' @param which a character which matrix to return,default is full matrix with fixed and
#' random terms, other options are "fixed" and "random"
#' @source Taken from predict.merMod in lme4
#' @import lme4
#' @keywords internal
buildModelMatrix <- function(model, newdata, which = "full"){
  X <- getME(model, "X")
  X.col.dropped <- attr(X, "col.dropped")
  if (is.null(newdata)) {
    newdata <- model@frame
  }
  RHS <- formula(substitute(~R,
                    list(R = RHSForm(formula(model, fixed.only=TRUE)))))
  Terms <- terms(model,fixed.only=TRUE)
  mf <- model.frame(model, fixed.only = FALSE)
  isFac <- vapply(mf, is.factor, FUN.VALUE = TRUE)
  isFac[attr(Terms,"response")] <- FALSE
  orig_levs <- if (length(isFac)==0) NULL else lapply(mf[isFac],levels)
  # Suppress warnings about non-factors classified as factors
  # These are false alarms related to grouping terms
  mfnew <- suppressWarnings(model.frame(delete.response(Terms),
                         newdata,
                         na.action="na.pass",
                         xlev=orig_levs)
  )
  X <- model.matrix(RHS, data=mfnew,
                      contrasts.arg=attr(X,"contrasts"))
  offset <- 0 # rep(0, nrow(X))
  tt <- terms(model)
  if (!is.null(off.num <- attr(tt, "offset"))) {
    for (i in off.num)
      offset <- offset + eval(attr(tt,"variables")[[i + 1]], newdata)
  }
  fit.na.action <- attr(mfnew,"na.action")
  if(is.numeric(X.col.dropped) && length(X.col.dropped) > 0) {
    X <- X[, -X.col.dropped, drop=FALSE]
  }

  re.form <- reOnly(formula(model)) # RE formula only
  newRE <- mkNewReTrms(object = model,
                         newdata = newdata, re.form, na.action="na.pass",
                         allow.new.levels = TRUE)
  ## reMat <- t(as.matrix(newRE$Zt))
  ## reMat <- as.matrix(reMat)
  reMat <- Matrix::t(newRE$Zt)  ## what breaks if we keep this sparse???
  colnames(reMat) <- rownames(newRE$Zt)
  mm <- cbind(X, reMat)
  if(which == "full"){
    return(mm)
  } else if(which == "fixed"){
    return(X)
  } else if(which == "random"){
    return(reMat)
  }

}

#' Calculate the intraclass correlation using mixed effect models
#'
#' @param outcome a character representing the variable of the outcome
#' @param group a character representing the name of the grouping term
#' @param data a data.frame
#' @param subset an optional subset
#'
#' @return a numeric for the intraclass correlation
#' @export
#' @import lme4
#' @examples
#' data(sleepstudy)
#' ICC(outcome = "Reaction", group = "Subject", data = sleepstudy)
ICC <- function(outcome, group, data, subset=NULL){
  fm1 <- as.formula(paste(outcome, "~", "1 + (1|", group, ")"))
  if(length(table(data[, outcome])) == 2){
    nullmod <- glmer(fm1, data = data, subset = subset, family = 'binomial')
  } else {
    nullmod <- lmer(fm1, data = data, subset = subset)
  }
  between <- as.numeric(attr(VarCorr(nullmod)[[1]], "stddev"))
  within <- arm::sigma.hat(nullmod)$sigma$data
  ICC <- between^2 / (within^2 + between^2)
  return(ICC)
}


#' Utility function to make RE terms objects
#' @param object a model object
#' @param newdata a data.frame to build RE terms for
#' @param re.form a random effect formula to simulate, generated by
#' \code{\link{reOnly}}
#' @param na.action an object describing how NA values should be handled in newdata
#' @param allow.new.levels logical, should new levels be allowed in factor variables
#' @return a random effect terms object for a merMod
#' @import lme4
#' @keywords internal
mkNewReTrms <- function(object, newdata, re.form=NULL, na.action=na.pass,
                        allow.new.levels=FALSE)
{
  if (is.null(newdata)) {
    rfd <- mfnew <- model.frame(object)
  } else {
    mfnew <- model.frame(delete.response(terms(object, fixed.only=TRUE)),
                         newdata, na.action=na.action)
    if(packageVersion("lme4") < "1.1.9"){
      old <- TRUE
    } else{
      old <- FALSE
    }

    if (old) {
      rfd <- na.action(newdata)
      if (is.null(attr(rfd,"na.action")))
        attr(rfd,"na.action") <- na.action
    } else {
      newdata.NA <- newdata
      if (!is.null(fixed.na.action <- attr(mfnew,"na.action"))) {
        newdata.NA <- newdata.NA[-fixed.na.action,]
      }
      tt <- delete.response(terms(object,random.only=TRUE))
      ## need to let NAs in RE components go through -- they're handled downstream
      rfd <- model.frame(tt,newdata.NA,na.action=na.pass)
      if (!is.null(fixed.na.action))
        attr(rfd,"na.action") <- fixed.na.action
    }
  }
  if (inherits(re.form, "formula")) {
    ## DROP values with NAs in fixed effects
    if (length(fit.na.action <- attr(mfnew,"na.action")) > 0) {
      newdata <- newdata[-fit.na.action,]
    }
    ## note: mkReTrms automatically *drops* unused levels
    # rfd = model frame
    ReTrms <- mkReTrms(findbars(re.form[[2]]), rfd)
    ## update Lambdat (ugh, better way to do this?)
    ReTrms <- within(ReTrms, Lambdat@x <- unname(getME(object,"theta")[Lind]))
    #
    if (!allow.new.levels && any(vapply(ReTrms$flist, anyNA, NA)))
      stop("NAs are not allowed in prediction data",
           " for grouping variables unless allow.new.levels is TRUE")
    ns.re <- names(re <- ranef(object, condVar = FALSE))
    nRnms <- names(Rcnms <- ReTrms$cnms)
    if (!all(nRnms %in% ns.re))
      stop("grouping factors specified in re.form that were not present in original model")
    new_levels <- lapply(ReTrms$flist, function(x) levels(factor(x)))
    ## fill in/delete levels as appropriate
    re_x <- Map(function(r,n) levelfun(r,n,allow.new.levels=allow.new.levels),
                re[names(new_levels)], new_levels)
    re_new <- lapply(seq_along(nRnms), function(i) {
      rname <- nRnms[i]
      if (!all(Rcnms[[i]] %in% names(re[[rname]])))
        stop("random effects specified in re.form that were not present in original model")
      re_x[[rname]][,Rcnms[[i]]]
    })
    re_new <- unlist(lapply(re_new, t))
  }
  Zt <- ReTrms$Zt
  attr(Zt, "na.action") <- attr(re_new, "na.action") <- attr(mfnew, "na.action")
  list(Zt=Zt, b=re_new, Lambdat = ReTrms$Lambdat)
}

#' Parse merMod formulas
#' @keywords internal
RHSForm <- function(form,as.form=FALSE) {
  rhsf <- form[[length(form)]]
  if (as.form) reformulate(deparse(rhsf)) else rhsf
}


#' Parse merMod levels
#' @keywords internal
levelfun <- function(x,nl.n,allow.new.levels=FALSE) {
  if (!all(nl.n %in% rownames(x))) {
    if (!allow.new.levels) stop("new levels detected in newdata")
      newx <- as.data.frame(matrix(0, nrow=length(nl.n), ncol=ncol(x),
                                 dimnames=list(nl.n, names(x))))
      newx[rownames(x),] <- x
      x <- newx
  }
  if (!all(r.inn <- rownames(x) %in% nl.n)) {
    x <- x[r.inn,,drop=FALSE]
  }
  return(x)
}

# Test helper functions
set.seed(51315)
# Trimming data frame----

test_that("Trimming results in correct size", {
  skip_on_cran()
  data(InstEval)
  trimDat <- merTools:::trimModelFrame(InstEval)
  expect_gt(nrow(InstEval), nrow( merTools:::trimModelFrame(InstEval)))
  expect_equal(nrow(trimDat), 4065)
  cbpp$obs <- 1:nrow(cbpp)
  d1 <- cbpp
  d1$y <- d1$incidence / d1$size
  gm2 <- glmer(y ~ period +
                  (1 | herd),
                family = binomial, data = d1, nAGQ = 9, weights = d1$size)
  trimDat <- merTools:::trimModelFrame(gm2@frame)
  expect_s3_class(trimDat, "data.frame")
  expect_equal(nrow(trimDat), 18)
})

test_that("Trimming does not corrupt order", {
  skip_on_cran()
  tmp <- InstEval[1:10, ]
  trimDat <- merTools:::trimModelFrame(InstEval)
  trimDat <- rbind(tmp, trimDat)
  expect_lt(nrow(trimDat), nrow(tmp) + nrow(InstEval))
  row.names(tmp) <- NULL
  row.names(trimDat) <- NULL
  expect_identical(tmp, trimDat[1:10, ])
})

# subBoot and Theta----
# context("subBoot and Theta")

test_that("Can extract theta from a fit model", {
  skip_on_cran()
  set.seed(404)
  d <- expand.grid(fac1=LETTERS[1:5], grp=factor(1:10),
                   obs=1:100)

  suppressMessages({
  d$y <- simulate( ~ fac1 + (1|grp),
                  newdata = d,
                  family = gaussian,
                  newparams = list(
                    "theta" = 0.22,
                    "beta" = c(2,1,3,4,7),
                    "sigma" = 0.23))[[1]]
  })

  subD <- d[sample(row.names(d), 1000),]
  g1 <- lmer(y~fac1+(1|grp), data=subD)

  g1b <- lm(y ~ fac1, data = subD)

  expect_equal(thetaExtract(g1), 0.2285, tolerance = 0.1)
  expect_error(thetaExtract(g1b))

  z1 <- suppressMessages({
    subBoot(g1, 500, FUN = thetaExtract, R = 10)
  })
  expect_s3_class(z1, "data.frame")
  expect_equal(nrow(z1), 11)
  expect_equal(ncol(z1), 2)
})

# Test formula Build-----
# context("Test formula build")

test_that("Formula works for additive functions", {
  skip_on_cran()
  n <- 20
  x <- y <- rnorm(n)
  z <- rnorm(n)
  r <- sample(1:5, size=n, replace=TRUE)
  d <- data.frame(x,y,z,r)
  d2 <- expand.grid(a=factor(1:4),b=factor(1:4),rep=1:10)
  n <- nrow(d2)
  d2 <- transform(d2,r=sample(1:5, size=n, replace=TRUE),
                  z=rnorm(n))
  d2 <- subset(d2,!(a=="4" & b=="4"))
  fm <- lmer( z ~ a + b + (1|r), data=d2)
  expect_s3_class(merTools:::formulaBuild(fm), "formula")
  expect_identical(merTools:::formulaBuild(fm), as.formula("z ~ a + b"))
})


test_that("Formula works for interactions", {
  skip_on_cran()
  n <- 200
  x <- y <- rnorm(n)
  z <- rnorm(n)
  r <- sample(1:5, size=n, replace=TRUE)
  d <- data.frame(x,y,z,r)
  d2 <- expand.grid(a=factor(1:4),b=factor(1:4), c = factor(1:4), rep=1:10)
  n <- nrow(d2)
  d2 <- transform(d2,r=sample(1:5, size=n, replace=TRUE),
                  z=rnorm(n))
  d2 <- subset(d2,!(a=="4" & b=="4"))
  d2$x <- rnorm(nrow(d2))
  suppressMessages({
    fm <- lmer( z ~ a * b + c +  (1|r), data=d2)
  })

  expect_s3_class(merTools:::formulaBuild(fm), "formula")
  expect_identical(merTools:::formulaBuild(fm), as.formula("z ~ a * b + c"))
  suppressMessages({
    fm <- lmer( z ~ a * b * c +  (1|r), data=d2)
  })

  expect_s3_class(merTools:::formulaBuild(fm), "formula")
  expect_identical(merTools:::formulaBuild(fm), as.formula("z ~ a * b * c"))
  suppressMessages({
    fm <- lmer( z ~ a * b * c + x + I(x^2) + (1 + c|r), data=d2)
  })

  expect_s3_class(merTools:::formulaBuild(fm), "formula")
  expect_identical(merTools:::formulaBuild(fm), as.formula("z ~ a * b * c + x + I(x^2)"))
})


test_that("Build model matrix produces matrices of the right size", {
  skip_on_cran()
  d <- expand.grid(fac1 = LETTERS[1:5],
                   grp = letters[11:20],
                   obs = 1:50)
  suppressMessages({
    d$y <- simulate(~fac1 + (1 | grp), family = binomial,
                    newdata = d,
                    newparams = list( "theta" = c(.33),
                                      "beta" = c(2,-1,3,-2,1.2)),
                    seed =634)[[1]]

  })

  subD <- d[sample(row.names(d), 1200), ]

  g1 <- glmer(y~fac1+(1|grp), data=subD, family = 'binomial')
  d$fitted <- predict(g1, d)

  mm <- merTools:::buildModelMatrix(g1, newdata = d, which = "full")
  expect_true(inherits(mm, "matrix") || inherits(mm, "Matrix"))
  expect_equal(dim(mm), c(2500, 15))

}
)

# Precompiled vignettes that depend on API key
# Must manually move image files from eia/ to eia/vignettes/ after knit

library(knitr)
knit("vignettes/Using_predictInterval.Rmd.orig",
     "vignettes/Using_predictInterval.Rmd")

knit("vignettes/merToolsIntro.Rmd.orig",
     "vignettes/merToolsIntro.Rmd")

knit("vignettes/marginal_effects.Rmd.orig",
     "vignettes/marginal_effects.Rmd")
knit("vignettes/imputation.Rmd.orig",
     "vignettes/imputation.Rmd")

#' @title Extracts random effects
#' @name REextract
#' @description Extracts random effect terms from an lme4 model
#' @param merMod a merMod object from the lme4 package
#' @return a data frame with the following columns
#' \describe{
#'   \item{groupFctr}{The name of the grouping factor associated with the random effects}
#'   \item{groupID}{The level of the grouping factor associated with the random effects}
#'   \item{'term'}{One column per random effect, the name is derived from the merMod}
#'   \item{'term'_se}{One column per random effect, the name is derived from the merMod}
#' }
#' @examples
#' m2 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
#' rfx <- REextract(m2)
#' #Note the column names
#' head(rfx)
#' @export
REextract <- function(merMod){
  stopifnot(inherits(merMod, "merMod"))
  
  out <- lme4::ranef(merMod, condVar = TRUE)
  lvlNames <- names(out)
  reDims <- length(out)
  tmp.out <- vector("list", reDims)
  for(i in c(1:reDims)){
    tmp.out[[i]] <- out[[i]]
    tmp.out[[i]]$groupFctr <- lvlNames[i]
    tmp.out[[i]]$groupID <- row.names(out[[i]])
    if(ncol(out[[i]]) > 1){
      tmp.out.se <- apply(attr(out[[i]], which = "postVar"), 3,
                          function(x) sqrt(diag(x)))
      tmp.out.se <- as.data.frame(t(tmp.out.se))
      colnames(tmp.out.se) <- paste0(names(out[[i]]), "_se")
      tmp.out[[i]] <- cbind(tmp.out[[i]], tmp.out.se)
    } else {
      tmp.out.se <- sapply(attr(out[[i]], which = "postVar"), sqrt)
      names(tmp.out.se) <- paste0(names(out[[i]]), "_se")
      tmp.out[[i]] <- cbind(tmp.out[[i]], tmp.out.se)
      names(tmp.out[[i]])[4] <-  paste0(names(out[[i]]), "_se")
    }
  }
  dat <- dplyr::bind_rows(tmp.out)
  # reorg output
  dat <- dat[, c("groupFctr", "groupID",
          names(dat)[!names(dat) %in% c("groupFctr", "groupID")])]
  return(dat)
}

#' Simulate random effects from merMod
#' \code{REsim} simulates random effects from merMod object posterior distributions
#' @param merMod a merMod object from the lme4 package
#' @param n.sims number of simulations to use
#' @param oddsRatio logical, should parameters be converted to odds ratios?
#' @param seed numeric, optional argument to set seed for simulations
#' @importFrom arm sim
#' @import lme4
#' @return a data frame with the following columns
#' \describe{
#'   \item{\code{groupFctr}}{Name of the grouping factor}
#'   \item{\code{groupID}}{Level of the grouping factor}
#'   \item{\code{term}}{Name of random term (intercept/coefficient)}
#'   \item{\code{mean}}{Mean of the simulations}
#'   \item{\code{median}}{Median of the simulations}
#'   \item{\code{sd}}{Standard deviation of the simulations, \code{NA} if \code{oddsRatio=TRUE}}
#' }
#' @details Use the Gelman sim technique to build empirical Bayes estimates.
#'  Uses the sim function in the arm package
#' @examples
#' require(lme4)
#' m2 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
#' re2 <- REsim(m2, 25)
#' head(re2)
#' @export
REsim <- function(merMod, n.sims = 200, oddsRatio = FALSE, seed=NULL){
  stopifnot(inherits(merMod, "merMod"))
  if (!is.null(seed))
    set.seed(seed)
  else if (!exists(".Random.seed", envir = .GlobalEnv))
    runif(1)

  mysim <- arm::sim(merMod, n.sims = n.sims)
  reDims <- length(mysim@ranef)
  tmp.out <- vector("list", reDims)
  names(tmp.out) <- names(mysim@ranef)
  for(i in c(1:reDims)){
    zed <- apply(mysim@ranef[[i]], c(2, 3),
                 function(x) as.data.frame(x) %>% dplyr::summarise_all(.funs = c("mean", "median", "sd")))
    zed <- bind_rows(zed)
    zed$X1 <- rep(dimnames(mysim@ranef[[i]])[[2]], length(dimnames(mysim@ranef[[i]])[[3]]))
    zed$X2 <- rep(dimnames(mysim@ranef[[i]])[[3]], each = length(dimnames(mysim@ranef[[i]])[[2]]))
    tmp.out[[i]] <- zed; rm(zed)
    tmp.out[[i]]$groupFctr <- names(tmp.out)[i]
    tmp.out[[i]]$X1 <- as.character(tmp.out[[i]]$X1)
    tmp.out[[i]]$X2 <- as.character(tmp.out[[i]]$X2)
  }
  dat <- do.call(rbind, tmp.out)
  dat$groupID <- dat$X1; dat$X1 <- NULL
  dat$term <- dat$X2; dat$X2 <- NULL
  dat <- dat[, c("groupFctr", "groupID", "term", "mean", "median", "sd")]
  rownames(dat) <- NULL
  if(oddsRatio == TRUE){
    dat$median <- exp(dat$median)
    dat$mean <- exp(dat$mean)
    dat$sd <- NA # don't know how to do SE of odds ratios currently
    return(dat)
  } else{
    return(dat)
  }
}

#' Simulate fixed effects from merMod
#' \code{FEsim} simulates fixed effects from merMod object posterior distributions
#' @param merMod a merMod object from the lme4 package
#' @param n.sims number of simulations to use
#' @param oddsRatio logical, should parameters be converted to odds ratios?
#' @param seed numeric, optional argument to set seed for simulations
#' @importFrom arm sim
#' @import lme4
#' @return a data frame with the following columns
#' \describe{
#'   \item{\code{term}}{Name of fixed term (intercept/coefficient)}
#'   \item{\code{mean}}{Mean of the simulations}
#'   \item{\code{median}}{Median of the simulations}
#'   \item{\code{sd}}{Standard deviation of the simulations, \code{NA} if \code{oddsRatio=TRUE}}
#' }
#' @details Use the Gelman sim technique to build fixed effect estimates and
#' confidence intervals. Uses the sim function in the arm package
#' @examples
#' require(lme4)
#' m2 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
#' fe2 <- FEsim(m2, 25)
#' head(fe2)
#' @export
FEsim <- function(merMod, n.sims = 200, oddsRatio=FALSE, seed=NULL){
  stopifnot(inherits(merMod, "merMod"))
  if (!is.null(seed))
    set.seed(seed)
  else if (!exists(".Random.seed", envir = .GlobalEnv))
    runif(1)

  mysim <- arm::sim(merMod, n.sims = n.sims)
  means <- apply(mysim@fixef, MARGIN = 2, mean)
  medians <- apply(mysim@fixef, MARGIN = 2, median)
  sds <- apply(mysim@fixef, MARGIN =2, sd)
  dat <- data.frame(term = names(means), mean = means,
                    median = medians,
                    sd = sds, row.names=NULL)
  if(oddsRatio == TRUE){
    dat$median <- exp(dat$median)
    dat$mean <- exp(dat$mean)
    dat$sd <- NA # don't know how to do SE of odds ratios currently
    return(dat)
  } else{
    return(dat)
  }
}

#' @title Estimate the Root Mean Squared Error (RMSE) for a lmerMod
#' @name RMSE.merMod
#' @description Extract the Root Mean Squared Error for a lmerMod object
#' @param merMod a lmerMod object from the lme4 package
#' @param scale logical, should the result be returned on the scale of
#' response variable standard deviations?
#' @import lme4
#' @return a numeric which represents the RMSE
#' @examples
#' require(lme4)
#' m2 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
#' RMSE.merMod(m2)
#' @export
RMSE.merMod <- function(merMod, scale = FALSE){
  stopifnot(inherits(merMod, "lmerMod") ||  inherits(merMod, "blmerMod"))
  # Express RMSE as percentage of dependent variable standard deviation
  dvSD <- sd(merMod@frame[, 1])
  RMSE <- sqrt(mean(residuals(merMod)^2))
  if(scale == TRUE){
    return(RMSE/dvSD)
  } else{
    return(RMSE)
  }
}

# 'seed' options in draw, REsim, FEsim, predictInterval and subBoot----


test_that("Equivalent seeds return equivalent results", {
  fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)

  d1a <- draw(fm1, type="random", seed=1234)
  d2  <- draw(fm1, type="random", seed=456)
  d1b <- draw(fm1, type="random", seed=1234)

  r1a <- REsim(fm1, 25, seed=1234)
  r2  <- REsim(fm1, 25, seed=456)
  r1b <- REsim(fm1, 25, seed=1234)

  f1a <- FEsim(fm1, 25, seed=1234)
  f2  <- FEsim(fm1, 25, seed=456)
  f1b <- FEsim(fm1, 25, seed=1234)

  # TODO - subboot now returns warnings and needs to be checked
  p1a <- predictInterval(fm1, newdata=sleepstudy[1:10,], seed=1234)
  p2  <- predictInterval(fm1, newdata=sleepstudy[1:10,], seed=456)
  p1b <- predictInterval(fm1, newdata=sleepstudy[1:10,], seed=1234)

  s1a <- suppressMessages({
    subBoot(fm1, n = 160, FUN = thetaExtract, R = 20, seed=1234)
  })
  suppressMessages({
  s2  <- subBoot(fm1, n = 160, FUN = thetaExtract, R = 20, seed=456)
  })
  suppressMessages({
    s1b <- subBoot(fm1, n = 160, FUN = thetaExtract, R = 20, seed=1234)
  })


  expect_identical(d1a, d1b)
  expect_identical(r1a, r1b)
  expect_identical(f1a, f1b)
  expect_identical(p1a, p1b)
  expect_identical(s1a, s1b)

  expect_false(identical(d1a, d2))
  expect_false(identical(r1a, r2))
  expect_false(identical(f1a, f2))
  expect_false(identical(p1a, p2))
  expect_false(identical(s1a, s2))
})

# -----------------------------------------------------
# Test framework includes tests for multiple intercepts and
# multiple slopes to ensure extraction of random effects
# works in these scenarios
#-------------------------------------------------------

set.seed(51315)
library(lme4)
data(grouseticks)
grouseticks$HEIGHT <- scale(grouseticks$HEIGHT)
grouseticks <- merge(grouseticks, grouseticks_agg[, 1:3], by = "BROOD")

# Build out models
form <- TICKS ~ YEAR + HEIGHT +(1|BROOD) + (1|INDEX) + (1|LOCATION)
glmer3Lev  <- glmer(form, family="poisson",data=grouseticks,
                     control = glmerControl(optimizer="Nelder_Mead",
                                            optCtrl=list(maxfun = 1e5)))
# GLMER 3 level + slope
form <- TICKS ~ YEAR + HEIGHT +(1 + HEIGHT|BROOD) + (1|LOCATION) + (1|INDEX)
glmer3LevSlope  <- glmer(form, family="poisson",data=grouseticks,
                     control = glmerControl(optimizer="bobyqa",
                                            optCtrl=list(maxfun = 1e5)))

# Sleepstudy
lmerSlope1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)

# Wackier example
data(Orthodont,package="nlme")
Orthodont$nsex <- as.numeric(Orthodont$Sex=="Male")
Orthodont$nsexage <- with(Orthodont, nsex*age)
lmerSlope2 <- lmer(distance ~ age + (0 + age + nsex|Subject), data=Orthodont)

###############################################
# Extract Random Effects from merMod----
################################################

test_that("REextract pulls out a data frame", {
  expect_s3_class(REextract(lmerSlope2), "data.frame")
  expect_s3_class(REextract(glmer3LevSlope), "data.frame")
  expect_s3_class(REextract(glmer3Lev), "data.frame")
  expect_s3_class(REextract(lmerSlope1), "data.frame")
})

test_that("REextract issues error with non merMod objects", {
  expect_error(REextract(lm(Reaction ~ Days, sleepstudy)))
  expect_error(REextract(glm(TICKS ~ YEAR + HEIGHT,
                             family="poisson", data=grouseticks)))
})

test_that("REextract gets correct dimensions", {
  expect_equal(ncol(REextract(glmer3Lev)), 4)
  expect_equal(ncol(REextract(lmerSlope1)), 6)
  expect_equal(ncol(REextract(lmerSlope2)), 6)
  expect_equal(ncol(REextract(glmer3LevSlope)), 6)
  expect_equal(nrow(REextract(glmer3Lev)), 584)
  expect_equal(nrow(REextract(lmerSlope1)), 18)
  expect_equal(nrow(REextract(lmerSlope2)), 27)
  expect_equal(nrow(REextract(glmer3LevSlope)), 584)
})

# Check names
# Check numerics

###############################################
# Fixed effect estimates from posterior----

################################################

test_that("FEsim produces data.frames", {
  expect_s3_class(FEsim(lmerSlope1, n.sims=100), "data.frame")
  expect_s3_class(FEsim(lmerSlope2, n.sims=100), "data.frame")
  expect_s3_class(FEsim(glmer3Lev, n.sims=100), "data.frame")
  expect_s3_class(FEsim(glmer3LevSlope, n.sims=100), "data.frame")
})

test_that("n.sims changes simulation results", {
  expect_false(identical(FEsim(lmerSlope1, n.sims = 1000),
                         FEsim(lmerSlope1, n.sims = 10)))
})

# numeric checks

###############################################
# Random effect estimates from posterior----
################################################

test_that("REsim produces data.frames", {
  expect_s3_class(REsim(lmerSlope1, n.sims=100), "data.frame")
  expect_s3_class(REsim(lmerSlope2, n.sims=100), "data.frame")
  expect_s3_class(REsim(glmer3Lev, n.sims=100), "data.frame")
  expect_s3_class(REsim(glmer3LevSlope, n.sims=100), "data.frame")
})


###############################################
# RMSE estimates----
################################################

test_that("RMSE produces correct variable types", {
  expect_type(RMSE.merMod(lmerSlope1), "double")
  expect_type(RMSE.merMod(lmerSlope2), "double")
  expect_type(RMSE.merMod(lmerSlope1, scale = TRUE), "double")
  expect_type(RMSE.merMod(lmerSlope2, scale = TRUE), "double")
})


test_that("RMSE respects scale parameter", {
  expect_false(identical(RMSE.merMod(lmerSlope1),
                         RMSE.merMod(lmerSlope1, scale = TRUE)))
  expect_false(identical(RMSE.merMod(lmerSlope2),
                         RMSE.merMod(lmerSlope2, scale = TRUE)))
  expect_lt(RMSE.merMod(lmerSlope2, scale = TRUE),
                   RMSE.merMod(lmerSlope2))
  expect_lt(RMSE.merMod(lmerSlope1, scale = TRUE),
                   RMSE.merMod(lmerSlope1))
})

#------------------------------------------------------------------------------
# ShinyMer
#-----------------------------------------------------------------------------

## Data and models
n <- 20
x <- y <- rnorm(n)
z <- rnorm(n)
r <- sample(1:5, size=n, replace=TRUE)
d <- data.frame(x,y,z,r)
d2 <- expand.grid(a=factor(1:4),b=factor(1:4),rep=1:10)
n <- nrow(d2)
d2 <- transform(d2,r=sample(1:5, size=n, replace=TRUE),
                z=rnorm(n))
d2 <- subset(d2,!(a=="4" & b=="4"))
fm <- lmer( z ~ a*b + (1|r), data=d2)
fm2 <- lmer( z ~ a*b + (1+b|r), data=d2)


data(grouseticks)
grouseticks$HEIGHT <- scale(grouseticks$HEIGHT)
grouseticks <- merge(grouseticks, grouseticks_agg[, 1:3], by = "BROOD")
grouseticks$TICKS_BIN <- ifelse(grouseticks$TICKS >=1, 1, 0)
# GLMER 3 level + slope
form <- TICKS_BIN ~ YEAR + (1 + HEIGHT|BROOD) + (1|LOCATION) + (1|INDEX)
glmer3LevSlope  <- glmer(form, family="binomial",data=grouseticks,
                         control = glmerControl(optimizer="bobyqa",
                                                optCtrl=list(maxfun = 1e5)))

set.seed(3845)
d <- expand.grid(fac1=LETTERS[1:5], grp=factor(1:10), fac2 = LETTERS[10:20],
                 obs=1:25)
d$x <- runif(nrow(d))
d$y <- simulate(~ x + fac1 + fac2 + (1 + fac1|grp) + (1|obs), family = binomial,
                newdata=d,
                newparams=list(beta = rnorm(16),
                               theta = rnorm(16, 5, 1)))[[1]]
subD <- d[sample(row.names(d), 5000),]

g1 <- glmer(y ~ x + fac1 + fac2 + (1+fac1|grp) + (1|obs), data = subD, family = 'binomial')
g2 <- lmer(y ~ lectage + studage + (1+lectage|d) + (1|dept), data=InstEval)
# ----------------------------------------------------
# shinyMer blocks

shinyMer(fm)
shinyMer(fm2)
shinyMer(g1)
shinyMer(g2)
shinyMer(glmer3LevSlope)
shinyMer(glmer3LevSlope, simData = glmer3LevSlope@frame[22:25,])

#SERVER----
server = function(input, output){
  output$text1 <- renderText({
    paste("You have selected", input$stat)
  })

  predInput <- reactive({
    data <- switch(input$newdataType,
                   "orig" = merMod@frame,
                   "mean" = draw(merMod, type = "average"),
                   "rand" = draw(merMod, type = "random"),
                   "user" = newdata)
    cbind(predictInterval(merMod, newdata = data, level = input$alpha/100,
                          type = input$predMetric,
                          include.resid.var = input$resid.var,
                          n.sims = input$n.sims, stat = input$stat), data)
  })

  if ("DT" %in% rownames(installed.packages())) {
    output$dt <- renderDataTable({
      predInput()
    })
  } else {
    output$dt <- renderTable({
      predInput()
    })
  }

  output$downloadData <- shiny::downloadHandler(
    filename = "predictIntervalResults.csv",
    content = function(file) {
      write.csv(shiny::isolate(predInput()), file)
    }
  )

  output$predPlot <- renderPlot({
    data <- predInput()
    data$x <- factor(seq(1:nrow(data)))
    ggplot(data, aes(x = x, y = fit, ymin = lwr, ymax = upr)) +
      geom_pointrange() +
      theme_bw() + theme(axis.text.x = element_blank(),
                         panel.grid.major.x = element_blank(),
                         panel.grid.minor.x = element_blank(),
                         axis.ticks.x = element_blank())
  })

  feData <- reactive({
    data <- FEsim(merMod, n.sims = input$n.sims)
    return(data)
  })

  output$feplot <- renderPlot({
    plotdf <- feData()
    scale <- input$alpha/100
    vartmp <- input$stat
    plotFEsim(plotdf, level = scale, stat = vartmp, sd = TRUE,
              intercept = FALSE)
  })

  reData <- reactive({
    data <- REsim(merMod, n.sims = input$n.sims)
    return(data)
  })

  output$replot <- renderPlot({
    plotdf <- reData()
    scale <- input$alpha/100
    vartmp <- input$stat
    plotREsim(plotdf, level = scale, stat = vartmp, sd = TRUE)
  })

  output$call <- renderPrint({
    merMod@call
  })

  reEffInput <- reactive({
    data <- switch(input$newdataType,
                   "orig" = merMod@frame,
                   "mean" = draw(merMod, type = "average"),
                   "rand" = draw(merMod, type = "random"),
                   "user" = newdata)
    if(nrow(data) > 12){
      warning("Too much data selected, only using top 12 rows.")
      data <- data[1:12, ]
    }
    return(data)
  })

  groupData <- reactive({
    plotdf <- REimpact(merMod, newdata = reEffInput(),
                       groupFctr = input$group,
                       term = input$term,
                       level = input$alpha/100,
                       breaks = input$nbin,
                       type = input$predMetric,
                       include.resid.var = input$resid.var,
                       n.sims = input$n.sims, stat = input$stat)
    plotdf$upr <- qnorm(input$alpha/100) * plotdf$AvgFitSE
    plotdf$lwr <- qnorm(input$alpha/100) * plotdf$AvgFitSE
    plotdf$upr <- plotdf$AvgFit + plotdf$upr
    plotdf$lwr <- plotdf$AvgFit - plotdf$lwr
    plotdf$bin <- factor(plotdf$bin)
    return(plotdf)
  })

  output$gPlot <- renderPlot({
    ggplot(groupData(), aes(x = bin, y = AvgFit, ymin = lwr, ymax = upr)) +
      geom_pointrange() + facet_wrap(~case) +
      theme_bw() + labs(x = "Bin", y = "Value of DV",
                        title = "Impact of grouping term for selected case")
  })

  wiggleData <- reactive({
    valLookup <- unique(merMod@frame[, input$fixef])
    if(class(valLookup) %in% c("numeric", "integer")){
      newvals <- seq(min(valLookup), max(valLookup), length.out = 20)
    } else{
      if(length(valLookup) < 50){
        newvals <- newvals
      } else{
        newvals <- sample(newvals, 50)
      }
    }
    plotdf <- wiggle(reEffInput(), input$fixef, values = list(newvals))
    plotdf <- cbind(plotdf, predictInterval(merMod, newdata=plotdf,
                                            type = input$predMetric,
                                            level = input$alpha/100,
                                            include.resid.var = input$resid.var,
                                            n.sims = input$n.sims, stat = input$stat))
    plotdf$X <- plotdf[, input$fixef]
    plotdf$case <- rep(1:length(newvals), length = nrow(reEffInput()))
    return(plotdf)
  })

  output$re.ui <- renderUI({
    choices <- names(ranef(merMod)[[input$group]])
    selectInput("term", "Group Term:",
                choices = choices,
                selected = choices[1])
  })

  output$wigglePlot <- renderPlot({
    ggplot(wiggleData(), aes(x = X, y = fit, ymin = lwr,
                             ymax = upr)) +
      geom_pointrange() + facet_wrap(~case) +
      theme_bw() + labs(y = "Simulated Value of DV",
                        title = "Impact of selected fixed effect for
                        selected cases.")
  })

  }

#' Calculate the expected rank of random coefficients that account for
#' uncertainty.
#'
#' \code{expectedRank} calculates the expected rank and the percentile expected
#' rank of any random term in a merMod object.  A simple ranking of the estimated
#' random effects (as produced by \code{\link[lme4]{ranef}}) is not satisfactory
#' because it ignores any amount of uncertainty.
#'
#' Inspired by Lingsma et al. (2010, see also Laird and Louis 1989),
#' expectedRank sums the probability that each level of the grouping factor is
#' greater than every other level of the grouping factor, similar to a
#' two-sample t-test.
#'
#' The formula for the expected rank is:
#' \deqn{ExpectedRank_i = 1 + \sum \phi((\theta_i - \theta_k) / \sqrt(var(\theta_i)+var(\theta_k))}
#' where \eqn{\phi} is the standard normal distribution function, \eqn{\theta}
#' is the estimated random effect and \eqn{var(\theta)} is the posterior
#' variance of the estimated random effect. We add one to the sum so that the
#' minimum rank is one instead of zero so that in the case where there is no
#' overlap between the variances of the random effects (or if the variances are
#' zero), the expected rank equals the actual rank.  The ranks are ordered such
#' that the winners have ranks that are greater than the losers.
#'
#' The formula for the percentile expected rank is:
#' \deqn{100 * (ExpectedRank_i - 0.5) / N_grps}
#' where \eqn{N_grps} is the number of grouping factor levels. The percentile
#' expected rank can be interpreted as the fraction of levels that score at or
#' below the given level.
#'
#' NOTE: \code{expectedRank} will only work under conditions that \code{lme4::ranef}
#' will work. One current example of when this is \emph{not} the case is for
#' models when there are multiple terms specified per factor (e.g. uncorrelated random
#' coefficients for the same term, e.g.
#' \code{lmer(Reaction ~ Days + (1 | Subject) + (0 + Days | Subject), data = sleepstudy)})
#'
#' @param merMod An object of class merMod
#'
#' @param groupFctr An optional character vector specifying the name(s) the grouping factor(s)
#'   over which the random coefficient of interest varies.  This is the
#'   variable to the right of the pipe, \code{|}, in the [g]lmer formula.
#'   This parameter is optional. If none is specified all terms will be returned.
#'
#' @param term An optional character vector specifying the name(s) of the random coefficient of interest. This is the
#'   variable to the left of the pipe, \code{|}, in the [g]lmer formula. Partial
#'   matching is attempted on the intercept term so the following character
#'   strings will all return rankings based on the intercept (\emph{provided that
#'   they do not match the name of another random coefficient for that factor}):
#'   \code{c("(Intercept)", "Int", "intercep", ...)}.
#'
#' @return A data.frame with the following five columns:
#'   \describe{
#'     \item{groupFctr}{a character representing name of the grouping factor}
#'     \item{groupLevel}{a character representing the level of the grouping factor}
#'     \item{term}{a character representing the formula term for the group}
#'     \item{estimate}{effect estimate from \code{lme4::ranef(, condVar=TRUE)}).}
#'     \item{std.error}{the posterior variance of the estimate random effect
#'                      (from \code{lme4::ranef(, condVar=TRUE)}); named "\code{term}"_var.}
#'     \item{ER}{The expected rank.}
#'     \item{pctER}{The percentile expected rank.}
#'   }
#'
#' @references
#' Laird NM and Louis TA. Empirical Bayes Ranking Methods. \emph{Journal of
#' Education Statistics}. 1989;14(1)29-46. Available at
#' \url{http://www.jstor.org/stable/1164724}.
#'
#'
#' Lingsma HF, Steyerberg EW, Eijkemans MJC, et al. Comparing and
#' ranking hospitals based on outcome: results from The Netherlands Stroke Survey.
#' \emph{QJM: An International Journal of Medicine}. 2010;103(2):99-108.
#' doi:10.1093/qjmed/hcp169
#'
#' @examples
#' \donttest{
#' #For a one-level random intercept model
#' m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
#' (m1.er <- expectedRank(m1))
#'
#' #For a one-level random intercept model with multiple random terms
#' m2 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
#' #ranked by the random slope on Days
#' (m2.er1 <- expectedRank(m2, term="Days"))
#' #ranked by the random intercept
#' (m2.er2 <- expectedRank(m2, term="int"))
#'
#' #For a two-level model with random intercepts
#' m3 <- lmer(y ~ service * dept + (1|s) + (1|d), InstEval)
#' #Ranked by the random intercept on 's'
#' (m3.er1 <- expectedRank(m3, groupFctr="s", term="Intercept"))
#' }
#' @export
expectedRank <- function(merMod, groupFctr=NULL, term=NULL) {
  #Count random terms in merMod
  n.rfx <- lme4::getME(merMod, "k")
  n.rfac <- lme4::getME(merMod, "n_rfac")
  rfx <- lme4::ranef(merMod, condVar=TRUE)

  if(!is.null(groupFctr)){
    groupFctr <- groupFctr
  } else{
    groupFctr <- names(rfx)
  }
  out <- data.frame(groupFctr = NA, groupLevel = NA, term = NA,
                    estimate = NA, std.error = NA,
                    ER = NA, pctER = NA)

  for(i in groupFctr){
      rfx.names <- rownames(rfx[[i]])
      n.grps <- length(rfx.names)
      n.terms <- length(rfx[[i]])
      if(!is.null(term)){
        termIdx <- term
      } else{
        termIdx <- names(rfx[[i]])
      }
      for(j in termIdx){
        if (all(grepl("[iI]nt[a-z]*", j)) & is.na(match(j, names(rfx[[i]])))) {
           j <- "(Intercept)"
        }

        term.idx <- grep(j, names(rfx[[i]]), fixed=TRUE)
        theta <- rfx[[i]][,term.idx]
        var.theta <- attr(rfx[[i]], which="postVar")[term.idx, term.idx, 1:n.grps]
        #Calculate Expected Rank which is the sum of the probabilities that group i is greater than all
        #other groups j (assuming normal distribution of random effects)
        ER <- pctER <- rep(NA, n.grps)
        for (k in 1:n.grps) {
          ER[k] <- 1 + sum(pnorm((theta[k]-theta[-k]) / sqrt(var.theta[k] + var.theta[-k])))
        }
        #Calculated percentile expected rank ... the version of the formula I am using is
        #the percentage of groups that are ranked **equal to or less than** the selected
        #group ... if we just wanted percentage ranked less than then remove the 0.5
        pctER <- round(100 * (ER - 0.5) / n.grps)

        tmp <- data.frame(groupFctr = i, groupLevel = rfx.names, term = j,
                          estimate = theta, std.error = var.theta,
                          ER = ER, pctER = pctER)
        out <- rbind(out, tmp)
      }
  }

  out <- out[-1, ]
  # Avoid parentheses in parameter names
  out$term <- gsub("(Intercept)", "Intercept", out$term, fixed = TRUE)
  return(out)
}

# Test expected rank

#Using 2 of sample models from test_merExtract.R

set.seed(51315)
library(lme4)

###############################################
# Testing expected rank----
###############################################
test_that("expectedRank parameters work and dont work as intended", {
  skip_on_cran()
  #########################################

  # Sleepstudy
  m1  <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)

  m2 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)

  # Wackier example
  data(Orthodont,package = "nlme")
  Orthodont$nsex <- as.numeric(Orthodont$Sex == "Male")
  Orthodont$nsexage <- with(Orthodont, nsex*age)
  m3 <- lmer(distance ~ age + (0 + age + nsex|Subject), data=Orthodont)

  # two grouping factors
  data(grouseticks)
  grouseticks$HEIGHT <- scale(grouseticks$HEIGHT)
  grouseticks <- merge(grouseticks, grouseticks_agg[, 1:3], by = "BROOD")

  form <- TICKS ~ YEAR + HEIGHT +(HEIGHT|BROOD) + (1|INDEX)
  m4  <- glmer(form, family="poisson",data=grouseticks,
               control = glmerControl(optimizer="Nelder_Mead",
                                      optCtrl=list(maxfun = 1e5)))

  form <- TICKS ~ YEAR + HEIGHT + (0 + HEIGHT|BROOD) + (1|INDEX)
  m5  <- glmer(form, family="poisson",data=grouseticks,
               control = glmerControl(optimizer="Nelder_Mead",
                                      optCtrl=list(maxfun = 1e5)))

  #Custom Expectation Functions
  expect_correct_dim <- function(merMod, groupFctr=NULL, term=NULL) {
    if (is.null(groupFctr)) {
      n.levels <- nrow(ranef(merMod)[[1]])
    } else {
      n.levels <- nrow(ranef(merMod)[[groupFctr]])
    }
    ER <- expectedRank(merMod, groupFctr, term)
    testthat::expect_true(nrow(ER) == n.levels &
                            ncol(ER) == 7 &
                            all(colnames(ER)[6:7] == c("ER", "pctER")) &
                            class(ER) == "data.frame")
  }


  ##################################


  expect_correct_dim(m1)
  expect_correct_dim(m1, groupFctr="Subject")
  expect_correct_dim(m1, term="(Intercept)")
  expect_correct_dim(m1, groupFctr="Subject", term="(Intercept)")

  expect_correct_dim(m2, term="(Intercept)")
  expect_correct_dim(m2, term="Days")

  expect_correct_dim(m3, groupFctr="Subject", term="age")
  expect_correct_dim(m3, groupFctr="Subject", term="nsex")

  expect_correct_dim(m4, groupFctr="BROOD", term="(Intercept)")
  expect_correct_dim(m4, groupFctr="INDEX", term="(Intercept)")

  expect_correct_dim(m5, groupFctr="BROOD")
  expect_correct_dim(m5, groupFctr="INDEX")

  # expect_error(expectedRank(m4), "Must specify which grouping factor when there are more than one")
  # expect_error(expectedRank(m4, groupFctr="BROOD"), "Must specify which random coefficient when there are more than one per selected grouping factor")
  # expect_error(expectedRank(m3, groupFctr="Subject"), "Must specify which random coefficient when there are more than one per selected grouping factor")
  # expect_error(expectedRank(m3, term="int"), "undefined columns selected")
})

test_that("Ranks have the correct range", {
  skip_on_cran()
  # Sleepstudy
  m1  <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)


  numGrps <- nrow(ranef(m1)[[1]])
  expect_true(max(expectedRank(m1)$ER) <= numGrps)
  expect_true(min(expectedRank(m1)$ER) >= 1)
  expect_equal(cor(expectedRank(m1)$ER, rank(ranef(m1)[[1]])), 0.99, tolerance = .01)
})

test_that("Percentile ranks have the correct range", {
  m1  <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
   expect_true(max(expectedRank(m1)$pctER) <= 100)
   expect_true(min(expectedRank(m1)$pctER) >= 0)
})

#' @title Clean up variable names in data frames
#' @name sanitizeNames
#' @description Strips out transformations from variable names in data frames
#' @param data a data.frame
#' @return a data frame with variable names cleaned to remove factor() construction
sanitizeNames <- function(data){
  badFac <- grep("factor\\(", names(data))
  for(i in badFac){
    names(data)[i] <- gsub("factor\\(", "", names(data)[i])
    names(data)[i] <- gsub("\\)", "", names(data)[i])
  }
  row.names(data) <- NULL
  return(data)
}

#' @title Remove attributes from a data.frame
#' @name stripAttributes
#' @description Strips attributes off of a data frame that come with a merMod model.frame
#' @param data a data.frame
#' @return a data frame with variable names cleaned to remove all attributes except for
#' names, row.names, and class
stripAttributes <- function(data){
  attr <- names(attributes(data))
  good <- c("names", "row.names", "class")
  for(i in attr[!attr %in% good]){
    attr(data, i) <- NULL
  }
  return(data)
}


#' @title Draw a single observation out of an object matching some criteria
#' @name draw
#' @description Draw is used to select a single observation out of an R object.
#' Additional parameters allow the user to control how that observation is
#' chosen in order to manipulate that observation later. This is a generic
#' function with methods for a number of objects.
#' @param object the object to draw from
#' @param type what kind of draw to make. Options include random or average
#' @param varList a list specifying filters to subset the data by when making the
#' draw
#' @param seed numeric, optional argument to set seed for simulations, ignored if type="average"
#' @param ... additional arguments required by certain methods
#' @return a data.frame with a single row representing the desired observation
#' @details In cases of tie, ".", may be substituted for factors.
#' @export draw
#' @rdname draw
draw <- function(object, type = c("random", "average"),
                 varList = NULL, seed = NULL, ...){
  UseMethod("draw")
}

#' @title Draw an observation from a merMod object
#' @rdname draw
#' @method draw merMod
#' @export
#' @import lme4
#' @examples
#' fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
#' # Random case
#' draw(fm1, type = "random")
#' # Average
#' draw(fm1, type = "average")
#' # Subset
#' draw(fm1, type = "average", varList = list("Subject" = "308"))
#'
draw.merMod <- function(object, type = c("random", "average"),
                        varList = NULL, seed = NULL, ...){
  type <- match.arg(type, c("random", "average"), several.ok = FALSE)
  if(type == 'random'){
    out <- randomObs(object, varList, seed)
    return(out)
  } else if(type == 'average'){
    out <- averageObs(object, varList, ...)
    return(out)
  }
}

#' @title Select a random observation from model data
#' @name randomObs
#' @description Select a random observation from the model frame of a merMod
#' @param merMod an object of class merMod
#' @param varList optional, a named list of conditions to subset the data on
#' @param seed numeric, optional argument to set seed for simulations
#' @return a data frame with a single row for a random observation, but with full
#' factor levels. See details for more.
#' @details Each factor variable in the data frame has all factor levels from the
#' full model.frame stored so that the new data is compatible with predict.merMod
#' @export
randomObs <- function(merMod, varList, seed = NULL){
  if(!missing(varList)){
    data <- subsetList(merMod@frame, varList)
  }

  if (!is.null(seed))
    set.seed(seed)
  else if (!exists(".Random.seed", envir = .GlobalEnv))
    runif(1)

  out <- data[sample(1:nrow(data), 1),]
  chars <- !sapply(out, is.numeric)
  for(i in names(out[, chars])){
    out[, i] <- superFactor(out[, i], fullLev = unique(merMod@frame[, i]))
  }
  out <- stripAttributes(out)
  return(out)
}

#' @title Collapse a dataframe to a single average row
#' @name collapseFrame
#' @description Take an entire dataframe and summarize it in one row by using the
#' mean and mode.
#' @param data a data.frame
#' @return a data frame with a single row
#' @details Each character and factor variable in the data.frame is assigned to the
#' modal category and each numeric variable is collapsed to the mean. Currently if
#' mode is a tie, returns a "."
collapseFrame <- function(data){
  chars <- !sapply(data, is.numeric)
  chars <- names(data[, chars, drop = FALSE])
  nums <- sapply(data, is.numeric)
  nums <- names(data[, nums, drop = FALSE])
  numDat <- apply(data[, nums, drop = FALSE], 2, mean)
  statmode <- function(x){
    z <- table(as.vector(x))
    m <- names(z)[z == max(z)]
    if (length(m) == 1) {
      return(m)
    }
    return(".")
  }
  charDat <- apply(data[, chars, drop = FALSE], 2, statmode)
  cfdata <- cbind(as.data.frame(t(numDat)), as.data.frame(t(charDat)))
  cfdata <- cfdata[, names(data)]
  return(cfdata)
}


#' @title Subset a data.frame using a list of conditions
#' @name subsetList
#' @description Split a data.frame by elements in a list
#' @param data a data.frame
#' @param list a named list of splitting conditions
#' @return a data frame with values that match the conditions in the list
subsetList <- function(data, list){
  if("logical" %in% unlist(lapply(list, class))){
    stop("List is improperly formatted. Try using only `=` instead of `==` in subsets")
  }
  for(i in names(list)){
    data <- split(data, data[, i])
    data <- data[[list[[i]]]]
    data <- as.data.frame(data)
  }
  return(data)
}

#' \code{findFormFuns} used by \link[merTools]{averageObs} to calculate proper
#' averages
#'
#' The purpose is to properly derive data for the average observation in the
#' data by being 'aware' of formulas that contain interactions and/or function
#' calls. For example, in the old behavior, if the formula contained a square
#' term specified as \code{I(x^2)}, we were returning the mean of \code{x(^2)} not the
#' square of mean(x).
#'
#' @param merMod the merMod object from which to draw the average observation
#' @param origData (default=NULL) a data frame containing the original,
#'        untransformed data used to call the model. This MUST be specified if
#'        the original variables used in formula function calls are NOT present
#'        as 'main effects'.
#'
#' @return a data frame with a single row for the average observation, but with full
#' factor levels. See details for more.
#'
#' @export
findFormFuns <- function(merMod, origData = NULL) {
  form <- getCall(merMod)$formula
  form.rhs <- delete.response(terms(form))
  modFrame <- model.frame(merMod)
  if (identical(modFrame, origData)) {
    origData = NULL
  }
  modFrame.tt <- terms(modFrame)
  #This part is a bit kludgy but should work
  modFrame.labels <- unique(unlist(strsplit(attr(modFrame.tt, "term.labels"), split = ":", fixed = TRUE)))
  modFrame.resp <- setdiff(rownames(attr(modFrame.tt, "factors")),
                           unique(unlist(strsplit(colnames(attr(modFrame.tt, "factors")), split = ":", fixed = TRUE))))

  merMod.weights <- hasWeights(merMod)
  if (merMod.weights) {
    modFrame <- modFrame[, c(modFrame.resp, modFrame.labels, "(weights)")]
  } else {
    modFrame <- modFrame[, c(modFrame.resp, modFrame.labels)]
  }

  #Scan RHS of formula labels for parens -> exit if clean
  paren_terms <- grepl("[()]", c(modFrame.resp, modFrame.labels))

  if (!any(paren_terms)) {
    if(is.null(origData)){
      out <- collapseFrame(modFrame)
    } else{
      out <- collapseFrame(origData)
    }
    return(out)
  } else {
    rhs.vars <- all.vars(form.rhs)
    # if (merMod.weights) {
    #   rhs.vars <- c(rhs.vars, c("(weights)"))
    # }
    #Warning if functions are detected but neither MAIN EFFECTS NOR DATA are supplied
    if (is.null(origData))  {
      if (!all(rhs.vars %in% modFrame.labels)) {
        warning(paste("\n\n  Functions detected in formula without user supplied data",
                      "  or main effects of affected variables so returning means of",
                      "  transformed variables.\n",
                      "  Make sure that this is appropriate or supply untransformed",
                      "  data using the 'origData' argument. See ?merTools::findFormFuns", sep = "\n"))
        out <- collapseFrame(modFrame)
        return(out)
      } else {
        #Functions Detected and Main Effects Present
        out <- collapseFrame(modFrame)
        for (i in which(paren_terms)) {
          out[1,i] <- eval(parse(text = colnames(out)[i]), envir = out[, rhs.vars])
        }
        return(out)
      }
    } else {
      #Functions Detected and Not All Main Effects Present ... but Data supplied
      out <- collapseFrame(modFrame)
      outData <- collapseFrame(origData)
      for (i in which(paren_terms)) {
        out[1,i] <- eval(parse(text = colnames(out)[i]), envir = outData)
      }
      return(out)
    }
  }
}

#' Identify if a merMod has weights
#'
#' @param merMod the merMod object to test for weights
#'
#' @return TRUE or FALSE for whether the model has weights
hasWeights <- function(merMod) {
  if (all(merMod@resp$weights == 1)) {
    FALSE
  } else {
    TRUE
  }
}


#' @title Find the average observation for a merMod object
#' @name averageObs
#' @description Extract a data frame of a single row that represents the
#' average observation in a merMod object. This function also allows the
#' user to pass a series of conditioning argument to calculate the average
#' observation conditional on other characteristics.
#' @param merMod a merMod object
#' @param varList optional, a named list of conditions to subset the data on
#' @param origData (default=NULL) a data frame containing the original,
#'        untransformed data used to call the model. This MUST be specified if
#'        the original variables used in formula function calls are NOT present
#'        as 'main effects'.
#' @param ... not used currently
#' @return a data frame with a single row for the average observation, but with full
#' factor levels. See details for more.
#' @details Each character and factor variable in the data.frame is assigned to the
#' modal category and each numeric variable is collapsed to the mean. Currently if
#' mode is a tie, returns a "." Uses the collapseFrame function.
#' @export
averageObs <- function(merMod, varList = NULL, origData = NULL, ...){
  if(!missing(varList)){
    if (is.null(origData)) {
      data <- subsetList(merMod@frame, varList)
    } else {
      data <- subsetList(origData, varList)
    }
    if(nrow(data) < 20 & nrow(data) > 2){
      warning("Subset has less than 20 rows, averages may be problematic.")
    }
    if(nrow(data) <3){
      warning("Subset has fewer than 3 rows, computing global average instead.")
      if (is.null(origData)) {
        data <- merMod@frame
      } else {
        data <- origData
      }
    }
  } else{
    if (is.null(origData)) {
      data <- merMod@frame
    } else {
      data <- origData
    }
  }
  out <- findFormFuns(merMod, origData = data)
  reTerms <- names(ngrps(merMod))
  if(any(reTerms %in% names(varList))){
    reTerms <- reTerms[!reTerms %in% names(varList)]
  }
  if(length(reTerms) > 0){
    for(i in 1:length(reTerms)){
      out[, reTerms[i]] <- REquantile(merMod = merMod,
                                      quantile = 0.5, groupFctr = reTerms[[i]])
      out[, reTerms[i]] <- as.character(out[, reTerms[i]])
    }
  }
  chars <- !sapply(out, is.numeric)
  for(i in names(out[, chars, drop = FALSE])){ # drop = FALSE
    out[, i] <- try(superFactor(out[, i], fullLev = unique(merMod@frame[, i])), silent = TRUE)
  }
  out <- stripAttributes(out)
  out <- out[, names(merMod@frame)]
  return(out)
}


#' @title Create a factor with unobserved levels
#' @name superFactor
#' @description Create a factor variable and include unobserved levels
#' for compatibility with model prediction functions
#' @param x a vector to be converted to a factor
#' @param fullLev a vector of factor levels to be assigned to x
#' @return a factor variable with all observed levels of x and all levels
#' of x in fullLev
#' @export
#' @examples
#' \donttest{
#' regularFactor <- c("A", "B", "C")
#' regularFactor <- factor(regularFactor)
#' levels(regularFactor)
#' # Now make it super
#' newLevs <- c("D", "E", "F")
#' regularFactor <- superFactor(regularFactor, fullLev = newLevs)
#' levels(regularFactor) # now super
#' }
superFactor <- function(x, fullLev){
  x <- as.character(x)
  if("factor" %in% class(fullLev)){
    fullLev <- unique(levels(fullLev))
  }
  unobsLev <- unique(x)[!unique(x) %in% fullLev]
  x <- factor(x, levels = c(fullLev, unobsLev),
              labels = c(fullLev, unobsLev))
  return(x)
}

#' @title Randomly reorder a dataframe
#' @name shuffle
#' @description Randomly reorder a dataframe by row
#' @param data a data frame
#' @return a data frame of the same dimensions with the rows reordered
#' randomly
shuffle <- function(data){
  return(data[sample(nrow(data)),])
}


# wiggle for a single variable (var) and single set of changing values (values)
single_wiggle <- function(data, var, values) {
  tmp.data <- data
  data <- do.call("rbind", replicate(length(values), data, simplify= FALSE))
  data[, var] <- rep(values, each = nrow(tmp.data))
  if(any(class(tmp.data[, var]) %in% c("factor", "ordered"))){
    data[, var] <- superFactor(data[, var],
                               fullLev = levels(tmp.data[, var]))

  }
  return(data)
}

#' @title Assign an observation to different values
#' @name wiggle
#' @description Creates a new data.frame with copies of the original observation,
#' each assigned to a different user-specified value of a variable. Allows the
#' user to look at the effect on predicted values of changing either a single variable
#' or multiple variables.
#' @param data a data frame with one or more observations to be reassigned
#' @param varlist a character vector specifying the name(s) of the variable to adjust
#' @param valueslist a list of vectors with the values to assign to var
#' @return a \code{data.frame} with each row assigned to the one of the new variable combinations.
#' All variable combinations are returned, eg wiggling two variables with 3 and 4 variables
#' respectively will return a new dataset with \code{3 * 4 = 12} observations.
#' @details If the variable specified is a factor, then wiggle will return it
#' as a character.
#' @export
#' @examples
#' data(iris)
#' wiggle(iris[3,], varlist = "Sepal.Width", valueslist = list(c(1, 2, 3, 5)))
#' wiggle(iris[3:5,], "Sepal.Width", valueslist = list(c(1, 2, 3, 5)))
#' wiggle(iris[3,], c("Sepal.Width", "Petal.Length"), list(c(1,2,3,5), c(3,5,6)))
#' wiggle(iris[3:5,], c("Sepal.Width", "Petal.Length"), list(c(1,2,3,5), c(3,5,6)))
wiggle <- function(data, varlist, valueslist) {
  if (length(varlist) != length(valueslist)) stop("varlist and valueslist must be equi-length.")
  n_var <- length(varlist)
  if (n_var == 1) {
    return(single_wiggle(data, varlist[[1]], valueslist[[1]]))
  } else {
    temp <- single_wiggle(data, varlist[[1]], valueslist[[1]])
    temp <- split(temp, f= varlist[[1]])
    varlist <- varlist[-1]; valueslist <- valueslist[-1]
    return(do.call("rbind", lapply(temp, wiggle, varlist= varlist, valueslist= valueslist)))
  }
}

#' @title Identify group level associated with RE quantile
#' @name REquantile
#' @description For a user specified quantile (or quantiles) of the random effect
#' terms in a merMod object. This allows the user to easily identify the observation
#' associated with the nth percentile effect.
#' @param merMod a merMod object with one or more random effect levels
#' @param quantile a numeric vector with values between 0 and 100 for quantiles
#' @param groupFctr a character of the name of the random effect grouping factor to extract
#' quantiles from
#' @param term a character of the random effect to extract for the grouping factor
#' specified. Default is the intercept.
#' @return a vector of the level of the random effect grouping term that corresponds
#' to each quantile
#' @export
#' @examples
#' \donttest{
#' fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
#' REquantile(fm1, quantile = 0.25, groupFctr = "Subject")
#' REquantile(fm1, quantile = 0.25, groupFctr = "Subject", term = "Days")
#' }
REquantile <- function(merMod, quantile, groupFctr, term = "(Intercept)"){
  if(any(quantile > 1 | quantile < 0)){
    stop("Quantiles must be specified on the range 0-1")
  }
  myRE <- ranef(merMod)[[groupFctr]]
  if(is.null(myRE)){
    stop("Random effect group name not found. Please respecify grouping factor.")
  }
  myRE.tmp <- try(myRE[order(myRE[, term]), ,drop = FALSE], silent = TRUE)
  if(!is(myRE.tmp, "data.frame")){
    term1 <- names(myRE)[1]
    myRE.tmp <- try(myRE[order(myRE[, term1]), ,drop = FALSE], silent = TRUE)
    warning(paste0(term, " not found in random effect terms. Returning first term, ",
          term1,", for grouping factor, ", groupFctr, ", instead."))
  }
  myRE <- myRE.tmp; myRE.tmp <- NULL
  nobs <- nrow(myRE)
  if(nobs < 20){
    message("Number of observations < 20, random effect quantiles may not be well-defined.")
  }
  obsnum <- floor(quantile * nobs)
  return(rownames(myRE)[obsnum])
}

# -----------------------------------------------------
#-------------------------------------------------------
local_edition(3)
set.seed(51315)
library(lme4)
data(grouseticks)
grouseticks$HEIGHT <- scale(grouseticks$HEIGHT)
grouseticks <- merge(grouseticks, grouseticks_agg[, 1:3], by = "BROOD")


# Build out models
form <- TICKS ~ YEAR + HEIGHT +(1|BROOD) + (1|LOCATION) + (1|INDEX)
glmer3Lev  <- glmer(form, family="poisson",data=grouseticks,
                    control = glmerControl(optimizer="Nelder_Mead",
                                           optCtrl=list(maxfun = 1e5)))
# GLMER 3 level + slope
form <- TICKS ~ YEAR + HEIGHT +(1 + HEIGHT|BROOD) + (1|LOCATION) + (1|INDEX)
glmer3LevSlope  <- glmer(form, family="poisson",data=grouseticks,
                         control = glmerControl(optimizer="bobyqa",
                                                optCtrl=list(maxfun = 1e5)))

# GLMER 2 level
# data(VerbAgg)
# fmVA <- glmer(r2 ~ Anger + Gender + btype + situ +
#                 (1|id) + (1|item), family = binomial, data =
#                 VerbAgg)

# Sleepstudy
lmerSlope1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)

# Wackier example
data(Orthodont,package="nlme")
Orthodont$nsex <- as.numeric(Orthodont$Sex=="Male")
Orthodont$nsexage <- with(Orthodont, nsex*age)
lmerSlope2 <- lmer(distance ~ age + (0 + age + nsex|Subject), data=Orthodont)

###############################################
#Sanitize Names----
################################################

test_that("Sanitize names renames variables in data.frame", {
  badMod <- lmer(distance ~ factor(Sex) + (0 + age + nsex|Subject),
                 data=Orthodont, control =
                   lmerControl(check.conv.grad = .makeCC("warning", tol= 8e-3)))
  expect_false(identical(names(badMod@frame),
                         names(merTools:::sanitizeNames(badMod@frame))))
  expect_s3_class(merTools:::sanitizeNames(badMod@frame), "data.frame")
  expect_identical(names(merTools:::sanitizeNames(badMod@frame))[2], "Sex")
  expect_identical(names(badMod@frame)[2], "factor(Sex)")
})


###############################################
#Strip Attributes----
################################################

test_that("Attributes can be stripped from data.frame", {
  full <- names(attributes(lmerSlope1@frame))
  redu <- names(attributes(merTools:::stripAttributes(lmerSlope1@frame)))
  redu2 <- names(attributes(merTools:::stripAttributes(glmer3LevSlope@frame)))
  expect_true(length(full) > length(redu))
  expect_true(all(redu %in% full))
  expect_true(all(redu %in% c("names", "row.names", "class")))
  expect_true(all(redu2 %in% c("names", "row.names", "class")))
})


###############################################
#Random Observation----
################################################


test_that("A random observation can be sampled from a merMod", {
  data1 <- draw(glmer3Lev, type = 'random')
  data2 <- draw(lmerSlope2, type = 'random')
  data3 <- draw(lmerSlope1, type = 'random')
  data4 <- draw(glmer3LevSlope, type = 'random')
  expect_equal(nrow(data1), 1)
  expect_equal(nrow(data2), 1)
  expect_equal(nrow(data3), 1)
  expect_equal(nrow(data4), 1)
  expect_equal(ncol(data1), 6)
  expect_equal(ncol(data2), 4)
  expect_equal(ncol(data3), 3)
  expect_equal(ncol(data4), 6)
  expect_identical(names(data1), names(glmer3Lev@frame))
  expect_identical(names(data2), names(lmerSlope2@frame))
  expect_identical(names(data3), names(lmerSlope1@frame))
  expect_identical(names(data4), names(glmer3LevSlope@frame))
  expect_false(identical(names(attributes(data1)), names(attributes(glmer3Lev@frame))))
  expect_false(identical(names(attributes(data2)), names(attributes(lmerSlope2@frame))))
  expect_false(identical(names(attributes(data3)), names(attributes(lmerSlope1@frame))))
  expect_false(identical(names(attributes(data4)), names(attributes(glmer3LevSlope@frame))))
  expect_false("formula" %in% names(attributes(data1)))
  expect_false("formula" %in% names(attributes(data2)))
  expect_false("formula" %in% names(attributes(data3)))
  expect_false("formula" %in% names(attributes(data4)))
})

test_that("Random observation preserves factor levels", {
  data1 <- draw(glmer3Lev, type = 'random')
  data2 <- draw(lmerSlope2, type = 'random')
  data3 <- draw(lmerSlope1, type = 'random')
  data4 <- draw(glmer3LevSlope, type = 'random')
  expect_true(length(levels(data1$YEAR)) > length(unique(data1$YEAR)))
  expect_true(length(levels(data1$BROOD)) > length(unique(data1$BROOD)))
  expect_true(length(levels(data1$LOCATION)) > length(unique(data1$LOCATION)))
  expect_true(length(levels(data4$YEAR)) > length(unique(data4$YEAR)))
  expect_true(length(levels(data4$BROOD)) > length(unique(data4$BROOD)))
  expect_true(length(levels(data4$LOCATION)) > length(unique(data4$LOCATION)))
  # test levels are correct levels as well
})

###############################################
#Collapse frame----
################################################

test_that("Collapsing a dataframe results in single row", {
  data1 <- merTools:::collapseFrame(Orthodont)
  data2 <- merTools:::collapseFrame(grouseticks)
  expect_equal(length(data1), length(Orthodont))
  expect_equal(length(data2), length(grouseticks))
  expect_equal(nrow(data1), 1)
  expect_equal(nrow(data2), 1)
  expect_equal(data1$distance, mean(Orthodont$distance))
  expect_equal(data1$distance, mean(Orthodont$distance))
  expect_equal(data1$age, mean(Orthodont$age))
  expect_equal(data1$nsex, mean(Orthodont$nsex))
  expect_equal(data1$nsexage, mean(Orthodont$nsexage))
  expect_equal(data2$TICKS, mean(grouseticks$TICKS))
  expect_equal(data2$HEIGHT, mean(grouseticks$HEIGHT))
  expect_equal(data2$cHEIGHT, mean(grouseticks$cHEIGHT))
  expect_equal(data2$meanTICKS, mean(grouseticks$meanTICKS))
})

###############################################

################################################

test_that("Data can be subset by a list", {
  list11 <- list("Sex" = "Male")
  list12 <- list("Sex" = "Male", "Subject" = "M05")
  list13 <- list("Sex" == "Male")
  list14 <- list("Sex" == "Male", "Subject" == "M05")
  list15 <- list("Sex" = "Male", "Subject" == "M05")
  data11 <- merTools:::subsetList(Orthodont, list11)
  data12 <- merTools:::subsetList(Orthodont, list12)
  expect_error(merTools:::subsetList(Orthodont, list13))
  expect_error(merTools:::subsetList(Orthodont, list14))
  expect_error(merTools:::subsetList(Orthodont, list15))
  list21 <- list("YEAR" = "95")
  list22 <- list("LOCATION" = "32", "BROOD" = "503")
  data21 <- merTools:::subsetList(grouseticks, list21)
  data22 <- merTools:::subsetList(grouseticks, list22)
  expect_equal(length(data11), length(Orthodont))
  expect_equal(length(data21), length(grouseticks))
  expect_equal(length(data12), length(Orthodont))
  expect_equal(length(data22), length(grouseticks))
  expect_equal(nrow(data11), 64)
  expect_equal(nrow(data21), 117)
  expect_equal(nrow(data12), 4)
  expect_equal(nrow(data22), 0)
})

###############################################
#Super factor ----
################################################

test_that("Unobserved factor levels can be respected", {
  fac1 <- factor(c("502", "503"))
  fac1a <- superFactor(fac1, fullLev = unique(grouseticks$BROOD))
  fac2 <- factor(c("M16", "M02", "M05"))
  fac2a <- superFactor(fac2, fullLev = unique(Orthodont$Subject))
  expect_false(identical(levels(fac1), levels(fac1a)))
  expect_false(identical(levels(fac2), levels(fac2a)))
  expect_true(identical(levels(grouseticks$BROOD), levels(fac1a)))
  expect_true(identical(levels(Orthodont$Subject), levels(fac2a)))
  expect_equal(length(levels(fac1a)), 118)
  expect_equal(length(levels(fac2a)), 27)
})

test_that("SuperFactor handles new factor levels correctly", {
  fac1 <- factor(c("999", "888"))
  fac1a <- superFactor(fac1, fullLev = unique(grouseticks$BROOD))
  fac2 <- factor(c("Z16", "Z02", "Z05"))
  fac2a <- superFactor(fac2, fullLev = unique(Orthodont$Subject))
  expect_false(identical(levels(fac1), levels(fac1a)))
  expect_false(identical(levels(fac2), levels(fac2a)))
  expect_false(identical(levels(grouseticks$BROOD), levels(fac1a)))
  expect_false(identical(levels(Orthodont$Subject), levels(fac2a)))
  expect_equal(length(levels(fac1a)), length(levels(grouseticks$BROOD)) + 2)
  expect_equal(length(levels(fac2a)), length(levels(Orthodont$Subject)) + 3)
  expect_true(identical(levels(fac1a)[1:118], levels(grouseticks$BROOD)))
  expect_true(identical(levels(fac2a)[1:27], levels(Orthodont$Subject)))
})


###############################################
#Shuffle----
################################################

test_that("Data can be shuffled", {
  expect_equal(nrow(Orthodont), nrow(merTools:::shuffle(Orthodont)))
  expect_equal(ncol(Orthodont), ncol(merTools:::shuffle(Orthodont)))
  expect_equal(nrow(grouseticks), nrow(merTools:::shuffle(grouseticks)))
  expect_equal(ncol(grouseticks), ncol(merTools:::shuffle(grouseticks)))
})

###############################################
#Find RE Quantiles----
################################################

test_that("RE Quantile errors and messages are issued", {
  expect_error(REquantile(glmer3Lev, 23, groupFctr = "BROOD"))
  expect_warning(REquantile(glmer3Lev, .23, groupFctr = "BROOD", term = "Cat"), "Cat not found")
  expect_error(REquantile(glmer3Lev, .23, groupFctr = "Cat"))
  expect_error(REquantile(glmer3Lev, c(23, .56, .75), "BROOD"))
  expect_error(REquantile(glmer3Lev, c(.23, 56, .75), "BROOD"))
  expect_error(REquantile(glmer3Lev, c(.23, .56, 75), "BROOD"))
  expect_error(REquantile(glmer3Lev, c(.23, .56, 107), "BROOD"))
  expect_error(REquantile(glmer3Lev, c(-2, .56, .7), "BROOD"))
  expect_message(REquantile(lmerSlope1, .25, groupFctr = "Subject"), "Number of observations < 20")
  expect_warning(REquantile(lmerSlope2, c(.24), "Subject"), "not found in random effect terms")
  expect_warning(REquantile(lmerSlope2, c(.24), "Subject", term = "Cat"), "Cat not found")
})

# what to do without intercepts (REquantile(lmerSlope2), c(.24), "Subject")
# test_that("Quantiles are returned correctly", {
#   myRE <- ranef(glmer3Lev)[["BROOD"]]
#   myRE <- myRE[order(myRE[, "(Intercept)"]), ,drop = FALSE]
#   rownames(myRE)[floor(23 / nrow(myRE)*100)]
#
#
# })

###############################################
#Test observation wiggle----
################################################

test_that("Row and column lengths are correct -- single_wiggle", {
  data1 <- grouseticks[5:9, ]
data1a <- wiggle(data1, var = "BROOD", values = list(c("606", "602", "537")))
  data1b <- wiggle(data1a, var = "YEAR", values = list(c("96", "97")))
  data2 <- grouseticks[3, ]
  data2a <- wiggle(data2, var = "BROOD", values = list(c("606", "602", "537")))
  data2b <- wiggle(data2a, var = "YEAR", values = list(c("96", "97")))
  data3 <- grouseticks[12:14, ]
  data3a <- wiggle(data3, var = "BROOD", values = list(c("606")))
  data3b <- wiggle(data3a, var = "YEAR", values = list(c("96", "97")))
  expect_equal(nrow(data1), 5)
  expect_equal(nrow(data1a), 15)
  expect_equal(nrow(data1b), 30)
  expect_equal(nrow(data2), 1)
  expect_equal(nrow(data2a), 3)
  expect_equal(nrow(data2b), 6)
  expect_equal(nrow(data3), 3)
  expect_equal(nrow(data3a), 3)
  expect_equal(nrow(data3b), 6)
  expect_equal(length(data1), length(data1a))
  expect_equal(length(data1a), length(data1b))
  expect_equal(length(data2), length(data2a))
  expect_equal(length(data2a), length(data2b))
  expect_equal(length(data3), length(data3a))
  expect_equal(length(data3a), length(data3b))
  data4 <- wiggle(data3, var = "BROOD",
                     values = list(REquantile(glmer3Lev,
                                         quantile = c(0.25, 0.5, 0.75),
                                         group = "BROOD")))
  expect_true(all(table(as.character(data4$BROOD),
                        as.character(data4$INDEX)) ==1))

})


test_that("Values are placed correctly -- single_wiggle", {
  data1 <- grouseticks[5:9, ]
  data1a <- wiggle(data1, var = "BROOD", list(values = c("606", "602", "537")))
  data1b <- wiggle(data1a, var = "YEAR", values = list(c("96", "97")))
  data2 <- grouseticks[3, ]
  data2a <- wiggle(data2, var = "BROOD", values = list(c("606", "602", "537")))
  data2b <- wiggle(data2a, var = "YEAR", values = list(c("96", "97")))
  data3 <- grouseticks[12:14, ]
  data3a <- wiggle(data3, var = "BROOD", values = list(c("606")))
  data3b <- wiggle(data3a, var = "YEAR", values = list(c("96", "97")))
  data4 <- Orthodont[15, ]
  data4a <- wiggle(data4, var = "age", values = list(c(10, 11, 12)))
  data4b <- wiggle(data4a, var = "Sex", values = list(c("Male", "Female")))
  expect_false(any(unique(data1$BROOD) %in% unique(data1a$BROOD)))
  expect_false(any(unique(data1$BROOD) %in% unique(data1b$BROOD)))
  expect_false(any(unique(data1a$YEAR) %in% unique(data1b$YEAR)))
  expect_false(any(unique(data2$BROOD) %in% unique(data2a$BROOD)))
  expect_false(any(unique(data2$BROOD) %in% unique(data2b$BROOD)))
  expect_false(any(unique(data2a$YEAR) %in% unique(data2b$YEAR)))
  expect_false(any(unique(data3$BROOD) %in% unique(data3a$BROOD)))
  expect_false(any(unique(data3$BROOD) %in% unique(data3b$BROOD)))
  expect_false(any(unique(data3a$YEAR) %in% unique(data3b$YEAR)))
  expect_true(all(unique(data1a$BROOD) %in% c("606", "602", "537")))
  expect_true(all(unique(data1b$BROOD) %in% c("606", "602", "537")))
  expect_true(all(unique(data2a$BROOD) %in% c("606", "602", "537")))
  expect_true(all(unique(data2b$BROOD) %in% c("606", "602", "537")))
  expect_true(all(unique(data3a$BROOD) %in% c("606")))
  expect_true(all(unique(data3b$BROOD) %in% c("606")))
  expect_true(all(unique(data4a$age) %in% c(10, 11, 12)))
  expect_true(all(unique(data4b$age) %in% c(10, 11, 12)))
  expect_true(all(!unique(data1a$YEAR) %in% c("96", "97")))
  expect_true(all(unique(data1b$YEAR) %in% c("96", "97")))
  expect_true(all(!unique(data2a$YEAR) %in% c("96", "97")))
  expect_true(all(unique(data2b$YEAR) %in% c("96", "97")))
  expect_true(all(!unique(data3a$YEAR) %in% c("96", "97")))
  expect_true(all(unique(data3b$YEAR) %in% c("96", "97")))
  expect_true(all(unique(data4a$Sex) %in% c("Male", "Female")))
  expect_true(all(unique(data4b$Sex) %in% c("Male", "Female")))
})

test_that("we can use wiggle for multiple variables", {
  data1 <- grouseticks[5:9, ]
  data1a <- wiggle(data1, var = c("BROOD", "YEAR"),
                   list(c("606", "602", "537"), c("96", "97")))
  data3 <- grouseticks[12:14, ]
  data3a <- wiggle(data3, var = c("BROOD", "YEAR"), list(c("606"), c("96", "97")))
  data4 <- Orthodont[15, ]
  data4a <- wiggle(data4, var = c("age", "Sex"), list(c(10, 11, 12), c("Male", "Female")))

  # tests 1 -- row and columns
  expect_equal(nrow(data1a), nrow(data1) * 3  * 2)
  expect_equal(nrow(data3a), nrow(data3) * 1  * 2)
  expect_equal(nrow(data4a), nrow(data4) * 3  * 2)
  expect_equal(ncol(data1a), ncol(data1))
  expect_equal(ncol(data3a), ncol(data3))
  expect_equal(ncol(data4a), ncol(data4))

  # tests 2 -- values
  expect_false(any(unique(data1$BROOD) %in% unique(data1a$BROOD)))
  expect_false(any(unique(data1$YEAR) %in% unique(data1a$YEAR)))
  expect_true(all.equal(sort(as.character(unique(data1a$BROOD))), c("537", "602", "606")))
  expect_true(all.equal(sort(as.character(unique(data1a$YEAR))), c("96", "97")))

  expect_false(any(unique(data3$BROOD) %in% unique(data3a$BROOD)))
  expect_false(any(unique(data3$YEAR) %in% unique(data3a$YEAR)))
  expect_true(all.equal(sort(as.character(unique(data3a$BROOD))), "606"))
  expect_true(all.equal(sort(as.character(unique(data3a$YEAR))), c("96", "97")))

  expect_true(all(unique(data4a$age) %in% 10:12))
  expect_true(all(unique(data4a$Sex) %in% (c("Female", "Male"))))
  expect_true(all.equal(sort(unique(data4a$age)), 10:12))
  expect_true(all.equal(sort(as.character(unique(data4a$Sex))), c("Female", "Male")))
})


###############################################
#Test average observation extraction----
################################################

test_that("Test averageobs returns a single row", {
  data1 <- draw(glmer3Lev, type = 'average')
  data1a <- draw(glmer3LevSlope, type = 'average')
  suppressMessages({
    data2 <- draw(lmerSlope1, type = 'average')
  })

  expect_equal(nrow(data1), 1)
  expect_equal(nrow(data1a), 1)
  expect_equal(nrow(data2), 1)
})

test_that("Draw warnings and errors are correct", {
  expect_message(draw(lmerSlope1, type = 'average'), "Number of observations < 20")
  expect_warning(draw(lmerSlope2, type = 'average'), "not found in random effect terms")
  mylist2 <- list("YEAR" = "97", "LOCATION" = "16")
  expect_warning(draw(glmer3LevSlope, type = 'average', varList = mylist2), "Subset has less than 20")
  mylist3 <- list("YEAR" = "97", "LOCATION" = c("16", "56"))
  expect_warning(draw(glmer3LevSlope, type = 'average', varList = mylist3), "Subset has fewer than 3 rows")
})

test_that("Subsets work", {
  mylist1 <- list("YEAR" = "97")
  data1 <- draw(glmer3LevSlope, type = 'average', varList = mylist1)
  data1a <- draw(glmer3LevSlope, type = 'average')
  expect_false(identical(data1, data1a))
  expect_equal(data1$TICKS, mean(grouseticks$TICKS[grouseticks$YEAR == "97"]))
  expect_equal(data1a$TICKS, mean(grouseticks$TICKS))
  mylist2 <- list("YEAR" = "97", "LOCATION" = "16")
  expect_warning(draw(glmer3LevSlope, type = 'average', varList = mylist2),
                 "less than 20 rows, averages may be problematic")
  mylist3 <- list("YEAR" = "97", "LOCATION" = c("16", "56"))
  expect_warning(draw(glmer3LevSlope, type = 'average', varList = mylist3),
                 "fewer than 3 rows, computing global average instead")
})

test_that("Nested specifications work", {
  library(ggplot2)
  mod1 <- lmer(sleep_total ~ bodywt + (1|vore/order), data=msleep)
  data1 <- draw(mod1, "random")
  expect_s3_class(data1, "data.frame")
  suppressMessages({
    data2 <- draw(mod1, "average")
  })

  expect_s3_class(data2, "data.frame")
  mylist1 <- list("vore" = "carni")
  mylist2 <- list("order" = "Cetacea")
  data1 <- draw(mod1, "random", varList = mylist1)
  expect_s3_class(data1, "data.frame")
  expect_identical(as.character(data1$vore), "carni")
  data1 <- draw(mod1, "random", varList = mylist2)
  expect_s3_class(data1, "data.frame")
  expect_identical(as.character(data1$order), "Cetacea")
  data1 <- suppressWarnings(draw(mod1, "average", varList = mylist1))
  expect_s3_class(data1, "data.frame")
  expect_identical(as.character(data1$vore), "carni")
  data1 <- suppressWarnings(draw(mod1, "average", varList = mylist2))
  expect_s3_class(data1, "data.frame")
  expect_identical(as.character(data1$order), "Cetacea")
  fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
  data1 <- suppressWarnings(draw(fm1, type = "average", varList = list("Subject" = "308")))
  expect_s3_class(data1, "data.frame")
  expect_identical(as.character(data1$Subject), "308")
})

test_that("findFormFuns works", {
  #Replicable toy data
  set.seed(72167)
  play <- data.frame(
    a = runif(1000),
    b = rnorm(1000),
    c = rbinom(1000, 1, .35),
    d = rpois(1000, 2)
  )
  play$d <- factor(play$d, labels = LETTERS[seq_along(unique(play$d))])
  play$y <- play$a + 0.5*play$b + 2*play$c -1.8*(play$d=="B") + .43*(play$d == "C") + runif(100, 0, .35)
  play$grp <- factor(sample(x = paste("Group", 1:43), size = 1000, replace = TRUE))
  statmode <- function(x){
    z <- table(as.vector(x))
    m <- names(z)[z == max(z)]
    if (length(m) == 1) {
      return(m)
    }
    return(".")
  }
  trueMeans <- merTools:::collapseFrame(play)
  #Estimate toy models
  ##. Scenario 1: I()
  suppressMessages({
    s1 <- lmer(y ~ a + b + I(b^2) + c + d + (1|grp), data=play)
  } )

  expect_equal(findFormFuns(s1)[names(trueMeans)], trueMeans)
  expect_equal(findFormFuns(s1)$b^2, findFormFuns(s1)$`I(b^2)`)
  expect_length(findFormFuns(s1), 7L)

  ##. Scenario 2: log and no regular a
  suppressMessages({
    s2 <- lmer(y ~ log(a) + b + c + d + (1|grp), data=play)
  })

  expect_warning(findFormFuns(s2))
  expect_false(suppressWarnings(findFormFuns(s2)$`log(a)` == log(trueMeans$a)))
  expect_silent(findFormFuns(s2, origData = play))
  expect_equal(findFormFuns(s2, origData = play)$`log(a)`, log(trueMeans$a))
  ##. Scenario 3: 2 continuous interaction with *
  suppressMessages({
    s3 <- lmer(y ~ a*b + c + d + (1|grp), data=play)
  })
  expect_equal(findFormFuns(s3)[names(trueMeans)], trueMeans)
  expect_length(findFormFuns(s3), 6L)
  ##. Scenario 4: 2 continuous interaction with :
  s4 <- lmer(y ~ a:b + c + d + (1|grp), data=play)
  expect_equal(findFormFuns(s4)[names(trueMeans)], trueMeans)
  expect_length(findFormFuns(s4), 6L)
  ##. Scenario 5: 1 cont 1 cat interaction with *
  suppressMessages({
    s5 <- lmer(y ~ a + c + b*d + (1|grp), data = play)
  })

  expect_equal(findFormFuns(s5)[names(trueMeans)], trueMeans)
  expect_length(findFormFuns(s5), 6L)
  ##. Scenario 6: 1 cont 1 cat interaction with :
  suppressMessages({
    s6 <- lmer(y ~ a + c + b:d + (1|grp), data = play)
  })

  expect_equal(findFormFuns(s6)[names(trueMeans)], trueMeans)
  expect_length(findFormFuns(s6), 6L)
  ##. Scenario 7: 2 cat interaction with *
  suppressMessages({
    s7 <- lmer(y ~ a + b + c*d + (1|grp), data = play)
  })

  expect_equal(findFormFuns(s7)[names(trueMeans)], trueMeans)
  expect_length(findFormFuns(s7), 6L)
  ##. Scenario 8: 2 cat interaction with :
  suppressMessages({
    s8 <- lmer(y ~ a + b + c:d + (1|grp), data = play)
  })
  expect_equal(findFormFuns(s8)[names(trueMeans)], trueMeans)
  expect_length(findFormFuns(s8), 6L)
  ##. Scenario 9: function in random slope
  suppressMessages({
    s9 <- lmer(y ~ a + b + c + d + (1 + sqrt(abs(b))|grp), data = play)
  })
  expect_equal(findFormFuns(s9)[names(trueMeans)], trueMeans)
  expect_equal(findFormFuns(s9)$`sqrt(abs(b))`, sqrt(abs(trueMeans$b)))
  expect_length(findFormFuns(s9), 7L)
  ##. Scenario 10: two columns in I with no main effects
  suppressMessages({
    s10 <- lmer(y ~ I(log(a) + b^3) + c + d + (1|grp), data=play)
  })
  expect_warning(findFormFuns(s10))
  expect_false(suppressWarnings(findFormFuns(s10)$`I(log(a) + b^3)`) == log(trueMeans$a) + trueMeans$b^3)
  expect_silent(findFormFuns(s10, origData = play))
  expect_equal(findFormFuns(s10, origData = play)$`I(log(a) + b^3)`, log(trueMeans$a) + trueMeans$b^3)
  ##. Test that draw, draw.merMod and averageObs accept origData and issue warning if appropriate
  expect_warning(averageObs(s10))
  expect_silent(averageObs(s10, origData = play))
  expect_warning(merTools:::draw.merMod(s10, type = "average"))
  expect_silent(merTools:::draw.merMod(s10, origData = play, type = "average"))
  expect_silent(merTools:::draw.merMod(s10, type = "random"))
})


test_that("weights work for averageObs", {
  m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy, weights = Days)
  m2 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)

  out1 <- averageObs(m1)
  out2 <- averageObs(m2)

  expect_equal(nrow(out1), 1)
  expect_equal(nrow(out2), 1)
  expect_equal(ncol(out1), 4)
  expect_equal(ncol(out2), 3)


})

# Import variables from function
library(ggplot2)
library(shiny)

merMod <<- .shinyMerPar$merMod
if(is.null(.shinyMerPar$simData)){

} else {
  newdata <<- .shinyMerPar$simData
}

if (!exists("newdata")) {
  df.choices <- c("Model Frame"   = "orig",
                  "Random Obs"    = "rand",
                  "Average Obs"   = "mean")
} else {
  df.choices <- c("User Supplied" = "user",
                  "Model Frame"   = "orig",
                  "Random Obs"    = "rand",
                  "Average Obs"   = "mean")
}

#' Display model fit summary of x or x like objects, fast
#'
#' Faster than the implementation in the arm package because it avoids refitting
#'
#' @title fastdisp: faster display of model summaries
#' @param x a model object
#' @param ... additional arguments to pass to \code{arm::\link[arm]{display}}
#' including number of digits
#' @details The time saving is only noticeable for large, time-consuming (g)lmer
#' fits.
#' @import arm
#' @return A printed summary of a x object
#' @examples
#' \donttest{
#' #Compare the time for displaying this modest model
#' require(arm)
#' m1 <- lmer(y ~ lectage + studage + (1|d) + (1|s), data=InstEval)
#' system.time(display(m1))
#' system.time(fastdisp(m1))
#' }
#' @seealso \code{\link[arm]{display}}
#' @rdname fastdisp
#' @export fastdisp
fastdisp <- function (x, ...) {
  UseMethod("fastdisp", x)
}

#' @rdname fastdisp
#' @importFrom stats df pt
#' @export
fastdisp.merMod <- function (x, ...)
{
  .local <- function (x, digits = 2, detail = FALSE)
  {
    out <- NULL
    out$call <- x@call
    print(out$call)
    fcoef <- fixef(x)
    useScale <- getME(x, "devcomp")$dims["useSc"]
    corF <- lme4::vcov.merMod(x)@factors$correlation
    coefs <- cbind(fcoef, corF@sd)
    if (length(fcoef) > 0) {
      if (!useScale) {
        coefs <- coefs[, 1:2, drop = FALSE]
        out$z.value <- coefs[, 1]/coefs[, 2]
        out$p.value <- 2 * pnorm(abs(out$z.value), lower.tail = FALSE)
        coefs <- cbind(coefs, `z value` = out$z.value,
                       `Pr(>|z|)` = out$p.value)
      }
      else {
        out$t.value <- coefs[, 1]/coefs[, 2]
        coefs <- cbind(coefs, `t value` = out$t.value)
      }
      dimnames(coefs)[[2]][1:2] <- c("coef.est", "coef.se")
      if (detail) {
        pfround(coefs, digits)
      }
      else {
        pfround(coefs[, 1:2], digits)
      }
    }
    out$coef <- coefs[, "coef.est"]
    out$se <- coefs[, "coef.se"]
    cat("\nError terms:\n")
    vc <- easyVarCorr(VarCorr(x), useScale = useScale,
                            digits)
    print(vc[, c(1:2, 4:ncol(vc))], quote = FALSE)
    out$ngrps <- lapply(x@flist, function(x) length(levels(x)))
    is_REML <- isREML(x)
    llik <- logLik(x, REML = is_REML)
    out$AIC <- AIC(llik)
    # out$deviance <- deviance(refitML(x))
    out$n <- getME(x, "devcomp")$dims["n"]
    # Dhat <- -2 * (llik)
    # pD <- out$deviance - Dhat
    # out$DIC <- out$deviance + pD
    cat("---\n")
    cat(sprintf("number of obs: %d, groups: ", out$n))
    cat(paste(paste(names(out$ngrps), out$ngrps, sep = ", "),
              collapse = "; "))
    cat(sprintf("\nAIC = %g", round(out$AIC, 1)))
    # cat(round(out$DIC, 1))
    # cat("\ndeviance =", fround(out$deviance, 1), "\n")
    if (useScale < 0) {
      out$sigma.hat <- sigma(x)
      cat("overdispersion parameter =", fround(out$sigma.hat,
                                               1), "\n")
    }
    return(invisible(out))
  }
  .local(x, ...)
}

#' @rdname fastdisp
#' @export
fastdisp.merModList <- function(x, ...){
  .local <- function (x, digits = 2, detail = FALSE)
  {
  out <- NULL
  useScale <- getME(x[[1]], "devcomp")$dims["useSc"]
  #useScale <- TRUE
  out$call <- x[[1]]@call
  listFE <- modelFixedEff(x)
  row.names(listFE) <- listFE$term
  out$t.value <- listFE$statistic
  out$coef <- listFE$estimate
  out$se <- listFE$std.error
  listRE <- modelRandEffStats(x)
  out$ngrps <- lapply(x[[1]]@flist, function(x) length(levels(x)))
  is_REML <- isREML(x[[1]])
  llik <- lapply(x, logLik, REML = is_REML)
  out$AIC <- mean(unlist(lapply(llik, AIC)))
  out$n <- round(mean(unlist(lapply(lapply(lapply(x, getME, "devcomp"),
                                     "[[", "dims"), "[", 2))), 0) # round to nearest integer
  print(out$call)
  if (!detail) {
    pfround(listFE[, 2:3], digits)
  }
  else {
    listFE$p.value <- 2 * pt(abs(listFE$statistic), listFE$df, lower.tail = FALSE)
    pfround(listFE[, 2:6], digits)
  }
  cat("\nError terms:\n")
  vc <- easyVarCorr(VarCorr(x[[1]]), useScale = useScale,
                    digits)
  # Resort the output of the random effect summary
  listRE <- listRE[grep("cor_", listRE$term, invert=TRUE), ]
  resid <- listRE[listRE$group == "Residual", ]
  listRE <- listRE[listRE$group != "Residual", ]
  listRE <- rbind(listRE, resid)
  #
  vc[, 3] <- as.character(round(listRE$estimate^2, digits = digits))
  vc[, 4] <- as.character(round(listRE$estimate, digits = digits))
  print(vc[, c(1:2, 4:ncol(vc))], quote = FALSE)
  cat("---\n")
  cat(sprintf("number of obs: %d, groups: ", out$n))
  cat(paste(paste(names(out$ngrps), out$ngrps, sep = ", "),
            collapse = "; "))
  cat(sprintf("\nAIC = %g", round(out$AIC, 1)))
  cat("---\n")
  # cat(round(out$DIC, 1))
  # cat("\ndeviance =", fround(out$deviance, 1), "\n")
  if (useScale < 0) {
    out$sigma.hat <- sigma(x)
    cat("overdispersion parameter =", fround(out$sigma.hat,
                                             1), "\n")
    cat("---\n")
  }
  return(invisible(out))
  }
  .local(x, ...)
}

# Test fastdisplay

set.seed(51315)
library(lme4); library(arm)
# Sleepstudy
lmerSlope1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)


###############################################
context("Fast display")
################################################

test_that("fastdisp pulls out a list", {
  # hack to avoid console output
  {sink("NUL"); zz <- fastdisp(lmerSlope1); sink()}
  expect_is(zz, "list")
  expect_identical(names(zz), c("call", "t.value", "coef", "se",
                                "ngrps", "AIC", "n"))
})

test_that("fastdisp speed is good", {
   {sink("NUL"); t1 <- system.time(force(fastdisp(lmerSlope1)))["elapsed"];
    sink()}
   {sink("NUL"); t2 <- system.time(force(display(lmerSlope1)))["elapsed"];
    sink()}
   expect_lt(t1, t2)
   expect_lt(t1, 0.1)
})

# ShinyMer
shinyUI(fluidPage(
  shiny::titlePanel("Explore your merMod interactively"),
  shiny::sidebarLayout(
    shiny::sidebarPanel(
      shiny::radioButtons("newdataType",
                          "Simulated data scenario",
                          choices=df.choices,
                          selected=NULL),
      #                  conditionalPanel(condition = "input.newdataType!='orig'",
      #                                   selectInput("filter", "Filter",
      #                                               choices = names(merMod@frame))
      #                                   ),
      shiny::numericInput("n.sims",
                          label="Simulations (Max=1,000)",
                          value=100,
                          min=1,
                          max=1000),
      conditionalPanel(condition="input.conditionedPanels==3",
                       helpText("Here you can compare impact of
                                changing input variables on the
                                outcome variable for selected cases."),
                       selectInput("group", "Group Factor:",
                                   choices = names(ranef(merMod)),
                                   selected = NULL),
                       uiOutput("re.ui"),
                       sliderInput("nbin", "Effect Bins", min = 3,
                                   max = 10, value = 4, step = 1),
                       helpText("And modify your fixed effects"),
                       selectInput("fixef", "Fixed Effect:",
                                   choices = all.vars(nobars(formula(merMod)))[-1],
                                   selected = NULL)
                       ),
      shiny::numericInput("alpha",
                          label="Credible Interval (%)",
                          value=95,
                          min=0,
                          max=100),
      shiny::radioButtons("stat",
                          "Measure of central tendency",
                          choices=c("Median"="median", "Mean"="mean"),
                          selected=NULL),
      shiny::radioButtons("predMetric",
                          "Prediction metric",
                          choices=c("Linear Predictor"="linear.prediction",
                                    "Probability"="probability"),
                          selected=NULL),
      shiny::checkboxInput("resid.var",
                           label="Include Residual Variation",
                           value=TRUE)
    ),
    shiny::mainPanel(
      shiny::tabsetPanel(type="tabs",
                         shiny::tabPanel("Prediction uncertainty",
                                         shiny::h3("Prediction Intervals:"),
                                         plotOutput("predPlot"),
                                         shiny::h3("All Predictions"),
                                         if ("DT" %in% rownames(installed.packages())) {
                                            dataTableOutput("dt") } else {
                                            tableOutput("shiny")
                                            },
                                         shiny::downloadButton("downloadData", "Download predict interval data")
                                         ,value = 1
                         ),
                         shiny::tabPanel("Parameters",
                                         shiny::h3("Original call"),
                                         textOutput("call"),
                                         shiny::h3("Fixed Effects"),
                                         plotOutput("feplot"),
                                         shiny::h3("Group Effects"),
                                         plotOutput("replot"), value = 2
                         ),
                         shiny::tabPanel("Substantive Effect",
                                         shiny::h3("Effect Sizes"),
                                         plotOutput("gPlot"),
                                         shiny::h3("Fixef Effect Impact"),
                                         plotOutput("wigglePlot"), value = 3
                         ), id = "conditionedPanels"
      )
    )
)
  ))

#' Launch a shiny app to explore your merMod interactively
#'
#' \code{shinyMer} launches a shiny app that allows you to interactively
#' explore an estimated merMod using functions from \code{merTools}.
#'
#' @param merMod An object of class "merMod".
#'
#' @param simData A data.frame to make predictions from (optional). If
#'   NULL, then the user can only make predictions using the data in
#'   the frame slot of the merMod object.
#'
#' @param pos The position of the environment to export function arguments to.
#' Defaults to 1, the global environment, to allow shiny to run.
#'
#' @return A shiny app
#'
#' @import ggplot2
#' @importFrom shiny shinyApp
#' @importFrom shiny fluidPage
#' @importFrom shiny titlePanel
#' @importFrom shiny sidebarLayout
#' @importFrom shiny sidebarPanel
#' @importFrom shiny radioButtons
#' @importFrom shiny numericInput
#' @importFrom shiny checkboxInput
#' @importFrom shiny actionButton
#' @importFrom shiny mainPanel
#' @importFrom shiny tabsetPanel
#' @importFrom shiny tabPanel
#' @importFrom shiny h3
#' @importFrom shiny textOutput
#' @importFrom shiny plotOutput
#' @importFrom shiny downloadButton
#' @importFrom shiny em
#' @importFrom shiny reactiveValues
#' @importFrom shiny eventReactive
#' @importFrom shiny observeEvent
#' @importFrom shiny reactive
#' @importFrom shiny renderPrint
#' @importFrom shiny renderPlot
#' @importFrom shiny isolate
#' @importFrom shiny renderPrint
#' @importFrom shiny downloadHandler
#' @importFrom shiny strong
#' @importFrom shiny runApp
#' @export

shinyMer <- function(merMod, simData = NULL, pos = 1) {
  envir = as.environment(pos)
  if(exists("simData")){
    expParm <- function(x, y) assign(".shinyMerPar", list("merMod" = x, "simData" = y), envir = envir)
    expParm(x = merMod, y = simData)
  } else{
    expParm2 <- function(x) assign(".shinyMerPar", list("merMod" = x, "simData" = NULL), envir = envir)
    expParm2(x = merMod)
  }
  appDir <- system.file("shiny-apps", "shinyMer", package = "merTools")
  shiny::runApp(appDir, display.mode = "normal")
}

library(microbenchmark)

# ClassFilter <- function(x) inherits(get(x), 'lm' ) & !inherits(get(x), 'gl
set.seed(101)


# Small

lmerSlope1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)

bench1 <- microbenchmark(
  predict(lmerSlope1, newdata = sleepstudy[1:100,]),
  predictInterval(lmerSlope1, newdata = sleepstudy[1:100,],
                  level = 0.9, n.sims = 100, stat = 'median',
                  include.resid.var = TRUE),
  predictInterval(lmerSlope1, newdata = sleepstudy[1:100,],
                  level = 0.9, n.sims = 100, stat = 'mean',
                  include.resid.var = TRUE),
  predictInterval(lmerSlope1, newdata = sleepstudy[1:100,],
                  level = 0.9, n.sims = 100, stat = 'median',
                  include.resid.var = FALSE),
  predictInterval(lmerSlope1, newdata = sleepstudy[1:100,],
                  level = 0.9, n.sims = 1000, stat = 'median',
                  include.resid.var = TRUE),
  predictInterval(lmerSlope1, newdata = sleepstudy[1:100,],
                  level = 0.8, n.sims = 1000, stat = 'median',
                  include.resid.var = TRUE),
  times = 10, unit = "s"
)

bench2 <- microbenchmark(
  predict(lmerSlope1, newdata = sleepstudy[1:100,]),
  predictInterval(lmerSlope1, newdata = sleepstudy[1:100,],
                  level = 0.9, n.sims = 100, stat = 'median',
                  include.resid.var = TRUE),
  predictInterval(lmerSlope1, newdata = sleepstudy[1:100,],
                  level = 0.9, n.sims = 200, stat = 'mean',
                  include.resid.var = TRUE),
  predictInterval(lmerSlope1, newdata = sleepstudy[1:100,],
                  level = 0.9, n.sims = 400, stat = 'median',
                  include.resid.var = FALSE),
  predictInterval(lmerSlope1, newdata = sleepstudy[1:100,],
                  level = 0.9, n.sims = 800, stat = 'median',
                  include.resid.var = TRUE),
  predictInterval(lmerSlope1, newdata = sleepstudy[1:100,],
                  level = 0.9, n.sims = 1600, stat = 'median',
                  include.resid.var = TRUE),
  times = 10, unit = "s"
)

# Medium
d <- expand.grid(fac1=LETTERS[1:5], grp=factor(1:10),
                 obs=1:100)
d$y <- simulate(~fac1+(1|grp),family = gaussian,
                newdata=d,
                newparams=list(beta=c(2,1,3,4,7), theta=c(.25),
                               sigma = c(.23)))[[1]]
g1 <- lmer(y~fac1+(1|grp), data=d)


bench3 <- microbenchmark(predictInterval(g1, newdata = d[1:100, ], level = 0.9,
                               n.sims = 50,
                               stat = 'mean', include.resid.var = TRUE),
               predictInterval(g1, newdata = d[1:200, ], level = 0.9,
                               n.sims = 50,
                               stat = 'mean', include.resid.var = TRUE),
               predictInterval(g1, newdata = d[1:400, ], level = 0.9,
                               n.sims = 50,
                               stat = 'mean', include.resid.var = TRUE),
               predictInterval(g1, newdata = d[1:800, ], level = 0.9,
                               n.sims = 50,
                               stat = 'mean', include.resid.var = TRUE),
               times = 10, unit = "s")


# Large

g2 <- lmer(y ~ lectage + studage + (1+lectage|d) + (1|dept), data=InstEval)
d2 <- InstEval[1:1000, ]

bench4 <- microbenchmark(predictInterval(g2, newdata = d2[1:100, ], level = 0.9,
                                         n.sims = 500,
                                         stat = 'mean', include.resid.var = TRUE),
                         predictInterval(g2, newdata = d2[1:200, ], level = 0.9,
                                         n.sims = 500,
                                         stat = 'mean', include.resid.var = TRUE),
                         predictInterval(g2, newdata = d2[1:400, ], level = 0.9,
                                         n.sims = 500,
                                         stat = 'mean', include.resid.var = TRUE),
                         predictInterval(g2, newdata = d2[1:800, ], level = 0.9,
                                         n.sims = 500,
                                         stat = 'mean', include.resid.var = TRUE),
                         times = 10, unit = "s")


d2 <- d2[order(d2$d, d2$dept),]

bench5 <- microbenchmark(predictInterval(g2, newdata = d2[1:100, ], level = 0.9,
                                         n.sims = 500,
                                         stat = 'mean', include.resid.var = TRUE),
                         predictInterval(g2, newdata = d2[1:200, ], level = 0.9,
                                         n.sims = 500,
                                         stat = 'mean', include.resid.var = TRUE),
                         predictInterval(g2, newdata = d2[1:400, ], level = 0.9,
                                         n.sims = 500,
                                         stat = 'mean', include.resid.var = TRUE),
                         predictInterval(g2, newdata = d2[1:800, ], level = 0.9,
                                         n.sims = 500,
                                         stat = 'mean', include.resid.var = TRUE),
                         times = 10, unit = "s")

g3 <- lmer(y ~ lectage + studage + (1|s) + (1+lectage|d) + (1|dept), data=InstEval)
g2 <- lmer(y ~ lectage + studage + (1+lectage|d) + (1|dept), data=InstEval)
p2 <- profvis({
  predictInterval(g2, level = 0.9, newdata = InstEval[1:100,],
                  n.sims = 7500,
                  stat = 'mean', include.resid.var = TRUE)
})
# View it with:
p2

library(doParallel)
cl <- makeCluster(4)
registerDoParallel(cl, 4)
zzz <- predictInterval(g3, level = 0.9, newdata = InstEval,
                n.sims = 7500,
                stat = 'mean', include.resid.var = TRUE, .parallel = TRUE)

# set.seed(101)
# d <- expand.grid(fac1=LETTERS[1:5], grp=factor(1:10),
#                  obs=1:50)
# d$y <- simulate(~fac1+(1|grp),family = binomial,
#                 newdata=d,
#                 newparams=list(beta=c(2,-1,3,-2,1.2), theta=c(.33)))[[1]]
# subD <- d[sample(row.names(d), 1200),]
# g1 <- glmer(y~fac1+(1|grp), data=subD, family = 'binomial')
# d$fitted <- predict(g1, d)
#
#
# outs <- predictInterval(g1, newdata = d, level = 0.95, n.sims = 500,
#                         stat = 'mean', include.resid.var = FALSE,
#                         type = 'linear.prediction')
#
#
# g2 <- lmer(y ~ lectage + studage + (1|d) + (1|s), data=InstEval)
# d1 <- InstEval[1:100, ]
#

#' Extract theta parameters from a merMod model
#' @description A convenience function that returns the theta parameters for a
#' \code{\link{merMod}} object.
#' @param merMod a valid merMod object
#'
#' @return a vector of the covariance, theta, parameters from a \code{\link{merMod}}
#' @seealso merMod
#' @export
#' @examples
#' (fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy))
#' thetaExtract(fm1) #(a numeric vector of the covariance parameters)
thetaExtract <- function(merMod){
  stopifnot(class(merMod) %in% c("lmerMod", "glmerMod", "blmerMod",
                                 "bglmerMod"))
  return(merMod@theta)
}

#' Bootstrap a subset of an lme4 model
#'
#' @param merMod a valid merMod object
#' @param n the number of rows to sample from the original data
#' in the merMod object, by default will resample the entire model frame
#' @param FUN the function to apply to each bootstrapped model
#' @param R the number of bootstrap replicates, default is 100
#' @param seed numeric, optional argument to set seed for simulations
#' @param warn logical, if TRUE, warnings from lmer will be issued, otherwise they will be suppressed
#' default is FALSE
#' @return a data.frame of parameters extracted from each of the R replications.
#' The original values are appended to the top of the matrix.
#' @details This function allows users to estimate parameters of a
#' large merMod object using bootstraps on a subset of the data.
#' @examples
#' \donttest{
#' (fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy))
#' resultMatrix <- subBoot(fm1, n = 160, FUN = thetaExtract, R = 20)
#' }
#' @export
subBoot <- function(merMod, n = NULL, FUN, R = 100, seed = NULL, warn = FALSE){
  if (missing(n)) {
    n <- nrow(merMod@frame)
  }

  if (!warn) {
    message("Warnings set to off by default, not all submodels may have converged.")
  }

  resultMat <- matrix(FUN(merMod), nrow = 1)
  tmp <- matrix(data=NA, nrow=R, ncol=ncol(resultMat))
  resultMat <- rbind(resultMat, tmp); rm(tmp)

  if (!is.null(seed))
    set.seed(seed)
  else if (!exists(".Random.seed", envir = .GlobalEnv))
    runif(1)

    for(i in 1:R){
      rows <- as.numeric(row.names(merMod@frame))
      mysamp <- as.character(sample(rows,  n, replace=TRUE))
      # http://proteo.me.uk/2013/12/fast-subset-selection-by-row-name-in-r/
      newdata <- merMod@frame[match(mysamp, rownames(merMod@frame)),]
      # Only for lmerMod
      if (!warn) {
        suppressWarnings({
          if (is(merMod, "lmerMod")) {
            tmpMod <- lmer(formula(merMod), data = newdata)
          } else if (is(merMod, "glmerMod")) {
            tmpMod <- glmer(formula(merMod), data = newdata, family = merMod@call$family)
          }

        })
      } else {
        if (is(merMod, "lmerMod")) {
          tmpMod <- lmer(formula(merMod), data = newdata)
        } else if (is(merMod, "glmerMod")) {
          tmpMod <- glmer(formula(merMod), data = newdata, family = merMod@call$family)
        }
      }
      resultMat[i + 1, ] <- FUN(tmpMod)
    }
  resultMat <- data.frame(param=resultMat)
  resultMat$replicate <- c("original", 1:R)
  return(resultMat)
}

# test subboot

set.seed(51315)
library(lme4)
# Sleepstudy
lmerSlope1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)


###############################################
# Extract theta----
#context("Extract theta using subBoot")
################################################

test_that("extract theta produces a vector", {
  expect_type(thetaExtract(lmerSlope1), "double")
  expect_equal(length(thetaExtract(lmerSlope1)), 3)
})

test_that("thetaExtract throws errors for non-merMod objects", {
  expect_error(thetaExtract(lmerSlope1@frame))
  m1 <- lm(mpg ~ disp + hp, data = mtcars)
  expect_error(thetaExtract(m1))
})

###############################################
# subBoot----
#context("subBoot")
################################################

test_that("subBoot produces correct output", {
  skip_on_cran()
  # Subbooot returns errors here
  out1 <- subBoot(lmerSlope1, n = 100, FUN = thetaExtract, R = 100)
  expect_s3_class(out1, "data.frame")
  expect_equal(ncol(out1), 4)
  expect_equal(nrow(out1), 101)

  out2 <- subBoot(lmerSlope1, n = 100,
                  FUN = function(x) getME(x, "fixef"),
                  R = 100)
  expect_s3_class(out2, "data.frame")
  expect_equal(ncol(out2), 3)
  expect_equal(nrow(out2), 101)
})

#context("subBoot glmer models")



test_that("subBoot produces correct glmer output", {
  skip_on_cran()
  d <- expand.grid(fac1 = LETTERS[1:5],
                   grp = letters[11:20],
                   obs = 1:50)
  d$y <- simulate(~fac1 + (1 | grp), family = binomial,
                  newdata = d,
                  newparams = list( beta = c(2,-1,3,-2,1.2),
                                    theta = c(.33)),
                  seed =634)[[1]]
  subD <- d[sample(row.names(d), 1200), ]

  g1 <- glmer(y~fac1+(1|grp), data=subD, family = 'binomial')

  out1 <- subBoot(g1, n = 1000, FUN = thetaExtract, R = 10)
  expect_s3_class(out1, "data.frame")
  expect_equal(ncol(out1), 2)
  expect_equal(nrow(out1), 11)
  #
  out2 <- subBoot(g1, n = 500,
                  FUN = function(x) getME(x, "fixef"),
                  R = 10)
  expect_s3_class(out2, "data.frame")
  expect_equal(ncol(out2), 6)
  expect_equal(nrow(out2), 11)

})

# ---------------------------------------------------------------------------
# predictInterval-regression.R
#
# Cross-version numeric regression harness for predictInterval().
#
# Generates a deterministic bundle of predictInterval() outputs for a pinned
# set of canonical models, inputs, and argument combinations. Two package
# versions can then be diffed bit-for-bit, making it easy to confirm that a
# refactor did not change numeric behavior for users who rely on a fixed seed.
#
# This script is NOT part of R CMD check. It is a manual tool invoked
# explicitly when someone touches simulation internals.
#
# Usage:
#
#   # (1) Generate an output bundle for one package version
#   Rscript tests/comparisons/predictInterval-regression.R harness \
#           <pkg_path> <output_rds>
#
#   # (2) Diff two bundles
#   Rscript tests/comparisons/predictInterval-regression.R diff \
#           <rds_a> <rds_b>
#
# Typical workflow (current checkout vs. origin/master):
#
#   git worktree add /tmp/mT-old origin/master
#   Rscript tests/comparisons/predictInterval-regression.R harness \
#           /tmp/mT-old  /tmp/old.rds
#   Rscript tests/comparisons/predictInterval-regression.R harness \
#           .           /tmp/new.rds
#   Rscript tests/comparisons/predictInterval-regression.R diff  \
#           /tmp/old.rds /tmp/new.rds
#   git worktree remove /tmp/mT-old
#
# Interpretation:
#   - All cases IDENTICAL (max_abs = 0)  -> refactor preserves behavior.
#   - Numeric differences                -> behavior changed; inspect whether
#                                           the change is intentional
#                                           (e.g., the binomial GLMM residual
#                                           simulation fix introduced in
#                                           merTools 0.9.0 legitimately alters
#                                           the two glmm_bin_with_resid_*
#                                           cases).
#
# Requires: pkgload, lme4.
# ---------------------------------------------------------------------------

SEED  <- 11213
NSIMS <- 1000

# -- Harness -----------------------------------------------------------------

run_harness <- function(pkg_path, out_path) {
  # Pin RNG so R version does not perturb stream behavior
  RNGversion("4.1.0")
  RNGkind(kind = "Mersenne-Twister",
          normal.kind = "Inversion",
          sample.kind = "Rejection")

  suppressPackageStartupMessages({
    library(pkgload)
    load_all(pkg_path, quiet = TRUE)
    library(lme4)
  })

  data(sleepstudy, package = "lme4")
  data(cbpp,       package = "lme4")

  cases <- list()

  # LMM: random slope + random intercept
  m_slope <- suppressMessages(
    lmer(Reaction ~ Days + (Days | Subject), data = sleepstudy)
  )
  cases$lmm_slope_default <- predictInterval(
    m_slope, newdata = sleepstudy[1:10, ],
    seed = SEED, n.sims = NSIMS)
  cases$lmm_slope_no_resid <- predictInterval(
    m_slope, newdata = sleepstudy[1:10, ],
    seed = SEED, n.sims = NSIMS, include.resid.var = FALSE)
  cases$lmm_slope_fixed_only <- predictInterval(
    m_slope, newdata = sleepstudy[1:10, ],
    seed = SEED, n.sims = NSIMS, which = "fixed")
  cases$lmm_slope_random_only <- predictInterval(
    m_slope, newdata = sleepstudy[1:10, ],
    seed = SEED, n.sims = NSIMS, which = "random")
  cases$lmm_slope_all <- predictInterval(
    m_slope, newdata = sleepstudy[1:10, ],
    seed = SEED, n.sims = NSIMS, which = "all")

  # LMM: random intercept only
  m_int <- suppressMessages(
    lmer(Reaction ~ Days + (1 | Subject), data = sleepstudy)
  )
  cases$lmm_int_default <- predictInterval(
    m_int, newdata = sleepstudy[1:10, ],
    seed = SEED, n.sims = NSIMS)
  cases$lmm_int_no_resid <- predictInterval(
    m_int, newdata = sleepstudy[1:10, ],
    seed = SEED, n.sims = NSIMS, include.resid.var = FALSE)
  cases$lmm_int_mean_stat <- predictInterval(
    m_int, newdata = sleepstudy[1:10, ],
    seed = SEED, n.sims = NSIMS, stat = "mean")
  cases$lmm_int_level_99 <- predictInterval(
    m_int, newdata = sleepstudy[1:10, ],
    seed = SEED, n.sims = NSIMS, level = 0.99)

  # LMM: single-row newdata
  cases$lmm_single_row <- predictInterval(
    m_slope, newdata = sleepstudy[1, , drop = FALSE],
    seed = SEED, n.sims = 500)

  # LMM: ignore.fixed.terms + fix.intercept.variance
  cases$lmm_ignore_fixed <- predictInterval(
    m_int, newdata = sleepstudy[1:10, ],
    seed = SEED, n.sims = NSIMS, include.resid.var = FALSE,
    ignore.fixed.terms = "(Intercept)")
  cases$lmm_fix_intercept_var <- predictInterval(
    m_int, newdata = sleepstudy[1:10, ],
    seed = SEED, n.sims = NSIMS, fix.intercept.variance = TRUE)

  # GLMM: binomial (the two with_resid cases differ intentionally in 0.9.0+
  # because of the n-trial binomial residual simulation fix)
  gm_bin <- suppressMessages(
    glmer(cbind(incidence, size - incidence) ~ period + (1 | herd),
          data = cbpp, family = binomial)
  )
  cases$glmm_bin_no_resid_linear <- predictInterval(
    gm_bin, newdata = cbpp[1:5, ],
    seed = SEED, n.sims = NSIMS,
    include.resid.var = FALSE, type = "linear.prediction")
  cases$glmm_bin_no_resid_prob <- predictInterval(
    gm_bin, newdata = cbpp[1:5, ],
    seed = SEED, n.sims = NSIMS,
    include.resid.var = FALSE, type = "probability")
  cases$glmm_bin_with_resid_linear <- suppressWarnings(predictInterval(
    gm_bin, newdata = cbpp[1:5, ],
    seed = SEED, n.sims = NSIMS,
    include.resid.var = TRUE, type = "linear.prediction"))
  cases$glmm_bin_with_resid_prob <- suppressWarnings(predictInterval(
    gm_bin, newdata = cbpp[1:5, ],
    seed = SEED, n.sims = NSIMS,
    include.resid.var = TRUE, type = "probability"))

  attr(cases, "metadata") <- list(
    pkg_path        = normalizePath(pkg_path),
    package_version = as.character(utils::packageVersion("merTools")),
    R_version       = R.version.string,
    sysname         = Sys.info()[["sysname"]],
    machine         = Sys.info()[["machine"]],
    timestamp       = format(Sys.time(), "%Y-%m-%dT%H:%M:%S%z"),
    seed            = SEED,
    n_sims          = NSIMS,
    n_cases         = length(cases)
  )

  saveRDS(cases, out_path)
  cat("Saved", length(cases), "cases to", out_path, "\n")
  cat("Package version:", attr(cases, "metadata")$package_version, "\n")
  invisible(cases)
}

# -- Diff --------------------------------------------------------------------

summarize_case <- function(ma, mb) {
  cols <- intersect(names(ma), names(mb))
  if (length(cols) == 0L) {
    return(data.frame(col = "<no common cols>",
                      max_abs = NA_real_, rel_err = NA_real_))
  }
  do.call(rbind, lapply(cols, function(col) {
    x <- suppressWarnings(as.numeric(ma[[col]]))
    y <- suppressWarnings(as.numeric(mb[[col]]))
    if (length(x) != length(y)) {
      return(data.frame(col = col, max_abs = NA_real_, rel_err = NA_real_))
    }
    d <- abs(x - y)
    s <- pmax(abs(x), abs(y), 1e-12)
    data.frame(col = col,
               max_abs = max(d, na.rm = TRUE),
               rel_err = max(d / s, na.rm = TRUE))
  }))
}

classify <- function(max_abs) {
  if (!is.finite(max_abs))  return("<all NA>")
  if (max_abs < 1e-10)      return("IDENTICAL")
  if (max_abs < 1e-6)       return("numeric-equal")
  if (max_abs < 1e-3)       return("tiny drift")
  if (max_abs < 0.5)        return("small drift")
  if (max_abs < 5)          return("moderate drift")
                            return("LARGE DIFF")
}

run_diff <- function(rds_a, rds_b) {
  a <- readRDS(rds_a); b <- readRDS(rds_b)
  meta_a <- attr(a, "metadata"); meta_b <- attr(b, "metadata")

  cat(sprintf("A: %s  (%s)\n", rds_a, meta_a$package_version %||% "?"))
  cat(sprintf("B: %s  (%s)\n", rds_b, meta_b$package_version %||% "?"))
  cat(sprintf("R: %s\n\n", meta_a$R_version %||% R.version.string))

  worst <- 0
  for (nm in union(names(a), names(b))) {
    if (is.null(a[[nm]]) || is.null(b[[nm]])) {
      cat(sprintf("%-32s [missing in one bundle]\n", nm)); next
    }
    ma <- a[[nm]]; mb <- b[[nm]]
    if (!identical(dim(ma), dim(mb))) {
      cat(sprintf("%-32s [structure differs]\n", nm)); next
    }
    r <- summarize_case(ma, mb)
    maxa <- max(r$max_abs, na.rm = TRUE)
    maxr <- max(r$rel_err, na.rm = TRUE)
    worst <- max(worst, if (is.finite(maxa)) maxa else 0)
    cat(sprintf("%-32s  max_abs=%.3e  max_rel=%.3e  %s\n",
                nm, maxa, maxr, classify(maxa)))
  }
  cat(sprintf("\nOverall max absolute diff: %.6e\n", worst))
  invisible(worst)
}

`%||%` <- function(a, b) if (is.null(a)) b else a

# -- Dispatch ----------------------------------------------------------------

args <- commandArgs(trailingOnly = TRUE)
if (length(args) < 1L) {
  stop("Usage: Rscript predictInterval-regression.R harness <pkg> <out.rds>\n",
       "   or: Rscript predictInterval-regression.R diff    <a.rds> <b.rds>")
}
mode <- args[[1]]
if (mode == "harness") {
  if (length(args) != 3L) stop("harness mode requires <pkg_path> <output_rds>")
  run_harness(args[[2]], args[[3]])
} else if (mode == "diff") {
  if (length(args) != 3L) stop("diff mode requires <rds_a> <rds_b>")
  run_diff(args[[2]], args[[3]])
} else {
  stop("Unknown mode: ", mode)
}

#What exactly are the values spit out by fitted(sim(merMod), merMod)???
#Does it produce what we are trying to produce with predictInterval?

#The answer is no because it doesn't re-sample the RFX (I am fairly certain)
#The following does show the arm::sim() side-by-side with our method.  It looks
#like the effect of the differences is that our within-group regression lines are
#closer to the population average regression line.

library(arm); library(abind); library(mvtnorm);
set.seed(95371)

data(sleepstudy)

model <- lmer(Reaction ~ Days + (Days|Subject), data=sleepstudy)

#Take 5 simulations
test.sim <- sim(model, 1000)

#Display simulated coefs
#coef(test.sim)
ranef.array <- coef(test.sim)$ranef[[1]]
fixef.matrix <- coef(test.sim)$fixef

#Expand and permute fixefs to conform to ranefs for elementwise addition
dim(ranef.array)
dim(fixef.matrix)

#For now do it by hand
fixef.array <- array(data = rep(fixef.matrix, dim(ranef.array)[2]),
                     dim  = c(dim(fixef.matrix), dim(ranef.array)[2]))
fixef.array <- aperm(fixef.array, c(1,3,2))
dim(fixef.array)

combo.array <- fixef.array+ranef.array

#Extract model.matrix for our own multiplication
model.matrix <- lFormula(model@call, data=model@frame)$X
dim(model.matrix)
expanded.combo.array <- combo.array[,model@flist[[1]],]
expanded.combo.array <- aperm(expanded.combo.array, c(3,2,1))

#Matrix multiplication with arrays ???

##The following yeilds a 180x180 matrix for each prediction because I have
##multiplied ALL obs by ALL possible group coefficients
myCalc.large <- abind(
                  lapply(1:dim(expanded.combo.array)[3],
                         function(i) model.matrix %*% expanded.combo.array[,,i]),
                  along=3)

##I only need the diagonal of each 180x180 matrix so we get the value of
##the values for obs i and the coefficients for obs i
myCalc <- abind(lapply(1:dim(myCalc.large)[3],
                       function(x) diag(myCalc.large[,,x])), along=2)

isTRUE(all.equal(myCalc,fitted(test.sim, model), check.attributes=FALSE))

##Compare with predictInterval()
checkPI <- predictInterval(model, sleepstudy, nsim = 1000, predict="link")$yhat

myCalc.lwr <- apply(myCalc,1,function(x) as.numeric(quantile(x, .025)))
myCalc.fit <- apply(myCalc,1,function(x) as.numeric(quantile(x, .500)))
myCalc.upr <- apply(myCalc,1,function(x) as.numeric(quantile(x, .975)))

checkPI.lwr <- apply(checkPI,1,function(x) as.numeric(quantile(x, .025)))
checkPI.fit <- apply(checkPI,1,function(x) as.numeric(quantile(x, .500)))
checkPI.upr <- apply(checkPI,1,function(x) as.numeric(quantile(x, .975)))

plot.data <- rbind(
               data.frame(model="Arm.sim",
                          x=(1:180)-.15,
                          lwr=myCalc.lwr,
                          fit=myCalc.fit,
                          upr=myCalc.upr),
               data.frame(model="predictInterval",
                          x=(1:180)+.15,
                          lwr=checkPI.lwr,
                          fit=checkPI.fit,
                          upr=checkPI.upr))

ggplot(aes(x=x, y=fit, ymin=lwr, ymax=upr, color=model, position="dodge"), data=plot.data) +
geom_point() +
geom_linerange()

library(dplyr)
calc.sigma <- sqrt(1/rgamma(1000, 0.5*lme4:::df.residual.merMod(model), 0.5*getME(model, "devcomp")$cmp[["pwrss"]]))
sim.sigma <- test.sim@sigma
data.frame(
  model=c(rep("ARM",1000),rep("Carl",1000)),
  sigma=c(sim.sigma, calc.sigma)
  ) %>%
  qplot(sigma, color=model, data=., geom="density")

#' Predict from merMod objects with a prediction interval
#'
#' @description This function provides a way to capture model uncertainty in
#' predictions from multi-level models fit with \code{lme4}. By drawing a sampling
#' distribution for the random and the fixed effects and then estimating the fitted
#' value across that distribution, it is possible to generate a prediction interval
#' for fitted values that includes all variation in the model except for variation
#' in the covariance parameters, theta. This is a much faster alternative than
#' bootstrapping for models fit to medium to large datasets.
#'
#' @param merMod a merMod object from lme4
#' @param newdata a data.frame of new data to predict
#' @param which a character specifying what to return, by default it returns the
#'   full interval, but you can also select to return only the fixed variation or
#'   the random component variation. If full is selected the resulting data.frame
#'   will be \code{nrow(newdata) * number of model levels} long
#' @param level the width of the prediction interval
#' @param n.sims number of simulation samples to construct
#' @param stat take the median or mean of simulated intervals
#' @param type type of prediction to develop
#' @param include.resid.var logical, include or exclude the residual variance for
#'   linear models
#' @param returnSims logical, should all n.sims simulations be returned?
#' @param seed numeric, optional argument to set seed for simulations
#' @param .parallel logical, use parallel computation (default FALSE)
#' @param fix.intercept.variance logical; should the variance of the intercept
#'   term be adjusted downwards to roughly correct for its covariance with the
#'   random effects, as if all the random effects are intercept effects?
#' @param ignore.fixed.terms a numeric or string vector of indexes or names of
#'   fixed effects which should be considered as fully known (zero variance).
#' @param new.levels character, how to treat grouping levels in \code{newdata}
#'   that were not present when the model was fit. \code{"zero"} (the default and
#'   the historical behavior) drops the random effect for such levels, so the
#'   prediction rests on the fixed effects plus residual variation. \code{"draw"}
#'   instead samples each unobserved group's effect from the estimated
#'   random-effect covariance (\code{VarCorr}), so the interval reflects
#'   between-group uncertainty -- the analogue of
#'   \code{brms::posterior_predict(allow_new_levels = TRUE)}. Observations that
#'   share an unobserved level share the same sampled effect.
#'
#' @return a data.frame with three columns: fit, lwr and upr. If `returnSims`
#'   is TRUE the attribute `sim.results` contains the full simulation array.
#' @export
#' @import lme4
#' @importFrom abind abind
#' @importFrom mvtnorm rmvnorm
#' @importFrom foreach %dopar% foreach
predictInterval <- function(
  merMod,
  newdata = NULL,
  which = c("full", "fixed", "random", "all"),
  level = 0.8,
  n.sims = 1000,
  stat = c("median", "mean"),
  type = c("linear.prediction", "probability"),
  include.resid.var = TRUE,
  returnSims = FALSE,
  seed = NULL,
  .parallel = FALSE,
  fix.intercept.variance = FALSE,
  ignore.fixed.terms = NULL,
  new.levels = c("zero", "draw")
) {
  #--- Argument handling -------------------------------------------------------
  new.levels <- match.arg(new.levels, c("zero", "draw"), several.ok = FALSE)
  if (inherits(merMod, "nlmerMod")) {
    stop("predictInterval() does not support nlmer models.", call. = FALSE)
  }
  if (missing(newdata)) {
    newdata <- merMod@frame
  }
  if (any(c("tbl_df", "tbl") %in% class(newdata))) {
    newdata <- as.data.frame(newdata)
    warning("newdata is tbl_df or tbl object from dplyr package and has been coerced to a data.frame")
  } else if (!inherits(newdata, "data.frame")) {
    newdata <- as.data.frame(newdata)
  }

  predict.type <- match.arg(type, c("linear.prediction", "probability"),
                            several.ok = FALSE)
  stat.type    <- match.arg(stat, c("median", "mean"), several.ok = FALSE)
  which.eff    <- match.arg(which, c("full", "fixed", "random", "all"),
                            several.ok = FALSE)

  # Set random seed for reproducibility
  if (!is.null(seed)) {
    set.seed(seed)
  } else if (!exists(".Random.seed", envir = .GlobalEnv)) {
    runif(1)
  }

  # Check for probability type with linear models
  merMod.devcomp <- getME(merMod, "devcomp")
  if (merMod.devcomp$dims[["GLMM"]] == 0 &&
      merMod.devcomp$dims[["NLMM"]] == 0 &&
      predict.type == "probability") {
    predict.type <- "linear.prediction"
    warning("Asking for predictions on the probability scale makes no sense, resetting predict.type to linear.prediction",
            call. = FALSE)
  }

  # Detect if an intercept is present in random effects
  # This is checked for every call to maintain backward compatibility with
  # the message about intercept detection
  groupNames <- names(ngrps(merMod))
  hasREIntercept <- FALSE
  for (j in groupNames) {
    if (!is.na(names(attr(VarCorr(merMod)[[j]], "stddev")["(Intercept)"]))) {
      hasREIntercept <- TRUE
      break
    }
  }
  if (!hasREIntercept) {
    if (fix.intercept.variance) {
      fix.intercept.variance <- FALSE
    }
    message("No intercept detected, setting fix.intercept.variance to FALSE")
  }

  # Check for intercept when fix.intercept.variance is still TRUE
  if (fix.intercept.variance) {
    # Check if intercept is in fixed effects
    if (!"(Intercept)" %in% names(fixef(merMod))) {
      warning("No fixed-effect intercept detected. Variance adjustment may be unreliable.")
    }
  }

  # Check GLMM support
  check_glmm_support <- function(merMod, include.resid.var) {
    devcomp <- getME(merMod, 'devcomp')
    
    if (devcomp$dims[["GLMM"]] == TRUE) {
      family <- merMod@resp$family$family
      supported_families <- c("binomial", "poisson", "Gamma")
      
      if (!(family %in% supported_families)) {
        msg <- paste0(
          "Prediction for ", family, " GLMMs is not supported.\n",
          "Supported families: binomial, poisson, Gamma\n",
          "Try include.resid.var = FALSE as a fallback."
        )
        stop(msg, call. = FALSE)
      }
      
      if (include.resid.var && family == "binomial") {
        warning(paste0(
          "For binomial GLMMs, include.resid.var = TRUE simulates from the\n",
          "conditional binomial distribution (n-trial binomial simulation).\n",
          "This is the theoretically correct approach.\n",
          "To get predictions without residual variance, set include.resid.var = FALSE."
        ), call. = FALSE)
      }
    }
  }
  
  check_glmm_support(merMod, include.resid.var)

  #--- Simulations ------------------------------------------------------------
  # Draw from a single RNG stream seeded once above: sigma -> random effects
  # -> fixed effects. Passing seed = NULL to the helpers means they use the
  # ambient RNG state rather than reset it, which would otherwise correlate
  # FE and RE draws and diverge from the pre-refactor numeric output for
  # any user-supplied seed.
  sigma_vec <- simulate_residual_variance(merMod, n.sims)
  random_list <- simulate_random_effects(
    merMod,
    newdata,
    n.sims,
    .parallel = .parallel,
    seed = NULL,
    new.levels = new.levels
  )
  fixed_mat <- simulate_fixed_effects(
    merMod,
    newdata,
    n.sims,
    ignore.fixed.terms = ignore.fixed.terms,
    fix.intercept.variance = fix.intercept.variance,
    .parallel = .parallel,
    seed = NULL
  )

  # Extract family/link info for GLMM support
  devcomp <- getME(merMod, 'devcomp')
  family_info <- NULL
  weights <- NULL
  
  if (devcomp$dims[["GLMM"]] == TRUE) {
    family_info <- list(
      family = merMod@resp$family$family,
      link = merMod@resp$family$link
    )
    
    # Handle weights for binomial GLMMs
    if (family_info$family == "binomial") {
      # First check for explicit weights
      if (!is.null(merMod@frame$weights)) {
        weights <- as.numeric(merMod@frame$weights)
      } else if (any(grepl("cbind", names(merMod@frame)))) {
        # For cbind syntax, extract weights from response columns
        cbind_cols <- names(merMod@frame)[grepl("cbind", names(merMod@frame))]
        if (length(cbind_cols) == 1) {
          weights <- rowSums(merMod@frame[, cbind_cols, drop = FALSE])
        }
      }
    }
  }

  #--- Combine components ------------------------------------------------------
combined_result <- combine_components(
     fixed_mat = fixed_mat,
     random_list = random_list,
     sigma_vec = sigma_vec,
     include.resid.var = include.resid.var,
     which = which.eff,
     family = if (!is.null(family_info)) family_info$family else NULL,
     link = if (!is.null(family_info)) family_info$link else NULL,
     weights = weights,
     use.probability = (include.resid.var && !is.null(family_info) &&
                        family_info$family %in% c("binomial", "poisson", "Gamma") &&
                        predict.type == "probability")
   )

  # Handle which = "all" case separately
  if (which.eff == "all") {
    yhat_arr <- combined_result$yhat
    pi.comps <- combined_result$components
  } else {
    yhat_arr <- combined_result
    pi.comps <- NULL
  }

  #--- Summarise --------------------------------------------------------------
  # For GLMMs with include.resid.var=TRUE, simulations are in response scale
  # so skip linkinv. For LMMs or GLMMs without residual variance, use original logic.
  devcomp <- getME(merMod, "devcomp")
  is.glmm.with.response <- (devcomp$dims[["GLMM"]] == TRUE && include.resid.var)
  
  outs <- summarise_predictions(
    yhat_arr = yhat_arr,
    level = level,
    stat.type = stat.type,
    predict.type = predict.type,
    N = nrow(newdata),
    merMod = merMod,
    which.eff = which.eff,
    pi.comps = pi.comps,
    is.glmm.with.response = is.glmm.with.response
  )

  # Attach simulation results if requested
  if (returnSims) {
    if (which.eff == "all") {
      attr(outs, "sim.results") <- pi.comps
    } else {
      attr(outs, "sim.results") <- yhat_arr
    }
  }

  outs
}

test_that("Prediction intervals work for multiple parameters per level", {
  skip_on_ci()
  skip_on_cran()
  set.seed(5150)
  data(grouseticks)
  grouseticks$HEIGHT <- scale(grouseticks$HEIGHT)
  grouseticks <- merge(grouseticks, grouseticks_agg[, 1:3], by = "BROOD")
  grouseticks$TICKS_BIN <- ifelse(grouseticks$TICKS >=1, 1, 0)
  # GLMER 3 level + slope
  form <- TICKS_BIN ~ YEAR + HEIGHT +(1 + HEIGHT|BROOD) + (1|LOCATION) + (1|INDEX)
  suppressMessages({
    glmer3LevSlope  <- glmer(form, family="binomial",data=grouseticks,
                             control = glmerControl(optimizer="bobyqa",
                                                    optCtrl=list(maxfun = 1e5)))
  })

  #In the call below we are getting warnings because our call to mvtnorm::rmvnorm
  #is shotting a warning when mean and sigma of multivariate distribution are
  #zero using the method="chol
  outs1 <- suppressWarnings(predictInterval(glmer3LevSlope, newdata = grouseticks[1:10,]))
  expect_s3_class(outs1, "data.frame")
})

test_that("Prediction works for random slopes not in fixed", {
  skip_on_ci()
  skip_on_cran()
  set.seed(5150)
  data(grouseticks)
  grouseticks$HEIGHT <- scale(grouseticks$HEIGHT)
  grouseticks <- merge(grouseticks, grouseticks_agg[, 1:3], by = "BROOD")
  grouseticks$TICKS_BIN <- ifelse(grouseticks$TICKS >=1, 1, 0)
  # GLMER 3 level + slope
  form <- TICKS_BIN ~ YEAR + (1 + HEIGHT|BROOD) + (1|LOCATION) + (1|INDEX)
  suppressMessages({
    glmer3LevSlope  <- glmer(form, family="binomial",data=grouseticks,
                             control = glmerControl(optimizer="bobyqa",
                                                    optCtrl=list(maxfun = 1e5)))
  })

  zNew <- grouseticks[1:10,]
  #In the call below we are getting warnings because our call to mvtnorm::rmvnorm
  #is shotting a warning when mean and sigma of multivariate distribution are
  #zero using the method="chol
  outs1 <- suppressWarnings(predictInterval(glmer3LevSlope, newdata = zNew))
  expect_s3_class(outs1, "data.frame")
  # Message may not be necessary any more
  # expect_message(predictInterval(glmer3LevSlope, newdata = zNew))
})

# Test for new factor levels----

test_that("Prediction intervals work with new factor levels added", {
  skip_on_ci()
  skip_on_cran()
  set.seed(5150)
  data(grouseticks)
  grouseticks$HEIGHT <- scale(grouseticks$HEIGHT)
  grouseticks <- merge(grouseticks, grouseticks_agg[, 1:3], by = "BROOD")
  grouseticks$TICKS_BIN <- ifelse(grouseticks$TICKS >=1, 1, 0)
  # GLMER 3 level + slope
  form <- TICKS_BIN ~ YEAR + HEIGHT +(1 + HEIGHT|BROOD) + (1|LOCATION) + (1|INDEX)
  suppressMessages({
    glmer3LevSlope  <- glmer(form, family="binomial",data=grouseticks,
                             control = glmerControl(optimizer="bobyqa",
                                                    optCtrl=list(maxfun = 1e5)))

  })

  zNew <- grouseticks[1:10,]
  zNew$BROOD <- as.character(zNew$BROOD)
  zNew$BROOD[1:9] <- "100"
  zNew$BROOD[10] <- "101"
  #In the call below we are getting warnings because our call to mvtnorm::rmvnorm
  #is shouting a warning when mean and sigma of multivariate distribution are
  #zero using the method="chol
  outs1 <- suppressWarnings(predictInterval(glmer3LevSlope, newdata = zNew))
  expect_s3_class(outs1, "data.frame")
  # Test all warnings and messages here
  expect_snapshot(predictInterval(glmer3LevSlope, newdata = zNew))
})


test_that("Prediction works for factor as a random slope not in fixed", {
  skip_on_ci()
  skip_on_cran()
  set.seed(5150)
  data(grouseticks)
  grouseticks$HEIGHT <- scale(grouseticks$HEIGHT)
  grouseticks <- merge(grouseticks, grouseticks_agg[, 1:3], by = "BROOD")
  grouseticks$TICKS_BIN <- ifelse(grouseticks$TICKS >=1, 1, 0)
  # GLMER 3 level + slope
  form <- TICKS_BIN ~ HEIGHT +(1 + YEAR|BROOD) + (1|LOCATION)
  #Suppressing warning for known degenerate model below
  suppressMessages({
    glmer3LevSlope  <- suppressWarnings(glmer(form, family="binomial",data=grouseticks,
                                              control = glmerControl(optimizer="bobyqa",
                                                                     optCtrl=list(maxfun = 1e5))))
  })
  zNew <- grouseticks[1:10,]
  zNew$BROOD <- as.character(zNew$BROOD)
  zNew$BROOD[1:9] <- "100"
  zNew$BROOD[10] <- "101"

  predictInterval(glmer3LevSlope, newdata = zNew) |>
     expect_warning("Currently, predictions for these values are based only on the") |>
    suppressWarnings()

  outs1 <- suppressWarnings(predictInterval(glmer3LevSlope, newdata = zNew))
  zNew <- grouseticks[1:10,]
  # Expect warnings because we have 0 for our reMeans on location in this model
  # mvtnorm now issues a warning about the indefinite/rank-deficient matrix
  predictInterval(glmer3LevSlope, newdata = zNew) |>
    expect_warning("the matrix is either rank-deficient or not positive definite") |>
    suppressWarnings()
  # Add test to confirm this
  suppressWarnings({
    outs2 <- predictInterval(glmer3LevSlope, newdata = zNew)
  })

  expect_s3_class(outs1, "data.frame")
  expect_s3_class(outs2, "data.frame")
  expect_identical(dim(outs1), dim(outs2))
})

# Numeric accuracy----

# Cases
# new factor level for group term

test_that("Median of prediction interval is close to predict.lmer for single group models", {
  skip_on_cran()
  set.seed(2311)
  fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
  truPred <- predict(fm1, newdata = sleepstudy)
  newPred <- predictInterval(fm1, newdata = sleepstudy, n.sims = 500,
                             level = 0.9, stat = c("median"),
                             include.resid.var = FALSE, seed = 4563)
  expect_equal(mean(newPred$fit - truPred), 0, tolerance = sd(truPred)/50)

  fm1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
  truPred <- predict(fm1, newdata = sleepstudy)
  newPred <- predictInterval(fm1, newdata = sleepstudy, n.sims = 1500,
                             level = 0.9, stat = c("median"),
                             include.resid.var = FALSE, seed = 9598)
  expect_equal(mean(newPred$fit - truPred), 0, tolerance = sd(truPred)/100)
})

test_that("Median of PI is close to predict.lmer for complex group models", {
  skip_on_cran()
  skip_on_ci()
  set.seed(101)
  moddf <- InstEval[sample(rownames(InstEval), 10000), ]
  g1 <- lmer(y ~ lectage + studage + (1|d) + (1|s), data=moddf)
  d1 <- moddf[1:200, ]
  newPred <- predictInterval(g1, newdata = d1, level = 0.8, n.sims = 500,
                             stat = 'median', include.resid.var = FALSE,
                             seed = 4563)
  truPred <- predict(g1, newdata = d1)
  expect_equal(mean(newPred$fit - truPred), 0, tolerance = sd(truPred)/100)
  rm(list=ls())
})

test_that("Median of PI is close to predict.glmer for basic and complex grouping", {
  skip_on_cran()
  skip_on_ci()
  set.seed(8496)
  d <- expand.grid(fac1=LETTERS[1:5], grp=factor(1:8), fac2 = LETTERS[12:19],
                   obs=1:20)
  d$x <- runif(nrow(d))
  suppressMessages({
    d$y <- simulate(~ x + fac1 + fac2 + (1 + fac1|grp) + (1|obs), family = binomial,
                    newdata=d,
                    newparams=list(beta = rnorm(13),
                                   theta = rnorm(16, 5, 1)), seed = 4563)[[1]]
  })
    subD <- d[sample(row.names(d), 1500),]
  # TOO SLOW
  g1 <- glmer(y ~ x + fac1 + fac2 + (1+fac1|grp) + (1|obs), data = subD,
              family = 'binomial',
              control = glmerControl(optimizer="bobyqa",
                                     optCtrl=list(maxfun = 1e5)))
  truPred <- predict(g1, subD)
  newPred <- suppressWarnings(predictInterval(g1, newdata = subD, level = 0.95, n.sims = 2000,
                                              stat = 'median', include.resid.var = FALSE,
                                              type = 'linear.prediction', seed = 3252))
  expect_equal(mean(newPred$fit - truPred), 0, tolerance = sd(truPred)/20)
  # This test fails currently
  #   g1 <- glmer(y ~ x +  fac2 + (1 + fac1|grp) + (1|obs), data = subD, family = 'binomial')
  #   truPred <- predict(g1, subD, type = "response")
  #   newPred <- predictInterval(g1, newdata = subD, level = 0.8, n.sims = 500,
  #                              stat = 'median', include.resid.var = FALSE,
  #                              type = 'probability')
  #   expect_equal(mean(newPred$fit - truPred), 0, tolerance = sd(truPred)/20)
  rm(list = ls())
})


test_that("Prediction intervals work with new factor levels added", {
  skip_on_cran()
  skip_on_ci()
  data(grouseticks)
  grouseticks$HEIGHT <- scale(grouseticks$HEIGHT)
  grouseticks <- merge(grouseticks, grouseticks_agg[, 1:3], by = "BROOD")
  grouseticks$TICKS_BIN <- ifelse(grouseticks$TICKS >=1, 1, 0)
  # GLMER 3 level + slope
  form <- TICKS_BIN ~ YEAR + HEIGHT +(1 + HEIGHT|BROOD) + (1|LOCATION) + (1|INDEX)
  glmer3LevSlope  <- glmer(form, family="binomial",data=grouseticks,
                           control = glmerControl(optimizer="bobyqa",
                                                  optCtrl=list(maxfun = 1e5)))

  zNew <- grouseticks
  zNew$BROOD <- as.character(zNew$BROOD)
  zNew$BROOD[1:99] <- "100"
  zNew$BROOD[100] <- "101"
  #In the call below we are getting warnings because our call to mvtnorm::rmvnorm
  #is shotting a warning when mean and sigma of multivariate distribution are
  #zero using the method="chol
  newPred <- suppressWarnings(predictInterval(glmer3LevSlope, newdata = zNew, level = 0.95,
                                              n.sims = 500, stat = 'median',
                                              include.resid.var = TRUE, seed = 4563))
  truPred <- predict(glmer3LevSlope, newdata = zNew, allow.new.levels = TRUE)
  expect_equal(mean(newPred$fit - truPred), 0, tolerance = sd(truPred)/40)
})

test_that("Prediction intervals work with slope not in fixed effects and data reordered", {
  skip_on_ci()
  skip_on_cran()
  data(grouseticks)
  grouseticks$HEIGHT <- scale(grouseticks$HEIGHT)
  grouseticks <- merge(grouseticks, grouseticks_agg[, 1:3], by = "BROOD")
  grouseticks$TICKS_BIN <- ifelse(grouseticks$TICKS >=1, 1, 0)
  # GLMER 3 level + slope
  form <- TICKS_BIN ~ YEAR + (1 + HEIGHT|BROOD) + (1|LOCATION) + (1|INDEX)
  glmer3LevSlope  <- glmer(form, family="binomial",data=grouseticks,
                           control = glmerControl(optimizer="bobyqa",
                                                  optCtrl=list(maxfun = 1e5)))
  zNew <- grouseticks
  zNew$BROOD <- as.character(zNew$BROOD)
  zNew$BROOD[1:99] <- "100"
  zNew$BROOD[100] <- "101"
  zNew <- zNew[, c(10, 9, 8, 7, 1, 2, 3, 4, 5, 6, 10)]
  #In the call below we are getting warnings because our call to mvtnorm::rmvnorm
  #is shotting a warning when mean and sigma of multivariate distribution are
  #zero using the method="chol
  newPred <- suppressWarnings(predictInterval(glmer3LevSlope, newdata = zNew, level = 0.95,
                                              n.sims = 500, stat = 'median',
                                              include.resid.var = TRUE, seed = 4563))
  truPred <- predict(glmer3LevSlope, newdata = zNew, allow.new.levels = TRUE)
  expect_equal(mean(newPred$fit - truPred), 0, tolerance = sd(truPred)/20)
})

# Test Parallel----


test_that("parallelization does not throw errors and generates good results", {
  skip_on_cran()
  skip_on_ci()
  library(foreach)
  set.seed(1241)
  m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
  predA <- predictInterval(m1, newdata = m1@frame, n.sims = 2200, seed = 54,
                           include.resid.var = FALSE, stat = "median") |> suppressWarnings()
  predB <- predictInterval(m1, newdata = m1@frame, n.sims = 1750, seed = 54,
                           include.resid.var = FALSE, stat = "median")
  expect_equal(mean(predA$fit - predB$fit), 0 , tolerance = .2)
  predA <- predictInterval(m1, newdata = m1@frame, n.sims = 2500, seed = 2141,
                           include.resid.var = FALSE)
  predB <- predictInterval(m1, newdata = m1@frame, n.sims = 1500, seed = 2141,
                           include.resid.var = FALSE)
  expect_equal(mean(predA$fit - predB$fit), 0 , tolerance = .01)
  moddf <- InstEval[sample(rownames(InstEval), 5000),]
  g1 <- lmer(y ~ lectage + studage + (1|d) + (1|s), data = moddf)
  predA <- predictInterval(g1, newdata = g1@frame, n.sims = 2500, seed = 2141,
                           include.resid.var = FALSE)
  predB <- predictInterval(g1, newdata = g1@frame, n.sims = 1500, seed = 2141,
                           include.resid.var = FALSE)
  expect_equal(mean(predA$fit - predB$fit), 0 , tolerance = .01)
  predA <- predictInterval(g1, newdata = g1@frame[1:499,], n.sims = 2500, seed = 2141,
                           include.resid.var = TRUE)
  predB <- predictInterval(g1, newdata = g1@frame[1:501,], n.sims = 2500, seed = 2141,
                           include.resid.var = TRUE)
  expect_equal(mean(predA$fit[1:499] - predB$fit[1:499]), 0 , tolerance = .0025)
  detach("package:foreach", character.only=TRUE)
})

#context("Test returning predict interval components")


test_that("Output is correct dimensions", {
  skip_on_cran()
  ###########################################
  # Test the option to return different predictInterval components
  m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
  m2 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
  form <- TICKS_BIN ~ YEAR + HEIGHT +(1 + HEIGHT|BROOD) + (1|LOCATION) + (1|INDEX)
  data(grouseticks)
  grouseticks$TICKS_BIN <- ifelse(grouseticks$TICKS >=1, 1, 0)
  grouseticks$YEAR  <- as.numeric(grouseticks$YEAR)
  grouseticks$HEIGHT  <- grouseticks$HEIGHT - 462.2
  glmer3LevSlope  <- glmer(form, family="binomial",data=grouseticks) |>
    suppressMessages()

  #############################################
  pred1 <- predictInterval(m1, which = "random")
  pred2 <- predictInterval(m2, which = "fixed")
  pred3 <- predictInterval(m2, which = "all")
  expect_equal(nrow(pred1), nrow(pred2))
  expect_equal(ncol(pred1), 3)
  expect_equal(ncol(pred2), 3)
  expect_equal(ncol(pred3), 5)
  expect_equal(nrow(pred3), 180*3)
  expect_equal(nrow(pred1), 180)
  expect_false(nrow(pred3) == nrow(pred2))
  expect_true(nrow(pred3) > nrow(pred2))
  pred1 <- predictInterval(m1, which = "random", stat = "mean")
  pred2 <- predictInterval(m2, which = "fixed", stat = "mean")
  pred3 <- predictInterval(m2, which = "all", stat = "mean")
  expect_equal(nrow(pred1), nrow(pred2))
  expect_equal(ncol(pred1), 3)
  expect_equal(ncol(pred2), 3)
  expect_equal(ncol(pred3), 5)
  expect_equal(nrow(pred3), 180*3)
  expect_equal(nrow(pred1), 180)
  expect_false(nrow(pred3) == nrow(pred2))
  expect_true(nrow(pred3) > nrow(pred2))
  pred1 <- suppressWarnings(predictInterval(glmer3LevSlope, which = "random",
                                            type = "linear.prediction"))
  pred2 <- suppressWarnings(predictInterval(glmer3LevSlope, which = "random",
                                            type = "probability"))
  pred3 <- suppressWarnings(predictInterval(glmer3LevSlope, which = "all",
                                            type = "linear.prediction"))
  pred4 <- suppressWarnings(predictInterval(glmer3LevSlope, which = "all",
                                            type = "probability"))
  expect_equal(nrow(pred1), nrow(pred2))
  expect_equal(nrow(pred3), nrow(pred4))
  expect_equal(ncol(pred1), 3)
  expect_equal(ncol(pred2), 3)
  expect_equal(ncol(pred3), 5)
  expect_equal(ncol(pred3), 5)
  expect_true(mean(pred2$fit) > mean(pred1$fit))
  expect_equal(mean(pred2$fit), 0.5, tolerance = 0.05)
  expect_equal(mean(pred1$fit), 0.00, tolerance = 0.05)
  # Tolerance meaning has changed
  expect_equal(mean(pred4$fit), mean(grouseticks$TICKS_BIN), tolerance = 0.2)
  expect_false(mean(pred4$fit) == mean(pred3$fit))
  expect_equal(nrow(pred1), 403)
  expect_equal(nrow(pred2), 403)
  expect_equal(nrow(pred3), 403*5)
  expect_equal(nrow(pred4), 403*5)
})



test_that("Compare random, fixed, include-resid", {
  skip_on_cran()
  ######################################
  # Test the option to return different predictInterval components
  m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
  m2 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
  form <- TICKS_BIN ~ YEAR + HEIGHT +(1 + HEIGHT|BROOD) + (1|LOCATION) + (1|INDEX)
  data(grouseticks)
  grouseticks$TICKS_BIN <- ifelse(grouseticks$TICKS >=1, 1, 0)
  grouseticks$YEAR  <- as.numeric(grouseticks$YEAR)
  grouseticks$HEIGHT  <- grouseticks$HEIGHT - 462.2
  glmer3LevSlope  <- glmer(form, family="binomial",data=grouseticks) |>
    suppressMessages()
  ################################################
  predInt1 <- predictInterval(m1)
  predInt1$width <- predInt1[, 2] - predInt1[, 3]
  predInt2 <- predictInterval(m1, which = "random")
  predInt2$width <- predInt2[, 2] - predInt2[, 3]
  predInt3 <- predictInterval(m1, which = "fixed")
  predInt3$width <- predInt3[, 2] - predInt3[, 3]
  predInt4 <- predictInterval(m1, which = "all")
  predInt4$width <- predInt4[, 3] - predInt4[, 4]
  predInt1b <- predictInterval(m1, include.resid.var = FALSE)
  predInt1b$width <- predInt1b[, 2] - predInt1b[, 3]
  predInt2b <- predictInterval(m1, which = "random", include.resid.var = FALSE)
  predInt2b$width <- predInt2b[, 2] - predInt2b[, 3]
  predInt3b <- predictInterval(m1, which = "fixed", include.resid.var = FALSE)
  predInt3b$width <- predInt3b[, 2] - predInt3b[, 3]
  predInt4b <- predictInterval(m1, which = "all", include.resid.var = FALSE)
  predInt4b$width <- predInt4b[, 3] - predInt4b[, 4]
  # These should be fairly close now since residual variance is the biggest
  expect_true(!all(predInt1$width > predInt2$width))
  expect_true(!all(predInt3$width > predInt2$width))
  expect_true(!all(predInt4$width[predInt4$effect == "combined"] > predInt1$width))
  expect_true(!all(predInt4$width[predInt4$effect == "combined"] > predInt2$width))
  expect_true(!all(predInt4$width[predInt4$effect == "combined"] > predInt3$width))
  #
  expect_true(all(predInt1b$width > predInt2b$width))
  expect_true(all(predInt3b$width != predInt2b$width))
  expect_true(all(predInt4b$width[predInt4b$effect == "combined"] > predInt2b$width))
  expect_true(!all(predInt4b$width[predInt4b$effect == "combined"] > predInt1b$width))
  # Fits
  expect_true(all(predInt1$fit > predInt2$fit))
  expect_true(all(predInt3$fit > predInt2$fit))
  expect_true(all(predInt1$upr > predInt1b$upr))
  expect_true(all(predInt1$lwr < predInt1b$lwr))
  expect_true(all(predInt2$upr > predInt2b$upr))
  expect_true(all(predInt2$lwr < predInt2b$lwr))
  expect_true(all(predInt2$upr > predInt2b$upr))
  expect_true(all(predInt3$lwr < predInt3b$lwr))
  expect_true(all(predInt4$upr > predInt4b$upr))
  expect_true(all(predInt4$lwr < predInt4b$lwr))

  predInt1 <- predictInterval(m2)
  predInt1$width <- predInt1[, 2] - predInt1[, 3]
  predInt2 <- predictInterval(m2, which = "random")
  predInt2$width <- predInt2[, 2] - predInt2[, 3]
  predInt3 <- predictInterval(m2, which = "fixed")
  predInt3$width <- predInt3[, 2] - predInt3[, 3]
  predInt4 <- predictInterval(m2, which = "all")
  predInt4$width <- predInt4[, 3] - predInt4[, 4]
  predInt1b <- predictInterval(m2, include.resid.var = FALSE)
  predInt1b$width <- predInt1b[, 2] - predInt1b[, 3]
  predInt2b <- predictInterval(m2, which = "random", include.resid.var = FALSE)
  predInt2b$width <- predInt2b[, 2] - predInt2b[, 3]
  predInt3b <- predictInterval(m2, which = "fixed", include.resid.var = FALSE)
  predInt3b$width <- predInt3b[, 2] - predInt3b[, 3]
  predInt4b <- predictInterval(m2, which = "all", include.resid.var = FALSE)
  predInt4b$width <- predInt4b[, 3] - predInt4b[, 4]
  # These should be fairly close now since residual variance is the biggest variance
  expect_true(!all(predInt1$width > predInt2$width))
  expect_true(!all(predInt3$width > predInt2$width))
  expect_true(!all(predInt4$width[predInt4$effect == "combined"] > predInt1$width))
  expect_true(!all(predInt4$width[predInt4$effect == "combined"] > predInt2$width))
  expect_true(!all(predInt4$width[predInt4$effect == "combined"] > predInt3$width))
  #
  expect_true(all(predInt1b$width > predInt2b$width))
  expect_true(all(predInt3b$width != predInt2b$width))
  expect_true(all(predInt4b$width[predInt4b$effect == "combined"] > predInt2b$width))
  expect_true(!all(predInt4b$width[predInt4b$effect == "combined"] > predInt1b$width))
  # Fits
  expect_true(all(predInt1$fit > predInt2$fit))
  expect_true(all(predInt3$fit > predInt2$fit))
  expect_true(all(predInt1$upr > predInt1b$upr))
  expect_true(all(predInt1$lwr < predInt1b$lwr))
  expect_true(all(predInt2$upr > predInt2b$upr))
  expect_true(all(predInt2$lwr < predInt2b$lwr))
  expect_true(all(predInt2$upr > predInt2b$upr))
  expect_true(all(predInt3$lwr < predInt3b$lwr))
  expect_true(all(predInt4$upr > predInt4b$upr))
  expect_true(all(predInt4$lwr < predInt4b$lwr))
})

test_that("Default is set to all effects", {
  skip_on_cran()
  ######################################
  # Test the option to return different predictInterval components
  m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
  m2 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
  form <- TICKS_BIN ~ YEAR + HEIGHT +(1 + HEIGHT|BROOD) + (1|LOCATION) + (1|INDEX)
  data(grouseticks)
  grouseticks$TICKS_BIN <- ifelse(grouseticks$TICKS >=1, 1, 0)
  grouseticks$YEAR  <- as.numeric(grouseticks$YEAR)
  grouseticks$HEIGHT  <- grouseticks$HEIGHT - 462.2
  glmer3LevSlope  <- glmer(form, family="binomial",data=grouseticks) |>
    suppressMessages()
  ################################################
  predInt1 <- predictInterval(m1, seed = 8231)
  predInt2 <- predictInterval(m1, which = "full", seed = 8231)
  expect_identical(predInt1, predInt2)
  predInt1 <- predictInterval(m1, seed = 8231, include.resid.var = TRUE)
  predInt2 <- predictInterval(m1, which = "full", seed = 8231,
                              include.resid.var = TRUE)
  expect_identical(predInt1, predInt2)
  predInt1 <- predictInterval(m1, seed = 8231, include.resid.var = TRUE)
  predInt2 <- predictInterval(m1, which = "all", seed = 8231,
                              include.resid.var = TRUE)
  expect_equal(mean(predInt1$fit - predInt2$fit[predInt2$effect == "combined"]),
               0, tolerance =0.14)
  expect_equal(mean(predInt1$lwr - predInt2$lwr[predInt2$effect == "combined"]),
               0, tolerance =0.1)
  expect_equal(mean(predInt1$upr - predInt2$upr[predInt2$effect == "combined"]),
               0, tolerance=0.1)
})

# Test nested effect specifications----

test_that("Nested effects can work", {
  skip_on_cran()
  library(ggplot2)
  mod1 <- lmer(sleep_total ~ bodywt + (1|vore/order), data=msleep)
  msleep$combn <- paste(msleep$vore, msleep$order, sep = "__")
  mod2 <- lmer(sleep_total ~ bodywt +  (1|combn) + (1|vore), data=msleep)
  #Suppressing warnings we already tested (coerce tbl and new levels)
  predInt1 <- suppressWarnings(predictInterval(merMod=mod1, newdata=msleep, seed = 548,
                              n.sims = 2000, include.resid.var = FALSE,
                              stat = "median", level = 0.8))
  predInt2 <- suppressWarnings(predictInterval(merMod=mod2, newdata=msleep, seed = 548,
                              n.sims = 2000, include.resid.var = FALSE,
                              stat = "median", level = 0.8))
  expect_s3_class(predInt1, "data.frame")
  expect_s3_class(predInt2, "data.frame")
  expect_equal(mean(predInt1[,1] - predInt2[,1]), 0, tolerance = sd(predInt1[,1])/20)
  expect_equal(mean(predInt1[,2] - predInt2[,2]), 0, tolerance = sd(predInt1[,2])/10)
  expect_equal(mean(predInt1[,3] - predInt2[,3]), 0, tolerance = sd(predInt1[,3])/20)
})

# context("Interactions without intercepts")

test_that("Models with cross-level interaction and no random intercept work", {
  skip_on_cran()
  #################################
  sleepstudy$Test <- rep(sample(c(TRUE, FALSE), length(unique(sleepstudy$Subject)),
                                replace = TRUE), each = 10)
  m1 <- lmer(Reaction ~ Days:Test + (0 + Days | Subject), data = sleepstudy)

  sleepstudy$cars <- sleepstudy$Days*3
  m2 <- lmer(Reaction ~ cars:Test + (0 + Days | Subject), data = sleepstudy)
  m3 <- lmer(Reaction ~ cars:Test + (1 | Subject), data = sleepstudy)
  m4 <- lmer(Reaction ~ cars:Test + (0 + cars | Subject), data = sleepstudy)

  ###################################
  suppressMessages({
    preds1 <- predictInterval(m1)
  })
  expect_equal(nrow(preds1), 180)
  expect_equal(ncol(preds1), 3)
  expect_message(predictInterval(m1))
  suppressMessages({
    preds1 <- predictInterval(m1, newdata = sleepstudy[1:10, ],
                              level = 0.9, n.sims = 500, include.resid.var = FALSE,
                              fix.intercept.variance = TRUE)
  })
  expect_equal(nrow(preds1), 10)
  expect_equal(ncol(preds1), 3)
  suppressMessages({
    preds1 <- predictInterval(m1, newdata = sleepstudy[1:10, ],
                              level = 0.9, n.sims = 500, include.resid.var = FALSE,
                              ignore.fixed.terms = TRUE)
  })

  expect_equal(nrow(preds1), 10)
  expect_equal(ncol(preds1), 3)
  suppressMessages({
    preds2 <- predictInterval(m1, newdata = sleepstudy[1:10, ],
                              level = 0.9, n.sims = 500, include.resid.var = FALSE,
                              ignore.fixed.terms = FALSE)
  })
  expect_equal(nrow(preds2), 10)
  expect_equal(ncol(preds2), 3)
  expect_false(any(preds1$fit == preds2$fit))
  rm(preds1, preds2)
  suppressMessages({
    preds1 <- predictInterval(m2)
  })

  expect_equal(nrow(preds1), 180)
  expect_equal(ncol(preds1), 3)
  truPred <- predict(m2)
  expect_equal(mean(preds1$fit - truPred), 0, tolerance = sd(truPred)/100)

  #
  preds1 <- predictInterval(m3)
  expect_equal(nrow(preds1), 180)
  expect_equal(ncol(preds1), 3)
  truPred <- predict(m3)
  expect_equal(mean(preds1$fit - truPred), 0, tolerance = sd(truPred)/100)
  #
  suppressMessages({
    preds1 <- predictInterval(m4)
  })

  expect_equal(nrow(preds1), 180)
  expect_equal(ncol(preds1), 3)
  truPred <- predict(m4)
  expect_equal(mean(preds1$fit - truPred), 0, tolerance = sd(truPred)/100)
})



test_that("Models with cross-level interaction and no random intercept work", {
  skip_on_cran()
  suppressMessages({
    m1 <- lmer(Reaction ~ 0 + Days + Days:Subject + (1 | Days), data = sleepstudy)
  })
  preds1 <- predictInterval(m1)
  expect_equal(nrow(preds1), 180)
  expect_equal(ncol(preds1), 3)
  predictInterval(m1, fix.intercept.variance = TRUE) |>
    expect_warning("No fixed-effect intercept detected") |>
    suppressWarnings()

  preds1 <- predictInterval(m1, newdata = sleepstudy[1:10, ],
                            level = 0.9, n.sims = 500, include.resid.var = FALSE,
                            ignore.fixed.terms = TRUE)
  expect_equal(nrow(preds1), 10)
  expect_equal(ncol(preds1), 3)
  truPred <- predict(m1, newdata = sleepstudy[1:10,])
  expect_equal(mean(preds1$fit - truPred), 0, tolerance = sd(truPred)/100)

  # This is less close
  preds1 <- predictInterval(m1, newdata = sleepstudy[1:50, ],
                            level = 0.9, n.sims = 500, include.resid.var = FALSE,
                            ignore.fixed.terms = FALSE)
  expect_equal(nrow(preds1), 50)
  expect_equal(ncol(preds1), 3)
  truPred <- predict(m1, newdata = sleepstudy[1:50,])
  expect_equal(mean(preds1$fit - truPred), 0, tolerance = sd(truPred)/25)


  predictInterval(m1, newdata = sleepstudy[1:50, ],
                            level = 0.9, n.sims = 500, include.resid.var = FALSE,
                            fix.intercept.variance = TRUE) |>
    expect_warning("No fixed-effect intercept") |>
    suppressWarnings()

    expect_failure({
    expect_equal(mean(preds1$fit - truPred), 0, tolerance = sd(truPred)/100)
  })


})

# -----------------------------------------------------------------------------
# Layer 2 numeric regression tests for predictInterval().
#
# These tests pin the *exact* predictInterval() output for a canonical set of
# LMM and GLMM inputs via testthat snapshots. Any change to simulation
# internals that alters these outputs produces a reviewable snapshot diff,
# not a mysterious random-tolerance failure.
#
# Scope: Linux only. Apple-silicon and Windows produce bit-level drift from
# different BLAS/LAPACK, which is benign but would pollute the snapshots.
# Cross-platform numerical equivalence is checked manually by the harness at
# tests/comparisons/predictInterval-regression.R. Linux is the canonical
# platform for the committed snapshot contract.
#
# If a change is intentional, regenerate with:
#   testthat::snapshot_accept("predictInterval-snapshot")
# and inspect the diff carefully before committing.
# -----------------------------------------------------------------------------

SEED  <- 11213
NSIMS <- 1000
TOL   <- 1e-6

snapshot_platform_ok <- function() {
  skip_on_os(c("mac", "windows", "solaris"))
  skip_on_cran()
}

# LMM: random slope + random intercept --------------------------------------

test_that("LMM random-slope + intercept: predictInterval output is stable", {
  snapshot_platform_ok()

  data(sleepstudy, package = "lme4")
  m  <- suppressMessages(
    lmer(Reaction ~ Days + (Days | Subject), data = sleepstudy)
  )
  nd <- sleepstudy[1:10, ]

  expect_snapshot_value(
    predictInterval(m, newdata = nd, seed = SEED, n.sims = NSIMS),
    style = "json2", tolerance = TOL
  )
  expect_snapshot_value(
    predictInterval(m, newdata = nd, seed = SEED, n.sims = NSIMS,
                    include.resid.var = FALSE),
    style = "json2", tolerance = TOL
  )
  expect_snapshot_value(
    predictInterval(m, newdata = nd, seed = SEED, n.sims = NSIMS,
                    which = "fixed"),
    style = "json2", tolerance = TOL
  )
  expect_snapshot_value(
    predictInterval(m, newdata = nd, seed = SEED, n.sims = NSIMS,
                    which = "random"),
    style = "json2", tolerance = TOL
  )
  expect_snapshot_value(
    predictInterval(m, newdata = nd, seed = SEED, n.sims = NSIMS,
                    which = "all"),
    style = "json2", tolerance = TOL
  )
})

# LMM: random intercept only ------------------------------------------------

test_that("LMM random-intercept: predictInterval output is stable", {
  snapshot_platform_ok()

  data(sleepstudy, package = "lme4")
  m  <- suppressMessages(
    lmer(Reaction ~ Days + (1 | Subject), data = sleepstudy)
  )
  nd <- sleepstudy[1:10, ]

  expect_snapshot_value(
    predictInterval(m, newdata = nd, seed = SEED, n.sims = NSIMS),
    style = "json2", tolerance = TOL
  )
  expect_snapshot_value(
    predictInterval(m, newdata = nd, seed = SEED, n.sims = NSIMS,
                    include.resid.var = FALSE),
    style = "json2", tolerance = TOL
  )
  expect_snapshot_value(
    predictInterval(m, newdata = nd, seed = SEED, n.sims = NSIMS,
                    stat = "mean"),
    style = "json2", tolerance = TOL
  )
  expect_snapshot_value(
    predictInterval(m, newdata = nd, seed = SEED, n.sims = NSIMS,
                    level = 0.99),
    style = "json2", tolerance = TOL
  )
})

# LMM: edge-case inputs ------------------------------------------------------

test_that("LMM: single-row newdata and variance-adjustment options are stable", {
  snapshot_platform_ok()

  data(sleepstudy, package = "lme4")
  m_slope <- suppressMessages(
    lmer(Reaction ~ Days + (Days | Subject), data = sleepstudy)
  )
  m_int <- suppressMessages(
    lmer(Reaction ~ Days + (1 | Subject), data = sleepstudy)
  )
  nd <- sleepstudy[1:10, ]

  # Single-row newdata
  expect_snapshot_value(
    predictInterval(m_slope, newdata = sleepstudy[1, , drop = FALSE],
                    seed = SEED, n.sims = 500),
    style = "json2", tolerance = TOL
  )

  # ignore.fixed.terms treats a fixed term as known (zero variance)
  expect_snapshot_value(
    predictInterval(m_int, newdata = nd, seed = SEED, n.sims = NSIMS,
                    include.resid.var = FALSE,
                    ignore.fixed.terms = "(Intercept)"),
    style = "json2", tolerance = TOL
  )

  # fix.intercept.variance adjusts intercept variance for RE covariance
  expect_snapshot_value(
    predictInterval(m_int, newdata = nd, seed = SEED, n.sims = NSIMS,
                    fix.intercept.variance = TRUE),
    style = "json2", tolerance = TOL
  )
})

# NOTE: pathological / rank-deficient models (e.g.,
# `Reaction ~ 0 + Days + Days:Subject + (1 | Days)`) are deliberately NOT in
# Layer 2. `mvtnorm::rmvnorm(method = "chol")` on a near-singular covariance
# matrix gives output that differs across BLAS/LAPACK implementations by
# enough to blow past any reasonable snapshot tolerance, even on the same
# OS across R versions. Layer 2 contracts require numerical stability across
# every target runner; if the model cannot deliver that, snapshot it is
# misplaced there. Such models are covered by:
#   - Layer 1 structural invariants in test-predict.R (dim, type, ordering)
#   - Layer 3 manual regression harness at
#     tests/comparisons/predictInterval-regression.R
#
# Rule of thumb: before adding a model to this file, ask whether the design
# matrix is rank-deficient or near-singular. If yes, skip Layer 2.

# GLMM: binomial ------------------------------------------------------------

test_that("GLMM binomial: predictInterval output is stable", {
  snapshot_platform_ok()

  data(cbpp, package = "lme4")
  m <- suppressMessages(
    glmer(
      cbind(incidence, size - incidence) ~ period + (1 | herd),
      data = cbpp, family = binomial
    )
  )
  nd <- cbpp[1:5, ]

  expect_snapshot_value(
    predictInterval(m, newdata = nd, seed = SEED, n.sims = NSIMS,
                    include.resid.var = FALSE, type = "linear.prediction"),
    style = "json2", tolerance = TOL
  )
  expect_snapshot_value(
    predictInterval(m, newdata = nd, seed = SEED, n.sims = NSIMS,
                    include.resid.var = FALSE, type = "probability"),
    style = "json2", tolerance = TOL
  )
  expect_snapshot_value(
    suppressWarnings(predictInterval(
      m, newdata = nd, seed = SEED, n.sims = NSIMS,
      include.resid.var = TRUE, type = "linear.prediction"
    )),
    style = "json2", tolerance = TOL
  )
  expect_snapshot_value(
    suppressWarnings(predictInterval(
      m, newdata = nd, seed = SEED, n.sims = NSIMS,
      include.resid.var = TRUE, type = "probability"
    )),
    style = "json2", tolerance = TOL
  )
})

# Special cases - rank deficiency----

test_that("Prediction intervals are accurate with interaction terms and rank deficiency", {
  skip_on_ci()
  skip_on_cran()
  set.seed(54656)
  n <- 20
  x <- y <- rnorm(n)
  z <- rnorm(n)
  r <- sample(1:5, size=n, replace=TRUE)
  d <- data.frame(x,y,z,r)
  d2 <- expand.grid(a=factor(1:4),b=factor(1:4),rep=1:10)
  n <- nrow(d2)
  d2 <- transform(d2,r=sample(1:5, size=n, replace=TRUE),
                  z=rnorm(n))
  d2 <- subset(d2,!(a=="4" & b=="4"))

  fm <- lmer( z ~ a*b + (1|r), data=d2) |> suppressMessages()
  expect_s3_class(predictInterval(fm, newdata = d2[1:10, ]), "data.frame")

  newPred <- predictInterval(fm, newdata = d2, level = 0.8, n.sims = 1500,
                             stat = 'median', include.resid.var = FALSE,
                             seed = 2342)
  truPred <- predict(fm, newdata = d2)
  expect_equal(mean(newPred$fit - truPred), 0, tolerance = sd(truPred)/15)
  fm2 <- lmer( z ~ a*b + (1+b|r), data=d2) |> suppressMessages()
  #In the call below we are getting warnings because our call to mvtnorm::rmvnorm
  #is shotting a warning when mean and sigma of multivariate distribution are
  #zero using the method="chol
  newPred <- suppressWarnings(predictInterval(fm2, newdata = d2, level = 0.8,
                                              n.sims = 1000,
                                              stat = 'median', include.resid.var = FALSE))
  truPred <- predict(fm2, newdata = d2)
  expect_s3_class(newPred, "data.frame")
  expect_equal(mean(newPred$fit - truPred), 0, tolerance = sd(truPred)/10)
})

# Test the simResults----


test_that("simResults option behaves", {
  skip_on_cran()
  m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
  preds1 <- predictInterval(m1, newdata = sleepstudy[1:5, ])
  preds2 <- predictInterval(m1, newdata = sleepstudy[1:5, ],
                            returnSims = TRUE)
  expect_null(attr(preds1, "sim.results"))
  expect_true(is.matrix(attr(preds2, "sim.results")))
  out <- attr(preds2, "sim.results")
  expect_equal(ncol(out), 1000)
  expect_equal(nrow(out), 5)
  m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
  preds1 <- predictInterval(m1, newdata = sleepstudy[1:5, ],
                            returnSims = TRUE,
                            which = "random", seed = 23151)
  preds2 <- predictInterval(m1, newdata = sleepstudy[1:5, ], which = "fixed",
                            returnSims = TRUE, seed = 23151)
  preds3 <- predictInterval(m1, newdata = sleepstudy[1:5, ], which = "all",
                            returnSims = TRUE, seed = 23151)
  preds4 <- predictInterval(m1, newdata = sleepstudy[1:5, ],
                            returnSims = TRUE, seed = 23151)
  preds1b <- predictInterval(m1, newdata = sleepstudy[1:5, ],
                             returnSims = TRUE,
                             which = "random", seed = 23151,
                             include.resid.var = FALSE)
  preds2b <- predictInterval(m1, newdata = sleepstudy[1:5, ], which = "fixed",
                             returnSims = TRUE, seed = 23151,
                             include.resid.var = FALSE)
  preds3b <- predictInterval(m1, newdata = sleepstudy[1:5, ], which = "all",
                             returnSims = TRUE, seed = 23151,
                             include.resid.var = FALSE)
  preds4b <- predictInterval(m1, newdata = sleepstudy[1:5, ],
                             returnSims = TRUE, seed = 23151,
                             include.resid.var = FALSE)
  expect_true(is.matrix(attr(preds1, "sim.results")))
  expect_gt(abs(mean(attr(preds1, "sim.results") - attr(preds2, "sim.results"))),
            200)
  expect_true(is.matrix(attr(preds2, "sim.results")))
  expect_gt(abs(mean(attr(preds4, "sim.results") - attr(preds1, "sim.results"))),
            200)
  expect_gt(abs(mean(attr(preds4, "sim.results") - attr(preds1, "sim.results"))),
            abs(mean(attr(preds1, "sim.results") - attr(preds2, "sim.results"))))
  expect_type(attr(preds3, "sim.results"), "list")
  expect_true(is.list(attr(preds3, "sim.results")))
  expect_true(is.matrix(attr(preds1b, "sim.results")))
  expect_true(is.matrix(attr(preds2b, "sim.results")))
  expect_true(is.list(attr(preds3b, "sim.results")))
  expect_true(is.matrix(attr(preds4b, "sim.results")))
  # Check that samples are wider for include.resid.var = TRUE
  expect_gt(quantile(attr(preds1, "sim.results"), probs = 0.9) - quantile(attr(preds1b, "sim.results"), probs = 0.9),
            20)
  expect_lt(quantile(attr(preds1, "sim.results"), probs = 0.1) - quantile(attr(preds1b, "sim.results"), probs = 0.1),
            -20)
  expect_gt(quantile(attr(preds2, "sim.results"), probs = 0.9) - quantile(attr(preds2b, "sim.results"), probs = 0.9),
            20)
  expect_lt(quantile(attr(preds2, "sim.results"), probs = 0.1) - quantile(attr(preds2b, "sim.results"), probs = 0.1),
            -20)
  expect_gt(quantile(attr(preds4, "sim.results"), probs = 0.9) - quantile(attr(preds4b, "sim.results"), probs = 0.9),
            15)
  expect_lt(quantile(attr(preds4, "sim.results"), probs = 0.1) - quantile(attr(preds4b, "sim.results"), probs = 0.1),
            -15)
})

# Test out of sample predictions----

test_that("predictInterval makes predictions without observed outcome", {
  skip_on_ci()
  skip_on_cran()
  possNames <- expand.grid(letters,LETTERS)
  possNames <- paste(possNames[, 1], possNames[, 2])
  newFac <- sample(possNames, 32)
  modData <- data.frame(
    y = rnorm(500),
    x = rnorm(500),
    team_name = sample(newFac, 500, replace = TRUE)
  )
  modData$y[251:500] <- rep(NA, 250)
  m0 <- lmer(y ~ x + (1|team_name), data = modData[1:250,]) |> suppressMessages()
  #In the calls below we are getting warnings because our call to mvtnorm::rmvnorm
  #is shotting a warning when mean and sigma of multivariate distribution are
  #zero using the method="chol
  testPreds1 <- suppressWarnings(predictInterval(m0, newdata = modData[, c(3, 2, 1)]))
  testPreds2 <- suppressWarnings(predictInterval(m0, newdata = modData[1:250, c(2, 3, 1)]))
  testPreds3 <- suppressWarnings(predictInterval(m0, newdata = modData[251:500,]))
  expect_s3_class(testPreds1, "data.frame")
  expect_s3_class(testPreds2, "data.frame")
  expect_s3_class(testPreds3, "data.frame")
})

# Input validation checks----



test_that("dplyr objects are successfully coerced", {
  skip_on_cran()
  set.seed(101)
  data(sleepstudy)
  m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
  predData <- sleepstudy |>
    dplyr::group_by(Subject) |>
    dplyr::summarise(Days = mean(Days))
  expect_warning(predictInterval(m1, newdata = predData),
                 regexp = "newdata is tbl_df or tbl object from dplyr package")
  #Suppress the warning that we tested for above
  preds2 <- suppressWarnings(predictInterval(m1, newdata = predData, n.sims=2000))
  expect_s3_class(preds2, "data.frame")
  predData2 <- as.data.frame(predData)
  preds1 <- predictInterval(m1, newdata = predData2, n.sims=2000)
  expect_true(sum(preds1$fit - preds2$fit) > -50 & sum(preds1$fit - preds2$fit) < 50)
})

# Model type warnings for non-binomial GLMM----


test_that("Warnings issued", {
  skip_on_cran()
  d <- expand.grid(fac1=LETTERS[1:5], grp=factor(1:10),
                   obs=1:50)
  d$y <- simulate(~fac1+(1|grp),family = poisson,
                  newdata=d,
                  newparams=list(beta=c(2,-1,3,-2,1.2), theta=c(.33)),
                  seed = 5636)[[1]] |> suppressMessages()
  g1 <- glmer(y~fac1+(1|grp), data=d, family = 'poisson')
  expect_warning(predictInterval(g1, newdata = d[1:100,]),
                 regexp = "Prediction for NLMMs or GLMMs that are not mixed")
  rm(list = ls())
})

#Prediction intervals cover for simulated problems----

test_that("Prediction intervals work for simple linear example", {
  set.seed(51315)
  skip_on_ci()
  skip_on_cran()
  d <- expand.grid(fac1=LETTERS[1:5], grp=factor(1:10),
                   obs=1:100)
  d$y <- simulate(~fac1+(1|grp),family = gaussian,
                  newdata=d,
                  newparams=list(beta=c(2,1,3,4,7), theta=c(.25),
                                 sigma = c(.23)), seed = 4548)[[1]]
  subD <- d[sample(row.names(d), 1000),]

  g1 <- lmer(y~fac1+(1|grp), data=subD)
  d$fitted <- predict(g1, d)
  #This suppresses the warning about no parallel backend registered
  outs <- suppressWarnings(
    predictInterval(g1, newdata = d, level = 0.9, n.sims = 1000,
                    seed = 468,
                    stat = 'mean', include.resid.var = TRUE)
  )
  outs <- cbind(d, outs); outs$coverage <- FALSE
  outs$coverage <- outs$fitted <= outs$upr & outs$fitted >= outs$lwr
  expect_true(all(outs$coverage))
  expect_lt(abs(mean(outs$fit - outs$fitted)), .0005)
  expect_lt(abs(mean(outs$fit - outs$y)), .01)
  rm(outs)
})


test_that("Prediction intervals work for simple GLMM example", {
  skip_on_ci()
  skip_on_cran()
  set.seed(101)
  d <- expand.grid(fac1=LETTERS[1:5], grp=factor(1:10),
                   obs=1:50)
  d$y <- simulate(~fac1+(1|grp),family = binomial,
                  newdata=d,
                  newparams=list(beta=c(2,-1,3,-2,1.2), theta=c(.33)),
                  seed =634)[[1]]
  subD <- d[sample(row.names(d), 1200),]

  g1 <- glmer(y~fac1+(1|grp), data=subD, family = 'binomial')
  d$fitted <- predict(g1, d)
  outs <- predictInterval(g1, newdata = d, level = 0.9, n.sims = 500,
                          stat = 'mean', include.resid.var = TRUE,
                          type = 'linear.prediction', seed = 4563)
  outs <- cbind(d, outs); outs$coverage <- FALSE
  outs$coverage <- outs$fitted <= outs$upr & outs$fitted >= outs$lwr
  expect_true(all(outs$coverage))
  expect_lt(abs(mean(outs$fit - outs$fitted)), .1)
  expect_lt(abs(mean(outs$fit - outs$y)), 2)

  outs2 <- predictInterval(g1, newdata = d, level = 0.95, n.sims = 500,
                           stat = 'mean', include.resid.var = FALSE,
                           type = 'probability')
  expect_false(identical(outs, outs2))
  expect_true(max(outs2$fit) <= 1)
  expect_true(min(outs2$fit) >= 0)
  expect_true(max(outs2$lwr) <= 1)
  expect_true(min(outs2$lwr) >= 0)
  expect_true(max(outs2$upr) <= 1)
  expect_true(min(outs2$upr) >= 0)
  expect_false(max(outs$fit) <= 1)
  # expect_true(min(outs$fit) < 0)
  expect_false(max(outs$lwr) <= 1)
  expect_false(min(outs$lwr) >= 0)
  expect_false(max(outs$upr) <= 1)
  rm(outs)
})

test_that("Prediction interval respects user input", {
  skip_on_ci()
  skip_on_cran()
  set.seed(101)
  d <- expand.grid(fac1=LETTERS[1:5], grp=factor(1:10),
                   obs=1:25)
  d$y <- simulate(~fac1+(1|grp),family = gaussian,
                  newdata=d,
                  newparams=list(beta=c(2,1,3,4,7), theta=c(.25),
                                 sigma = c(.23)), seed =463)[[1]]
  subD <- d[sample(row.names(d), 1000),]

  g1 <- lmer(y~fac1+(1|grp), data=subD)
  d$fitted <- predict(g1, d)
  outs1 <- predictInterval(g1, newdata = d, level = 0.8, n.sims = 500,
                           stat = 'mean', include.resid.var = TRUE, seed=643)
  outs2 <- predictInterval(g1, newdata = d, level = 0.95, n.sims = 500,
                           stat = 'mean', include.resid.var = TRUE, seed=643)
  outs1a <- predictInterval(g1, newdata = d, level = 0.8, n.sims = 1500,
                            stat = 'mean', include.resid.var = TRUE, seed=643)
  outs2a <- predictInterval(g1, newdata = d, level = 0.95, n.sims = 1500,
                            stat = 'mean', include.resid.var = TRUE, seed=643)
  outs3 <- predictInterval(g1, newdata = d, level = 0.8, n.sims = 500,
                           stat = 'mean', include.resid.var = FALSE, seed=643)
  outs3b <- predictInterval(g1, newdata = d, level = 0.8, n.sims = 500,
                            stat = 'median', include.resid.var = FALSE, seed=643)
  outs3c <- predictInterval(g1, newdata = d[1, ], level = 0.8, n.sims = 500,
                            stat = 'median', include.resid.var = FALSE, seed=643)

  expect_gt(median(outs2$upr - outs1$upr), 0.1)
  expect_gt(median(outs2a$upr - outs1a$upr), 0.1)
  expect_lt(median(outs3$upr - outs1$upr), -.2)
  expect_lt(median(outs3b$upr - outs1a$upr), -.2)
  expect_lt(mean(outs1$upr - outs1$lwr), mean(outs2$upr - outs2$lwr))
  expect_lt(mean(outs1$upr - outs1$lwr), mean(outs1a$upr - outs1a$lwr))
  expect_lt(mean(outs2$upr - outs2$lwr), mean(outs2a$upr - outs2a$lwr))
  expect_false(median(outs3$fit) == median(outs3b$fit))
  expect_equal(nrow(outs3c), 1)
})

# Prediction works for all combinations of slopes and intercepts----
#context("Prediction works for all combinations of slopes and intercepts")

test_that("Predict handles unused and subset of factor levels", {
  skip_on_cran()
  skip_on_ci()
  set.seed(101)
  moddf <- InstEval[sample(rownames(InstEval), 10000), ]
  g1 <- lmer(y ~ lectage + studage + (1|d) + (1|s), data=moddf)
  d1 <- InstEval[1:100, ]
  outs1 <- suppressWarnings(predictInterval(g1, newdata = d1, level = 0.8, n.sims = 500,
                                            stat = 'mean', include.resid.var = TRUE, seed = 4632))
  d2 <- rbind(d1, InstEval[670:900,])
  outs1a <- suppressWarnings(predictInterval(g1, newdata = d2, level = 0.8, n.sims = 500,
                                             stat = 'mean', include.resid.var=TRUE, seed = 4632)[1:100,])
  expect_s3_class(outs1, "data.frame")
  expect_s3_class(outs1a, "data.frame")
  expect_equal(nrow(outs1), 100)
  expect_equal(nrow(outs1a), 100)

  g2 <- lmer(y ~ lectage + studage + (1+lectage|d) + (1|dept), data=moddf) |>
    suppresWarnings()

  d2 <- InstEval[670:900,]
  outs1a <- suppressWarnings(predictInterval(g2, newdata = d2, level = 0.8, n.sims = 500,
                                             stat = 'mean', include.resid.var=TRUE, seed = 4632))
  expect_s3_class(outs1a, "data.frame")
  expect_equal(nrow(outs1a), 231)
})

rm(list = ls())

# Regression tests for #124:
# https://github.com/jknowles/merTools/issues/124
#
# For a nested model `y ~ 1 + (1|a/b)`, a prediction frame that mixes observed
# and unobserved levels of the *interaction* grouping factor (`a:b`) used to
# produce non-reproducible results. The mapping from the model matrix back to a
# grouping level used `max.col()`, which (a) returns a column even for an
# all-zero row (an unobserved interaction level) and (b) breaks the resulting
# tie *at random*. So an unobserved `a:b` level silently borrowed a randomly
# chosen observed level's random effect, making the point estimate depend on the
# RNG seed and differ between batch vs. row-by-row prediction.
#
# The documented, correct behavior is that an unobserved grouping level
# contributes zero from that random-effect term (predictions fall back to the
# fixed effects plus any *observed* higher-level random effects, plus residual
# variation). These tests pin that behavior.

make_nested_fit <- function() {
  training_df <- data.frame(
    class = c("c1", "c1", "c1", "c1", "c1", "c1", "c1", "c1", "c2"),
    pupil = c("p1", "p1", "p1", "p2", "p3", "p3", "p4", "p5", "p7"),
    grade = c(4.5, 4.8, 4.7, 5.0, 4.1, 4.0, 3.5, 5.5, 5.5)
  )
  suppressMessages(suppressWarnings(
    lme4::lmer(grade ~ 1 + (1 | class / pupil), data = training_df)
  ))
}

test_that("predictInterval is reproducible for partially-observed nested levels (#124)", {
  model <- make_nested_fit()
  # Row 3 (class = c1, pupil = p6) is unobserved at the class:pupil level.
  query_df <- data.frame(
    class = c("c1", "c1", "c1"),
    pupil = c("p1", "p2", "p6")
  )

  s1 <- suppressWarnings(
    predictInterval(model, query_df, seed = 1, n.sims = 4000)
  )
  s2 <- suppressWarnings(
    predictInterval(model, query_df, seed = 2, n.sims = 4000)
  )

  # The unobserved-level row's point estimate must be stable across seeds.
  # Before the fix this swung by ~0.3 (random column selection); after the fix
  # it varies only by ordinary Monte Carlo noise (< 0.1 here).
  expect_lt(abs(s1$fit[3] - s2$fit[3]), 0.1)

  # The genuinely observed rows were always stable; confirm they still are.
  expect_lt(abs(s1$fit[1] - s2$fit[1]), 0.1)
  expect_lt(abs(s1$fit[2] - s2$fit[2]), 0.1)
})

test_that("unobserved nested level falls back to fixed + observed higher-level RE (#124)", {
  model <- make_nested_fit()
  query_df <- data.frame(
    class = c("c1", "c1", "c1"),
    pupil = c("p1", "p2", "p6")
  )

  pred <- suppressWarnings(
    predictInterval(model, query_df, seed = 1, n.sims = 4000)
  )

  # For an unobserved class:pupil level, the class:pupil contribution is zero,
  # so the fitted value should equal the intercept plus the (observed) class
  # random effect for c1 -- it must NOT borrow a random pupil effect.
  expected_fit <- unname(
    lme4::fixef(model)[["(Intercept)"]] +
      lme4::ranef(model)$class["c1", "(Intercept)"]
  )
  expect_equal(pred$fit[3], expected_fit, tolerance = 0.1)
})

test_that("nested-level prediction is consistent batch vs. row-by-row (#124)", {
  model <- make_nested_fit()
  query_df <- data.frame(
    class = c("c1", "c1", "c1"),
    pupil = c("p1", "p2", "p6")
  )

  batch <- suppressWarnings(
    predictInterval(model, query_df, seed = 1, n.sims = 4000)
  )
  alone <- suppressWarnings(
    predictInterval(model, query_df[3, , drop = FALSE], seed = 1, n.sims = 4000)
  )

  # The unobserved-level row should yield essentially the same estimate whether
  # predicted in a batch with observed rows or on its own.
  expect_lt(abs(batch$fit[3] - alone$fit[1]), 0.1)
})

# Tests for the new.levels argument of predictInterval().
#
# For a grouping level not present in the fitted model, the historical (and
# default) behavior is new.levels = "zero": the random effect is dropped and the
# prediction rests on the fixed effects plus residual variation. new.levels =
# "draw" instead samples that group's effect from the estimated random-effect
# covariance (VarCorr), so the interval reflects between-group uncertainty --
# the analogue of brms::posterior_predict(allow_new_levels = TRUE).

make_slope_fit <- function() {
  suppressMessages(suppressWarnings(
    lme4::lmer(Reaction ~ Days + (Days | Subject), data = lme4::sleepstudy)
  ))
}

test_that("new.levels defaults to 'zero' (backward compatible)", {
  m <- make_slope_fit()
  nd <- data.frame(Days = 0:4, Subject = "NEW_SUBJECT")
  def  <- suppressWarnings(predictInterval(m, nd, level = 0.9, n.sims = 2000, seed = 1))
  zero <- suppressWarnings(predictInterval(m, nd, level = 0.9, n.sims = 2000, seed = 1,
                                           new.levels = "zero"))
  expect_identical(def, zero)
})

test_that("new.levels='draw' is wider than 'zero' for an unobserved group", {
  m <- make_slope_fit()
  nd <- data.frame(Days = 0:4, Subject = "NEW_SUBJECT")
  zero <- suppressWarnings(predictInterval(m, nd, level = 0.9, n.sims = 4000, seed = 1,
                                           new.levels = "zero"))
  draw <- suppressWarnings(predictInterval(m, nd, level = 0.9, n.sims = 4000, seed = 1,
                                           new.levels = "draw"))
  expect_gt(mean(draw$upr - draw$lwr), mean(zero$upr - zero$lwr))
})

test_that("new.levels='draw' width matches the theoretical predictive SD", {
  m <- make_slope_fit()
  nd <- data.frame(Days = 0:4, Subject = "NEW_SUBJECT")
  draw <- suppressWarnings(predictInterval(m, nd, level = 0.9, n.sims = 8000, seed = 1,
                                           new.levels = "draw"))
  Sigma <- as.matrix(lme4::VarCorr(m)$Subject)        # RE covariance (intercept, Days)
  sig2  <- sigma(m)^2
  z <- qnorm(0.95)
  # for each Days value, total predictive variance = z' Sigma z + residual
  expected_width <- vapply(nd$Days, function(d) {
    zvec <- c(1, d)
    2 * z * sqrt(as.numeric(t(zvec) %*% Sigma %*% zvec) + sig2)
  }, numeric(1))
  observed_width <- draw$upr - draw$lwr
  # Monte Carlo, so allow a modest relative tolerance
  expect_equal(observed_width, expected_width, tolerance = 0.08)
})

test_that("observations sharing a new level share the sampled effect", {
  m <- make_slope_fit()
  # two rows, same new subject, same Days -> identical sampled random effect.
  # Exclude residual variance so only the (shared) RE component remains; with
  # residual noise on, each row gets an independent residual draw and they would
  # differ even though the underlying RE draw is shared.
  nd <- data.frame(Days = c(3, 3), Subject = c("NEW_A", "NEW_A"))
  pi <- suppressWarnings(predictInterval(m, nd, which = "random", level = 0.9,
                                         n.sims = 1000, seed = 1, new.levels = "draw",
                                         include.resid.var = FALSE, returnSims = TRUE))
  sims <- attr(pi, "sim.results")
  if (nrow(sims) == nrow(nd)) sims <- t(sims)
  # same new level + same covariates -> the two rows' draws coincide
  expect_equal(sims[, 1], sims[, 2])
})

# -----------------------------------------------------------------------------
# Layer 1 structural invariants for predictInterval().
#
# These are properties that MUST hold for every valid predictInterval() output,
# regardless of seed, simulation noise, platform, model type, or option
# combination. If any of these fail, the function is broken — not the
# tolerance and not the test.
#
# These tests run on every CI platform (no skip_on_os). They complement:
#
#   - test-predictInterval-snapshot.R  (Layer 2, Linux-only, pinned numerics)
#   - tests/comparisons/*              (Layer 3, manual regression harness)
#
# Adding a new invariant: derive the property from the function's contract
# (documentation, mathematical definition, or design intent). Do NOT snapshot
# numeric output here; that belongs in Layer 2. Invariants should be true by
# definition — e.g., "upr >= fit" is true by quantile ordering.
# -----------------------------------------------------------------------------

SEED <- 11213

# Build a small, well-conditioned fixture set covering the main algorithmic
# branches: random-slope LMM, random-intercept LMM, binomial GLMM.
make_fixtures <- function() {
  data(sleepstudy, package = "lme4")
  data(cbpp,       package = "lme4")

  list(
    lmm_slope = list(
      fit = suppressMessages(
        lmer(Reaction ~ Days + (Days | Subject), data = sleepstudy)
      ),
      newdata = sleepstudy[1:20, ]
    ),
    lmm_int = list(
      fit = suppressMessages(
        lmer(Reaction ~ Days + (1 | Subject), data = sleepstudy)
      ),
      newdata = sleepstudy[1:20, ]
    ),
    glmm_bin = list(
      fit = suppressMessages(
        glmer(cbind(incidence, size - incidence) ~ period + (1 | herd),
              data = cbpp, family = binomial)
      ),
      newdata = cbpp[1:10, ]
    )
  )
}

# --------------------------------------------------------------------------
# Invariant 1: output is a data.frame with the expected column names.
# --------------------------------------------------------------------------

test_that("output is a data.frame with columns fit, upr, lwr", {
  skip_on_cran()
  fx <- make_fixtures()
  for (nm in names(fx)) {
    out <- suppressWarnings(
      predictInterval(fx[[nm]]$fit, newdata = fx[[nm]]$newdata,
                      seed = SEED, n.sims = 200)
    )
    expect_true(is.data.frame(out), info = nm)
    expect_identical(names(out), c("fit", "upr", "lwr"), info = nm)
  }
})

# --------------------------------------------------------------------------
# Invariant 2: interval ordering — upr >= fit >= lwr for every row.
# Ties are acceptable; strict inequality is not guaranteed (e.g., single-
# level factor with zero variance contribution).
# --------------------------------------------------------------------------

test_that("upr >= fit >= lwr for every row", {
  skip_on_cran()
  fx <- make_fixtures()
  for (nm in names(fx)) {
    out <- suppressWarnings(
      predictInterval(fx[[nm]]$fit, newdata = fx[[nm]]$newdata,
                      seed = SEED, n.sims = 200)
    )
    expect_true(all(out$upr >= out$fit), info = nm)
    expect_true(all(out$fit >= out$lwr), info = nm)
  }
})

# --------------------------------------------------------------------------
# Invariant 3: row-wise correspondence with newdata for which != "all".
# --------------------------------------------------------------------------

test_that("nrow(output) matches nrow(newdata) for which = full/fixed/random", {
  skip_on_cran()
  fx <- make_fixtures()
  for (nm in names(fx)) {
    nd <- fx[[nm]]$newdata
    for (w in c("full", "fixed", "random")) {
      out <- suppressWarnings(
        predictInterval(fx[[nm]]$fit, newdata = nd,
                        seed = SEED, n.sims = 200, which = w)
      )
      expect_equal(nrow(out), nrow(nd),
                   info = paste0(nm, " / which = ", w))
    }
  }
})

# --------------------------------------------------------------------------
# Invariant 4: no NAs, NaNs, or Infs in output for well-conditioned inputs.
# --------------------------------------------------------------------------

test_that("no NA / NaN / Inf in output for well-conditioned inputs", {
  skip_on_cran()
  fx <- make_fixtures()
  for (nm in names(fx)) {
    out <- suppressWarnings(
      predictInterval(fx[[nm]]$fit, newdata = fx[[nm]]$newdata,
                      seed = SEED, n.sims = 200)
    )
    expect_true(all(is.finite(out$fit)), info = nm)
    expect_true(all(is.finite(out$upr)), info = nm)
    expect_true(all(is.finite(out$lwr)), info = nm)
  }
})

# --------------------------------------------------------------------------
# Invariant 5: seed reproducibility — same seed produces bit-identical
# output, different seeds produce different output.
# --------------------------------------------------------------------------

test_that("same seed produces identical output; different seeds do not", {
  skip_on_cran()
  fx <- make_fixtures()
  for (nm in names(fx)) {
    a <- suppressWarnings(
      predictInterval(fx[[nm]]$fit, newdata = fx[[nm]]$newdata,
                      seed = SEED, n.sims = 200)
    )
    b <- suppressWarnings(
      predictInterval(fx[[nm]]$fit, newdata = fx[[nm]]$newdata,
                      seed = SEED, n.sims = 200)
    )
    c <- suppressWarnings(
      predictInterval(fx[[nm]]$fit, newdata = fx[[nm]]$newdata,
                      seed = SEED + 1L, n.sims = 200)
    )
    expect_identical(a, b, info = paste0(nm, " / same seed"))
    expect_false(identical(a, c), info = paste0(nm, " / different seed"))
  }
})

# --------------------------------------------------------------------------
# Invariant 6: quantile monotonicity — a higher `level` produces intervals
# that contain the intervals at any lower `level`, since both are computed
# from the same simulation draws.
# --------------------------------------------------------------------------

test_that("higher level produces wider-or-equal intervals on same draws", {
  skip_on_cran()
  fx <- make_fixtures()
  for (nm in names(fx)) {
    i80 <- suppressWarnings(
      predictInterval(fx[[nm]]$fit, newdata = fx[[nm]]$newdata,
                      seed = SEED, n.sims = 500, level = 0.80)
    )
    i99 <- suppressWarnings(
      predictInterval(fx[[nm]]$fit, newdata = fx[[nm]]$newdata,
                      seed = SEED, n.sims = 500, level = 0.99)
    )
    expect_true(all(i99$upr >= i80$upr), info = paste0(nm, " / upr"))
    expect_true(all(i99$lwr <= i80$lwr), info = paste0(nm, " / lwr"))
  }
})

# --------------------------------------------------------------------------
# Invariant 7: GLMM probability predictions must lie in [0, 1].
# --------------------------------------------------------------------------

test_that("GLMM type='probability' output is in [0, 1]", {
  skip_on_cran()
  fx <- make_fixtures()
  out <- suppressWarnings(
    predictInterval(fx$glmm_bin$fit, newdata = fx$glmm_bin$newdata,
                    seed = SEED, n.sims = 500,
                    type = "probability", include.resid.var = FALSE)
  )
  expect_true(all(out$fit >= 0 & out$fit <= 1))
  expect_true(all(out$upr >= 0 & out$upr <= 1))
  expect_true(all(out$lwr >= 0 & out$lwr <= 1))
})

# --------------------------------------------------------------------------
# Regression test: GitHub issue #101
# predictInterval() must not crash on no-intercept random-slope models.
# --------------------------------------------------------------------------

test_that("no-intercept random-slope models do not crash (#101)", {
  skip_on_cran()
  data(sleepstudy, package = "lme4")
  set.seed(SEED)
  sleepstudy$Test <- rep(
    sample(c(TRUE, FALSE), length(unique(sleepstudy$Subject)), replace = TRUE),
    each = 10
  )
  fit <- suppressMessages(
    lmer(Reaction ~ Days:Test + (0 + Days | Subject), data = sleepstudy)
  )
  out <- suppressMessages(
    predictInterval(fit, newdata = head(sleepstudy), n.sims = 200, seed = SEED)
  )
  expect_s3_class(out, "data.frame")
  expect_equal(nrow(out), 6L)
  expect_named(out, c("fit", "upr", "lwr"))
  expect_true(all(out$upr >= out$fit))
  expect_true(all(out$fit >= out$lwr))
})

# -----------------------------------------------------------------------------
# Regression tests for issue #118:
# predictInterval() fails on models with multiple random-effect term blocks
# for the same grouping factor (double-bar syntax, explicit splits, mixed
# correlated + uncorrelated). Root cause: lme4 returns `postVar` as a list
# of arrays in those cases, and the indexing code assumed a 3-D array.
# Fixed by combine_postvar_blocks() which normalizes to a block-diagonal
# array before the level-filtering and rmvnorm path runs.
#
# Structural invariants only — numeric output for uncorrelated specs is
# sensitive to lme4's internal parameterization and is not pinned here.
# -----------------------------------------------------------------------------

SEED <- 11213
N_SIMS <- 500

make_sleep_fixture <- function() {
  data(sleepstudy, package = "lme4")
  set.seed(SEED)
  sleepstudy$av <- rnorm(nrow(sleepstudy))
  sleepstudy
}

expect_valid_predictInterval <- function(pi, n_rows) {
  expect_s3_class(pi, "data.frame")
  expect_equal(nrow(pi), n_rows)
  expect_named(pi, c("fit", "upr", "lwr"))
  expect_true(all(is.finite(unlist(pi))))
  expect_true(all(pi$upr >= pi$fit))
  expect_true(all(pi$fit >= pi$lwr))
}

# 1. Exact reproducer from issue #118 ------------------------------------------
test_that("predictInterval works with double-bar uncorrelated random slopes (#118)", {
  skip_on_cran()
  sleep <- make_sleep_fixture()
  fm <- suppressMessages(suppressWarnings(
    lme4::lmer(Reaction ~ Days + (Days + av || Subject), data = sleep)
  ))
  pi <- suppressWarnings(
    predictInterval(fm, newdata = sleep, n.sims = N_SIMS, seed = SEED)
  )
  expect_valid_predictInterval(pi, nrow(sleep))
})

# 2. Double-bar with intercept -------------------------------------------------
test_that("predictInterval works with (1 + Days || Subject)", {
  skip_on_cran()
  data(sleepstudy, package = "lme4")
  fm <- suppressMessages(suppressWarnings(
    lme4::lmer(Reaction ~ Days + (1 + Days || Subject), data = sleepstudy)
  ))
  pi <- suppressWarnings(
    predictInterval(fm, newdata = sleepstudy, n.sims = N_SIMS, seed = SEED)
  )
  expect_valid_predictInterval(pi, nrow(sleepstudy))
})

# 3. Explicit split equivalent -------------------------------------------------
test_that("predictInterval works with explicit split (1|Subject) + (0 + Days|Subject)", {
  skip_on_cran()
  data(sleepstudy, package = "lme4")
  fm <- suppressMessages(suppressWarnings(
    lme4::lmer(
      Reaction ~ Days + (1 | Subject) + (0 + Days | Subject),
      data = sleepstudy
    )
  ))
  pi <- suppressWarnings(
    predictInterval(fm, newdata = sleepstudy, n.sims = N_SIMS, seed = SEED)
  )
  expect_valid_predictInterval(pi, nrow(sleepstudy))
})

# 4. Mixed correlated + uncorrelated -------------------------------------------
test_that("predictInterval works with mixed (Days|Subject) + (0 + av|Subject)", {
  skip_on_cran()
  sleep <- make_sleep_fixture()
  fm <- suppressMessages(suppressWarnings(
    lme4::lmer(
      Reaction ~ Days + (Days | Subject) + (0 + av | Subject),
      data = sleep
    )
  ))
  pi <- suppressWarnings(
    predictInterval(fm, newdata = sleep, n.sims = N_SIMS, seed = SEED)
  )
  expect_valid_predictInterval(pi, nrow(sleep))
})

# 5. GLMM with uncorrelated random slopes --------------------------------------
test_that("predictInterval works on GLMM with uncorrelated random slopes", {
  skip_on_cran()
  data(cbpp, package = "lme4")
  cbpp$period_num <- as.numeric(cbpp$period)
  gm <- suppressMessages(suppressWarnings(
    lme4::glmer(
      cbind(incidence, size - incidence) ~ period + (1 + period_num || herd),
      data = cbpp, family = binomial
    )
  ))
  pi <- suppressWarnings(
    predictInterval(
      gm, newdata = cbpp, n.sims = N_SIMS, seed = SEED,
      type = "linear.prediction", include.resid.var = FALSE
    )
  )
  expect_valid_predictInterval(pi, nrow(cbpp))
})

# 6. Correlated-only case (regression guard) -----------------------------------
test_that("correlated-only random slopes still work (regression guard)", {
  skip_on_cran()
  data(sleepstudy, package = "lme4")
  fm <- suppressMessages(suppressWarnings(
    lme4::lmer(Reaction ~ Days + (Days | Subject), data = sleepstudy)
  ))
  pi <- suppressWarnings(
    predictInterval(fm, newdata = sleepstudy, n.sims = N_SIMS, seed = SEED)
  )
  expect_valid_predictInterval(pi, nrow(sleepstudy))
})

# 7. Unit test on combine_postvar_blocks() -------------------------------------
test_that("combine_postvar_blocks builds correct block-diagonal array", {
  # Two blocks: a 2x2 with level-specific covariance, and a 1x1 scalar.
  # Three levels. Fill with distinctive values so block-diagonal structure
  # is verifiable element-wise.
  n_levels <- 3L
  block1 <- array(0, dim = c(2L, 2L, n_levels))
  block2 <- array(0, dim = c(1L, 1L, n_levels))
  for (k in seq_len(n_levels)) {
    block1[, , k] <- matrix(c(k, 0.1 * k, 0.1 * k, 2 * k), 2L, 2L)
    block2[, , k] <- matrix(10 * k, 1L, 1L)
  }

  # Mock reMeans: 3 levels x 3 columns, names carried to dimnames of output.
  reMeans <- matrix(0, nrow = n_levels, ncol = 3L,
                    dimnames = list(c("L1", "L2", "L3"),
                                    c("(Intercept)", "x", "y")))

  out <- combine_postvar_blocks(list(block1, block2), reMeans)

  expect_true(is.array(out))
  expect_equal(dim(out), c(3L, 3L, 3L))
  expect_equal(dimnames(out)[[1]], c("(Intercept)", "x", "y"))
  expect_equal(dimnames(out)[[3]], c("L1", "L2", "L3"))

  for (k in seq_len(n_levels)) {
    # Block 1 occupies rows/cols 1:2
    expect_equal(out[1:2, 1:2, k], block1[, , k], ignore_attr = TRUE)
    # Block 2 occupies row/col 3
    expect_equal(out[3, 3, k], 10 * k, ignore_attr = TRUE)
    # Off-diagonal entries between blocks are zero
    expect_equal(out[1:2, 3, k], c(0, 0), ignore_attr = TRUE)
    expect_equal(out[3, 1:2, k], c(0, 0), ignore_attr = TRUE)
  }
})

# 8. Helper raises informative error on malformed input ------------------------
test_that("combine_postvar_blocks stops when block dims don't sum to ncol(reMeans)", {
  block <- array(0, dim = c(1L, 1L, 2L))
  reMeans <- matrix(0, nrow = 2L, ncol = 3L,
                    dimnames = list(c("L1", "L2"), c("a", "b", "c")))
  expect_error(
    combine_postvar_blocks(list(block), reMeans),
    "Unexpected lme4 postVar structure"
  )
})

# Pin RNG algorithm and version across R releases so that results generated
# with set.seed() are reproducible regardless of which R version runs the
# tests. Without this, R-release vs R-devel vs R-oldrel-1 on CI can drift on
# the same seed because default RNG defaults have shifted across R versions
# (e.g., sample.kind changed in R 3.6).
RNGversion("4.1.0")
RNGkind(kind = "Mersenne-Twister",
        normal.kind = "Inversion",
        sample.kind = "Rejection")

#' Simulate residual variance for merMod objects
#'
#' Internal helper that reproduces the sigma-drawing logic from the original
#' `predictInterval()` implementation. It returns a numeric vector of length
#' `n.sims` containing the simulated residual standard deviations.
#'
#' @param merMod A merMod object from lme4.
#' @param n.sims Number of simulation draws.
#' @return Numeric vector of length `n.sims`.
#' @keywords internal
simulate_residual_variance <- function(merMod, n.sims) {
  devcomp <- getME(merMod, 'devcomp')
  
  # GLMM: No residual variance to simulate (dispersion = 1 by definition)
  if (devcomp$dims[["GLMM"]] == TRUE) {
    return(NULL)
  }
  
  # NLMM: No residual variance either
  if (devcomp$dims[["NLMM"]] == TRUE) {
    return(NULL)
  }
  
  # LMM: Gamma-based sigma simulation
  if (devcomp$dims[["GLMM"]] == 0 && devcomp$dims[["NLMM"]] == 0) {
    sigma <- sqrt(
      1 / rgamma(
        n.sims,
        shape = 0.5 * residDF.merMod(merMod),
        rate = 0.5 * devcomp$cmp[["pwrss"]]
      )
    )
    return(sigma)
  }
  
  stop("Unknown model type in simulate_residual_variance")
}

#' Simulate fixed-effect predictions
#'
#' Draws `n.sims` samples from the multivariate normal distribution defined by
#' the fixed-effect estimates and their variance-covariance matrix, then
#' multiplies by the model matrix to produce predictions. It also respects
#' `ignore.fixed.terms` and optional intercept variance adjustment.
#'
#' @param merMod A merMod object.
#' @param newdata Data frame containing the prediction covariates.
#' @param n.sims Number of simulations.
#' @param ignore.fixed.terms Vector of fixed-effect names or indices to treat as known.
#' @param fix.intercept.variance Logical flag for intercept variance correction.
#' @param .parallel Logical flag to enable parallel execution via foreach.
#' @return Matrix of dimensions `nrow(newdata)` x `n.sims` containing simulated
#'   fixed-effect predictions.
#' @keywords internal
simulate_fixed_effects <- function(
  merMod,
  newdata,
  n.sims,
  ignore.fixed.terms = NULL,
  fix.intercept.variance = FALSE,
  .parallel = FALSE,
  seed = NULL
) {
  if (!is.null(seed)) set.seed(seed)
  else if (!exists('.Random.seed', envir = .GlobalEnv)) runif(1)
  fe.tmp <- fixef(merMod)
  vcov.tmp <- as.matrix(vcov(merMod))

  # Adjust intercept variance if requested (mirrors original logic)
  if (fix.intercept.variance) {
    rr <- ranef(merMod, condVar = TRUE)
    intercept.var <- vcov.tmp[1, 1]
    groupsizes <- ngrps(merMod)
    for (j in names(groupsizes)) {
      reMatrix <- attr(rr[[j]], which = 'postVar')
      groupVar <- (attr(VarCorr(merMod)[[j]], 'stddev')['(Intercept)'])^2
      extraPrec <- 0
      for (eff in seq_len(dim(reMatrix)[3])) {
        term <- 1 / (reMatrix[1, 1, eff] + groupVar)
        if (term > 0) {
          extraPrec <- extraPrec + term
        } else {
          warning("fix.intercept.variance got negative precision; better turn it off.")
        }
      }
      intercept.var <- intercept.var - 1 / extraPrec
    }
    if (intercept.var < 0) {
      warning("fix.intercept.variance got negative variance; better turn it off.")
    }
    ratio <- intercept.var / vcov.tmp[1, 1]
    prec.tmp <- solve(vcov.tmp)
    prec.tmp[1, 1] <- prec.tmp[1, 1] / ratio
    vcov.tmp[1, ] <- vcov.tmp[1, ] * ratio
    vcov.tmp <- solve(prec.tmp, tol = 1e-50)
  }

  # Inflate precision for ignored fixed terms (effectively zero variance)
  if (!is.null(ignore.fixed.terms)) {
    prec.tmp <- solve(vcov.tmp)
    for (term in ignore.fixed.terms) {
      prec.tmp[term, term] <- prec.tmp[term, term] * 1e15
    }
    vcov.tmp <- solve(prec.tmp, tol = 1e-50)
  }

  # Simulate coefficients - use parallel for large n.sims
  if (n.sims > 2000 || .parallel) {
    if (.parallel) {
      setup_parallel()
    }
    i <- 1:n.sims
    fe_call <- as.call(c(list(quote(foreach::foreach), i = i, .combine = 'rbind')))
    fe <- eval(fe_call)
    betaSim <- foreach::`%dopar%`(
      fe,
      mvtnorm::rmvnorm(n = 1, mean = fe.tmp, sigma = vcov.tmp, method = 'chol')
    )
  } else {
    betaSim <- abind::abind(
      lapply(1:n.sims, function(x) {
        mvtnorm::rmvnorm(n = 1, mean = fe.tmp, sigma = vcov.tmp, method = 'chol')
      }),
      along = 1
    )
  }
  colnames(betaSim) <- names(fe.tmp)
  rownames(betaSim) <- 1:n.sims

  # Build model matrix for newdata
  newdata.modelMatrix <- buildModelMatrix(merMod, newdata = newdata, which = "fixed")

  # Pad betaSim if model matrix has more columns (random slopes not in fixed effects)
  if (ncol(newdata.modelMatrix) > ncol(betaSim)) {
    pad <- matrix(
      rep(0),
      nrow = nrow(betaSim),
      ncol = ncol(newdata.modelMatrix) - ncol(betaSim)
    )
    if (ncol(pad) > 0) {
      message("Fixed effect matrix has been padded with 0 coefficients
            for random slopes not included in the fixed effects and interaction terms.")
    }
    colnames(pad) <- setdiff(colnames(newdata.modelMatrix), colnames(betaSim))
    betaSim <- cbind(betaSim, pad)
  }

  # Align columns between model matrix and coefficient matrix
  keep <- intersect(colnames(newdata.modelMatrix), colnames(betaSim))
  newdata.modelMatrix <- newdata.modelMatrix[, keep, drop = FALSE]
  betaSim <- betaSim[, keep, drop = FALSE]

  # Compute predictions: X %*% t(beta) -> (N x n.sims)
  fixed.xb <- newdata.modelMatrix %*% t(betaSim)
  fixed.xb
}

#' Compute random effect predictions for a single grouping factor
#'
#' Internal helper that multiplies predictor data by simulated random effect
#' coefficients for each grouping level. Handles new levels not present in the
#' original model data by returning zeros for their random effect contributions.
#'
#' @param data data.frame with predictor columns and a grouping factor column.
#'   The last column must be the grouping factor.
#' @param coefs 3D array of simulated coefficients with dimensions
#'   (n.sims x terms x levels).
#' @param group character name of the grouping factor column in data.
#' @return matrix of predicted values with dimensions (nrow(data) x n.sims).
#' @keywords internal
tmp.pred <- function(data, coefs, group) {
  # Identify new levels not in the model data
  new.levels <- unique(
    as.character(data[, group])[
      !as.character(data[, group]) %in% dimnames(coefs)[[3]]
    ]
  )


  if (length(new.levels) > 0) {
    msg <- paste0(
      "     The following levels of ", group, " from newdata \n -- ",
      paste0(new.levels, collapse = ", "),
      " -- are not in the model data. \n",
      "     Currently, predictions for these values are based only on the \n",
      " fixed coefficients and the observation-level error."
    )
    warning(msg, call. = FALSE)
  }

  # Initialize output array

  yhatTmp <- array(data = NA, dim = c(nrow(data), dim(coefs)[1]))
  colIdx <- ncol(data) - 1  # Number of predictor columns (excluding group)

  # Handle coefficient dimension mismatch

  # When predictor columns exceed coefficient terms, expand coefficients
  colLL <- length(1:colIdx)
  if (colLL > dim(coefs)[2]) {
    coefs_new <- array(
      NA,
      dim = c(dim(coefs)[1], colLL, dim(coefs)[3])
    )
    dimnames(coefs_new)[c(1, 3)] <- dimnames(coefs)[c(1, 3)]
    dimnames(coefs_new)[[2]] <- rep(dimnames(coefs)[[2]], dim(coefs_new)[2])
    for (k in 1:colLL) {
      coefs_new[, k, 1:dim(coefs)[3]] <- coefs[, 1, 1:dim(coefs)[3]]
    }
    coefs <- coefs_new
  }

  # Compute predictions for each observation
  for (i in 1:nrow(data)) {
    lvl <- as.character(data[, group][i])
    if (!lvl %in% new.levels) {
      # Known level: multiply predictors by coefficients
      yhatTmp[i, ] <- as.numeric(data[i, 1:colIdx]) %*% t(coefs[, 1:colIdx, lvl])
    } else {
      # New level: zero out the random effect contribution
      yhatTmp[i, ] <- rep(0, colIdx) %*% t(coefs[, 1:colIdx, 1])
    }
  }

  rownames(yhatTmp) <- rownames(data)
  return(yhatTmp)
}

#' Combine a per-term list of postVar arrays into a block-diagonal array
#'
#' When a grouping factor has multiple random-effect term blocks — from the
#' double-bar syntax (`(x + y || g)`), explicit splits (`(1|g) + (0+x|g)`),
#' or mixed correlated + uncorrelated specs — `lme4::ranef(..., condVar =
#' TRUE)` returns `postVar` as a list of arrays, one per term block. Each
#' element has shape `d_m × d_m × n_levels`. This helper concatenates the
#' blocks into a single `d × d × n_levels` array with block-diagonal
#' structure per level (zeros off-diagonal because uncorrelated). The block
#' ordering and column ordering match `colnames(reMeans)` / lme4's `cnms`.
#' @param pv_list List of per-term postVar arrays.
#' @param reMeans The `as.matrix(ranef(m)[[j]])` matrix for the same
#'   grouping factor; used to size and name the combined array.
#' @return 3-D array of shape `ncol(reMeans) × ncol(reMeans) × nrow(reMeans)`.
#' @keywords internal
combine_postvar_blocks <- function(pv_list, reMeans) {
  n_levels <- dim(pv_list[[1L]])[3L]
  d_total  <- ncol(reMeans)
  block_dims <- vapply(pv_list, function(b) dim(b)[1L], integer(1L))
  if (sum(block_dims) != d_total) {
    stop(
      "Unexpected lme4 postVar structure: sum of block dimensions (",
      sum(block_dims), ") does not match ncol(reMeans) (", d_total,
      "). Please report this with your model specification and lme4 version.",
      call. = FALSE
    )
  }
  out <- array(0, dim = c(d_total, d_total, n_levels))
  pos <- 1L
  for (block in pv_list) {
    d_m <- dim(block)[1L]
    idx <- seq.int(pos, length.out = d_m)
    for (lvl in seq_len(n_levels)) {
      out[idx, idx, lvl] <- block[, , lvl]
    }
    pos <- pos + d_m
  }
  dimnames(out) <- list(colnames(reMeans),
                        colnames(reMeans),
                        rownames(reMeans))
  out
}

#' Simulate random‑effect contributions for all grouping factors
#'
#' For each random effect term the function draws `n.sims` samples from the
#' conditional multivariate normal distribution (BLUPs + post‑var). The result is a
#' named list where each element is an array of dimensions `nrow(newdata)` ×
#' `n.sims`.
#' @param merMod A merMod object.
#' @param newdata Data frame containing the prediction covariates.
#' @param n.sims Number of simulations.
#' @param .parallel Logical flag to enable parallel execution via foreach.
#' @param seed Optional random seed for reproducibility.
#' @param new.levels Character; how to treat grouping levels present in
#'   `newdata` but absent from the fitted model. `"zero"` drops the random
#'   effect for such levels; `"draw"` samples each unobserved level's effect
#'   from the estimated random-effect covariance (`VarCorr`).
#' @return List of matrices (rows = observations, cols = simulations).
simulate_random_effects <- function(
  merMod,
  newdata,
  n.sims,
  .parallel = FALSE,
  seed = NULL,
  new.levels = c("zero", "draw")
) {
  new.levels <- match.arg(new.levels, c("zero", "draw"), several.ok = FALSE)
  if (!is.null(seed)) set.seed(seed)
  else if (!exists('.Random.seed', envir = .GlobalEnv)) runif(1)
  rr <- ranef(merMod, condVar = TRUE)
  # Marginal random-effect (co)variance per grouping factor, used to sample
  # effects for unobserved levels when new.levels = "draw".
  reVarCov <- if (identical(new.levels, "draw")) lme4::VarCorr(merMod) else NULL
  re.xb <- vector(getME(merMod, 'n_rfacs'), mode = 'list')
  names(re.xb) <- names(ngrps(merMod))
  newdata.modelMatrix <- buildModelMatrix(model = merMod, newdata = newdata)

  for (j in names(re.xb)) {
    reMeans <- as.matrix(rr[[j]])
    reMatrix <- attr(rr[[j]], which = 'postVar')
    # When a grouping factor has multiple RE term blocks (|| or explicit
    # splits), lme4 returns postVar as a list of per-block arrays. Convert
    # to a single block-diagonal array so downstream indexing works (#118).
    if (is.list(reMatrix)) {
      reMatrix <- combine_postvar_blocks(reMatrix, reMeans)
    }
    # Determine levels to keep based on newdata
    if (j %in% names(newdata)) {
      obslvl <- unique(as.character(newdata[, j]))
      alllvl <- rownames(reMeans)
      keep <- intersect(obslvl, alllvl)
    } else {
      obslvl <- colnames(newdata.modelMatrix)
      alllvl <- rownames(reMeans)
      keep <- intersect(obslvl, alllvl)
    }
    if (length(keep) > 0 && !identical(keep, alllvl)) {
      reMeans <- reMeans[keep, , drop = FALSE]
      dimnames(reMatrix)[[3]] <- alllvl
      reMatrix <- reMatrix[, , keep, drop = FALSE]
    } else if (length(keep) > 0 && identical(keep, alllvl)) {
      dimnames(reMatrix)[[3]] <- alllvl
      reMatrix <- reMatrix[, , keep, drop = FALSE]
    } else {
      reMeans <- reMeans[1, , drop = FALSE]
      reMatrix <- reMatrix[, , 1, drop = FALSE]
    }

    # Simulate draws for each level
    tmpList <- vector('list', nrow(reMeans))
    for (k in seq_len(nrow(reMeans))) {
      meanTmp <- reMeans[k, ]
      matrixTmp <- as.matrix(reMatrix[, , k])
      tmpList[[k]] <- mvtnorm::rmvnorm(
        n = n.sims,
        mean = meanTmp,
        sigma = matrixTmp,
        method = 'chol'
      )
    }
    REcoefs <- sapply(tmpList, identity, simplify = 'array')
    dimnames(REcoefs) <- list(1:n.sims,
                               attr(reMeans, 'dimnames')[[2]],
                               attr(reMeans, 'dimnames')[[1]])

    # new.levels = "draw": for grouping levels in newdata that were not in the
    # fitted model, sample an effect for each from the estimated random-effect
    # covariance (one draw per unobserved level, shared by its observations) and
    # append it to REcoefs. Downstream code then treats these levels as known,
    # so the prediction carries the between-group uncertainty rather than
    # dropping the effect. The default ("zero") skips this and lets such levels
    # fall back to the fixed effects plus residual variation in tmp.pred().
    if (identical(new.levels, "draw") && j %in% names(newdata)) {
      new_lv <- setdiff(unique(as.character(newdata[, j])), alllvl)
      terms_o <- dimnames(REcoefs)[[2]]
      Sigma <- as.matrix(reVarCov[[j]])
      if (length(new_lv) > 0 && all(terms_o %in% rownames(Sigma))) {
        Sigma <- Sigma[terms_o, terms_o, drop = FALSE]
        add <- array(NA_real_, dim = c(n.sims, length(terms_o), length(new_lv)),
                     dimnames = list(1:n.sims, terms_o, new_lv))
        for (k in seq_along(new_lv)) {
          add[, , k] <- mvtnorm::rmvnorm(n.sims, mean = rep(0, length(terms_o)),
                                         sigma = Sigma, method = 'chol')
        }
        REcoefs <- abind::abind(REcoefs, add, along = 3)
      }
    }

    # Build temporary data frame for prediction as in original code
    if (j %in% names(newdata)) {
      newdata.modelMatrix <- as.matrix(newdata.modelMatrix)
      tmp <- cbind(as.data.frame(newdata.modelMatrix), var = newdata[, j])
      tmp <- tmp[, !duplicated(colnames(tmp))]
      keepCols <- names(tmp)[names(tmp) %in% colnames(REcoefs)]
      if (length(keepCols) == 0) {
        keepCols <- grep(dimnames(REcoefs)[[2]], names(tmp), value = TRUE)
      }
      if (length(keepCols) == 0) {
        tmp <- cbind(model.frame(reformulas::subbars(formula(merMod)), data = newdata),
                     var = newdata[, j])
        keepCols <- grep(dimnames(REcoefs)[[2]], names(tmp), value = TRUE)
      }
      if (length(keepCols) == 0) {
        # Add in an intercept for RE purposes
        tmp <- cbind(as.data.frame(newdata.modelMatrix), var = newdata[, j])
        tmp <- tmp[, !duplicated(colnames(tmp))]
        tmp <- cbind(data.frame(1), tmp)
        names(tmp)[1] <- "(Intercept)"
        keepCols <- "(Intercept)"
      }
      tmp <- tmp[, c(keepCols, "var"), drop = FALSE]
      tmp[, "var"] <- as.character(tmp[, "var"])
      colnames(tmp)[which(names(tmp) == "var")] <- names(newdata[, j, drop = FALSE])
      if (all(grepl(":", keepCols))) {
        # Strip out the interaction after
        keepCols <- unique(gsub("(.*):.*", "\\1", keepCols))
      }
    } else {
      # If newdata.modelMatrix is still sparse at this point, we need to convert it safely
      if (is(newdata.modelMatrix, 'dgCMatrix')) {
        newdata.modelMatrix <- as.matrix(newdata.modelMatrix)
        tmp <- as.data.frame(newdata.modelMatrix)
      } else {
        tmp <- as.data.frame(newdata.modelMatrix)
      }
      tmp <- tmp[, !duplicated(colnames(tmp))]
      # Assign the grouping level for each row based on which model-matrix
      # column is active. Rows whose interaction level is not among the
      # observed `keep` levels are all-zero across tmp[keep]; max.col() would
      # otherwise pick a column at random (an all-zero row is a tie), silently
      # borrowing an unrelated level's random effect and making the result
      # depend on the RNG seed (#124). Detect those rows and assign NA so that
      # tmp.pred() treats them as new levels and zeroes the random-effect
      # contribution, matching the documented out-of-sample behavior.
      keepMat <- as.matrix(tmp[keep])
      activeLevel <- names(tmp[keep])[max.col(keepMat, ties.method = "first")]
      activeLevel[rowSums(keepMat != 0) == 0] <- NA_character_
      tmp$var <- activeLevel
      keepCols <- names(tmp)[names(tmp) %in% dimnames(REcoefs)[[2]]]
      tmp <- tmp[, c(keepCols, "var"), drop = FALSE]
      tmp[, "var"] <- as.character(tmp[, "var"])
      colnames(tmp)[which(names(tmp) == "var")] <- j
    }

    # Compute contribution matrix using the internal prediction helper
    # Use parallel execution for large datasets
    if (nrow(tmp) > 1000 || .parallel) {
      if (requireNamespace("foreach", quietly = TRUE)) {
        if (.parallel) {
          setup_parallel()
        }
        # Split data into chunks of 500 rows
        tmp2 <- split(tmp, (seq_len(nrow(tmp)) - 1) %/% 500)
        tmp2 <- tmp2[vapply(tmp2, function(x) nrow(x) > 0, logical(1))]

        fe_call <- as.call(c(
          list(quote(foreach::foreach), i = seq_along(tmp2), .combine = 'rbind')
        ))
        fe <- eval(fe_call)

        re.xb[[j]] <- foreach::`%dopar%`(
          fe,
          tmp.pred(data = tmp2[[i]], coefs = REcoefs[, keepCols, , drop = FALSE], group = j)
        )
      } else {
        warning("foreach package is unavailable, parallel computing not available")
        re.xb[[j]] <- tmp.pred(data = tmp, coefs = REcoefs[, keepCols, , drop = FALSE], group = j)
      }
    } else {
      re.xb[[j]] <- tmp.pred(data = tmp, coefs = REcoefs[, keepCols, , drop = FALSE], group = j)
    }
  }
  re.xb
}

#' Combine fixed, random and residual components into a prediction array
#'
#' @param fixed_mat Matrix of dimensions `nrow(newdata)` × `n.sims`.
#' @param random_list List returned by `simulate_random_effects()`.
#' @param sigma_vec Residual standard deviations (output of `simulate_residual_variance`).
#' @param include.resid.var Logical, whether to add residual variance.
#' @param which Character: "full", "fixed", "random" or "all".
#' @return Array `nrow(newdata)` × `n.sims` (or list when `which="all"`).
#' @keywords internal
combine_components <- function(
  fixed_mat,
  random_list,
  sigma_vec,
  include.resid.var = TRUE,
  which = c('full', 'fixed', 'random', 'all'),
  family = NULL,
  link = NULL,
  weights = NULL,
  use.probability = FALSE
) {
  which <- match.arg(which)
  re.xb <- random_list
  re.xb$fixed <- fixed_mat

  if (which == 'fixed') {
    yhat <- re.xb[['fixed']]
    if (include.resid.var) {
      N <- nrow(yhat)
      
if (use.probability) {
         yhat <- simulate_glmm_response(yhat, family = family, link = link,
                                         weights = weights, use.probability = TRUE)
       } else if (!is.null(sigma_vec)) {
         yhat <- abind::abind(
           lapply(seq_len(length(sigma_vec)), function(x) rnorm(N, yhat[, x], sigma_vec[x])),
           along = 2
         )
       }
     }
     return(yhat)
   }
 
   if (which == 'random') {
    re.xb[['fixed']] <- NULL
    yhat <- Reduce('+', re.xb)
    if (include.resid.var) {
      N <- nrow(yhat)
      
if (use.probability) {
         yhat <- simulate_glmm_response(yhat, family = family, link = link,
                                         weights = weights, use.probability = TRUE)
       } else if (!is.null(sigma_vec)) {
         yhat <- abind::abind(
           lapply(seq_len(length(sigma_vec)), function(x) rnorm(N, yhat[, x], sigma_vec[x])),
           along = 2
         )
       }
     }
     return(yhat)
   }
 
   if (which == 'full') {
    yhat <- Reduce('+', re.xb)
    if (include.resid.var) {
      N <- nrow(yhat)
      
      if (use.probability) {
        yhat <- simulate_glmm_response(yhat, family = family, link = link,
                                        weights = weights, use.probability = TRUE)
      } else if (!is.null(sigma_vec)) {
        yhat <- abind::abind(
          lapply(seq_len(length(sigma_vec)), function(x) rnorm(N, yhat[, x], sigma_vec[x])),
          along = 2
        )
      }
    }
    return(yhat)
  }

  if (which == 'all') {
    yhat <- Reduce('+', re.xb)
    
    for (i in seq_along(re.xb)) {
      if (include.resid.var) {
        N <- nrow(yhat)
        
if (!is.null(family) && family %in% c("binomial", "poisson", "Gamma")) {
           re.xb[[i]] <- simulate_glmm_response(re.xb[[i]], family = family, link = link,
                                                 weights = weights, use.probability = use.probability)
         } else if (!is.null(sigma_vec)) {
           re.xb[[i]] <- abind::abind(
             lapply(seq_len(length(sigma_vec)), function(x) rnorm(N, re.xb[[i]][, x], sigma_vec[x])),
             along = 2
           )
         }
      }
    }
    
    if (include.resid.var) {
      N <- nrow(yhat)
      
      if (use.probability) {
        yhat <- simulate_glmm_response(yhat, family = family, link = link,
                                        weights = weights, use.probability = TRUE)
      } else if (!is.null(sigma_vec)) {
        yhat <- abind::abind(
          lapply(seq_len(length(sigma_vec)), function(x) rnorm(N, yhat[, x], sigma_vec[x])),
          along = 2
        )
      }
    }
    return(list(yhat = yhat, components = re.xb))
  }
}

#' Summarise prediction simulations into intervals
#'
#' Takes the simulated prediction array and computes summary statistics
#' (fit, lower bound, upper bound) based on the specified statistic type
#' and confidence level. Handles probability transformation and the
#' `which = "all"` case with component-wise summaries.
#'
#' @param yhat_arr matrix of simulated predictions with dimensions (N x n.sims),
#'   or for `which = "all"`, the combined yhat from `combine_components()`.
#' @param level numeric confidence level (e.g., 0.8 for 80
#' @param stat.type character either "median" or "mean" for the central estimate.
#' @param predict.type character either "linear.prediction" or "probability".
#' @param N integer number of observations in the original newdata.
#' @param merMod merMod object, required when `predict.type = "probability"` to
#'   access the link function.
#' @param which.eff character the effect type: "full", "fixed", "random", or "all".
#' @param pi.comps list of component prediction matrices (from `combine_components()`
#'   when `which = "all"`), or NULL for other cases.
#' @return data.frame with columns `fit`, `upr`, `lwr`. For `which = "all"`,
#'   includes additional columns `effect` and `obs`.
#' @keywords internal
summarise_predictions <- function(
  yhat_arr,
  level,
  stat.type,
  predict.type,
  N,
  merMod = NULL,
  which.eff = "full",
  pi.comps = NULL,
  is.glmm.with.response = FALSE
) {
  # Calculate quantile bounds

upCI <- 1 - ((1 - level) / 2)
  loCI <- (1 - level) / 2

  # Initialize output data frame
  outs <- data.frame(
    fit = rep(NA_real_, N),
    upr = rep(NA_real_, N),
    lwr = rep(NA_real_, N)
  )

  # Compute summary statistics based on stat.type
  if (stat.type == "median") {
    outs[, 1:3] <- t(apply(yhat_arr, 1, quantile,
                           prob = c(0.5, upCI, loCI),
                           na.rm = TRUE))
  } else if (stat.type == "mean") {
    outs$fit <- apply(yhat_arr, 1, mean, na.rm = TRUE)
    outs[, 2:3] <- t(apply(yhat_arr, 1, quantile,
                           prob = c(upCI, loCI),
                           na.rm = TRUE))
  }

  # Apply link function for probability predictions
  if (predict.type == "probability") {
    if (is.null(merMod)) {
      stop("merMod object required for probability predictions", call. = FALSE)
    }
    
    # Skip linkinv for ALL GLMM families with response-scale simulations
    devcomp <- getME(merMod, "devcomp")
    is_glmm <- (devcomp$dims[["GLMM"]] == TRUE)
    
    if (!(is.glmm.with.response && is_glmm)) {
      if (nrow(outs) == 1) {
        outs <- t(apply(outs, 2, merMod@resp$family$linkinv))
      } else {
        outs <- apply(outs, 2, merMod@resp$family$linkinv)
      }
    }
    
    outs <- as.data.frame(outs)
    names(outs) <- c("fit", "upr", "lwr")
  }

  # Handle which = "all" case with component breakdown
  if (which.eff == "all" && !is.null(pi.comps)) {
    # Process each component
    for (i in seq_along(pi.comps)) {
      if (stat.type == "median") {
        pi.comps[[i]] <- t(apply(pi.comps[[i]], 1, quantile,
                                 prob = c(0.5, upCI, loCI),
                                 na.rm = TRUE))
        pi.comps[[i]] <- as.data.frame(pi.comps[[i]])
        names(pi.comps[[i]]) <- c("fit", "upr", "lwr")
      } else if (stat.type == "mean") {
        tmp <- pi.comps[[i]]
        pi.comps[[i]] <- data.frame(
          fit = rep(NA_real_, N),
          upr = NA_real_,
          lwr = NA_real_
        )
        pi.comps[[i]]$fit <- apply(tmp, 1, mean, na.rm = TRUE)
        pi.comps[[i]][, 2:3] <- t(apply(tmp, 1, quantile,
                                        prob = c(upCI, loCI),
                                        na.rm = TRUE))
      }
      # Apply link function for probability predictions
      if (predict.type == "probability") {
        devcomp <- getME(merMod, "devcomp")
        is_binomial_glmm <- (devcomp$dims[["GLMM"]] == TRUE &&
                             merMod@resp$family$family == "binomial")
        
        if (!(is.glmm.with.response && is_binomial_glmm)) {
          pi.comps[[i]] <- apply(pi.comps[[i]], 2, merMod@resp$family$linkinv)
        }
        pi.comps[[i]] <- as.data.frame(pi.comps[[i]])
        names(pi.comps[[i]]) <- c("fit", "upr", "lwr")
      }
    }

    # Bind components together with effect labels
    componentOut <- dplyr::bind_rows(pi.comps, .id = "effect")
    outs <- cbind(data.frame(effect = "combined"), outs)
    outs <- suppressWarnings(dplyr::bind_rows(outs, componentOut))
    outs$obs <- rep(seq_len(N), nrow(outs) %/% N)
  }

  as.data.frame(outs)
}

#' Inverse link function for GLMM families
#'
#' Internal helper that applies the inverse link function to convert linear
#' predictors to distribution parameters.
#'
#' @param eta Linear predictor (fixed + random effects).
#' @param family Character string specifying the distribution family.
#' @param link Character string specifying the link function.
#' @return Vector of distribution parameters (prob, lambda, or rate).
#' @keywords internal
inverse_link <- function(eta, family, link) {
  if (family == "binomial") {
    if (link == "logit") return(1 / (1 + exp(-eta)))
    if (link == "probit") return(pnorm(eta))
    if (link == "cloglog") return(1 - exp(-exp(eta)))
  }
  
  if (family == "poisson") {
    if (link == "log") return(exp(eta))
    if (link == "sqrt") return(pmax(eta, 0)^2)
    if (link == "identity") return(pmax(eta, 0))
  }
  
  if (family == "Gamma") {
    eps <- .Machine$double.xmin
    if (link == "inverse") return(1 / pmax(eta, eps))
    if (link == "identity") return(pmax(eta, eps))
    if (link == "log") return(exp(eta))
  }
  
  stop("Unsupported family/link combination: ", family, "/", link)
}

#' Simulate GLMM responses from conditional distribution
#'
#' Internal helper that simulates responses from the conditional distribution
#' of Y|b (given random effects) for GLMMs. This implements the correct
#' theoretical approach where residual variance is inherent in the distribution.
#'
#' @param yhat Matrix of linear predictors (fixed + random effects).
#' @param family Character string specifying the distribution family.
#' @param link Character string specifying the link function.
#' @param weights Optional weights (for binomial with n-trial parameter).
#' @param use.probability For binomial, return proportions (0-1) instead of counts?
#' @return Matrix of simulated responses with same dimensions as yhat.
#' @keywords internal
simulate_glmm_response <- function(yhat, family, link, weights = NULL, use.probability = FALSE) {
  N <- nrow(yhat)
  n_sims <- ncol(yhat)
  
  if (family == "binomial") {
    if (is.null(weights)) weights <- rep(1, N)
    if (length(weights) == 1) weights <- rep(weights, N)
    
    eta <- as.matrix(yhat)
    prob <- inverse_link(eta, family, link)
    
    result <- matrix(nrow = N, ncol = n_sims)
    for (i in 1:N) {
      if (use.probability) {
        result[i, ] <- rbinom(n_sims, size = weights[i], prob = prob[i, ]) / weights[i]
      } else {
        result[i, ] <- rbinom(n_sims, size = weights[i], prob = prob[i, ])
      }
    }
    return(result)
  }
  
  if (family == "poisson") {
    lambda <- inverse_link(as.matrix(yhat), family, link)
    result <- matrix(nrow = N, ncol = n_sims)
    for (i in 1:N) {
      result[i, ] <- rpois(n_sims, lambda = lambda[i])
    }
    return(result)
  }
  
  if (family == "Gamma") {
    rate <- inverse_link(as.matrix(yhat), family, link)
    shape <- 1
    result <- matrix(nrow = N, ncol = n_sims)
    for (i in 1:N) {
      result[i, ] <- rgamma(n_sims, shape = shape, rate = rate[i])
    }
    return(result)
  }
  
  stop("Unsupported family for GLMM simulation: ", family)
}

# Tests for predictInterval helper functions

# Test simulate_residual_variance() ----

test_that("simulate_residual_variance returns correct length for LMM", {
  skip_on_cran()
  m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
  set.seed(11213)
  sigma_vec <- merTools:::simulate_residual_variance(m1, n.sims = 100)
  expect_length(sigma_vec, 100)
  expect_true(all(sigma_vec > 0))
})

test_that("simulate_residual_variance respects seed set externally", {
  skip_on_cran()
  m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
  set.seed(11213)
  sigma1 <- merTools:::simulate_residual_variance(m1, n.sims = 100)
  set.seed(11213)
  sigma2 <- merTools:::simulate_residual_variance(m1, n.sims = 100)
  set.seed(456)
  sigma3 <- merTools:::simulate_residual_variance(m1, n.sims = 100)
  expect_identical(sigma1, sigma2)
  expect_false(identical(sigma1, sigma3))
})

test_that("simulate_residual_variance returns NULL for binomial GLMM", {
   skip_on_cran()
   d1 <- cbpp
   d1$y <- d1$incidence / d1$size
   gm1 <- glmer(y ~ period + (1 | herd), family = binomial, data = d1, weights = d1$size)
   sigma_vec <- merTools:::simulate_residual_variance(gm1, n.sims = 100)
   expect_null(sigma_vec)
 })


# Test simulate_fixed_effects() ----

test_that("simulate_fixed_effects returns correct dimensions", {
  skip_on_cran()
  m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
  fixed_mat <- merTools:::simulate_fixed_effects(
    m1,
    newdata = sleepstudy[1:10, ],
    n.sims = 100
  )
  expect_equal(nrow(fixed_mat), 10)
  expect_equal(ncol(fixed_mat), 100)
})

test_that("simulate_fixed_effects respects ignore.fixed.terms", {
  skip_on_cran()
  m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
  set.seed(11213)
  fixed1 <- merTools:::simulate_fixed_effects(
    m1,
    newdata = sleepstudy[1:5, ],
    n.sims = 500,
    ignore.fixed.terms = NULL
  )
  set.seed(11213)
  fixed2 <- merTools:::simulate_fixed_effects(
    m1,
    newdata = sleepstudy[1:5, ],
    n.sims = 500,
    ignore.fixed.terms = "(Intercept)"
  )
  # With ignore.fixed.terms, variance should be reduced
  var1 <- apply(fixed1, 1, var)
  var2 <- apply(fixed2, 1, var)
  expect_true(all(var2 < var1))
})


# Test tmp.pred() ----

test_that("tmp.pred returns correct dimensions", {
  skip_on_cran()
  # Create mock data and coefficients
  data <- data.frame(
    x1 = c(1, 0, 1),
    x2 = c(0, 1, 0),
    group = c("A", "B", "A")
  )
  coefs <- array(
    rnorm(200),
    dim = c(100, 2, 2),
    dimnames = list(1:100, c("x1", "x2"), c("A", "B"))
  )
  result <- merTools:::tmp.pred(data, coefs, "group")
  expect_equal(nrow(result), 3)
  expect_equal(ncol(result), 100)
})

test_that("tmp.pred warns for new levels", {
  skip_on_cran()
  data <- data.frame(
    x1 = c(1, 0, 1),
    group = c("A", "C", "A")  # "C" is a new level
  )
  coefs <- array(
    rnorm(200),
    dim = c(100, 1, 2),
    dimnames = list(1:100, "x1", c("A", "B"))
  )
  expect_warning(
    merTools:::tmp.pred(data, coefs, "group"),
    "not in the model data"
  )
})

test_that("tmp.pred returns zeros for new levels", {
  skip_on_cran()
  data <- data.frame(
    x1 = c(1, 1),
    group = c("A", "C")  # "C" is a new level
  )
  coefs <- array(
    rep(5, 200),  # Non-zero coefficients
    dim = c(100, 1, 2),
    dimnames = list(1:100, "x1", c("A", "B"))
  )
  result <- suppressWarnings(merTools:::tmp.pred(data, coefs, "group"))
  # Row 1 (level A) should have non-zero predictions
  expect_true(all(result[1, ] != 0))
  # Row 2 (new level C) should have zero predictions
  expect_true(all(result[2, ] == 0))
})


# Test simulate_random_effects() ----

test_that("simulate_random_effects returns named list", {
  skip_on_cran()
  m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
  re_list <- merTools:::simulate_random_effects(
    m1,
    newdata = sleepstudy[1:10, ],
    n.sims = 100
  )
  expect_type(re_list, "list")
  expect_named(re_list, "Subject")
  expect_equal(nrow(re_list$Subject), 10)
  expect_equal(ncol(re_list$Subject), 100)
})

test_that("simulate_random_effects handles multiple grouping factors", {
  skip_on_cran()
  skip_on_ci()
  moddf <- InstEval[sample(rownames(InstEval), 5000), ]
  g1 <- lmer(y ~ lectage + studage + (1 | d) + (1 | s), data = moddf)
  re_list <- merTools:::simulate_random_effects(
    g1,
    newdata = moddf[1:10, ],
    n.sims = 50
  )
  expect_type(re_list, "list")
  expect_true(all(c("d", "s") %in% names(re_list)))
})


# Test combine_components() ----

test_that("combine_components returns correct structure for each which option", {
  skip_on_cran()
  # Create mock data
  fixed_mat <- matrix(rnorm(300), nrow = 3, ncol = 100)
  random_list <- list(
    Subject = matrix(rnorm(300), nrow = 3, ncol = 100)
  )
  sigma_vec <- rep(1, 100)

  # Test which = "full"
  result_full <- merTools:::combine_components(
    fixed_mat, random_list, sigma_vec,
    include.resid.var = FALSE, which = "full"
  )
  expect_true(is.matrix(result_full))
  expect_equal(dim(result_full), c(3, 100))

  # Test which = "fixed"
  result_fixed <- merTools:::combine_components(
    fixed_mat, random_list, sigma_vec,
    include.resid.var = FALSE, which = "fixed"
  )
  expect_true(is.matrix(result_fixed))
  expect_equal(dim(result_fixed), c(3, 100))

  # Test which = "random"
  result_random <- merTools:::combine_components(
    fixed_mat, random_list, sigma_vec,
    include.resid.var = FALSE, which = "random"
  )
  expect_true(is.matrix(result_random))
  expect_equal(dim(result_random), c(3, 100))

  # Test which = "all"
  result_all <- merTools:::combine_components(
    fixed_mat, random_list, sigma_vec,
    include.resid.var = FALSE, which = "all"
  )
  expect_type(result_all, "list")
  expect_named(result_all, c("yhat", "components"))
  expect_true(is.matrix(result_all$yhat))
})

test_that("combine_components applies residual variance", {
  skip_on_cran()
  set.seed(11213)
  fixed_mat <- matrix(rep(0, 300), nrow = 3, ncol = 100)
  random_list <- list(
    Subject = matrix(rep(0, 300), nrow = 3, ncol = 100)
  )
  sigma_vec <- rep(10, 100)  # Large sigma

  result_with <- merTools:::combine_components(
    fixed_mat, random_list, sigma_vec,
    include.resid.var = TRUE, which = "full"
  )
  result_without <- merTools:::combine_components(
    fixed_mat, random_list, sigma_vec,
    include.resid.var = FALSE, which = "full"
  )

  # With residual variance, results should have larger spread
  expect_gt(sd(result_with), sd(result_without))
})


# Test summarise_predictions() ----

test_that("summarise_predictions returns correct columns", {
  skip_on_cran()
  yhat_arr <- matrix(rnorm(300), nrow = 3, ncol = 100)
  result <- merTools:::summarise_predictions(
    yhat_arr,
    level = 0.8,
    stat.type = "median",
    predict.type = "linear.prediction",
    N = 3
  )
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 3)
  expect_named(result, c("fit", "upr", "lwr"))
})

test_that("summarise_predictions respects stat.type", {
  skip_on_cran()
  set.seed(11213)
  yhat_arr <- matrix(rnorm(3000, mean = 100), nrow = 3, ncol = 1000)
  result_median <- merTools:::summarise_predictions(
    yhat_arr,
    level = 0.8,
    stat.type = "median",
    predict.type = "linear.prediction",
    N = 3
  )
  result_mean <- merTools:::summarise_predictions(
    yhat_arr,
    level = 0.8,
    stat.type = "mean",
    predict.type = "linear.prediction",
    N = 3
  )
  # Mean and median should be close but not identical
  expect_equal(result_median$fit, result_mean$fit, tolerance = 1)
  expect_false(identical(result_median$fit, result_mean$fit))
})

test_that("summarise_predictions respects level parameter", {
  skip_on_cran()
  set.seed(11213)
  yhat_arr <- matrix(rnorm(3000), nrow = 3, ncol = 1000)
  result_80 <- merTools:::summarise_predictions(
    yhat_arr,
    level = 0.8,
    stat.type = "median",
    predict.type = "linear.prediction",
    N = 3
  )
  result_95 <- merTools:::summarise_predictions(
    yhat_arr,
    level = 0.95,
    stat.type = "median",
    predict.type = "linear.prediction",
    N = 3
  )
  # 95% interval should be wider than 80%
  width_80 <- result_80$upr - result_80$lwr
  width_95 <- result_95$upr - result_95$lwr
  expect_true(all(width_95 > width_80))
})

test_that("summarise_predictions applies link function for probability", {
  skip_on_cran()
  d1 <- cbpp
  d1$y <- d1$incidence / d1$size
  gm1 <- glmer(y ~ period + (1 | herd), family = binomial, data = d1, weights = d1$size)

  yhat_arr <- matrix(rnorm(300, mean = 0, sd = 1), nrow = 3, ncol = 100)
  result <- merTools:::summarise_predictions(
    yhat_arr,
    level = 0.8,
    stat.type = "median",
    predict.type = "probability",
    N = 3,
    merMod = gm1
  )
  # All values should be between 0 and 1 for probability
  expect_true(all(result$fit >= 0 & result$fit <= 1))
  expect_true(all(result$upr >= 0 & result$upr <= 1))
  expect_true(all(result$lwr >= 0 & result$lwr <= 1))
})

test_that("summarise_predictions handles which='all' correctly", {
  skip_on_cran()
  yhat_arr <- matrix(rnorm(300), nrow = 3, ncol = 100)
  pi.comps <- list(
    Subject = matrix(rnorm(300), nrow = 3, ncol = 100),
    fixed = matrix(rnorm(300), nrow = 3, ncol = 100)
  )
  result <- merTools:::summarise_predictions(
    yhat_arr,
    level = 0.8,
    stat.type = "median",
    predict.type = "linear.prediction",
    N = 3,
    which.eff = "all",
    pi.comps = pi.comps
  )
  # Should have effect and obs columns
  expect_true("effect" %in% names(result))
  expect_true("obs" %in% names(result))
  # Should have combined + 2 components = 9 rows (3 obs each)
  expect_equal(nrow(result), 9)
})

test_that("component reproducibility with seed", {
  data(sleepstudy, package = "lme4")
  m <- lmer(Reaction ~ Days + (Days | Subject), data = sleepstudy, REML = FALSE)
  
  seed <- 11213
  n.sims <- 100
  
  result1 <- predictInterval(
    m,
    newdata = sleepstudy[1:5, ],
    n.sims = n.sims,
    seed = seed
  )
  
  result2 <- predictInterval(
    m,
    newdata = sleepstudy[1:5, ],
    n.sims = n.sims,
    seed = seed
  )
  
  expect_equal(result1, result2)
})

test_that("single observation predictions (n=1)", {
  data(sleepstudy, package = "lme4")
  m <- lmer(Reaction ~ Days + (Days | Subject), data = sleepstudy, REML = FALSE)
  
  seed <- 11213
  result <- predictInterval(
    m,
    newdata = sleepstudy[1, ],
    n.sims = 500,
    seed = seed
  )
  
  expect_equal(nrow(result), 1)
  expect_identical(names(result), c("fit", "upr", "lwr"))
  expect_true(result$upr > result$fit)
  expect_true(result$fit > result$lwr)
})

test_that("extreme confidence levels (0.8 vs 0.99)", {
  data(sleepstudy, package = "lme4")
  m <- lmer(Reaction ~ Days + (Days | Subject), data = sleepstudy, REML = FALSE)
  
  seed <- 11213
  
  result_80 <- predictInterval(
    m,
    newdata = sleepstudy[1:3, ],
    level = 0.8,
    n.sims = 500,
    seed = seed
  )
  
  result_99 <- predictInterval(
    m,
    newdata = sleepstudy[1:3, ],
    level = 0.99,
    n.sims = 500,
    seed = seed
  )
  
  expect_equal(nrow(result_80), 3)
  expect_equal(nrow(result_99), 3)
  
  interval_width_80 <- result_80[,"upr"] - result_80[,"lwr"]
  interval_width_99 <- result_99[,"upr"] - result_99[,"lwr"]
  
  expect_true(all(interval_width_99 > interval_width_80))
})

test_that("GLMM probability predictions bounded [0,1]", {
  data(cbpp, package = "lme4")
  
  m <- glmer(
        cbind(incidence, size - incidence) ~ period + (1 | herd),
    data = cbpp,
    family = binomial
  )
  
  seed <- 11213
  result <- predictInterval(
    m,
    newdata = cbpp[1:10, ],
    type = "probability",
    n.sims = 200,
    seed = seed
  )
  
  expect_equal(nrow(result), 10)
  expect_true(all(result[, "fit"] >= 0 & result[, "fit"] <= 1))
  expect_true(all(result[, "lwr"] >= 0 & result[, "lwr"] <= 1))
  expect_true(all(result[, "upr"] >= 0 & result[, "upr"] <= 1))
})

test_that("parallelization does not throw errors", {
  skip_on_cran()
  
  data(sleepstudy, package = "lme4")
  m <- lmer(Reaction ~ Days + (Days | Subject), data = sleepstudy, REML = FALSE)
  
  seed <- 11213
  
  suppressWarnings(expect_no_error({
    predictInterval(
      m,
      newdata = sleepstudy[1:20, ],
      n.sims = 500,
      seed = seed,
      .parallel = TRUE
    )
  }))
})

test_that("all which options return correct dimensions", {
  data(sleepstudy, package = "lme4")
  m <- lmer(Reaction ~ Days + (Days | Subject), data = sleepstudy, REML = FALSE)
  
  seed <- 11213
  n.sims <- 200
  n.obs <- 5
  
  result_full <- predictInterval(
    m,
    newdata = sleepstudy[1:n.obs, ],
    which = "full",
    n.sims = n.sims,
    seed = seed
  )
  
  result_fixed <- predictInterval(
    m,
    newdata = sleepstudy[1:n.obs, ],
    which = "fixed",
    n.sims = n.sims,
    seed = seed
  )
  
  result_random <- predictInterval(
    m,
    newdata = sleepstudy[1:n.obs, ],
    which = "random",
    n.sims = n.sims,
    seed = seed
  )
  
  result_all <- predictInterval(
    m,
    newdata = sleepstudy[1:n.obs, ],
    which = "all",
    n.sims = n.sims,
    seed = seed
  )
  
  expect_equal(nrow(result_full), n.obs)
  expect_equal(nrow(result_fixed), n.obs)
  expect_equal(nrow(result_random), n.obs)
  
  expect_identical(names(result_full), c("fit", "upr", "lwr"))
  expect_identical(names(result_fixed), c("fit", "upr", "lwr"))
  expect_identical(names(result_random), c("fit", "upr", "lwr"))
  
  expect_true(nrow(result_all) > n.obs)
  expect_true("effect" %in% names(result_all))
})

test_that("large simulation count (n.sims = 10000)", {
  data(sleepstudy, package = "lme4")
  m <- lmer(Reaction ~ Days + (Days | Subject), data = sleepstudy, REML = FALSE)
  
  seed <- 11213
  n.sims <- 10000
  
  result <- predictInterval(
    m,
    newdata = sleepstudy[1:10, ],
    n.sims = n.sims,
    seed = seed,
    returnSims = TRUE
  )
  
  expect_equal(nrow(result), 10)
  
  attr_sims <- attr(result, "sim.results")
  expect_true(!is.null(attr_sims))
  
  interval_width <- result[, "upr"] - result[, "lwr"]
  expect_true(all(interval_width > 0))
})

test_that("component-by-component consistency with predictInterval wrapper", {
  data(sleepstudy, package = "lme4")
  m <- lmer(Reaction ~ Days + (Days | Subject), data = sleepstudy, REML = FALSE)
  
  seed <- 11213
  n.sims <- 200
  newdata <- sleepstudy[1:5, ]
  
  set.seed(seed)
  full_result <- predictInterval(
    m,
    newdata = newdata,
    n.sims = n.sims,
    seed = seed
  )
  
  sigma_vec <- merTools:::simulate_residual_variance(m, n.sims)
  random_list <- merTools:::simulate_random_effects(m, newdata, n.sims, seed = seed)
  fixed_mat <- merTools:::simulate_fixed_effects(m, newdata, n.sims, seed = seed)
  
  combined <- merTools:::combine_components(
    fixed_mat = fixed_mat,
    random_list = random_list,
    sigma_vec = sigma_vec,
    include.resid.var = TRUE,
    which = "full"
  )
  
  combined_df <- as.data.frame(combined)
  names(combined_df) <- c("fit", "upr", "lwr")
  
  expect_equal(full_result$fit, combined_df$fit, tolerance = sd(full_result$fit)/100)
  expect_equal(full_result$upr, combined_df$upr, tolerance = sd(full_result$upr)/100)
  expect_equal(full_result$lwr, combined_df$lwr, tolerance = sd(full_result$lwr)/100)
})

---
title: "Exploring Contextual Effects with merTools"
author: "Jared Knowles and Carl Frederick"
date: "Written 2026-05-29 (last updated 2026-05-29)"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Exploring Contextual Effects with merTools}
  %\VignetteEncoding{UTF-8}
---

<style type="text/css">pre, pre code { white-space: pre-wrap !important; word-break: break-word; overflow-wrap: anywhere; }</style>



## Contextual effects

In a multilevel model, a predictor measured at the individual level can also be
aggregated to the group level, and the two can carry *different* meanings. The
classic example is socio-economic status (SES): a student's own SES captures a
*within-school* relationship, while the average SES of a student's school
captures a *contextual* (or *compositional*) effect -- the additional influence
of attending a more or less advantaged school, over and above a student's own
background.

This vignette reproduces the kind of analysis described in the `modelbased`
package's article ["Studying some effects in context"][modelbased] using the
tools in `merTools`. The data and model follow the contextual-effects example
that has long been used to teach multilevel modeling (Raudenbush and Bryk,
2002; Gelman and Hill, 2007).

[modelbased]: https://easystats.github.io/modelbased/articles/practical_context_effect.html

## The data

`merTools` ships with `hsb`, a subset of the 1982 *High School and Beyond*
survey: 7,185 students nested in 160
schools. It contains both a student-level SES measure (`ses`) and the
school-mean SES (`meanses`) -- exactly the pair we need to separate within-school
and contextual effects.


``` r
data(hsb)
str(hsb[, c("schid", "mathach", "ses", "meanses")])
```

```
#> 'data.frame':	7185 obs. of  4 variables:
#>  $ schid  : chr  "1224" "1224" "1224" "1224" ...
#>  $ mathach: num  5.88 19.71 20.35 8.78 17.9 ...
#>  $ ses    : num  -1.528 -0.588 -0.528 -0.668 -0.158 ...
#>  $ meanses: num  -0.428 -0.428 -0.428 -0.428 -0.428 -0.428 -0.428 -0.428 -0.428 -0.428 ...
```

## A contextual-effects model

We model math achievement as a function of a student's own SES and the average
SES of their school, allowing each school its own intercept:


``` r
cm <- lmer(mathach ~ ses + meanses + (1 | schid), data = hsb)
arm::display(cm)
```

```
#> lmer(formula = mathach ~ ses + meanses + (1 | schid), data = hsb)
#>             coef.est coef.se
#> (Intercept) 12.66     0.15  
#> ses          2.19     0.11  
#> meanses      3.68     0.38  
#> 
#> Error terms:
#>  Groups   Name        Std.Dev.
#>  schid    (Intercept) 1.64    
#>  Residual             6.08    
#> ---
#> number of obs: 7185, groups: schid, 160
#> AIC = 46578.6, DIC = 46559
#> deviance = 46563.8
```

The coefficient on `ses` is the *within-school* slope; the coefficient on
`meanses` is the *contextual* effect. `merTools::FEsim()` summarizes the
uncertainty in both, and `plotFEsim()` highlights the terms whose interval
excludes zero:


``` r
fe <- FEsim(cm, n.sims = 500)
plotFEsim(fe) +
  labs(title = "Within-school (ses) vs. contextual (meanses) effects")
```

<div class="figure" style="text-align: center">
<img src="context-feplot-1.png" alt=""  />
</div>

## Predictions that separate the two effects

Because `meanses` enters the model directly, we can ask how the predicted math
score changes as we move a student across schools of differing average SES while
holding their own SES fixed. `wiggle()` builds the counterfactual data and
`predictInterval()` attaches prediction intervals:


``` r
# A median student in an average school
base_case <- draw(cm, type = "average")
scenarios <- wiggle(base_case, varlist = "meanses",
                    valueslist = list(seq(-0.7, 0.8, by = 0.15)))
pi <- predictInterval(cm, newdata = scenarios, level = 0.9, n.sims = 500,
                      seed = 11213)
scenarios <- cbind(scenarios, pi)

ggplot(scenarios, aes(x = meanses, y = fit, ymin = lwr, ymax = upr)) +
  geom_ribbon(alpha = 0.2) +
  geom_line() +
  labs(x = "School-average SES (context)", y = "Predicted math achievement",
       title = "Contextual effect of school SES (own SES held at its mean)") +
  theme_minimal(base_size = 12)
```

<div class="figure" style="text-align: center">
<img src="context-wiggle-1.png" alt=""  />
</div>

## How much do schools matter across the rank distribution?

Finally, `REimpact()` summarizes how much being in a higher- versus
lower-performing school shifts the predicted outcome, binning schools by their
expected rank. `plotREimpact()` (new in this release) plots the result, and can
overlay several cases for comparison:


``` r
imp <- REimpact(cm, newdata = hsb[1, ], groupFctr = "schid",
                breaks = 4, n.sims = 500)
plotREimpact(imp) +
  labs(title = "Impact of school placement on predicted achievement")
```

<div class="figure" style="text-align: center">
<img src="context-reimpact-1.png" alt=""  />
</div>

## References

Bates, D., Maechler, M., Bolker, B., and Walker, S. (2015). Fitting Linear
Mixed-Effects Models Using lme4. *Journal of Statistical Software*, 67(1),
1-48.

Gelman, A. and Hill, J. (2007). *Data Analysis Using Regression and
Multilevel/Hierarchical Models*. Cambridge University Press.

Lüdecke, D., et al. *modelbased: Studying some effects in context.* easystats.
<https://easystats.github.io/modelbased/articles/practical_context_effect.html>

Raudenbush, S. W. and Bryk, A. S. (2002). *Hierarchical Linear Models:
Applications and Data Analysis Methods* (2nd ed.). SAGE.

---
title: "An Introduction to merTools"
author: "Jared Knowles and Carl Frederick"
date: "Written 2015-05-27 (last updated 2026-05-29)"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{An Introduction to merTools}
  %\VignetteEncoding{UTF-8}
---

<style type="text/css">pre, pre code { white-space: pre-wrap !important; word-break: break-word; overflow-wrap: anywhere; }</style>



## Introduction

Working with generalized linear mixed models (GLMM) and linear mixed models (LMM)
has become increasingly easy with the advances in the `lme4` package recently.
As we have found ourselves using these models more and more within our work, we,
the authors, have developed a set of tools for simplifying and speeding up common
tasks for interacting with `merMod` objects from `lme4`. This package provides
those tools.

## Illustrating Model Effects

As the complexity of the model fit grows, it becomes harder and harder to
interpret the substantive effect of parameters in the model.

Let's start with a medium-sized example model using the `InstEval` data provided
by the `lme4` package. These data represent university lecture evaluations at
ETH Zurich made by students. In this data, `s` is an individual student,
`d` is an individual lecturer, `studage` is the semester the student
is enrolled, `lectage` is how many semesters back the lecture with the
rating took place, `dept` is the department of the lecture, and `y` is an integer
1:5 representing the ratings of the lecture from "poor" to "very good":


``` r
library(lme4)
head(InstEval)
```

```
#>   s    d studage lectage service dept y
#> 1 1 1002       2       2       0    2 5
#> 2 1 1050       2       1       1    6 2
#> 3 1 1582       2       2       0    2 5
#> 4 1 2050       2       2       1    3 3
#> 5 2  115       2       1       0    5 2
#> 6 2  756       2       1       0    5 4
```

``` r
str(InstEval)
```

```
#> 'data.frame':	73421 obs. of  7 variables:
#>  $ s      : Factor w/ 2972 levels "1","2","3","4",..: 1 1 1 1 2 2 3 3 3 3 ...
#>  $ d      : Factor w/ 1128 levels "1","6","7","8",..: 525 560 832 1068 62 406 3 6 19 75 ...
#>  $ studage: Ord.factor w/ 4 levels "2"<"4"<"6"<"8": 1 1 1 1 1 1 1 1 1 1 ...
#>  $ lectage: Ord.factor w/ 6 levels "1"<"2"<"3"<"4"<..: 2 1 2 2 1 1 1 1 1 1 ...
#>  $ service: Factor w/ 2 levels "0","1": 1 2 1 2 1 1 2 1 1 1 ...
#>  $ dept   : Factor w/ 14 levels "15","5","10",..: 14 5 14 12 2 2 13 3 3 3 ...
#>  $ y      : int  5 2 5 3 2 4 4 5 5 4 ...
```

Starting with a simple model:


``` r
m1 <- lmer(y ~ service + lectage + studage + (1|d) + (1|s), data=InstEval)
```

After fitting the model we can make use of the first function provided by
`merTools`, `fastdisp` which modifies the function `arm:::display` to more
quickly display a summary of the model without calculating the model sigma:


``` r
library(merTools)
fastdisp(m1)
```

```
#> lmer(formula = y ~ service + lectage + studage + (1 | d) + (1 | 
#>     s), data = InstEval)
#>             coef.est coef.se
#> (Intercept)  3.22     0.02  
#> service1    -0.07     0.01  
#> lectage.L   -0.19     0.02  
#> lectage.Q    0.02     0.01  
#> lectage.C   -0.02     0.01  
#> lectage^4   -0.02     0.01  
#> lectage^5   -0.04     0.02  
#> studage.L    0.10     0.02  
#> studage.Q    0.01     0.02  
#> studage.C    0.02     0.02  
#> 
#> Error terms:
#>  Groups   Name        Std.Dev.
#>  s        (Intercept) 0.33    
#>  d        (Intercept) 0.52    
#>  Residual             1.18    
#> ---
#> number of obs: 73421, groups: s, 2972; d, 1128
#> AIC = 237655
```

We see some interesting effects. First, our decision to include student and
lecturer effects seems justified as there is substantial variance within these
groups. Second, there do appear to be some effects by age and for lectures
given as a service by an outside lecturer. Let's look at these in more detail.
One way to do this would be to plot the coefficients together in a line to
see which deviate from 0 and in what direction. To get a confidence interval
for our fixed effect coefficients we have a number of options that represent a
tradeoff between coverage and computation time -- see `confint.merMod` for
details.

An alternative is to simulate values of the fixed effects from the posterior
using the function `arm::sim`. Our next tool, `FEsim`, is a convenience wrapper
to do this and provide an informative data frame of the results.


``` r
feEx <- FEsim(m1, 1000)
cbind(feEx[,1] , round(feEx[, 2:4], 3))
```

```
#>      feEx[, 1]   mean median    sd
#> 1  (Intercept)  3.226  3.226 0.020
#> 2     service1 -0.072 -0.071 0.013
#> 3    lectage.L -0.185 -0.186 0.017
#> 4    lectage.Q  0.024  0.024 0.013
#> 5    lectage.C -0.024 -0.024 0.013
#> 6    lectage^4 -0.020 -0.020 0.013
#> 7    lectage^5 -0.038 -0.039 0.015
#> 8    studage.L  0.096  0.095 0.019
#> 9    studage.Q  0.006  0.007 0.016
#> 10   studage.C  0.017  0.017 0.016
```

We can present these results graphically, using `ggplot2`:


``` r
library(ggplot2)
ggplot(feEx[feEx$term!= "(Intercept)", ]) +
  aes(x = term, ymin = median - 1.96 * sd,
      ymax = median + 1.96 * sd, y = median) +
  geom_pointrange() +
  geom_hline(yintercept = 0, size = I(1.1), color = I("red")) +
  coord_flip() +
  theme_bw() + labs(title = "Coefficient Plot of InstEval Model",
                    x = "Median Effect Estimate", y = "Evaluation Rating")
```

```
#> Warning: Using `size` aesthetic for lines was deprecated in ggplot2 3.4.0.
#> ℹ Please use `linewidth` instead.
#> This warning is displayed once per session.
#> Call `lifecycle::last_lifecycle_warnings()` to see where this warning was
#> generated.
```

![](mertoolsIntro-fixeffplot-1.png)

However, an easier option is:


``` r
plotFEsim(feEx) +
  theme_bw() + labs(title = "Coefficient Plot of InstEval Model",
                    x = "Median Effect Estimate", y = "Evaluation Rating")
```

![](mertoolsIntro-quickFEplot-1.png)

## Random Effects

Next, we might be interested in exploring the random effects. Again, we create a
dataframe of the values of the simulation of these effects for the individual
levels.


``` r
reEx <- REsim(m1)
head(reEx)
```

```
#>   groupFctr groupID        term       mean       median        sd
#> 1         s       1 (Intercept)  0.1867451 0.2108076314 0.3015230
#> 2         s       2 (Intercept) -0.0251678 0.0006431095 0.3390197
#> 3         s       3 (Intercept)  0.3098553 0.3219927161 0.2811772
#> 4         s       4 (Intercept)  0.2305368 0.2428994747 0.2838737
#> 5         s       5 (Intercept)  0.1085762 0.1061684358 0.3096078
#> 6         s       6 (Intercept)  0.1125153 0.0945106543 0.2257176
```

The result is a dataframe with estimates of the values of each of the random
effects provided by the `arm::sim()` function. *groupID* represents the identifiable
level for the variable for one random effect, *term* represents whether the
simulated values are for an intercept or which slope, and *groupFctr* identifies
which of the `(1|x)` terms the values represent. To make unique identifiers
for each term, we need to use both the `groupID` and the `groupFctr` term in case
these two variables use overlapping label names for their groups. In this case:


``` r
table(reEx$term)
```

```
#> 
#> (Intercept) 
#>        4100
```

``` r
table(reEx$groupFctr)
```

```
#> 
#>    d    s 
#> 1128 2972
```

Most important is producing caterpillar or dotplots of these terms to explore
their variation. This is easily accomplished with the `dotplot` function:


``` r
lattice::dotplot(ranef(m1, condVar=TRUE))
```

However, these graphics do not provide much control over the results. Instead,
we can use the `plotREsim` function in `merTools` to gain more control over
plotting of the random effect simulations.


``` r
p1 <- plotREsim(reEx)
p1
```

![](mertoolsIntro-refplot1-1.png)

The result is a ggplot2 object which can be modified however the user sees
fit. Here, we've established that most student and professor effects are
indistinguishable from zero, but there do exist extreme outliers with both
high and low averages that need to be accounted for.

## Substantive Effects

A logical next line of questioning is to see how much of the variation in a
rating can be caused by changing the student rater and how much is due to the
fixed effects we identified above. This is a very difficult problem to solve,
but using simulation we can examine the model behavior under a range of scenarios
to understand how the model is reflecting changes in the data. To do this,
we use another set of functions available in `merTools`.

The simplest option is to pick an observation at random and then modify its
values deliberately to see how the prediction changes in response. `merTools`
makes this task very simple:


``` r
example1 <- draw(m1, type = 'random')
head(example1)
```

```
#>      y service lectage studage    d   s
#> 5587 2       0       3       8 1212 204
```

The `draw` function takes a random observation from the data in the model
and extracts it as a dataframe. We can now do a number of operations to this
observation:


``` r
# predict it
predict(m1, newdata = example1)
```

```
#>     5587 
#> 3.324069
```

``` r
# change values
example1$service <- "1"
predict(m1, newdata = example1)
```

```
#>     5587 
#> 3.253225
```

More interesting, let's programmatically modify this observation to see how the
predicted value changes if we hold everything but one variable constant.


``` r
example2 <- wiggle(example1, varlist = "lectage",
          valueslist = list(c("1", "2", "3", "4", "5", "6")))

example2
```

```
#>       y service lectage studage    d   s
#> 5587  2       1       1       8 1212 204
#> 55871 2       1       2       8 1212 204
#> 55872 2       1       3       8 1212 204
#> 55873 2       1       4       8 1212 204
#> 55874 2       1       5       8 1212 204
#> 55875 2       1       6       8 1212 204
```

The function `wiggle` allows us to create a new dataframe with copies of the
variable that modify just one value. Chaining together `wiggle` calls, we
can see how the variable behaves under a number of different scenarios
simultaneously.


``` r
example2$yhat <- predict(m1, newdata = example2)

ggplot(example2, aes(x = lectage, y = yhat)) + geom_line(aes(group = 1)) +
  theme_bw() + ylim(c(1, 5)) +
  geom_hline(yintercept = mean(InstEval$y), linetype = 2) +
  geom_hline(yintercept = mean(InstEval$y) + sd(InstEval$y), linetype = 3) +
  geom_hline(yintercept = mean(InstEval$y) - sd(InstEval$y), linetype = 3)
```

![](mertoolsIntro-predictplotwiggle-1.png)

The result allows us to graphically display the effect of each level of `lectage`
on an observation that is otherwise identical. This is plotted here against a
horizontal line representing the mean of the observed ratings, and two finer
lines showing plus or minus one standard deviation of the mean.

This is nice, but selecting a random observation is not very satisfying as it
may not be very meaningful. To address this, we can instead take the average
observation:


``` r
example3 <- draw(m1, type = 'average')
example3
```

```
#>          y service lectage studage    d    s
#> 1 3.205745       0       1       6 1510 1014
```

Here, the average observation is identified based on either the modal observation
for factors or on the mean for numeric variables. Then, the random effect
terms are set to the level equivalent to the median effect -- very close
to 0.


``` r
example3 <- wiggle(example1, varlist = "service",
          valueslist = list(c("0", "1")))
example3$yhat <- predict(m1, newdata = example3)

ggplot(example3, aes(x = service, y = yhat)) + geom_line(aes(group = 1)) +
  theme_bw() + ylim(c(1, 5)) +
  geom_hline(yintercept = mean(InstEval$y), linetype = 2) +
  geom_hline(yintercept = mean(InstEval$y) + sd(InstEval$y), linetype = 3) +
  geom_hline(yintercept = mean(InstEval$y) - sd(InstEval$y), linetype = 3)
```

![](mertoolsIntro-wiggle2-1.png)

Here we can see that for the average observation, whether the lecture is outside
of the home department has a very slight negative effect on the overall rating.
Might the individual professor or student have more of an impact on the overall
rating? To answer this question we need to wiggle the same observation across
a wide range of student or lecturer effects.

How do we identify this range? `merTools` provides the `REquantile` function
which helps to identify which levels of the grouping terms correspond to which
quantile of the magnitude of the random effects:


``` r
REquantile(m1, quantile = 0.25, groupFctr = "s")
```

```
#> [1] "2518"
```

``` r
REquantile(m1, quantile = 0.25, groupFctr = "d")
```

```
#> [1] "18"
```

Here we can see that group level 2518
corresponds to the 25th percentile of the effect for the student groups, and level
`REquantile(m1, quantile = 0.25, groupFctr = "d")` corresponds to the 25th
percentile for the instructor group. Using this information we can reassign
a specific observation to varying magnitudes of grouping term effects to see
how much they might influence our final prediction.


``` r
example4 <- draw(m1, type = 'average')
example4 <- wiggle(example4, varlist = "s",
                      list(REquantile(m1, quantile = seq(0.1, 0.9, .1),
                                 groupFctr = "s")))

example4$yhat <- predict(m1, newdata = example4)

ggplot(example4, aes(x = reorder(s, -yhat), y = yhat)) +
  geom_line(aes(group = 1)) +
  theme_bw() + ylim(c(1, 5)) +
  geom_hline(yintercept = mean(InstEval$y), linetype = 2) +
  geom_hline(yintercept = mean(InstEval$y) + sd(InstEval$y), linetype = 3) +
  geom_hline(yintercept = mean(InstEval$y) - sd(InstEval$y), linetype = 3)
```

![](mertoolsIntro-wiggleanddraw-1.png)

This figure is very interesting because it shows that moving across the range
of student effects can have a larger impact on the score than the fixed effects
we observed above. That is, getting a "generous" or a "stingy" rater can have
a substantial impact on the final rating.

But, we can do even better. First, we can move beyond the average observation
by taking advantage of the `varList` option to the function which allows us to
specify a subset of the data to compute an average for.


``` r
subExample <- list(studage = "2", lectage = "4")
example5 <- draw(m1, type = 'average', varList = subExample)
example5
```

```
#>          y service lectage studage    d    s
#> 1 3.087193       0       4       2 1510 1014
```

Now we have the average observation with a student age of 2 and a lecture age
of 4. We can then follow the same procedure as before to explore the effects
on our subsamples. Before we do that, let's fit a slightly more complex model
that includes a random slope.


``` r
data(VerbAgg)
m2 <- glmer(r2 ~ Anger + Gender + btype + situ +
                (1|id) + (1 + Gender|item), family = binomial,
              data = VerbAgg)
```

```
#> Warning in checkConv(attr(opt, "derivs"), opt$par, ctrl = control$checkConv, : Model failed to converge with max|grad| = 0.0882443 (tol = 0.002, component 1)
#>   See ?lme4::convergence and ?lme4::troubleshooting.
```

``` r
example6 <- draw(m2, type = 'average', varList = list("id" = "149"))
example6$btype <- "scold"
example6$situ <- "self"


tempdf <- wiggle(example6, varlist = "Gender", list(c("M", "F")))
tempdf <- wiggle(tempdf, varlist = "item",
                    list(unique(VerbAgg$item)))
tempdf$yhat <- predict(m2, newdata = tempdf, type = "response",
                       allow.new.levels = TRUE)

ggplot(tempdf, aes(x = item, y = yhat, group = Gender)) +
  geom_line(aes(color = Gender))+
  theme_bw() + ylim(c(0, 1)) +
  theme(axis.text.x = element_text(angle = 20, hjust=1),
        legend.position = "bottom") + labs(x = "Item", y = "Probability")
```

![](mertoolsIntro-wigglesubsamples-1.png)

Here we've shown that the effect of both the intercept and the gender slope on
item simultaneously affect our predicted value. This results in the two lines
for predicted values across the items not being parallel. While we can see this
by looking at the results of the summary of the model object, using `fastdisp`
in the `merTools` package for larger models, it is not intuitive what that
effect looks like across different scenarios. `merTools` has given us the
machinery to investigate this.

## Uncertainty

The above examples make use of simulation to show the model behavior after
changing some values in a dataset. However, until now, we've focused on using
point estimates to represent these changes. The use of predicted point estimates
without incorporating any uncertainty can lead to overconfidence in the precision
of the model.

In the `predictInterval` function, discussed in more detail in another package
vignette, we provide a way to incorporate three out of the four types of
uncertainty inherent in a model. These are:

1. Overall model uncertainty
2. Uncertainty in fixed effect values
3. Uncertainty in random effect values
4. Uncertainty in the distribution of the random effects

1-3 are incorporated in the results of `predictInterval`, while capturing 4
would require making use of the `bootMer` function -- options discussed in
greater detail elsewhere. The main advantage of `predictInterval` is that it is
fast. By leveraging the power of the `arm::sim()` function, we are able to
generate prediction intervals for individual observations from very large models
very quickly. And, it works a lot like `predict`:


``` r
exampPreds <- predictInterval(m2, newdata = tempdf,
                              type = "probability", level = 0.8)
```

```
#> Warning: For binomial GLMMs, include.resid.var = TRUE simulates from the
#> conditional binomial distribution (n-trial binomial simulation).
#> This is the theoretically correct approach.
#> To get predictions without residual variance, set include.resid.var = FALSE.
```

``` r
tempdf <- cbind(tempdf, exampPreds)

ggplot(tempdf, aes(x = item, y = fit, ymin = lwr, ymax = upr,
                   group = Gender)) +
  geom_ribbon(aes(fill = Gender), alpha = I(0.2), color = I("black"))+
  theme_bw() + ylim(c(0, 1)) +
  theme(axis.text.x = element_text(angle = 20),
        legend.position = "bottom")+ labs(x = "Item", y = "Probability")
```

![](mertoolsIntro-speedexample-1.png)

Here we can see there is barely any gender difference in terms of area of
potential prediction intervals. However, by default, this approach includes the
residual variance of the model. If we instead focus just on the uncertainty of
the random and fixed effects, we get:


``` r
exampPreds <- predictInterval(m2, newdata = tempdf,
                              type = "probability",
                              include.resid.var = FALSE, level = 0.8)
tempdf <- cbind(tempdf[, 1:8], exampPreds)

ggplot(tempdf, aes(x = item, y = fit, ymin = lwr, ymax = upr,
                   group = Gender)) +
  geom_ribbon(aes(fill = Gender), alpha = I(0.2), color = I("black"))+
  geom_line(aes(color = Gender)) +
  theme_bw() + ylim(c(0, 1)) +
  theme(axis.text.x = element_text(angle = 20),
        legend.position = "bottom") + labs(x = "Item", y = "Probability")
```

![](mertoolsIntro-excluderesidvar-1.png)

Here, more difference emerges, but we see that the differences are not very
precise.

---
title: "Using merTools to Marginalize Over Random Effect Levels"
author: "Jared Knowles and Carl Frederick"
date: "Written 2019-05-11 (last updated 2026-05-29)"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Marginalizing Random Effect Levels}
  %\VignetteEncoding{UTF-8}
---

<style type="text/css">pre, pre code { white-space: pre-wrap !important; word-break: break-word; overflow-wrap: anywhere; }</style>



# Marginalizing Random Effects

One of the most common questions about multilevel models is how much influence grouping terms have
on the outcome. One way to explore this is to simulate the predicted values of an observation
across the distribution of random effects for a specific grouping variable and term. This can
be described as "marginalizing" predictions over the distribution of random effects. This allows
you to explore the influence of the grouping term and grouping levels on the outcome scale by
simulating predictions for simulated values of each observation across the distribution of
effect sizes.

The `REmargins()` function allows you to do this. Here, we take the example `sleepstudy` model and
marginalize predictions for all of the random effect terms (Subject:Intercept, Subject:Days). By
default, the function will marginalize over the *quartiles* of the expected rank (see expected
rank vignette) of the effect distribution for each term.


``` r
fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
mfx <- REmargins(merMod = fm1, newdata = sleepstudy[1:10,])
head(mfx)
```

```
#>   Reaction Days Subject case grouping_var      term breaks original_group_level
#> 1   249.56    0     309    1      Subject Intercept      1                  308
#> 2   249.56    0     334    1      Subject      Days      1                  308
#> 3   249.56    0     350    1      Subject Intercept      2                  308
#> 4   249.56    0     330    1      Subject      Days      2                  308
#> 5   249.56    0     308    1      Subject Intercept      3                  308
#> 6   249.56    0     332    1      Subject      Days      3                  308
#>   fit_combined upr_combined lwr_combined fit_Subject upr_Subject lwr_Subject
#> 1     211.9200     250.8137     173.6175  -37.958384   -2.680876   -76.85914
#> 2     245.6775     283.3831     207.7996   -6.238617   25.164606   -46.69958
#> 3     236.9896     275.3616     201.5464  -14.117588   21.020768   -51.19104
#> 4     273.9680     311.3997     236.0823   23.409186   59.360085   -13.18793
#> 5     254.5492     290.2708     218.4986    2.253723   35.755512   -32.79586
#> 6     263.1717     296.7733     225.2313    8.055660   43.478246   -27.57725
#>   fit_fixed upr_fixed lwr_fixed
#> 1  252.4671  288.5500  219.3783
#> 2  250.5605  287.2176  215.9889
#> 3  252.2250  290.5297  217.9524
#> 4  253.0412  287.3219  217.2943
#> 5  250.6383  284.8410  216.6404
#> 6  252.7486  286.7932  218.2970
```

The new data frame output from `REmargins` contains a lot of information. The first few columns
contain the original data passed to `newdata`. Each observation in `newdata` is identified by a
`case` number, because the function repeats each observation by the number of random effect terms
and number of breaks to simulate each term over. Then the `grouping_var`


# Summarizing




# Plotting

Finally - you can plot the results marginalization to evaluate the effect of the random effect terms
graphically.


``` r
ggplot(mfx) + aes(x = breaks, y = fit_Subject, group = case) +
  geom_line() +
  facet_wrap(~term)
```

![](mfx-mfxplot1-1.png)

---
title: "Prediction Intervals from merMod Objects"
author: "Jared Knowles and Carl Frederick"
date: "Written 2015-05-21 (last updated 2026-05-29)"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Prediction Intervals from merMod Objects}
  %\VignetteEncoding{UTF-8}
---

<style type="text/css">pre, pre code { white-space: pre-wrap !important; word-break: break-word; overflow-wrap: anywhere; }</style>



## Introduction

Fitting (generalized) linear mixed models, (G)LMM, to very large data sets is
becoming increasingly easy, but understanding and communicating the uncertainty
inherent in those models is not. As the documentation for
`lme4::predict.merMod()` notes:

> There is no option for computing standard errors of predictions because it is
> difficult to define an efficient method that incorporates uncertainty in the
> variance parameters; we recommend `lme4::bootMer()` for this task.

We agree that, short of a fully Bayesian analysis, bootstrapping is the
gold-standard for deriving a prediction interval predictions from a (G)LMM, but
the time required to obtain even a respectable number of replications from
`bootMer()` quickly becomes prohibitive when the initial model fit is on the
order of hours instead of seconds. The only other alternative we have identified
for these situations is to use the `arm::sim()` function to simulate values.
Unfortunately, this only takes variation of the fixed coefficients and residuals
into account, and assumes the conditional modes of the random effects are fixed.

We developed the `predictInterval()` function to incorporate the variation in
the conditional modes of the random effects (CMRE, a.k.a. BLUPs in the LMM case)
into calculating prediction intervals. Ignoring the variance in the CMRE results
in overly confident estimates of predicted values and in cases where the precision
of the grouping term varies across levels of grouping terms, creates the illusion
of difference where none may exist. The importance of accounting for this variance
comes into play sharply when comparing the predictions of different models across
observations.

We take the warning from `lme4::predict.merMod()`
seriously, but view this method as a decent first approximation the full bootstrap
analysis for (G)LMMs fit to very large data sets.

## Conceptual description

In order to generate a proper prediction interval, a prediction must account for
three sources of uncertainty in mixed models:

1. the residual (observation-level) variance,
2. the uncertainty in the fixed coefficients, and
3. the uncertainty in the variance parameters for the grouping factors.

A fourth, uncertainty about the data, is beyond the scope of any prediction method.

As we mentioned above, the `arm:sim()` function incorporates the first two sources
of variation but not the third , while bootstrapping using `lme4::bootMer()`
does incorporate all three sources of uncertainty because it re-estimates the
model using random samples of the data.

When inference about the values of the CMREs is of interest, it would be nice to
incorporate some degree of uncertainty in those estimates when comparing
observations across groups. `predictInterval()` does this by drawing values of
the CMREs from the conditional variance-covariance matrix of the random affects
accessible from `lme4::ranef(model, condVar=TRUE)`. Thus, `predictInterval()`
incorporates all of the uncertainty from sources one and two, and part of the
variance from source 3, but the variance parameters themselves are treated as
fixed.

To do this, `predictInterval()` takes an estimated model of class `merMod` and,
like `predict()`, a data.frame upon which to make those predictions and:

1. extracts the fixed and random coefficients
2. takes `n` draws from the multivariate normal distribution of the fixed and
random coefficients (separately)
3. calculates the linear predictor for each row in `newdata` based on these draws,
and
4. optionally incorporates the residual variation (per the `arm::sim()` function),
and,
5. returns newdata with the lower and upper limits of the prediction interval
and the mean or median of the simulated predictions

Currently, the supported model types are linear mixed models and mixed logistic
regression models.

The prediction data set *can* include levels that are not in the estimation model
frame. The prediction intervals for such observations only incorporate
uncertainty from fixed coefficient estimates and the residual level of variation.

## Comparison to existing methods

What do the differences between `predictInterval()` and the other methods for
constructing prediction intervals mean in practice? We would expect to see
`predictInterval()` to produce prediction intervals that are wider than all methods
except for the `bootMer()` method. We would also hope that the prediction point
estimate from other methods falls within the prediction interval produced by
`predictInterval()`. Ideally, the predicted point estimate produced by
`predictInterval()`  would fall close to that produced by `bootMer()`.

This section compares the results of `predictInterval()` with those obtained
using `arm::sim()` and `lme4::bootMer()` using the sleepstudy data from
`lme4`.  These data contain reaction time observations for 10 days on 18 subjects.
The data are sorted such that the first 10 observations are days one through
ten for subject 1, the next 10 are days one through ten for subject 2 and so on.
The example model that we are estimating below estimates random intercepts and a
random slope for the number of days.

### Step 1: Estimating the model and using `predictInterval()`

First, we will load the required packages and data and estimate the model:


``` r
set.seed(271828)
data(sleepstudy)
fm1 <- lmer(Reaction ~ Days + (Days|Subject), data=sleepstudy)
arm::display(fm1)
```

```
#> lmer(formula = Reaction ~ Days + (Days | Subject), data = sleepstudy)
#>             coef.est coef.se
#> (Intercept) 251.41     6.82 
#> Days         10.47     1.55 
#> 
#> Error terms:
#>  Groups   Name        Std.Dev. Corr 
#>  Subject  (Intercept) 24.74         
#>           Days         5.92    0.07 
#>  Residual             25.59         
#> ---
#> number of obs: 180, groups: Subject, 18
#> AIC = 1755.6, DIC = 1760.3
#> deviance = 1751.9
```

Then, calculate prediction intervals using `predictInterval()`. The `predictInterval`
function has a number of user configurable options. In this example, we use the
original data `sleepstudy` as the newdata. We pass the function the `fm1` model
we fit above. We also choose a 95% interval with `level = 0.95`, though
we could choose a less conservative prediction interval. We make 1,000 simulations
for each observation `n.sims = 1000`. We set the point estimate to be the median
of the simulated values, instead of the mean. We ask for the linear predictor back,
if we fit a logistic regression, we could have asked instead for our predictions
on the probability scale instead. Finally, we indicate that we want the predictions
to incorporate the residual variance from the model -- an option only available
for `lmerMod` objects.


``` r
PI.time <- system.time(
  PI <- predictInterval(merMod = fm1, newdata = sleepstudy,
                        level = 0.95, n.sims = 1000,
                        stat = "median", type="linear.prediction",
                        include.resid.var = TRUE)
)
```

Here is the first few rows of the object `PI`:



|      fit|      upr|      lwr|
|--------:|--------:|--------:|
| 251.6685| 311.3171| 196.4096|
| 271.4802| 330.9195| 214.2175|
| 292.6809| 350.9867| 237.7714|
| 311.6967| 369.2911| 254.2237|
| 331.8318| 389.7439| 278.1857|
| 350.7450| 408.1386| 294.8506|



The three columns are the median (`fit`) and limits of the 95% prediction
interval (`upr` and `lwr`) because we set `level=0.95`. The following figure
displays the output graphically for the first 30 observations.


``` r
library(ggplot2);
ggplot(aes(x=1:30, y=fit, ymin=lwr, ymax=upr), data=PI[1:30,]) +
  geom_point() +
  geom_linerange() +
  labs(x="Index", y="Prediction w/ 95% PI") + theme_bw()
```

<div class="figure" style="text-align: center">
<img src="usage-Inspect_predInt_2-1.png" alt=""  />
</div>

#### Step 1a: Adjusting for correlation between fixed and random effects

The prediction intervals above do not correct for correlations between fixed
and random effects. This tends to lead to predictive intervals that are too
conservative, especially for existing groups when there is a lot of data on
relatively few groups. In that case, a significant portion of the uncertainty
in the prediction can be due to variance in the fixed intercept which is
anti-correlated with variance in the random intercept effects. For instance,
it does not actually matter if the fixed intercept is 5 and the random
intercept effects are -2, 1, and 1, versus a fixed intercept of 6 and
random intercept effects of -3, 0, and 0. (The latter situation will never be
the MLE, but it can occur in this package's simulations.)

To show this issue, we'll use the sleep study model, predicting the reaction times
of subjects after experiencing sleep deprivation:


``` r
fm1 <- lmer(Reaction ~ Days + (Days|Subject), data=sleepstudy)
arm::display(fm1)
```

```
#> lmer(formula = Reaction ~ Days + (Days | Subject), data = sleepstudy)
#>             coef.est coef.se
#> (Intercept) 251.41     6.82 
#> Days         10.47     1.55 
#> 
#> Error terms:
#>  Groups   Name        Std.Dev. Corr 
#>  Subject  (Intercept) 24.74         
#>           Days         5.92    0.07 
#>  Residual             25.59         
#> ---
#> number of obs: 180, groups: Subject, 18
#> AIC = 1755.6, DIC = 1760.3
#> deviance = 1751.9
```

Let's use the model to give an interval for the true average body fat of a
large group of students like the first one in the study — a 196cm female
baseball player:


``` r
sleepstudy[1,]
```

```
#>   Reaction Days Subject
#> 1   249.56    0     308
```

``` r
predictInterval(fm1, sleepstudy[1,], include.resid.var=0) #predict the average body fat for a group of 196cm female baseball players
```

```
#>        fit      upr      lwr
#> 1 253.9977 270.7438 236.2829
```

There are two ways to get predictInterval to create less-conservative intervals
to deal with this. The first is just to tell it to consider certain fixed
effects as fully-known (that is, with an effectively 0 variance.) This is done
using the `ignore.fixed.effects` argument.


``` r
predictInterval(fm1, sleepstudy[1,], include.resid.var=0, ignore.fixed.terms = 1)
```

```
#>        fit      upr      lwr
#> 1 253.8537 268.5299 239.6275
```

``` r
# predict the average reaction time for a subject at day 0, taking the global intercept
# (mean reaction time) as fully known
predictInterval(fm1, sleepstudy[1,], include.resid.var=0, ignore.fixed.terms = "(Intercept)")
```

```
#>        fit      upr      lwr
#> 1 254.2354 269.3875 239.3116
```

``` r
#Same as above
predictInterval(fm1, sleepstudy[1,], include.resid.var=0, ignore.fixed.terms = 1:2)
```

```
#>        fit      upr     lwr
#> 1 253.6743 269.8776 237.984
```

``` r
# as above, taking the first two fixed effects (intercept and days effect) as fully known
```

The second way is to use an ad-hoc variance adjustment, with the
`fix.intercept.variance` argument. This takes the model's intercept variance
$\hat\sigma^2_\mu$ and adjusts it to:

$$\hat\sigma\prime^2_\mu = \hat\sigma^2_\mu-\Sigma_{levels}\frac{1}{\Sigma_{groups(level)}1/(\hat\sigma^2_{level}+sigma^2_{group})}$$

In other words, it assumes the given intercept variance incorporates spurious
variance for each level, where each of the spurious variance terms has a
precision equal to the of the precisions due to the individual groups at that
level.


``` r
predictInterval(fm1, sleepstudy[1,], include.resid.var=0,
                fix.intercept.variance = TRUE)
```

```
#>        fit      upr      lwr
#> 1 253.5872 268.8683 236.6639
```

``` r
# predict the average reaction time for a subject at day 0,, using an ad-hoc
# correction for the covariance of the intercept with the random intercept effects.
```

A few notes about these two arguments:

* `fix.intercept.variance=T` is redundant with `ignore.fixed.effects=1`, but not
vice versa.
* These corrections should NOT be used when predicting outcomes for groups not
present in the original data.

### Step 2: Comparison with `arm::sim()`

How does the output above compare to what we could get from `arm::sim()`?


``` r
PI.arm.time <- system.time(
  PI.arm.sims <- arm::sim(fm1, 1000)
)

PI.arm <- data.frame(
  fit=apply(fitted(PI.arm.sims, fm1), 1, function(x) quantile(x, 0.500)),
  upr=apply(fitted(PI.arm.sims, fm1), 1, function(x) quantile(x, 0.975)),
  lwr=apply(fitted(PI.arm.sims, fm1), 1, function(x) quantile(x, 0.025))
)

comp.data <- rbind(data.frame(Predict.Method="predictInterval()", x=(1:nrow(PI))-0.1, PI),
                   data.frame(Predict.Method="arm::sim()", x=(1:nrow(PI.arm))+0.1, PI.arm))

comp.data$Predict.Method <- factor(comp.data$Predict.Method, levels = unique(comp.data$Predict.Method))
ggplot(aes(x=x, y=fit, ymin=lwr, ymax=upr, color=Predict.Method), data=comp.data[c(1:30,181:210),]) +
  geom_point() +
  geom_linerange() +
  labs(x="Index", y="Prediction w/ 95% PI") +
  theme_bw() +  theme(legend.position="bottom") +
  scale_color_brewer(type = "qual", palette = 2)
```

<div class="figure" style="text-align: center">
<img src="usage-arm.Sim-1.png" alt=""  />
</div>

The prediction intervals from `arm:sim()` are much smaller and the random slope
for days vary more than they do for `predictInterval`. Both results are as
expected, given the small number of subjects and observations per subject in
these data. Because `predictInterval()` is incorporating uncertainty in the CMFEs
(but not the variance parameters of the random coefficients themselves), the
Days slopes are closer to the overall or pooled regression slope.

### Step 3: Comparison with `lme4::bootMer()`

As quoted above, the developers of lme4 suggest that users interested in
uncertainty estimates around their predictions use `lme4::bootmer()` to calculate
them. The documentation for `lme4::bootMer()` goes on to describe three
implemented flavors of bootstrapped estimates:

1. parametrically resampling both the *"spherical"* random effects
*u* and the i.i.d. errors $\epsilon$
2. treating the random effects as fixed and parametrically resampling the
i.i.d. errors
3. treating the random effects as fixed and semi-parametrically resampling
the i.i.d. errors from the distribution of residuals.

We will compare the results from `predictInterval()` with each method, in turn.

#### Step 3a: `lme4::bootMer()` method 1


``` r
##Functions for bootMer() and objects
####Return predicted values from bootstrap
mySumm <- function(.) {
  predict(., newdata=sleepstudy, re.form=NULL)
}
####Collapse bootstrap into median, 95% PI
sumBoot <- function(merBoot) {
  return(
    data.frame(fit = apply(merBoot$t, 2, function(x) as.numeric(quantile(x, probs=.5, na.rm=TRUE))),
               lwr = apply(merBoot$t, 2, function(x) as.numeric(quantile(x, probs=.025, na.rm=TRUE))),
               upr = apply(merBoot$t, 2, function(x) as.numeric(quantile(x, probs=.975, na.rm=TRUE)))
    )
  )
}

##lme4::bootMer() method 1
PI.boot1.time <- system.time(
  boot1 <- lme4::bootMer(fm1, mySumm, nsim=250, use.u=FALSE, type="parametric")
)

PI.boot1 <- sumBoot(boot1)

comp.data <- rbind(data.frame(Predict.Method="predictInterval()", x=(1:nrow(PI))-0.1, PI),
                   data.frame(Predict.Method="lme4::bootMer() - Method 1", x=(1:nrow(PI.boot1))+0.1, PI.boot1))

comp.data$Predict.Method <- factor(comp.data$Predict.Method, levels = unique(comp.data$Predict.Method))
ggplot(aes(x=x, y=fit, ymin=lwr, ymax=upr, color=Predict.Method), data=comp.data[c(1:30,181:210),]) +
  geom_point() +
  geom_linerange() +
  labs(x="Index", y="Prediction w/ 95% PI") +
  theme_bw() +  theme(legend.position="bottom") +
  scale_color_brewer(type = "qual", palette = 2)
```

<div class="figure" style="text-align: center">
<img src="usage-bootMer.1-1.png" alt=""  />
</div>

Method 1 (`use.u = FALSE`) re-simulates the random effects on every
bootstrap replicate instead of conditioning on the values estimated from the
data, so its point predictions are pulled toward the population average and do
*not* track individual subjects. For subjects whose intercept is far from the
mean (indices 11-30 above -- subjects 309 and 310, the lowest-reaction-time
subjects) the `bootMer` points therefore sit well above the subject-specific
predictions from `predictInterval()` (shown in green), which condition on the
estimated random effects. This divergence is expected: the two methods are
answering slightly different questions. The conditional comparisons in Methods
2 and 3 below, which hold the random effects fixed (`use.u = TRUE`), align
closely with `predictInterval()`.

#### Step 3b: `lme4::bootMer()` method 2


``` r
##lme4::bootMer() method 2
PI.boot2.time <- system.time(
  boot2 <- lme4::bootMer(fm1, mySumm, nsim=250, use.u=TRUE, type="parametric")
)

PI.boot2 <- sumBoot(boot2)

comp.data <- rbind(data.frame(Predict.Method="predictInterval()", x=(1:nrow(PI))-0.1, PI),
                   data.frame(Predict.Method="lme4::bootMer() - Method 2", x=(1:nrow(PI.boot2))+0.1, PI.boot2))

comp.data$Predict.Method <- factor(comp.data$Predict.Method, levels = unique(comp.data$Predict.Method))
ggplot(aes(x=x, y=fit, ymin=lwr, ymax=upr, color=Predict.Method), data=comp.data[c(1:30,181:210),]) +
  geom_point() +
  geom_linerange() +
  labs(x="Index", y="Prediction w/ 95% PI") +
  theme_bw() +  theme(legend.position="bottom") +
  scale_color_brewer(type = "qual", palette = 2)
```

<div class="figure" style="text-align: center">
<img src="usage-bootMer.2-1.png" alt=""  />
</div>

Here, the results for `predictInterval` in green again encompass the results from
`bootMer`, but are much wider. The `bootMer` estimates are ignoring the variance
in the group effects, and as such, are only incorporating the residual variance
and the variance in the fixed effects -- similar to the `arm::sim()` function.

#### Step 3c: `lme4::bootMer()` method 3


``` r
##lme4::bootMer() method 3
PI.boot3.time <- system.time(
  boot3 <- lme4::bootMer(fm1, mySumm, nsim=250, use.u=TRUE, type="semiparametric")
)

PI.boot3 <- sumBoot(boot3)

comp.data <- rbind(data.frame(Predict.Method="predictInterval()", x=(1:nrow(PI))-0.1, PI),
                   data.frame(Predict.Method="lme4::bootMer() - Method 3", x=(1:nrow(PI.boot3))+0.1, PI.boot3))

comp.data$Predict.Method <- factor(comp.data$Predict.Method, levels = unique(comp.data$Predict.Method))
ggplot(aes(x=x, y=fit, ymin=lwr, ymax=upr, color=Predict.Method), data=comp.data[c(1:30,181:210),]) +
  geom_point() +
  geom_linerange() +
  labs(x="Index", y="Prediction w/ 95% PI") +
  theme_bw() +  theme(legend.position="bottom") +
  scale_color_brewer(type = "qual", palette = 2)
```

<div class="figure" style="text-align: center">
<img src="usage-bootMer.3-1.png" alt=""  />
</div>

These results are virtually identical to those above.

#### Step 3d: Comparison to rstanarm





``` r
PI.time.stan <- system.time({
  fm_stan <- stan_lmer(Reaction ~ Days + (Days|Subject), data = sleepstudy,
                       verbose = FALSE, open_progress = FALSE, refresh = -1,
                       show_messages=FALSE, chains = 1)
  zed <- posterior_predict(fm_stan)
  PI.stan <- cbind(apply(zed, 2, median), central_intervals(zed, prob=0.95))
})
```

```
#> Chain 1: 
#> Chain 1: Gradient evaluation took 5.5e-05 seconds
#> Chain 1: 1000 transitions using 10 leapfrog steps per transition would take 0.55 seconds.
#> Chain 1: Adjust your expectations accordingly!
#> Chain 1: 
#> Chain 1: 
#> Chain 1: 
#> Chain 1:  Elapsed Time: 4.058 seconds (Warm-up)
#> Chain 1:                1.222 seconds (Sampling)
#> Chain 1:                5.28 seconds (Total)
#> Chain 1:
```

``` r
print(fm_stan)
```

```
#> stan_lmer
#>  family:       gaussian [identity]
#>  formula:      Reaction ~ Days + (Days | Subject)
#>  observations: 180
#> ------
#>             Median MAD_SD
#> (Intercept) 251.6    6.7 
#> Days         10.6    1.6 
#> 
#> Auxiliary parameter(s):
#>       Median MAD_SD
#> sigma 25.9    1.6  
#> 
#> Error terms:
#>  Groups   Name        Std.Dev. Corr 
#>  Subject  (Intercept) 24.3          
#>           Days         6.8     0.06 
#>  Residual             26.0          
#> Num. levels: Subject 18 
#> 
#> ------
#> * For help interpreting the printed output see ?print.stanreg
#> * For info on the priors used see ?prior_summary.stanreg
```

``` r
PI.stan <- as.data.frame(PI.stan)
names(PI.stan) <- c("fit", "lwr", "upr")
PI.stan <- PI.stan[, c("fit", "upr", "lwr")]
comp.data <- rbind(data.frame(Predict.Method="predictInterval()", x=(1:nrow(PI))-0.1, PI),
                   data.frame(Predict.Method="rstanArm", x=(1:nrow(PI.stan))+0.1, PI.stan))

comp.data$Predict.Method <- factor(comp.data$Predict.Method, levels = unique(comp.data$Predict.Method))
ggplot(aes(x=x, y=fit, ymin=lwr, ymax=upr, color=Predict.Method), data=comp.data[c(1:30,181:210),]) +
  geom_point() +
  geom_linerange() +
  labs(x="Index", y="Prediction w/ 95% PI") +
  theme_bw() +  theme(legend.position="bottom") +
  scale_color_brewer(type = "qual", palette = 2)
```

<div class="figure" style="text-align: center">
<img src="usage-stancomp-1.png" alt=""  />
</div>


### Computation time

Our initial motivation for writing this function was to develop a method for
incorporating uncertainty in the CMFEs  for mixed models estimated on very large
samples. Even for models with only modest degrees of complexity, using
`lme4::bootMer()` quickly becomes time prohibitive because it involves
re-estimating the model for each simulation. We have seen how each alternative
compares to `predictInterval()` substantively, but how do they compare in terms
of computational time? The table below lists the output of `system.time()` for
all five methods for calculating prediction intervals for `merMod` objects.



|                         | user.self| sys.self| elapsed|
|:------------------------|---------:|--------:|-------:|
|predictInterval()        |     0.196|    0.000|   0.196|
|arm::sim()               |     0.386|    0.000|   0.386|
|lme4::bootMer()-Method 1 |     3.886|    0.000|   3.889|
|lme4::bootMer()-Method 2 |     3.699|    0.000|   3.700|
|lme4::bootMer()-Method 3 |     3.711|    0.000|   3.711|
|rstanarm:predict         |     5.765|    0.009|   5.781|



For this simple example, we see that `arm::sim()` is the fastest--nearly five
times faster than `predictInterval()`. However, `predictInterval()` is nearly
six times faster than any of the bootstrapping options via `lme4::bootMer`.  This
may not seem like a lot, but consider that the computational time for required
for bootstrapping is roughly proportional to the number of bootstrapped simulations
requested ... `predictInterval()` is not because it is just a series of draws from
various multivariate normal distributions, so the time ratios in the table below
represents the lowest bound of the computation time ratio of bootstrapping
to `predictInterval()`.

## Simulation

TBC.

---
title: "Validating predictInterval() against brms"
author: "Jared Knowles"
date: "Written 2026-05-29 (last updated 2026-05-29)"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Validating predictInterval() against brms}
  %\VignetteEncoding{UTF-8}
---

<style type="text/css">pre, pre code { white-space: pre-wrap !important; word-break: break-word; overflow-wrap: anywhere; }</style>



## Why compare to brms?

`merTools::predictInterval()` was written to put prediction intervals on
mixed-effect models that are too large or too slow to bootstrap. It does this by
simulating from the estimated distribution of the fixed and random effects --
fast, but an approximation. Since then, [brms](https://paulbuerkner.com/brms/)
has become a standard for *fully* Bayesian multilevel modeling, and its
posterior predictive distribution is about as close to a gold standard as we
have. This vignette asks a simple question: **how close are
`predictInterval()`'s intervals to the ones brms produces?**

We look at three things: a Gaussian random-slopes model, what happens for an
entirely unobserved group, and a binomial GLMM. This vignette is *precompiled* --
reproducing it requires the `brms` package and a working Stan toolchain.

## A random-slopes model

We use the bundled `hsb` data (7,185 students in 160 schools) and model math
achievement from student SES, school-mean SES, and a school-varying SES slope.
We hold out 20% of students for testing, plus six whole schools to stand in for
"new groups" later.


``` r
data(hsb)
hsb$schid <- factor(hsb$schid)
new_schools <- sample(levels(hsb$schid), 6)
hsb$.set <- "train"
hsb$.set[hsb$schid %in% new_schools] <- "test_new"
seen <- which(hsb$.set == "train")
hsb$.set[sample(seen, round(0.2 * length(seen)))] <- "test_seen"
train     <- droplevels(hsb[hsb$.set == "train", ])
test_seen <- hsb[hsb$.set == "test_seen" & hsb$schid %in% levels(train$schid), ]
test_new  <- hsb[hsb$.set == "test_new", ]

f1 <- mathach ~ ses + meanses + (ses | schid)
m_lme <- lmer(f1, data = train)
brms_fit_seconds <- system.time(
  m_brm <- fit_brm(f1, train, gaussian(), "brms_hsb_meanses"))[["elapsed"]]
```

We draw a predictive distribution for each held-out student from each method --
`predictInterval(returnSims = TRUE)` and `brms::posterior_predict()` -- and
compute everything (point estimates, intervals, coverage) from those draws.


``` r
mt_seen <- mt_draws(m_lme, test_seen)
pp_seen <- posterior_predict(m_brm, newdata = test_seen, allow_new_levels = TRUE)
```

### Point estimates are essentially identical

Both methods condition on the same estimated fixed effects and school BLUPs, so
the conditional-mean predictions should agree almost exactly -- and they do.


``` r
lme_mean <- predict(m_lme, newdata = test_seen)
brm_mean <- colMeans(posterior_epred(m_brm, newdata = test_seen, allow_new_levels = TRUE))
c(correlation = cor(lme_mean, brm_mean),
  mean_abs_diff = mean(abs(lme_mean - brm_mean)),
  response_sd = sd(test_seen$mathach))
```

```
#>   correlation mean_abs_diff   response_sd 
#>    0.99984045    0.03670134    6.71745425
```

``` r
ggplot(data.frame(merTools = lme_mean, brms = brm_mean), aes(merTools, brms)) +
  geom_abline(slope = 1, intercept = 0, linetype = 2, color = "grey50") +
  geom_point(alpha = 0.3, size = 0.8, color = "#1B9E77") + coord_equal() +
  labs(x = "lme4 predict()", y = "brms posterior_epred()") +
  theme_minimal(base_size = 12)
```

<div class="figure" style="text-align: center">
<img src="brms-point-1.png" alt=""  />
</div>

### The prediction intervals agree


``` r
b_mt <- qband(mt_seen, 0.90); b_brm <- qband(pp_seen, 0.90)
c(width_merTools = median(b_mt$upr - b_mt$lwr),
  width_brms = median(b_brm$upr - b_brm$lwr),
  cor_lower = cor(b_mt$lwr, b_brm$lwr),
  cor_upper = cor(b_mt$upr, b_brm$upr))
```

```
#> width_merTools     width_brms      cor_lower      cor_upper 
#>     20.1789145     20.2344075      0.9901934      0.9899334
```

``` r
endpts <- rbind(data.frame(bound = "lower", merTools = b_mt$lwr, brms = b_brm$lwr),
                data.frame(bound = "upper", merTools = b_mt$upr, brms = b_brm$upr))
ggplot(endpts, aes(merTools, brms, color = bound)) +
  geom_abline(slope = 1, intercept = 0, linetype = 2, color = "grey50") +
  geom_point(alpha = 0.3, size = 0.8) + coord_equal() +
  labs(title = "90% prediction-interval endpoints", x = "merTools", y = "brms") +
  theme_minimal(base_size = 12)
```

<div class="figure" style="text-align: center">
<img src="brms-intervals-1.png" alt=""  />
</div>

### And they are equally well calibrated

The real test is out-of-sample: do nominal intervals cover held-out scores at
the nominal rate? Both methods track the diagonal, and each other.


``` r
data.frame(nominal = NOMINAL,
           merTools = coverage(mt_seen, test_seen$mathach, NOMINAL),
           brms     = coverage(pp_seen, test_seen$mathach, NOMINAL))
```

```
#>   nominal  merTools      brms
#> 1    0.50 0.4569776 0.4598698
#> 2    0.80 0.7895879 0.7932032
#> 3    0.90 0.9168474 0.9168474
#> 4    0.95 0.9660159 0.9703543
```

### At a fraction of the cost


``` r
t_lme <- system.time(lmer(f1, data = train))[["elapsed"]]
t_mt  <- system.time(mt_draws(m_lme, test_seen))[["elapsed"]]
t_pp  <- system.time(posterior_predict(m_brm, newdata = test_seen,
                                       allow_new_levels = TRUE))[["elapsed"]]
data.frame(  # brms_fit_seconds was captured when the model was fit, above
  step = c("lme4 fit", "predictInterval", "brms fit (compile+sample)", "posterior_predict"),
  seconds = round(c(t_lme, t_mt, brms_fit_seconds, t_pp), 2))
```

```
#>                        step seconds
#> 1                  lme4 fit    0.05
#> 2           predictInterval    0.53
#> 3 brms fit (compile+sample)  398.97
#> 4         posterior_predict    1.52
```

For this model, the entire lme4 + `predictInterval()` path runs in about a
second, while a single brms fit takes several minutes -- a couple of orders of
magnitude difference, before the (much larger) models `predictInterval()` was
really built for.

## What about an entirely new group?

For a school that was never seen during fitting there is a genuine modeling
choice: do you *drop* the school's effect and predict from the fixed effects
only, or *sample* a plausible effect for it from the estimated school-level
distribution? Both packages support either, and the comparison is only fair when
the choices match:

| treatment of an unseen group | `predictInterval()` | brms |
|---|---|---|
| drop the effect (population prediction) | `new.levels = "zero"` *(default)* | `re_formula = NA` |
| sample a new effect | `new.levels = "draw"` | `allow_new_levels = TRUE` *(default)* |

Matched up, the two tools agree on both the width and the coverage of the
new-school intervals:


``` r
nd <- test_new
pp_pop  <- posterior_predict(m_brm, newdata = nd, re_formula = NA)
pp_draw <- posterior_predict(m_brm, newdata = nd, allow_new_levels = TRUE)
mt_zero <- mt_draws(m_lme, nd, new.levels = "zero")
mt_draw <- mt_draws(m_lme, nd, new.levels = "draw")
w <- function(d) median(qband(d, 0.90)$upr - qband(d, 0.90)$lwr)
data.frame(
  unseen_group   = c("drop effect", "sample effect"),
  merTools_width = c(w(mt_zero), w(mt_draw)),
  brms_width     = c(w(pp_pop),  w(pp_draw)),
  merTools_cov90 = c(coverage(mt_zero, nd$mathach, 0.90), coverage(mt_draw, nd$mathach, 0.90)),
  brms_cov90     = c(coverage(pp_pop,  nd$mathach, 0.90), coverage(pp_draw, nd$mathach, 0.90)))
```

```
#>    unseen_group merTools_width brms_width merTools_cov90 brms_cov90
#> 1   drop effect       19.96801   20.02855      0.8708487  0.8745387
#> 2 sample effect       20.67910   20.76924      0.8892989  0.8856089
```

So the interval width for an unseen group is set by the *choice*, not the method.
Earlier I compared `predictInterval()`'s default against brms's default, which is
an apples-to-oranges pairing -- the apparent "gap" was the two packages making
different default assumptions, not a deficiency in either.

How much the choice matters depends on how much group variance is left after the
fixed effects. Here `meanses` already absorbs most of the between-school
variation, so dropping vs. sampling the school effect barely changes the width.
Remove `meanses` and the school SD roughly doubles, so the two options diverge:


``` r
m0 <- lmer(mathach ~ ses + (ses | schid), data = train)
data.frame(
  model = c("with meanses", "without meanses"),
  school_SD = c(attr(VarCorr(m_lme)$schid, "stddev")[["(Intercept)"]],
                attr(VarCorr(m0)$schid,    "stddev")[["(Intercept)"]]),
  width_zero = c(w(mt_draws(m_lme, nd, new.levels = "zero")),
                 w(mt_draws(m0,    nd, new.levels = "zero"))),
  width_draw = c(w(mt_draws(m_lme, nd, new.levels = "draw")),
                 w(mt_draws(m0,    nd, new.levels = "draw"))))
```

```
#>             model school_SD width_zero width_draw
#> 1    with meanses  1.598503   19.96801   20.67910
#> 2 without meanses  2.125770   19.94993   21.22576
```

The practical guidance: `predictInterval()`'s default (`new.levels = "zero"`)
answers "what is the expected outcome for a typical member of an unseen group,"
while `new.levels = "draw"` answers "what is the outcome for an unseen group
drawn from the population" -- the same quantity brms returns by default. Pick the
one that matches your question.

## A generalized linear mixed model

The story carries over to GLMMs. We fit a binomial model for whether grouse had
ticks and compare predicted probabilities -- `predictInterval(type =
"probability")` against `brms::posterior_epred()`.


``` r
data(grouseticks)
grouseticks$TICKS_BIN <- as.integer(grouseticks$TICKS >= 1)
grouseticks$cHEIGHT   <- as.numeric(scale(grouseticks$HEIGHT))
gk <- grouseticks
gk$.set <- ifelse(runif(nrow(gk)) < 0.2, "test", "train")
gk_tr <- droplevels(gk[gk$.set == "train", ])
gk_te <- gk[gk$.set == "test" & gk$BROOD %in% levels(gk_tr$BROOD) &
            gk$LOCATION %in% unique(gk_tr$LOCATION), ]

gf <- TICKS_BIN ~ YEAR + cHEIGHT + (1 | BROOD) + (1 | LOCATION)
g_lme <- glmer(gf, family = binomial, data = gk_tr,
               control = glmerControl(optimizer = "bobyqa"))
g_brm <- fit_brm(gf, gk_tr, bernoulli(), "brms_grouseticks")

mt_pr <- mt_draws(g_lme, gk_te, type = "probability", include.resid.var = FALSE)
if (max(mt_pr) > 1 || min(mt_pr) < 0) mt_pr <- plogis(mt_pr)
brm_pr <- posterior_epred(g_brm, newdata = gk_te, allow_new_levels = TRUE)
mt_p <- apply(mt_pr, 2, median); brm_p <- apply(brm_pr, 2, median)
c(correlation = cor(mt_p, brm_p), mean_abs_diff = mean(abs(mt_p - brm_p)))
```

```
#>   correlation mean_abs_diff 
#>    0.99488587    0.02704796
```

``` r
ggplot(data.frame(merTools = mt_p, brms = brm_p), aes(merTools, brms)) +
  geom_abline(slope = 1, intercept = 0, linetype = 2, color = "grey50") +
  geom_point(alpha = 0.4, color = "#7570B3") + coord_equal(xlim = c(0, 1), ylim = c(0, 1)) +
  labs(title = "Predicted P(tick): merTools vs brms",
       x = "predictInterval(type = 'probability')", y = "posterior_epred()") +
  theme_minimal(base_size = 12)
```

<div class="figure" style="text-align: center">
<img src="brms-glmm-1.png" alt=""  />
</div>

## Takeaways

- For **observed groups**, `predictInterval()` reproduces brms's point estimates
  and prediction intervals almost exactly, with the same out-of-sample coverage,
  at a tiny fraction of the computational cost.
- For **new groups**, you choose how to treat the unseen effect:
  `new.levels = "zero"` (the default) drops it and predicts from the fixed
  effects, while `new.levels = "draw"` samples it from the estimated covariance
  -- matching brms's `re_formula = NA` and `allow_new_levels = TRUE`
  respectively. Either way the two packages agree; pick the convention that fits
  your question.
- The same agreement holds for **GLMMs** on the probability scale.

In short, when bootstrapping or full MCMC is impractical, `predictInterval()` is
a fast, well-calibrated stand-in for the cases it was designed to handle.

---
title: "An Introduction to merTools"
author: "Jared Knowles and Carl Frederick"
date: "2020-06-22"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{An Introduction to merTools}
  %\VignetteEncoding{UTF-8}
---



## Introduction

Working with generalized linear mixed models (GLMM) and linear mixed models (LMM)
has become increasingly easy with the advances in the `lme4` package recently.
As we have found ourselves using these models more and more within our work, we,
the authors, have developed a set of tools for simplifying and speeding up common
tasks for interacting with `merMod` objects from `lme4`. This package provides
those tools.

## Illustrating Model Effects

As the complexity of the model fit grows, it becomes harder and harder to
interpret the substantive effect of parameters in the model.

Let's start with a medium-sized example model using the `InstEval` data provided
by the `lme4` package. These data represent university lecture evaluations at
ETH Zurich made by students. In this data, `s` is an individual student,
`d` is an individual lecturer, `studage` is the semester the student
is enrolled, `lectage` is how many semesters back the lecture with the
rating took place, `dept` is the department of the lecture, and `y` is an integer
1:5 representing the ratings of the lecture from "poor" to "very good":


```r
library(lme4)
head(InstEval)
#>   s    d studage lectage service dept y
#> 1 1 1002       2       2       0    2 5
#> 2 1 1050       2       1       1    6 2
#> 3 1 1582       2       2       0    2 5
#> 4 1 2050       2       2       1    3 3
#> 5 2  115       2       1       0    5 2
#> 6 2  756       2       1       0    5 4
str(InstEval)
#> 'data.frame':	73421 obs. of  7 variables:
#>  $ s      : Factor w/ 2972 levels "1","2","3","4",..: 1 1 1 1 2 2 3 3 3 3 ...
#>  $ d      : Factor w/ 1128 levels "1","6","7","8",..: 525 560 832 1068 62 406 3 6 19 75 ...
#>  $ studage: Ord.factor w/ 4 levels "2"<"4"<"6"<"8": 1 1 1 1 1 1 1 1 1 1 ...
#>  $ lectage: Ord.factor w/ 6 levels "1"<"2"<"3"<"4"<..: 2 1 2 2 1 1 1 1 1 1 ...
#>  $ service: Factor w/ 2 levels "0","1": 1 2 1 2 1 1 2 1 1 1 ...
#>  $ dept   : Factor w/ 14 levels "15","5","10",..: 14 5 14 12 2 2 13 3 3 3 ...
#>  $ y      : int  5 2 5 3 2 4 4 5 5 4 ...
```

Starting with a simple model:


```r
m1 <- lmer(y ~ service + lectage + studage + (1|d) + (1|s), data=InstEval)
```

After fitting the model we can make use of the first function provided by
`merTools`, `fastdisp` which modifies the function `arm:::display` to more
quickly display a summary of the model without calculating the model sigma:


```r
library(merTools)
fastdisp(m1)
#> lmer(formula = y ~ service + lectage + studage + (1 | d) + (1 | 
#>     s), data = InstEval)
#>             coef.est coef.se
#> (Intercept)  3.22     0.02  
#> service1    -0.07     0.01  
#> lectage.L   -0.19     0.02  
#> lectage.Q    0.02     0.01  
#> lectage.C   -0.02     0.01  
#> lectage^4   -0.02     0.01  
#> lectage^5   -0.04     0.02  
#> studage.L    0.10     0.02  
#> studage.Q    0.01     0.02  
#> studage.C    0.02     0.02  
#> 
#> Error terms:
#>  Groups   Name        Std.Dev.
#>  s        (Intercept) 0.33    
#>  d        (Intercept) 0.52    
#>  Residual             1.18    
#> ---
#> number of obs: 73421, groups: s, 2972; d, 1128
#> AIC = 237655
```

We see some interesting effects. First, our decision to include student and
lecturer effects seems justified as there is substantial variance within these
groups. Second, there do appear to be some effects by age and for lectures
given as a service by an outside lecturer. Let's look at these in more detail.
One way to do this would be to plot the coefficients together in a line to
see which deviate from 0 and in what direction. To get a confidence interval
for our fixed effect coefficients we have a number of options that represent a
tradeoff between coverage and computation time -- see `confint.merMod` for
details.

An alternative is to simulate values of the fixed effects from the posterior
using the function `arm::sim`. Our next tool, `FEsim`, is a convenience wrapper
to do this and provide an informative data frame of the results.


```r
feEx <- FEsim(m1, 1000)
cbind(feEx[,1] , round(feEx[, 2:4], 3))
#>      feEx[, 1]   mean median    sd
#> 1  (Intercept)  3.225  3.225 0.020
#> 2     service1 -0.070 -0.070 0.013
#> 3    lectage.L -0.186 -0.186 0.017
#> 4    lectage.Q  0.024  0.024 0.012
#> 5    lectage.C -0.025 -0.025 0.013
#> 6    lectage^4 -0.020 -0.019 0.014
#> 7    lectage^5 -0.039 -0.039 0.015
#> 8    studage.L  0.096  0.096 0.018
#> 9    studage.Q  0.005  0.005 0.017
#> 10   studage.C  0.017  0.017 0.016
```

We can present these results graphically, using `ggplot2`:


```r
library(ggplot2)
ggplot(feEx[feEx$term!= "(Intercept)", ]) +
  aes(x = term, ymin = median - 1.96 * sd,
      ymax = median + 1.96 * sd, y = median) +
  geom_pointrange() +
  geom_hline(yintercept = 0, size = I(1.1), color = I("red")) +
  coord_flip() +
  theme_bw() + labs(title = "Coefficient Plot of InstEval Model",
                    x = "Median Effect Estimate", y = "Evaluation Rating")
```

![plot of chunk fixeffplot](mertoolsIntro-fixeffplot-1.png)

However, an easier option is:


```r
plotFEsim(feEx) +
  theme_bw() + labs(title = "Coefficient Plot of InstEval Model",
                    x = "Median Effect Estimate", y = "Evaluation Rating")
```

![plot of chunk quickFEplot](mertoolsIntro-quickFEplot-1.png)

## Random Effects

Next, we might be interested in exploring the random effects. Again, we create a
dataframe of the values of the simulation of these effects for the individual
levels.


```r
reEx <- REsim(m1)
head(reEx)
#>   groupFctr groupID        term        mean      median        sd
#> 1         s       1 (Intercept)  0.18042888  0.21906223 0.3145710
#> 2         s       2 (Intercept) -0.07034954 -0.06339508 0.2972897
#> 3         s       3 (Intercept)  0.32105622  0.33625741 0.3187445
#> 4         s       4 (Intercept)  0.23713963  0.23271723 0.2761635
#> 5         s       5 (Intercept)  0.02613185  0.02878794 0.3054642
#> 6         s       6 (Intercept)  0.10806580  0.11082677 0.2429651
```

The result is a dataframe with estimates of the values of each of the random
effects provided by the `arm::sim()` function. *groupID* represents the identfiable
level for the variable for one random effect, *term* represents whether the
simulated values are for an intercept or which slope, and *groupFctr* identifies
which of the `(1|x)` terms the values represent. To make unique identifiers
for each term, we need to use both the `groupID` and the `groupFctr` term in case
these two variables use overlapping label names for their groups. In this case:


```r
table(reEx$term)
#> 
#> (Intercept) 
#>        4100
table(reEx$groupFctr)
#> 
#>    d    s 
#> 1128 2972
```

Most important is producing caterpillar or dotplots of these terms to explore
their variation. This is easily accomplished with the `dotplot` function:


```r
lattice::dotplot(ranef(m1, condVar=TRUE))
```

However, these graphics do not provide much control over the results. Instead,
we can use the `plotREsim` function in `merTools` to gain more control over
plotting of the random effect simulations.


```r
p1 <- plotREsim(reEx)
p1
```

![plot of chunk refplot1](mertoolsIntro-refplot1-1.png)

The result is a ggplot2 object which can be modified however the user sees
fit. Here, we've established that most student and professor effects are
indistinguishable from zero, but there do exist extreme outliers with both
high and low averages that need to be accounted for.

## Subtantive Effects

A logical next line of questioning is to see how much of the variation in a
rating can be caused by changing the student rater and how much is due to the
fixed effects we identified above. This is a very difficult problem to solve,
but using simulation we can examine the model behavior under a range of scenarios
to understand how the model is reflecting changes in the data. To do this,
we use another set of functions available in `merTools`.

The simplest option is to pick an observation at random and then modify its
values deliberately to see how the prediction changes in response. `merTools`
makes this task very simple:


```r
example1 <- draw(m1, type = 'random')
head(example1)
#>       y service lectage studage   d    s
#> 29762 1       0       1       4 403 1208
```

The `draw` function takes a random observation from the data in the model
and extracts it as a dataframe. We can now do a number of operations to this
observation:


```r
# predict it
predict(m1, newdata = example1)
#>    29762 
#> 3.742122
# change values
example1$service <- "1"
predict(m1, newdata = example1)
#>    29762 
#> 3.671278
```

More interesting, let's programatically modify this observation to see how the
predicted value changes if we hold everything but one variable constant.


```r
example2 <- wiggle(example1, varlist = "lectage",
          valueslist = list(c("1", "2", "3", "4", "5", "6")))

example2
#>        y service lectage studage   d    s
#> 29762  1       1       1       4 403 1208
#> 297621 1       1       2       4 403 1208
#> 297622 1       1       3       4 403 1208
#> 297623 1       1       4       4 403 1208
#> 297624 1       1       5       4 403 1208
#> 297625 1       1       6       4 403 1208
```

The function `wiggle` allows us to create a new dataframe with copies of the
variable that modify just one value. Chaining together `wiggle` calls, we
can see how the variable behaves under a number of different scenarios
simultaneously.


```r
example2$yhat <- predict(m1, newdata = example2)

ggplot(example2, aes(x = lectage, y = yhat)) + geom_line(aes(group = 1)) +
  theme_bw() + ylim(c(1, 5)) +
  geom_hline(yintercept = mean(InstEval$y), linetype = 2) +
  geom_hline(yintercept = mean(InstEval$y) + sd(InstEval$y), linetype = 3) +
  geom_hline(yintercept = mean(InstEval$y) - sd(InstEval$y), linetype = 3)
```

![plot of chunk predictplotwiggle](mertoolsIntro-predictplotwiggle-1.png)

The result allows us to graphically display the effect of each level of `lectage`
on an observation that is otherwise identical. This is plotted here against a
horizontal line representing the mean of the observed ratings, and two finer
lines showing plus or minus one standard deviation of the mean.

This is nice, but selecting a random observation is not very satisfying as it
may not be very meaningful. To address this, we can instead take the average
observation:


```r
example3 <- draw(m1, type = 'average')
example3
#>          y service lectage studage    d    s
#> 1 3.205745       0       1       6 1510 2237
```

Here, the average observation is identified based on either the modal observation
for factors or on the mean for numeric variables. Then, the random effect
terms are set to the level equivalent to the median effect -- very close
to 0.


```r
example3 <- wiggle(example1, varlist = "service",
          valueslist = list(c("0", "1")))
example3$yhat <- predict(m1, newdata = example3)

ggplot(example3, aes(x = service, y = yhat)) + geom_line(aes(group = 1)) +
  theme_bw() + ylim(c(1, 5)) +
  geom_hline(yintercept = mean(InstEval$y), linetype = 2) +
  geom_hline(yintercept = mean(InstEval$y) + sd(InstEval$y), linetype = 3) +
  geom_hline(yintercept = mean(InstEval$y) - sd(InstEval$y), linetype = 3)
```

![plot of chunk wiggle2](mertoolsIntro-wiggle2-1.png)

Here we can see that for the average observation, whether the lecture is outside
of the home department has a very slight negative effect on the overall rating.
Might the individual professor or student have more of an impact on the overall
rating? To answer this question we need to wiggle the same observation across
a wide range of student or lecturer effects.

How do we identify this range? `merTools` provides the `REquantile` function
which helps to identify which levels of the grouping terms correspond to which
quantile of the magnitude of the random effects:


```r
REquantile(m1, quantile = 0.25, groupFctr = "s")
#> [1] "446"
REquantile(m1, quantile = 0.25, groupFctr = "d")
#> [1] "18"
```

Here we can see that group level 446
corresponds to the 25th percentile of the effect for the student groups, and level
`REquantile(m1, quantile = 0.25, groupFctr = "d")` corresponds to the 25th
percentile for the instructor group. Using this information we can reassign
a specific observation to varying magnitudes of grouping term effects to see
how much they might influence our final prediction.


```r
example4 <- draw(m1, type = 'average')
example4 <- wiggle(example4, varlist = "s",
                      list(REquantile(m1, quantile = seq(0.1, 0.9, .1),
                                 groupFctr = "s")))

example4$yhat <- predict(m1, newdata = example4)

ggplot(example4, aes(x = reorder(s, -yhat), y = yhat)) +
  geom_line(aes(group = 1)) +
  theme_bw() + ylim(c(1, 5)) +
  geom_hline(yintercept = mean(InstEval$y), linetype = 2) +
  geom_hline(yintercept = mean(InstEval$y) + sd(InstEval$y), linetype = 3) +
  geom_hline(yintercept = mean(InstEval$y) - sd(InstEval$y), linetype = 3)
```

![plot of chunk wiggleanddraw](mertoolsIntro-wiggleanddraw-1.png)

This figure is very interesting because it shows that moving across the range
of student effects can have a larger impact on the score than the fixed effects
we observed above. That is, getting a "generous" or a "stingy" rater can have
a substantial impact on the final rating.

But, we can do even better. First, we can move beyond the average observation
by taking advantage of the `varList` option to the function which allows us to
specify a subset of the data to compute an average for.


```r
subExample <- list(studage = "2", lectage = "4")
example5 <- draw(m1, type = 'average', varList = subExample)
example5
#>          y service lectage studage    d    s
#> 1 3.087193       0       4       2 1510 2237
```

Now we have the average observation with a student age of 2 and a lecture age
of 4. We can then follow the same procedure as before to explore the effects
on our subsamples. Before we do that, let's fit a slightly more complex model
that includes a random slope.


```r
data(VerbAgg)
m2 <- glmer(r2 ~ Anger + Gender + btype + situ +
                (1|id) + (1 + Gender|item), family = binomial,
              data = VerbAgg)

example6 <- draw(m2, type = 'average', varList = list("id" = "149"))
example6$btype <- "scold"
example6$situ <- "self"


tempdf <- wiggle(example6, varlist = "Gender", list(c("M", "F")))
tempdf <- wiggle(tempdf, varlist = "item",
                    list(unique(VerbAgg$item)))
tempdf$yhat <- predict(m2, newdata = tempdf, type = "response",
                       allow.new.levels = TRUE)

ggplot(tempdf, aes(x = item, y = yhat, group = Gender)) +
  geom_line(aes(color = Gender))+
  theme_bw() + ylim(c(0, 1)) +
  theme(axis.text.x = element_text(angle = 20, hjust=1),
        legend.position = "bottom") + labs(x = "Item", y = "Probability")
```

![plot of chunk wigglesubsamples](mertoolsIntro-wigglesubsamples-1.png)

Here we've shown that the effect of both the intercept and the gender slope on
item simultaneously affect our predicted value. This results in the two lines
for predicted values across the items not being parallel. While we can see this
by looking at the results of the summary of the model object, using `fastdisp`
in the `merTools` package for larger models, it is not intuitive what that
effect looks like across different scenarios. `merTools` has given us the
machinery to investigate this.

## Uncertainty

The above examples make use of simulation to show the model behavior after
changing some values in a dataset. However, until now, we've focused on using
point estimates to represent these changes. The use of predicted point estimates
without incorporating any uncertainty can lead to overconfidence in the precision
of the model.

In the `predictInterval` function, discussed in more detail in another package
vignette, we provide a way to incorporate three out of the four types of
uncertainty inherent in a model. These are:

1. Overall model uncertainty
2. Uncertainty in fixed effect values
3. Uncertainty in random effect values
4. Uncertainty in the distribution of the random effects

1-3 are incorporated in the results of `predictInterval`, while capturing 4
would require making use of the `bootMer` function -- options discussed in
greater detail elsewhere. The main advantage of `predictInterval` is that it is
fast. By leveraging the power of the `arm::sim()` function, we are able to
generate prediction intervals for individual observations from very large models
very quickly. And, it works a lot like `predict`:


```r
exampPreds <- predictInterval(m2, newdata = tempdf,
                              type = "probability", level = 0.8)

tempdf <- cbind(tempdf, exampPreds)

ggplot(tempdf, aes(x = item, y = fit, ymin = lwr, ymax = upr,
                   group = Gender)) +
  geom_ribbon(aes(fill = Gender), alpha = I(0.2), color = I("black"))+
  theme_bw() + ylim(c(0, 1)) +
  theme(axis.text.x = element_text(angle = 20),
        legend.position = "bottom")+ labs(x = "Item", y = "Probability")
```

![plot of chunk speedexample](mertoolsIntro-speedexample-1.png)

Here we can see there is barely any gender difference in terms of area of
potential prediction intervals. However, by default, this approach includes the
residual variance of the model. If we instead focus just on the uncertainty of
the random and fixed effects, we get:


```r
exampPreds <- predictInterval(m2, newdata = tempdf,
                              type = "probability",
                              include.resid.var = FALSE, level = 0.8)
tempdf <- cbind(tempdf[, 1:8], exampPreds)

ggplot(tempdf, aes(x = item, y = fit, ymin = lwr, ymax = upr,
                   group = Gender)) +
  geom_ribbon(aes(fill = Gender), alpha = I(0.2), color = I("black"))+
  geom_line(aes(color = Gender)) +
  theme_bw() + ylim(c(0, 1)) +
  theme(axis.text.x = element_text(angle = 20),
        legend.position = "bottom") + labs(x = "Item", y = "Probability")
```

![plot of chunk excluderesidvar](mertoolsIntro-excluderesidvar-1.png)

Here, more difference emerges, but we see that the differences are not very
precise.

---
title: "Prediction Intervals from merMod Objects"
author: "Jared Knowles and Carl Frederick"
date: "2020-06-22"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteIndexEntry{Prediction Intervals from merMod Objects}
  %\VignetteEncoding{UTF-8}
---



## Introduction

Fitting (generalized) linear mixed models, (G)LMM, to very large data sets is
becoming increasingly easy, but understanding and communicating the uncertainty
inherent in those models is not. As the documentation for
`lme4::predict.merMod()` notes:

> There is no option for computing standard errors of predictions because it is
> difficult to define an efficient method that incorporates uncertainty in the
> variance parameters; we recommend `lme4::bootMer()` for this task.

We agree that, short of a fully Bayesian analysis, bootstrapping is the
gold-standard for deriving a prediction interval predictions from a (G)LMM, but
the time required to obtain even a respectable number of replications from
`bootMer()` quickly becomes prohibitive when the initial model fit is on the
order of hours instead of seconds. The only other alternative we have identified
for these situations is to use the `arm::sim()` function to simulate values.
Unfortunately, this only takes variation of the fixed coefficients and residuals
into account, and assumes the conditional modes of the random effects are fixed.

We developed the `predictInterval()` function to incorporate the variation in
the conditional modes of the random effects (CMRE, a.k.a. BLUPs in the LMM case)
into calculating prediction intervals. Ignoring the variance in the CMRE results
in overly confident estimates of predicted values and in cases where the precision
of the grouping term varies across levels of grouping terms, creates the illusion
of difference where none may exist. The importance of accounting for this variance
comes into play sharply when comparing the predictions of different models across
observations.

We take the warning from `lme4::predict.merMod()`
seriously, but view this method as a decent first approximation the full bootstrap
analysis for (G)LMMs fit to very large data sets.

## Conceptual description

In order to generate a proper prediction interval, a prediction must account for
three sources of uncertainty in mixed models:

1. the residual (observation-level) variance,
2. the uncertainty in the fixed coefficients, and
3. the uncertainty in the variance parameters for the grouping factors.

A fourth, uncertainty about the data, is beyond the scope of any prediction method.

As we mentioned above, the `arm:sim()` function incorporates the first two sources
of variation but not the third , while bootstrapping using `lme4::bootMer()`
does incorporate all three sources of uncertainty because it re-estimates the
model using random samples of the data.

When inference about the values of the CMREs is of interest, it would be nice to
incorporate some degree of uncertainty in those estimates when comparing
observations across groups. `predictInterval()` does this by drawing values of
the CMREs from the conditional variance-covariance matrix of the random affects
accessible from `lme4::ranef(model, condVar=TRUE)`. Thus, `predictInterval()`
incorporates all of the uncertainty from sources one and two, and part of the
variance from source 3, but the variance parameters themselves are treated as
fixed.

To do this, `predictInterval()` takes an estimated model of class `merMod` and,
like `predict()`, a data.frame upon which to make those predictions and:

1. extracts the fixed and random coefficients
2. takes `n` draws from the multivariate normal distribution of the fixed and
random coefficients (separately)
3. calculates the linear predictor for each row in `newdata` based on these draws,
and
4. optionally incorporates the residual variation (per the `arm::sim()` function),
and,
5. returns newdata with the lower and upper limits of the prediction interval
and the mean or median of the simulated predictions

Currently, the supported model types are linear mixed models and mixed logistic
regression models.

The prediction data set *can* include levels that are not in the estimation model
frame. The prediction intervals for such observations only incorporate
uncertainty from fixed coefficient estimates and the residual level of variation.

## Comparison to existing methods

What do the differences between `predictInterval()` and the other methods for
constructing prediction intervals mean in practice? We would expect to see
`predictInterval()` to produce prediction intervals that are wider than all methods
except for the `bootMer()` method. We would also hope that the prediction point
estimate from other methods falls within the prediction interval produced by
`predictInterval()`. Ideally, the predicted point estimate produced by
`predictInterval()`  would fall close to that produced by `bootMer()`.

This section compares the results of `predictInterval()` with those obtained
using `arm::sim()` and `lme4::bootMer()` using the sleepstudy data from
`lme4`.  These data contain reaction time observations for 10 days on 18 subjects.
The data are sorted such that the first 10 observations are days one through
ten for subject 1, the next 10 are days one through ten for subject 2 and so on.
The example model that we are estimating below estimates random intercepts and a
random slope for the number of days.

###Step 1: Estimating the model and using `predictInterval()`

First, we will load the required packages and data and estimate the model:


```r
set.seed(271828)
data(sleepstudy)
fm1 <- lmer(Reaction ~ Days + (Days|Subject), data=sleepstudy)
display(fm1)
#> lmer(formula = Reaction ~ Days + (Days | Subject), data = sleepstudy)
#>             coef.est coef.se
#> (Intercept) 251.41     6.82 
#> Days         10.47     1.55 
#> 
#> Error terms:
#>  Groups   Name        Std.Dev. Corr 
#>  Subject  (Intercept) 24.74         
#>           Days         5.92    0.07 
#>  Residual             25.59         
#> ---
#> number of obs: 180, groups: Subject, 18
#> AIC = 1755.6, DIC = 1760.3
#> deviance = 1751.9
```

Then, calculate prediction intervals using `predictInterval()`. The `predictInterval`
function has a number of user configurable options. In this example, we use the
original data `sleepstudy` as the newdata. We pass the function the `fm1` model
we fit above. We also choose a 95% interval with `level = 0.95`, though
we could choose a less conservative prediction interval. We make 1,000 simulations
for each observation `n.sims = 1000`. We set the point estimate to be the median
of the simulated values, instead of the mean. We ask for the linear predictor back,
if we fit a logistic regression, we could have asked instead for our predictions
on the probability scale instead. Finally, we indicate that we want the predictions
to incorporate the residual variance from the model -- an option only available
for `lmerMod` objects.


```r
PI.time <- system.time(
  PI <- predictInterval(merMod = fm1, newdata = sleepstudy,
                        level = 0.95, n.sims = 1000,
                        stat = "median", type="linear.prediction",
                        include.resid.var = TRUE)
)
```

Here is the first few rows of the object `PI`:


|      fit|      upr|      lwr|
|--------:|--------:|--------:|
| 251.6685| 311.3171| 196.4096|
| 271.4802| 330.9195| 214.2175|
| 292.6809| 350.9867| 237.7714|
| 311.6967| 369.2911| 254.2237|
| 331.8318| 389.7439| 278.1857|
| 350.7450| 408.1386| 294.8506|

The three columns are the median (`fit`) and limits of the 95% prediction
interval (`upr` and `lwr`) because we set `level=0.95`. The following figure
displays the output graphically for the first 30 observations.


```r
library(ggplot2);
ggplot(aes(x=1:30, y=fit, ymin=lwr, ymax=upr), data=PI[1:30,]) +
  geom_point() +
  geom_linerange() +
  labs(x="Index", y="Prediction w/ 95% PI") + theme_bw()
```

<img src="usage-Inspect_predInt_2-1.png" title="plot of chunk Inspect_predInt_2" alt="plot of chunk Inspect_predInt_2" style="display: block; margin: auto;" />

#### Step 1a: Adjusting for correlation between fixed and random effects

The prediction intervals above do not correct for correlations between fixed
and random effects. This tends to lead to predictive intervals that are too
conservative, especially for existing groups when there is a lot of data on
relatively few groups. In that case, a significant portion of the uncertainty
in the prediction can be due to variance in the fixed intercept which is
anti-correlated with variance in the random intercept effects. For instance,
it does not actually matter if the fixed intercept is 5 and the random
intercept effects are -2, 1, and 1, versus a fixed intercept of 6 and
random intercept effects of -3, 0, and 0. (The latter situation will never be
the MLE, but it can occur in this package's simulations.)

To show this issue, we'll use the sleep study model, predicting the reaction times
of subjects after experiencing sleep deprivation:


```r
fm1 <- lmer(Reaction ~ Days + (Days|Subject), data=sleepstudy)
display(fm1)
#> lmer(formula = Reaction ~ Days + (Days | Subject), data = sleepstudy)
#>             coef.est coef.se
#> (Intercept) 251.41     6.82 
#> Days         10.47     1.55 
#> 
#> Error terms:
#>  Groups   Name        Std.Dev. Corr 
#>  Subject  (Intercept) 24.74         
#>           Days         5.92    0.07 
#>  Residual             25.59         
#> ---
#> number of obs: 180, groups: Subject, 18
#> AIC = 1755.6, DIC = 1760.3
#> deviance = 1751.9
```

Let's use the model to give an interval for the true average body fat of a
large group of students like the first one in the study — a 196cm female
baseball player:


```r
sleepstudy[1,]
#>   Reaction Days Subject
#> 1   249.56    0     308
predictInterval(fm1, sleepstudy[1,], include.resid.var=0) #predict the average body fat for a group of 196cm female baseball players
#>        fit      upr      lwr
#> 1 253.9977 270.7438 236.2829
```

There are two ways to get predictInterval to create less-conservative intervals
to deal with this. The first is just to tell it to consider certain fixed
effects as fully-known (that is, with an effectively 0 variance.) This is done
using the `ignore.fixed.effects` argument.


```r
predictInterval(fm1, sleepstudy[1,], include.resid.var=0, ignore.fixed.terms = 1)
#>        fit      upr      lwr
#> 1 253.8537 268.5299 239.6275
# predict the average reaction time for a subject at day 0, taking the global intercept
# (mean reaction time) as fully known
predictInterval(fm1, sleepstudy[1,], include.resid.var=0, ignore.fixed.terms = "(Intercept)")
#>        fit      upr      lwr
#> 1 254.2354 269.3875 239.3116
#Same as above
predictInterval(fm1, sleepstudy[1,], include.resid.var=0, ignore.fixed.terms = 1:2)
#>        fit      upr     lwr
#> 1 253.6743 269.8776 237.984
# as above, taking the first two fixed effects (intercept and days effect) as fully known
```

The second way is to use an ad-hoc variance adjustment, with the
`fix.intercept.variance` argument. This takes the model's intercept variance
$\hat\sigma^2_\mu$ and adjusts it to:

$$\hat\sigma\prime^2_\mu = \hat\sigma^2_\mu-\Sigma_{levels}\frac{1}{\Sigma_{groups(level)}1/(\hat\sigma^2_{level}+sigma^2_{group})}$$

In other words, it assumes the given intercept variance incorporates spurious
variance for each level, where each of the spurious variance terms has a
precision equal to the of the precisions due to the individual groups at that
level.


```r
predictInterval(fm1, sleepstudy[1,], include.resid.var=0,
                fix.intercept.variance = TRUE)
#>        fit      upr      lwr
#> 1 253.5872 268.8683 236.6639
# predict the average reaction time for a subject at day 0,, using an ad-hoc
# correction for the covariance of the intercept with the random intercept effects.
```

A few notes about these two arguments:

* `fix.intercept.variance=T` is redundant with `ignore.fixed.effects=1`, but not
vice versa.
* These corrections should NOT be used when predicting outcomes for groups not
present in the original data.

### Step 2: Comparison with `arm::sim()`

How does the output above compare to what we could get from `arm::sim()`?


```r
PI.arm.time <- system.time(
  PI.arm.sims <- arm::sim(fm1, 1000)
)

PI.arm <- data.frame(
  fit=apply(fitted(PI.arm.sims, fm1), 1, function(x) quantile(x, 0.500)),
  upr=apply(fitted(PI.arm.sims, fm1), 1, function(x) quantile(x, 0.975)),
  lwr=apply(fitted(PI.arm.sims, fm1), 1, function(x) quantile(x, 0.025))
)

comp.data <- rbind(data.frame(Predict.Method="predictInterval()", x=(1:nrow(PI))-0.1, PI),
                   data.frame(Predict.Method="arm::sim()", x=(1:nrow(PI.arm))+0.1, PI.arm))

ggplot(aes(x=x, y=fit, ymin=lwr, ymax=upr, color=Predict.Method), data=comp.data[c(1:30,181:210),]) +
  geom_point() +
  geom_linerange() +
  labs(x="Index", y="Prediction w/ 95% PI") +
  theme_bw() +  theme(legend.position="bottom") +
  scale_color_brewer(type = "qual", palette = 2)
```

<img src="usage-arm.Sim-1.png" title="plot of chunk arm.Sim" alt="plot of chunk arm.Sim" style="display: block; margin: auto;" />

The prediction intervals from `arm:sim()` are much smaller and the random slope
for days vary more than they do for `predictInterval`. Both results are as
expected, given the small number of subjects and observations per subject in
these data. Because `predictInterval()` is incorporating uncertainty in the CMFEs
(but not the variance parameters of the random coefficients themselves), the
Days slopes are closer to the overall or pooled regression slope.

###Step 3: Comparison with `lme4::bootMer()`

As quoted above, the developers of lme4 suggest that users interested in
uncertainty estimates around their predictions use `lme4::bootmer()` to calculate
them. The documentation for `lme4::bootMer()` goes on to describe three
implemented flavors of bootstrapped estimates:

1. parametrically resampling both the *"spherical"* random effects
*u* and the i.i.d. errors $\epsilon$
2. treating the random effects as fixed and parametrically resampling the
i.i.d. errors
3. treating the random effects as fixed and semi-parametrically resampling
the i.i.d. errors from the distribution of residuals.

We will compare the results from `predictInterval()` with each method, in turn.

#### Step 3a: `lme4::bootMer()` method 1


```r
##Functions for bootMer() and objects
####Return predicted values from bootstrap
mySumm <- function(.) {
  predict(., newdata=sleepstudy, re.form=NULL)
}
####Collapse bootstrap into median, 95% PI
sumBoot <- function(merBoot) {
  return(
    data.frame(fit = apply(merBoot$t, 2, function(x) as.numeric(quantile(x, probs=.5, na.rm=TRUE))),
               lwr = apply(merBoot$t, 2, function(x) as.numeric(quantile(x, probs=.025, na.rm=TRUE))),
               upr = apply(merBoot$t, 2, function(x) as.numeric(quantile(x, probs=.975, na.rm=TRUE)))
    )
  )
}

##lme4::bootMer() method 1
PI.boot1.time <- system.time(
  boot1 <- lme4::bootMer(fm1, mySumm, nsim=250, use.u=FALSE, type="parametric")
)

PI.boot1 <- sumBoot(boot1)

comp.data <- rbind(data.frame(Predict.Method="predictInterval()", x=(1:nrow(PI))-0.1, PI),
                   data.frame(Predict.Method="lme4::bootMer() - Method 1", x=(1:nrow(PI.boot1))+0.1, PI.boot1))

ggplot(aes(x=x, y=fit, ymin=lwr, ymax=upr, color=Predict.Method), data=comp.data[c(1:30,181:210),]) +
  geom_point() +
  geom_linerange() +
  labs(x="Index", y="Prediction w/ 95% PI") +
  theme_bw() +  theme(legend.position="bottom") +
  scale_color_brewer(type = "qual", palette = 2)
```

<img src="usage-bootMer.1-1.png" title="plot of chunk bootMer.1" alt="plot of chunk bootMer.1" style="display: block; margin: auto;" />

The intervals produced by `predictInterval`, represented in green, cover the
point estimates produced by `bootMer` in every case for these 30 observations.
Additionally, in almost every case, the `predictInterval` encompasses the entire
interval presented by `bootMer`. Here, the estimates produced by `bootMer` are
re-estimating the group terms, but by refitting the model, they are also taking
into account the conditional variance of these terms, or `theta`, and provide
tighter prediction intervals than the `predictInterval` method.

####Step 3b: `lme4::bootMer()` method 2


```r
##lme4::bootMer() method 2
PI.boot2.time <- system.time(
  boot2 <- lme4::bootMer(fm1, mySumm, nsim=250, use.u=TRUE, type="parametric")
)

PI.boot2 <- sumBoot(boot2)

comp.data <- rbind(data.frame(Predict.Method="predictInterval()", x=(1:nrow(PI))-0.1, PI),
                   data.frame(Predict.Method="lme4::bootMer() - Method 2", x=(1:nrow(PI.boot2))+0.1, PI.boot2))

ggplot(aes(x=x, y=fit, ymin=lwr, ymax=upr, color=Predict.Method), data=comp.data[c(1:30,181:210),]) +
  geom_point() +
  geom_linerange() +
  labs(x="Index", y="Prediction w/ 95% PI") +
  theme_bw() +  theme(legend.position="bottom") +
  scale_color_brewer(type = "qual", palette = 2)
```

<img src="usage-bootMer.2-1.png" title="plot of chunk bootMer.2" alt="plot of chunk bootMer.2" style="display: block; margin: auto;" />

Here, the results for `predictInterval` in green again encompass the results from
`bootMer`, but are much wider. The `bootMer` estimates are ignoring the variance
in the group effects, and as such, are only incorporating the residual variance
and the variance in the fixed effects -- similar to the `arm::sim()` function.

#### Step 3c: `lme4::bootMer()` method 3


```r
##lme4::bootMer() method 3
PI.boot3.time <- system.time(
  boot3 <- lme4::bootMer(fm1, mySumm, nsim=250, use.u=TRUE, type="semiparametric")
)

PI.boot3 <- sumBoot(boot3)

comp.data <- rbind(data.frame(Predict.Method="predictInterval()", x=(1:nrow(PI))-0.1, PI),
                   data.frame(Predict.Method="lme4::bootMer() - Method 3", x=(1:nrow(PI.boot3))+0.1, PI.boot3))

ggplot(aes(x=x, y=fit, ymin=lwr, ymax=upr, color=Predict.Method), data=comp.data[c(1:30,181:210),]) +
  geom_point() +
  geom_linerange() +
  labs(x="Index", y="Prediction w/ 95% PI") +
  theme_bw() +  theme(legend.position="bottom") +
  scale_color_brewer(type = "qual", palette = 2)
```

<img src="usage-bootMer.3-1.png" title="plot of chunk bootMer.3" alt="plot of chunk bootMer.3" style="display: block; margin: auto;" />

These results are virtually identical to those above.

#### Step 3c: Comparison to rstanarm





```r
PI.time.stan <- system.time({
  fm_stan <- stan_lmer(Reaction ~ Days + (Days|Subject), data = sleepstudy,
                       verbose = FALSE, open_progress = FALSE, refresh = -1,
                       show_messages=FALSE, chains = 1)
  zed <- posterior_predict(fm_stan)
  PI.stan <- cbind(apply(zed, 2, median), central_intervals(zed, prob=0.95))
})
#> Chain 1: 
#> Chain 1: Gradient evaluation took 0 seconds
#> Chain 1: 1000 transitions using 10 leapfrog steps per transition would take 0 seconds.
#> Chain 1: Adjust your expectations accordingly!
#> Chain 1: 
#> Chain 1: 
#> Chain 1: 
#> Chain 1:  Elapsed Time: 6.994 seconds (Warm-up)
#> Chain 1:                2.497 seconds (Sampling)
#> Chain 1:                9.491 seconds (Total)
#> Chain 1:


print(fm_stan)
#> stan_lmer
#>  family:       gaussian [identity]
#>  formula:      Reaction ~ Days + (Days | Subject)
#>  observations: 180
#> ------
#>             Median MAD_SD
#> (Intercept) 251.5    6.4 
#> Days         10.5    1.7 
#> 
#> Auxiliary parameter(s):
#>       Median MAD_SD
#> sigma 25.9    1.6  
#> 
#> Error terms:
#>  Groups   Name        Std.Dev. Corr
#>  Subject  (Intercept) 23.8         
#>           Days         6.9     0.09
#>  Residual             26.0         
#> Num. levels: Subject 18 
#> 
#> ------
#> * For help interpreting the printed output see ?print.stanreg
#> * For info on the priors used see ?prior_summary.stanreg

PI.stan <- as.data.frame(PI.stan)
names(PI.stan) <- c("fit", "lwr", "upr")
PI.stan <- PI.stan[, c("fit", "upr", "lwr")]
comp.data <- rbind(data.frame(Predict.Method="predictInterval()", x=(1:nrow(PI))-0.1, PI),
                   data.frame(Predict.Method="rstanArm", x=(1:nrow(PI.stan))+0.1, PI.stan))

ggplot(aes(x=x, y=fit, ymin=lwr, ymax=upr, color=Predict.Method), data=comp.data[c(1:30,181:210),]) +
  geom_point() +
  geom_linerange() +
  labs(x="Index", y="Prediction w/ 95% PI") +
  theme_bw() +  theme(legend.position="bottom") +
  scale_color_brewer(type = "qual", palette = 2)
```

<img src="usage-stancomp-1.png" title="plot of chunk stancomp" alt="plot of chunk stancomp" style="display: block; margin: auto;" />


### Computation time

Our initial motivation for writing this function was to develop a method for
incorporating uncertainty in the CMFEs  for mixed models estimated on very large
samples. Even for models with only modest degrees of complexity, using
`lme4::bootMer()` quickly becomes time prohibitive because it involves
re-estimating the model for each simulation. We have seen how each alternative
compares to `predictInterval()` substantively, but how do they compare in terms
of computational time? The table below lists the output of `system.time()` for
all five methods for calculating prediction intervals for `merMod` objects.


|                         | user.self| sys.self| elapsed|
|:------------------------|---------:|--------:|-------:|
|predictInterval()        |      0.30|     0.01|    0.31|
|arm::sim()               |      0.56|     0.00|    0.56|
|lme4::bootMer()-Method 1 |      5.79|     0.08|    5.92|
|lme4::bootMer()-Method 2 |      6.03|     0.05|    6.13|
|lme4::bootMer()-Method 3 |      5.93|     0.01|    6.05|
|rstanarm:predict         |     10.09|     0.05|   10.19|

For this simple example, we see that `arm::sim()` is the fastest--nearly five
times faster than `predictInterval()`. However, `predictInterval()` is nearly
six times faster than any of the bootstrapping options via `lme4::bootMer`.  This
may not seem like a lot, but consider that the computational time for required
for bootstrapping is roughly proportional to the number of bootstrapped simulations
requested ... `predictInterval()` is not because it is just a series of draws from
various multivariate normal distributions, so the time ratios in the table below
represents the lowest bound of the computation time ratio of bootstrapping
to `predictInterval()`.

## Simulation

TBC.

# Prediction intervals work with new factor levels added part 2

    Code
      predictInterval(glmer3LevSlope, newdata = zNew)
    Condition
      Warning:
      For binomial GLMMs, include.resid.var = TRUE simulates from the
      conditional binomial distribution (n-trial binomial simulation).
      This is the theoretically correct approach.
      To get predictions without residual variance, set include.resid.var = FALSE.
      Warning in `chol.default()`:
      the matrix is either rank-deficient or not positive definite
      Warning in `chol.default()`:
      the matrix is either rank-deficient or not positive definite
      Warning in `chol.default()`:
      the matrix is either rank-deficient or not positive definite
      Warning in `chol.default()`:
      the matrix is either rank-deficient or not positive definite
      Warning in `chol.default()`:
      the matrix is either rank-deficient or not positive definite
      Warning in `chol.default()`:
      the matrix is either rank-deficient or not positive definite
      Warning in `chol.default()`:
      the matrix is either rank-deficient or not positive definite
      Warning in `chol.default()`:
      the matrix is either rank-deficient or not positive definite
      Warning in `chol.default()`:
      the matrix is either rank-deficient or not positive definite
      Warning in `chol.default()`:
      the matrix is either rank-deficient or not positive definite
      Warning in `chol.default()`:
      the matrix is either rank-deficient or not positive definite
      Warning:
           The following levels of BROOD from newdata 
       -- 100, 101 -- are not in the model data. 
           Currently, predictions for these values are based only on the 
       fixed coefficients and the observation-level error.
    Output
                  fit       upr        lwr
      1   0.286184362 0.8728483 -0.3369354
      2   0.286184362 0.8728483 -0.3369354
      3  -0.004346961 0.6208335 -0.5510659
      4   0.032606238 0.6251944 -0.6043012
      5   0.032606238 0.6251944 -0.6043012
      6   0.032606238 0.6251944 -0.6043012
      7   0.032606238 0.6251944 -0.6043012
      8  -0.161448918 0.4544298 -0.8194496
      9  -0.161448918 0.4544298 -0.8194496
      10 -0.161448918 0.4544298 -0.8194496

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merList.R
\name{print.summary.merModList}
\alias{print.summary.merModList}
\title{Print the summary of a merMod list}
\usage{
\method{print}{summary.merModList}(x, ...)
}
\arguments{
\item{x}{a summary of amerModList object}

\item{...}{additional arguments}
}
\value{
summary content printed to console
}
\description{
Print the summary of a merMod list
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shinyMer.R
\name{shinyMer}
\alias{shinyMer}
\title{Launch a shiny app to explore your merMod interactively}
\usage{
shinyMer(merMod, simData = NULL, pos = 1)
}
\arguments{
\item{merMod}{An object of class "merMod".}

\item{simData}{A data.frame to make predictions from (optional). If
NULL, then the user can only make predictions using the data in
the frame slot of the merMod object.}

\item{pos}{The position of the environment to export function arguments to.
Defaults to 1, the global environment, to allow shiny to run.}
}
\value{
A shiny app
}
\description{
\code{shinyMer} launches a shiny app that allows you to interactively
explore an estimated merMod using functions from \code{merTools}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merList.R
\name{fixef.merModList}
\alias{fixef.merModList}
\title{Extract fixed-effects estimates for a merModList}
\usage{
\method{fixef}{merModList}(object, add.dropped = FALSE, ...)
}
\arguments{
\item{object}{any fitted model object from which fixed
  effects estimates can be extracted.}

\item{add.dropped}{for models with rank-deficient design
  matrix, reconstitute the full-length parameter vector by
  adding \code{NA} values in appropriate locations?}

\item{...}{optional additional arguments. Currently
  none are used in any methods.}
}
\value{
a named, numeric vector of fixed-effects estimates.
}
\description{
Extract fixed-effects estimates for a merModList
}
\details{
Extract the estimates of the fixed-effects parameters from a list of
fitted \code{merMod} models. Takes the mean of the individual \code{fixef}
objects for each of the component models in the \code{merModList}.
}
\examples{
\donttest{
sim_list <- replicate(n = 10,
        expr = sleepstudy[sample(row.names(sleepstudy), 180),],
        simplify=FALSE)
fml <- "Reaction ~ Days + (Days | Subject)"
mod <- lmerModList(fml, data = sim_list)
fixef(mod)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predictInterval_helpers.R
\name{simulate_glmm_response}
\alias{simulate_glmm_response}
\title{Simulate GLMM responses from conditional distribution}
\usage{
simulate_glmm_response(
  yhat,
  family,
  link,
  weights = NULL,
  use.probability = FALSE
)
}
\arguments{
\item{yhat}{Matrix of linear predictors (fixed + random effects).}

\item{family}{Character string specifying the distribution family.}

\item{link}{Character string specifying the link function.}

\item{weights}{Optional weights (for binomial with n-trial parameter).}

\item{use.probability}{For binomial, return proportions (0-1) instead of counts?}
}
\value{
Matrix of simulated responses with same dimensions as yhat.
}
\description{
Internal helper that simulates responses from the conditional distribution
of Y|b (given random effects) for GLMMs. This implements the correct
theoretical approach where residual variance is inherent in the distribution.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merSubstEff.R
\name{REimpact}
\alias{REimpact}
\title{Calculate the weighted mean of fitted values for various levels of
random effect terms.}
\usage{
REimpact(merMod, newdata, groupFctr = NULL, term = NULL, breaks = 3, ...)
}
\arguments{
\item{merMod}{An object of class merMod}

\item{newdata}{a data frame of observations to calculate group-level differences
for}

\item{groupFctr}{The name of the grouping factor over which the random
coefficient of interest varies.  This is the variable to the right of the
pipe, \code{|}, in the [g]lmer formula. This parameter is optional, if not
specified, it will perform the calculation for the first effect listed
by \code{ranef}.}

\item{term}{The name of the random coefficient of interest. This is the
variable to the left of the pipe, \code{|}, in the [g]lmer formula. Partial
matching is attempted on the intercept term so the following character
strings will all return rankings based on the intercept (\emph{provided that
they do not match the name of another random coefficient for that factor}):
\code{c("(Intercept)", "Int", "intercep", ...)}.}

\item{breaks}{an integer representing the number of bins to divide the group
effects into, the default is 3; alternatively it can specify breaks from 0-100
for how to cut the expected rank distribution}

\item{...}{additional arguments to pass to \code{\link{predictInterval}}}
}
\value{
A data.frame with all unique combinations of the number of cases, rows
in the newdata element, and number of bins:
  \describe{
    \item{case}{The row number of the observation from newdata.}
    \item{bin}{The ranking bin for the expected rank, the higher the bin number,
    the greater the expected rank of the groups in that bin.}
    \item{AvgFitWght}{The weighted mean of the fitted values for case i in bin k}
    \item{AvgFitWghtSE}{The standard deviation of the mean of the fitted values
    for case i in bin k.}
    \item{nobs}{The number of group effects contained in that bin.}
  }
}
\description{
\code{REimpact} calculates the average predicted value for each row of a
new data frame across the distribution of \code{\link{expectedRank}} for a
merMod object. This allows the user to make meaningful comparisons about the
influence of random effect terms on the scale of the response variable,
for user-defined inputs, and accounting for the variability in grouping terms.
}
\details{
The function predicts the response at every level in the random effect term
specified by the user. Then, the expected rank of each group level is binned
to the number of bins specified by the user. Finally, a weighted mean of the
fitted value for all observations in each bin of the expected ranks is
calculated using the inverse of the variance as the weight -- so that less
precise estimates are downweighted in the calculation of the mean for the bin.
Finally, a standard error for the bin mean is calculated.

This function uses the formula for variance of a weighted mean
recommended by Cochran (1977).
}
\examples{
\donttest{
#For a one-level random intercept model
m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
m1.er <- REimpact(m1, newdata = sleepstudy[1, ], breaks = 2)
#For a one-level random intercept model with multiple random terms
m2 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
#ranked by the random slope on Days
m2.er1 <- REimpact(m2,  newdata = sleepstudy[1, ],
           groupFctr = "Subject", term="Days")
#ranked by the random intercept
m2.er2 <- REimpact(m2, newdata = sleepstudy[1, ],
             groupFctr = "Subject", term="int")

# You can also pass additional arguments to predictInterval through REimpact
g1 <- lmer(y ~ lectage + studage + (1|d) + (1|s), data=InstEval)
zed <- REimpact(g1, newdata = InstEval[9:12, ], groupFctr = "d", n.sims = 50,
                include.resid.var = TRUE)
zed2 <- REimpact(g1, newdata = InstEval[9:12, ], groupFctr = "s", n.sims = 50,
                 include.resid.var = TRUE)
zed3 <- REimpact(g1, newdata = InstEval[9:12, ], groupFctr = "d", breaks = 5,
                n.sims = 50, include.resid.var = TRUE)
}
}
\references{
Gatz, DF and Smith, L. The Standard Error of a Weighted Mean Concentration.
I. Bootstrapping vs other methods. \emph{Atmospheric Environment}.
1995;11(2)1185-1193. Available at
\url{https://www.sciencedirect.com/science/article/pii/135223109400210C}

Cochran, WG. 1977. Sampling Techniques (3rd Edition). Wiley, New York.
}
\seealso{
\code{\link{expectedRank}}, \code{\link{predictInterval}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merList.R
\name{REsdExtract}
\alias{REsdExtract}
\title{Extract the standard deviation of the random effects from a merMod object}
\usage{
REsdExtract(model)
}
\arguments{
\item{model}{an object that inherits from class merMod}
}
\value{
a numeric vector for standard deviations of the random effects
}
\description{
Extract the standard deviation of the random effects from a merMod object
}
\examples{
fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
REsdExtract(fm1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merExtract.R
\name{RMSE.merMod}
\alias{RMSE.merMod}
\title{Estimate the Root Mean Squared Error (RMSE) for a lmerMod}
\usage{
RMSE.merMod(merMod, scale = FALSE)
}
\arguments{
\item{merMod}{a lmerMod object from the lme4 package}

\item{scale}{logical, should the result be returned on the scale of
response variable standard deviations?}
}
\value{
a numeric which represents the RMSE
}
\description{
Extract the Root Mean Squared Error for a lmerMod object
}
\examples{
require(lme4)
m2 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
RMSE.merMod(m2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merList.R
\name{modelRandEffStats}
\alias{modelRandEffStats}
\title{Extract data.frame of random effect statistics from merMod List}
\usage{
modelRandEffStats(modList)
}
\arguments{
\item{modList}{a list of multilevel models}
}
\value{
a data.frame
}
\description{
Extract data.frame of random effect statistics from merMod List
}
\examples{
\donttest{
sim_list <- replicate(n = 10,
        expr = sleepstudy[sample(row.names(sleepstudy), 180),],
        simplify=FALSE)
fml <- "Reaction ~ Days + (Days | Subject)"
mod <- lmerModList(fml, data = sim_list)
modelRandEffStats(mod)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merTools-package.r
\docType{data}
\name{hsb}
\alias{hsb}
\title{A subset of data from the 1982 High School and Beyond survey used as examples for HLM software}
\format{
A data frame with 7,185 observations on the following 8 variables.
\describe{
 \item{\code{schid}}{a numeric vector, 160 unique values}
 \item{\code{mathach}}{a numeric vector for the performance on a standardized math assessment}
 \item{\code{female}}{a numeric vector coded 0 for male and 1 for female}
 \item{\code{ses}}{a numeric measure of student socio-economic status}
 \item{\code{minority}}{a numeric vector coded 0 for white and 1 for non-white students}
 \item{\code{schtype}}{a numeric vector coded 0 for public and 1 for private schools}
 \item{\code{meanses}}{a numeric, the average SES for each school in the data set}
 \item{\code{size}}{a numeric for the number of students in the school}
}
}
\source{
Data made available by UCLA Institute for Digital Research and Education
(IDRE) online: \url{https://stats.oarc.ucla.edu/other/hlm/hlm-mlm/introduction-to-multilevel-modeling-using-hlm/}
}
\usage{
hsb
}
\description{
A key example dataset used for examples in the HLM software manual.
Included here for use in replicating HLM analyses in R.
}
\details{
The data file used for this presentation is a subsample from the
1982 High School and Beyond Survey and is used extensively in
Hierarchical Linear Models by Raudenbush and Bryk. It consists of 7,185 students
nested in 160 schools.
}
\examples{
data(hsb)
head(hsb)
}
\references{
Stephen W. Raudenbush and Anthony S. Bryk (2002). Hierarchical
Linear Models: Applications and Data Analysis Methods (2nd ed.). SAGE.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{famlink}
\alias{famlink}
\title{Find link function family}
\usage{
famlink(object, resp = object@resp)
}
\arguments{
\item{object}{a merMod object}

\item{resp}{the response vector}
}
\value{
the link function and family
}
\description{
Find link function family
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merList.R
\name{merModList-class}
\alias{merModList-class}
\alias{merModList}
\title{merModList S3 Class}
\description{
A list of fitted mixed-effects models from lme4
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/REmargins.R
\name{REmargins}
\alias{REmargins}
\title{Calculate the predicted value for each observation across the distribution
of the random effect terms.}
\usage{
REmargins(
  merMod,
  newdata = NULL,
  groupFctr = NULL,
  term = NULL,
  breaks = 4,
  .parallel = FALSE,
  ...
)
}
\arguments{
\item{merMod}{An object of class merMod}

\item{newdata}{a data frame of observations to calculate group-level differences
for}

\item{groupFctr}{The name of the grouping factor over which the random
coefficient of interest varies.  This is the variable to the right of the
pipe, \code{|}, in the [g]lmer formula. This parameter is optional, if not
specified, it will perform the calculation for the first effect listed
by \code{ranef}. If the length is > 1 then the combined effect of all
listed groups will calculated and marginalized over co-occurences of those
groups if desired.}

\item{term}{The name of the random coefficient of interest. This is the
variable to the left of the pipe, \code{|}, in the [g]lmer formula. Partial
matching is attempted on the intercept term so the following character
strings will all return rankings based on the intercept (\emph{provided that
they do not match the name of another random coefficient for that factor}):
\code{c("(Intercept)", "Int", "intercep", ...)}.}

\item{breaks}{an integer representing the number of bins to divide the group
effects into, the default is 3.}

\item{.parallel, }{logical should parallel computation be used, default is TRUE}

\item{...}{additional arguments to pass to \code{\link{predictInterval}}}
}
\value{
A data.frame with all unique combinations of the number of cases, rows
in the newdata element:
  \describe{
    \item{...}{The columns of the original data taken from \code{newdata}}
    \item{case}{The row number of the observation from newdata. Each row in newdata will be
    repeated for all unique levels of the grouping_var, term, and breaks.}
    \item{grouping_var}{The grouping variable the random effect is being marginalized over.}
    \item{term}{The term for the grouping variable the random effect is being marginalized over.}
    \item{breaks}{The ntile of the effect size for \code{grouping_var} and \code{term}}
    \item{original_group_level}{The original grouping value for this \code{case}}
    \item{fit_combined}{The predicted value from \code{predictInterval} for this case simulated
    at the Nth ntile of the expected rank distribution of \code{grouping_var} and \code{term}}
    \item{upr_combined}{The upper bound of the predicted value.}
    \item{lwr_combined}{The lower bound of the predicted value.}
    \item{fit_XX}{For each grouping term in newdata the predicted value is decomposed into its
    fit components via predictInterval and these are all returned here}
    \item{upr_XX}{The upper bound for the effect of each grouping term}
    \item{lwr_XX}{The lower bound for the effect of each grouping term}
    \item{fit_fixed}{The predicted fit with all the grouping terms set to 0 (average)}
    \item{upr_fixed}{The upper bound fit with all the grouping terms set to 0 (average)}
    \item{lwr_fixed}{The lower bound fit with all the grouping terms set to 0 (average)}
  }
}
\description{
\code{REmargins} calculates the average predicted value for each row of a
new data frame across the distribution of \code{\link{expectedRank}} for a
merMod object. This allows the user to make meaningful comparisons about the
influence of random effect terms on the scale of the response variable,
for user-defined inputs, and accounting for the variability in grouping terms.
}
\details{
The function simulates the

The function predicts the response at every level in the random effect term
specified by the user. Then, the expected rank of each group level is binned
to the number of bins specified by the user. Finally, a weighted mean of the
fitted value for all observations in each bin of the expected ranks is
calculated using the inverse of the variance as the weight -- so that less
precise estimates are downweighted in the calculation of the mean for the bin.
Finally, a standard error for the bin mean is calculated.
}
\examples{
\donttest{
fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
mfx <- REmargins(merMod = fm1, newdata = sleepstudy[1:10,])

# You can also pass additional arguments to predictInterval through REimpact
 g1 <- lmer(y ~ lectage + studage + (1|d) + (1|s), data=InstEval)
 margin_df <- REmargins(g1, newdata = InstEval[20:25, ], groupFctr = c("s"),
                        breaks = 4)
 margin_df <- REmargins(g1, newdata = InstEval[20:25, ], groupFctr = c("d"),
                         breaks = 3)
}
}
\references{
Gatz, DF and Smith, L. The Standard Error of a Weighted Mean Concentration.
I. Bootstrapping vs other methods. \emph{Atmospheric Environment}.
1995;11(2)1185-1193. Available at
\url{https://www.sciencedirect.com/science/article/pii/135223109400210C}

Cochran, WG. 1977. Sampling Techniques (3rd Edition). Wiley, New York.
}
\seealso{
\code{\link{expectedRank}}, \code{\link{predictInterval}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merExtract.R
\name{FEsim}
\alias{FEsim}
\title{Simulate fixed effects from merMod
\code{FEsim} simulates fixed effects from merMod object posterior distributions}
\usage{
FEsim(merMod, n.sims = 200, oddsRatio = FALSE, seed = NULL)
}
\arguments{
\item{merMod}{a merMod object from the lme4 package}

\item{n.sims}{number of simulations to use}

\item{oddsRatio}{logical, should parameters be converted to odds ratios?}

\item{seed}{numeric, optional argument to set seed for simulations}
}
\value{
a data frame with the following columns
\describe{
  \item{\code{term}}{Name of fixed term (intercept/coefficient)}
  \item{\code{mean}}{Mean of the simulations}
  \item{\code{median}}{Median of the simulations}
  \item{\code{sd}}{Standard deviation of the simulations, \code{NA} if \code{oddsRatio=TRUE}}
}
}
\description{
Simulate fixed effects from merMod
\code{FEsim} simulates fixed effects from merMod object posterior distributions
}
\details{
Use the Gelman sim technique to build fixed effect estimates and
confidence intervals. Uses the sim function in the arm package
}
\examples{
require(lme4)
m2 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
fe2 <- FEsim(m2, 25)
head(fe2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{plot_sim_error_chks}
\alias{plot_sim_error_chks}
\title{Extract all warning msgs from a merMod object}
\usage{
plot_sim_error_chks(
  type = c("FE", "RE"),
  level = 0.95,
  stat = c("mean", "median"),
  sd = TRUE,
  sigmaScale = NULL,
  oddsRatio = FALSE,
  labs = FALSE,
  facet = TRUE
)
}
\arguments{
\item{type}{check a fixed or random effect}

\item{level}{the width of the confidence interval}

\item{stat}{a character value indicating the variable name in data of the
midpoint of the estimated interval, e.g. "mean" or "median"}

\item{sd}{a logical indicating whether or not to plot error bars around
the estimates (default is TRUE). Calculates the width of the error bars
based on \code{level} and the variable named "sd" in \code{data}}

\item{sigmaScale}{a numeric value to divide the estimate and the standard
deviation by in the case of doing an effect size calculation}

\item{oddsRatio}{logical, should the parameters be converted to odds ratios
before plotting}

\item{labs}{logical, include the labels of the groups on the x-axis}

\item{facet}{Accepts either logical (\code{TRUE}) or \code{list} to specify which
random effects to plot. If \code{TRUE}, facets by both \code{groupFctr} and \code{term}.
If \code{list} selects the panel specified by the named elements of the list}
}
\description{
Extract all warning msgs from a merMod object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predictInterval_helpers.R
\name{combine_components}
\alias{combine_components}
\title{Combine fixed, random and residual components into a prediction array}
\usage{
combine_components(
  fixed_mat,
  random_list,
  sigma_vec,
  include.resid.var = TRUE,
  which = c("full", "fixed", "random", "all"),
  family = NULL,
  link = NULL,
  weights = NULL,
  use.probability = FALSE
)
}
\arguments{
\item{fixed_mat}{Matrix of dimensions `nrow(newdata)` × `n.sims`.}

\item{random_list}{List returned by `simulate_random_effects()`.}

\item{sigma_vec}{Residual standard deviations (output of `simulate_residual_variance`).}

\item{include.resid.var}{Logical, whether to add residual variance.}

\item{which}{Character: "full", "fixed", "random" or "all".}
}
\value{
Array `nrow(newdata)` × `n.sims` (or list when `which="all"`).
}
\description{
Combine fixed, random and residual components into a prediction array
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merData.R
\name{randomObs}
\alias{randomObs}
\title{Select a random observation from model data}
\usage{
randomObs(merMod, varList, seed = NULL)
}
\arguments{
\item{merMod}{an object of class merMod}

\item{varList}{optional, a named list of conditions to subset the data on}

\item{seed}{numeric, optional argument to set seed for simulations}
}
\value{
a data frame with a single row for a random observation, but with full
factor levels. See details for more.
}
\description{
Select a random observation from the model frame of a merMod
}
\details{
Each factor variable in the data frame has all factor levels from the
full model.frame stored so that the new data is compatible with predict.merMod
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merData.R
\name{sanitizeNames}
\alias{sanitizeNames}
\title{Clean up variable names in data frames}
\usage{
sanitizeNames(data)
}
\arguments{
\item{data}{a data.frame}
}
\value{
a data frame with variable names cleaned to remove factor() construction
}
\description{
Strips out transformations from variable names in data frames
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{levelfun}
\alias{levelfun}
\title{Parse merMod levels}
\usage{
levelfun(x, nl.n, allow.new.levels = FALSE)
}
\description{
Parse merMod levels
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merTools-package.r
\docType{package}
\name{merTools-package}
\alias{merTools}
\alias{merTools-package}
\title{merTools: Provides methods for extracting and exploring results from merMod
objects in the lme4 package.}
\description{
The merTools package contains convenience tools for extracting useful
information from and exploring the implications of merMod objects created by
the lme4 package.  These convenience functions are especially useful for
merMod objects that take a long time to estimate due to their complexity or
because they are estimated on very large samples.
}
\details{
See the vignettes for usage examples
}
\section{merMod extraction/utility functions}{


\itemize{
  \item \code{\link{fastdisp}}
  \item \code{\link{superFactor}}
  \item \code{\link{REextract}}
  \item \code{\link{REsim}}
  \item \code{\link{FEsim}}
  \item \code{\link{RMSE.merMod}}
  \item \code{\link{thetaExtract}}
  \item \code{\link{REquantile}}
}
}

\section{merMod exploration functions}{


\itemize{
  \item \code{\link{plotREsim}}
  \item \code{\link{plotFEsim}}
  \item \code{\link{plotREimpact}}
  \item \code{\link{draw}}
  \item \code{\link{wiggle}}
  \item \code{\link{subBoot}}
  \item \code{\link{predictInterval}}
  \item \code{\link{expectedRank}}
  \item \code{\link{REimpact}}
  \item \code{\link{shinyMer}}
}
}

\section{Influences and acknowledgements}{


The simulation approach used by \code{\link{predictInterval}} to propagate
uncertainty from both the fixed and random effects draws on the framework
described by Gelman and Hill (2007) and implemented in the \pkg{arm}
package's \code{sim()} function. The package is built on top of the model
objects fit by \pkg{lme4} (Bates et al. 2015) and is intended to complement
\code{lme4::bootMer()} by offering a faster, simulation-based alternative
for large models. We are grateful to the authors and maintainers of those
packages.
}

\references{
Gelman, A. and Hill, J. (2007). \emph{Data Analysis Using Regression and
Multilevel/Hierarchical Models}. Cambridge University Press.

Bates, D., Maechler, M., Bolker, B., and Walker, S. (2015). Fitting Linear
Mixed-Effects Models Using lme4. \emph{Journal of Statistical Software},
67(1), 1-48. \doi{10.18637/jss.v067.i01}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://jknowles.github.io/merTools/}
  \item \url{https://github.com/jknowles/merTools}
  \item Report bugs at \url{https://github.com/jknowles/merTools/issues}
}

}
\author{
\strong{Maintainer}: Jared E. Knowles \email{jared@civilytics.com}

Authors:
\itemize{
  \item Jared E. Knowles \email{jared@civilytics.com}
  \item Carl Frederick \email{carlbfrederick@gmail.com}
}

Other contributors:
\itemize{
  \item Alex Whitworth \email{whitworth.alex@gmail.com} [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{mkNewReTrms}
\alias{mkNewReTrms}
\title{Utility function to make RE terms objects}
\usage{
mkNewReTrms(
  object,
  newdata,
  re.form = NULL,
  na.action = na.pass,
  allow.new.levels = FALSE
)
}
\arguments{
\item{object}{a model object}

\item{newdata}{a data.frame to build RE terms for}

\item{re.form}{a random effect formula to simulate, generated by
\code{\link{reOnly}}}

\item{na.action}{an object describing how NA values should be handled in newdata}

\item{allow.new.levels}{logical, should new levels be allowed in factor variables}
}
\value{
a random effect terms object for a merMod
}
\description{
Utility function to make RE terms objects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merData.R
\name{collapseFrame}
\alias{collapseFrame}
\title{Collapse a dataframe to a single average row}
\usage{
collapseFrame(data)
}
\arguments{
\item{data}{a data.frame}
}
\value{
a data frame with a single row
}
\description{
Take an entire dataframe and summarize it in one row by using the
mean and mode.
}
\details{
Each character and factor variable in the data.frame is assigned to the
modal category and each numeric variable is collapsed to the mean. Currently if
mode is a tie, returns a "."
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predictInterval_helpers.R
\name{simulate_residual_variance}
\alias{simulate_residual_variance}
\title{Simulate residual variance for merMod objects}
\usage{
simulate_residual_variance(merMod, n.sims)
}
\arguments{
\item{merMod}{A merMod object from lme4.}

\item{n.sims}{Number of simulation draws.}
}
\value{
Numeric vector of length `n.sims`.
}
\description{
Internal helper that reproduces the sigma-drawing logic from the original
`predictInterval()` implementation. It returns a numeric vector of length
`n.sims` containing the simulated residual standard deviations.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merList.R
\name{modelInfo}
\alias{modelInfo}
\title{Extract model information from a merMod}
\usage{
modelInfo(object)
}
\arguments{
\item{object}{a merMod object}
}
\value{
Simple summary information about the object, number
of observations, number of grouping terms, AIC, and residual standard deviation
}
\description{
Extract model information from a merMod
}
\examples{
\donttest{
sim_list <- replicate(n = 10,
        expr = sleepstudy[sample(row.names(sleepstudy), 180),],
        simplify=FALSE)
fml <- "Reaction ~ Days + (Days | Subject)"
mod <- lmerModList(fml, data = sim_list)
modelInfo(mod[[1]])
lapply(mod, modelInfo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merExtract.R
\name{REextract}
\alias{REextract}
\title{Extracts random effects}
\usage{
REextract(merMod)
}
\arguments{
\item{merMod}{a merMod object from the lme4 package}
}
\value{
a data frame with the following columns
\describe{
  \item{groupFctr}{The name of the grouping factor associated with the random effects}
  \item{groupID}{The level of the grouping factor associated with the random effects}
  \item{'term'}{One column per random effect, the name is derived from the merMod}
  \item{'term'_se}{One column per random effect, the name is derived from the merMod}
}
}
\description{
Extracts random effect terms from an lme4 model
}
\examples{
m2 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
rfx <- REextract(m2)
#Note the column names
head(rfx)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merPredict.R
\name{predictInterval}
\alias{predictInterval}
\title{Predict from merMod objects with a prediction interval}
\usage{
predictInterval(
  merMod,
  newdata = NULL,
  which = c("full", "fixed", "random", "all"),
  level = 0.8,
  n.sims = 1000,
  stat = c("median", "mean"),
  type = c("linear.prediction", "probability"),
  include.resid.var = TRUE,
  returnSims = FALSE,
  seed = NULL,
  .parallel = FALSE,
  fix.intercept.variance = FALSE,
  ignore.fixed.terms = NULL,
  new.levels = c("zero", "draw")
)
}
\arguments{
\item{merMod}{a merMod object from lme4}

\item{newdata}{a data.frame of new data to predict}

\item{which}{a character specifying what to return, by default it returns the
full interval, but you can also select to return only the fixed variation or
the random component variation. If full is selected the resulting data.frame
will be \code{nrow(newdata) * number of model levels} long}

\item{level}{the width of the prediction interval}

\item{n.sims}{number of simulation samples to construct}

\item{stat}{take the median or mean of simulated intervals}

\item{type}{type of prediction to develop}

\item{include.resid.var}{logical, include or exclude the residual variance for
linear models}

\item{returnSims}{logical, should all n.sims simulations be returned?}

\item{seed}{numeric, optional argument to set seed for simulations}

\item{.parallel}{logical, use parallel computation (default FALSE)}

\item{fix.intercept.variance}{logical; should the variance of the intercept
term be adjusted downwards to roughly correct for its covariance with the
random effects, as if all the random effects are intercept effects?}

\item{ignore.fixed.terms}{a numeric or string vector of indexes or names of
fixed effects which should be considered as fully known (zero variance).}

\item{new.levels}{character, how to treat grouping levels in \code{newdata}
that were not present when the model was fit. \code{"zero"} (the default and
the historical behavior) drops the random effect for such levels, so the
prediction rests on the fixed effects plus residual variation. \code{"draw"}
instead samples each unobserved group's effect from the estimated
random-effect covariance (\code{VarCorr}), so the interval reflects
between-group uncertainty -- the analogue of
\code{brms::posterior_predict(allow_new_levels = TRUE)}. Observations that
share an unobserved level share the same sampled effect.}
}
\value{
a data.frame with three columns: fit, lwr and upr. If `returnSims`
  is TRUE the attribute `sim.results` contains the full simulation array.
}
\description{
This function provides a way to capture model uncertainty in
predictions from multi-level models fit with \code{lme4}. By drawing a sampling
distribution for the random and the fixed effects and then estimating the fitted
value across that distribution, it is possible to generate a prediction interval
for fitted values that includes all variation in the model except for variation
in the covariance parameters, theta. This is a much faster alternative than
bootstrapping for models fit to medium to large datasets.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merExpectedRank.R
\name{expectedRank}
\alias{expectedRank}
\title{Calculate the expected rank of random coefficients that account for
uncertainty.}
\usage{
expectedRank(merMod, groupFctr = NULL, term = NULL)
}
\arguments{
\item{merMod}{An object of class merMod}

\item{groupFctr}{An optional character vector specifying the name(s) the grouping factor(s)
over which the random coefficient of interest varies.  This is the
variable to the right of the pipe, \code{|}, in the [g]lmer formula.
This parameter is optional. If none is specified all terms will be returned.}

\item{term}{An optional character vector specifying the name(s) of the random coefficient of interest. This is the
variable to the left of the pipe, \code{|}, in the [g]lmer formula. Partial
matching is attempted on the intercept term so the following character
strings will all return rankings based on the intercept (\emph{provided that
they do not match the name of another random coefficient for that factor}):
\code{c("(Intercept)", "Int", "intercep", ...)}.}
}
\value{
A data.frame with the following five columns:
  \describe{
    \item{groupFctr}{a character representing name of the grouping factor}
    \item{groupLevel}{a character representing the level of the grouping factor}
    \item{term}{a character representing the formula term for the group}
    \item{estimate}{effect estimate from \code{lme4::ranef(, condVar=TRUE)}).}
    \item{std.error}{the posterior variance of the estimate random effect
                     (from \code{lme4::ranef(, condVar=TRUE)}); named "\code{term}"_var.}
    \item{ER}{The expected rank.}
    \item{pctER}{The percentile expected rank.}
  }
}
\description{
\code{expectedRank} calculates the expected rank and the percentile expected
rank of any random term in a merMod object.  A simple ranking of the estimated
random effects (as produced by \code{\link[lme4]{ranef}}) is not satisfactory
because it ignores any amount of uncertainty.
}
\details{
Inspired by Lingsma et al. (2010, see also Laird and Louis 1989),
expectedRank sums the probability that each level of the grouping factor is
greater than every other level of the grouping factor, similar to a
two-sample t-test.

The formula for the expected rank is:
\deqn{ExpectedRank_i = 1 + \sum \phi((\theta_i - \theta_k) / \sqrt(var(\theta_i)+var(\theta_k))}
where \eqn{\phi} is the standard normal distribution function, \eqn{\theta}
is the estimated random effect and \eqn{var(\theta)} is the posterior
variance of the estimated random effect. We add one to the sum so that the
minimum rank is one instead of zero so that in the case where there is no
overlap between the variances of the random effects (or if the variances are
zero), the expected rank equals the actual rank.  The ranks are ordered such
that the winners have ranks that are greater than the losers.

The formula for the percentile expected rank is:
\deqn{100 * (ExpectedRank_i - 0.5) / N_grps}
where \eqn{N_grps} is the number of grouping factor levels. The percentile
expected rank can be interpreted as the fraction of levels that score at or
below the given level.

NOTE: \code{expectedRank} will only work under conditions that \code{lme4::ranef}
will work. One current example of when this is \emph{not} the case is for
models when there are multiple terms specified per factor (e.g. uncorrelated random
coefficients for the same term, e.g.
\code{lmer(Reaction ~ Days + (1 | Subject) + (0 + Days | Subject), data = sleepstudy)})
}
\examples{
\donttest{
#For a one-level random intercept model
m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
(m1.er <- expectedRank(m1))

#For a one-level random intercept model with multiple random terms
m2 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
#ranked by the random slope on Days
(m2.er1 <- expectedRank(m2, term="Days"))
#ranked by the random intercept
(m2.er2 <- expectedRank(m2, term="int"))

#For a two-level model with random intercepts
m3 <- lmer(y ~ service * dept + (1|s) + (1|d), InstEval)
#Ranked by the random intercept on 's'
(m3.er1 <- expectedRank(m3, groupFctr="s", term="Intercept"))
}
}
\references{
Laird NM and Louis TA. Empirical Bayes Ranking Methods. \emph{Journal of
Education Statistics}. 1989;14(1)29-46. Available at
\url{http://www.jstor.org/stable/1164724}.


Lingsma HF, Steyerberg EW, Eijkemans MJC, et al. Comparing and
ranking hospitals based on outcome: results from The Netherlands Stroke Survey.
\emph{QJM: An International Journal of Medicine}. 2010;103(2):99-108.
doi:10.1093/qjmed/hcp169
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predictInterval_helpers.R
\name{summarise_predictions}
\alias{summarise_predictions}
\title{Summarise prediction simulations into intervals}
\usage{
summarise_predictions(
  yhat_arr,
  level,
  stat.type,
  predict.type,
  N,
  merMod = NULL,
  which.eff = "full",
  pi.comps = NULL,
  is.glmm.with.response = FALSE
)
}
\arguments{
\item{yhat_arr}{matrix of simulated predictions with dimensions (N x n.sims),
or for `which = "all"`, the combined yhat from `combine_components()`.}

\item{level}{numeric confidence level (e.g., 0.8 for 80}

\item{stat.type}{character either "median" or "mean" for the central estimate.}

\item{predict.type}{character either "linear.prediction" or "probability".}

\item{N}{integer number of observations in the original newdata.}

\item{merMod}{merMod object, required when `predict.type = "probability"` to
access the link function.}

\item{which.eff}{character the effect type: "full", "fixed", "random", or "all".}

\item{pi.comps}{list of component prediction matrices (from `combine_components()`
when `which = "all"`), or NULL for other cases.}
}
\value{
data.frame with columns `fit`, `upr`, `lwr`. For `which = "all"`,
  includes additional columns `effect` and `obs`.
}
\description{
Takes the simulated prediction array and computes summary statistics
(fit, lower bound, upper bound) based on the specified statistic type
and confidence level. Handles probability transformation and the
`which = "all"` case with component-wise summaries.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{reTermNames}
\alias{reTermNames}
\title{Get names of random effect terms in a model object}
\usage{
reTermNames(model)
}
\arguments{
\item{model}{a merMod object with random effect terms}
}
\value{
a data.frame with rows for each term with columns naming the grouping
term and the effect type
}
\description{
Get names of random effect terms in a model object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{formulaBuild}
\alias{formulaBuild}
\title{Clean formula}
\usage{
formulaBuild(model)
}
\arguments{
\item{model}{a merMod object from lme4}
}
\value{
a formula object
}
\description{
a function to modify the formula for a merMod object to create
a model matrix with all predictor terms in both the group level and fixed
effect level
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merData.R
\name{wiggle}
\alias{wiggle}
\title{Assign an observation to different values}
\usage{
wiggle(data, varlist, valueslist)
}
\arguments{
\item{data}{a data frame with one or more observations to be reassigned}

\item{varlist}{a character vector specifying the name(s) of the variable to adjust}

\item{valueslist}{a list of vectors with the values to assign to var}
}
\value{
a \code{data.frame} with each row assigned to the one of the new variable combinations.
All variable combinations are returned, eg wiggling two variables with 3 and 4 variables
respectively will return a new dataset with \code{3 * 4 = 12} observations.
}
\description{
Creates a new data.frame with copies of the original observation,
each assigned to a different user-specified value of a variable. Allows the
user to look at the effect on predicted values of changing either a single variable
or multiple variables.
}
\details{
If the variable specified is a factor, then wiggle will return it
as a character.
}
\examples{
data(iris)
wiggle(iris[3,], varlist = "Sepal.Width", valueslist = list(c(1, 2, 3, 5)))
wiggle(iris[3:5,], "Sepal.Width", valueslist = list(c(1, 2, 3, 5)))
wiggle(iris[3,], c("Sepal.Width", "Petal.Length"), list(c(1,2,3,5), c(3,5,6)))
wiggle(iris[3:5,], c("Sepal.Width", "Petal.Length"), list(c(1,2,3,5), c(3,5,6)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merList.R
\name{VarCorr.merModList}
\alias{VarCorr.merModList}
\title{Extract the variances and correlations for random effects from a merMod list}
\usage{
\method{VarCorr}{merModList}(x, sigma = 1, rdig = 3L, ...)
}
\arguments{
\item{x}{for \code{VarCorr}: a fitted model object, usually an object inheriting from
    class \code{\linkS4class{merMod}}. For \code{as.data.frame}, a
    \code{VarCorr.merMod} object returned from \code{VarCorr}.}

\item{sigma}{an optional numeric value used as a multiplier for the
    standard deviations.}

\item{rdig}{the number of digits to round to, integer}

\item{...}{Ignored for the \code{as.data.frame} method; passed to
    other \code{\link{print}()} methods for the \code{print()} method.}
}
\value{
a list with two elements "stddev" and "correlation" for the standard
deviations and correlations averaged across models in the list
}
\description{
Extract the variances and correlations for random effects from a merMod list
}
\examples{
sim_list <- replicate(n = 10,
        expr = sleepstudy[sample(row.names(sleepstudy), 180),],
        simplify=FALSE)
fml <- "Reaction ~ Days + (Days | Subject)"
mod <- lmerModList(fml, data = sim_list)
VarCorr(mod)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predictInterval_helpers.R
\name{tmp.pred}
\alias{tmp.pred}
\title{Compute random effect predictions for a single grouping factor}
\usage{
tmp.pred(data, coefs, group)
}
\arguments{
\item{data}{data.frame with predictor columns and a grouping factor column.
The last column must be the grouping factor.}

\item{coefs}{3D array of simulated coefficients with dimensions
(n.sims x terms x levels).}

\item{group}{character name of the grouping factor column in data.}
}
\value{
matrix of predicted values with dimensions (nrow(data) x n.sims).
}
\description{
Internal helper that multiplies predictor data by simulated random effect
coefficients for each grouping level. Handles new levels not present in the
original model data by returning zeros for their random effect contributions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merPlots.R
\name{plotFEsim}
\alias{plotFEsim}
\title{Plot the results of a simulation of the fixed effects}
\usage{
plotFEsim(
  data,
  level = 0.95,
  stat = "median",
  sd = TRUE,
  intercept = FALSE,
  sigmaScale = NULL,
  oddsRatio = FALSE
)
}
\arguments{
\item{data}{a data.frame generated by \code{\link{FEsim}} with simulations of
the fixed effects of a \code{\link{merMod}}}

\item{level}{the width of the confidence interval}

\item{stat}{a character value indicating the variable name in data of the
midpoint of the estimated interval, e.g. "mean" or "median"}

\item{sd}{logical, indicating whether or not to plot error bars around
the estimates (default is TRUE). Calculates the width of the error bars
based on \code{level} and the variable named "sd" in \code{data}}

\item{intercept}{logical, should the intercept be included, default is FALSE}

\item{sigmaScale}{a numeric value to divide the estimate and the standard
deviation by in the case of doing an effect size calculation}

\item{oddsRatio}{logical, should the parameters be converted to odds ratios
before plotting}
}
\value{
a ggplot2 plot of the coefficient effects
}
\description{
Plot the simulated fixed effects on a ggplot2 chart
}
\examples{
 fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
 (p1 <- plotFEsim(FEsim(fm1)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merData.R
\name{averageObs}
\alias{averageObs}
\title{Find the average observation for a merMod object}
\usage{
averageObs(merMod, varList = NULL, origData = NULL, ...)
}
\arguments{
\item{merMod}{a merMod object}

\item{varList}{optional, a named list of conditions to subset the data on}

\item{origData}{(default=NULL) a data frame containing the original,
untransformed data used to call the model. This MUST be specified if
the original variables used in formula function calls are NOT present
as 'main effects'.}

\item{...}{not used currently}
}
\value{
a data frame with a single row for the average observation, but with full
factor levels. See details for more.
}
\description{
Extract a data frame of a single row that represents the
average observation in a merMod object. This function also allows the
user to pass a series of conditioning argument to calculate the average
observation conditional on other characteristics.
}
\details{
Each character and factor variable in the data.frame is assigned to the
modal category and each numeric variable is collapsed to the mean. Currently if
mode is a tie, returns a "." Uses the collapseFrame function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merData.R
\name{hasWeights}
\alias{hasWeights}
\title{Identify if a merMod has weights}
\usage{
hasWeights(merMod)
}
\arguments{
\item{merMod}{the merMod object to test for weights}
}
\value{
TRUE or FALSE for whether the model has weights
}
\description{
Identify if a merMod has weights
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merFastDisplay.R
\name{fastdisp}
\alias{fastdisp}
\alias{fastdisp.merMod}
\alias{fastdisp.merModList}
\title{fastdisp: faster display of model summaries}
\usage{
fastdisp(x, ...)

\method{fastdisp}{merMod}(x, ...)

\method{fastdisp}{merModList}(x, ...)
}
\arguments{
\item{x}{a model object}

\item{...}{additional arguments to pass to \code{arm::\link[arm]{display}}
including number of digits}
}
\value{
A printed summary of a x object
}
\description{
Display model fit summary of x or x like objects, fast
}
\details{
Faster than the implementation in the arm package because it avoids refitting

The time saving is only noticeable for large, time-consuming (g)lmer
fits.
}
\examples{
\donttest{
#Compare the time for displaying this modest model
require(arm)
m1 <- lmer(y ~ lectage + studage + (1|d) + (1|s), data=InstEval)
system.time(display(m1))
system.time(fastdisp(m1))
}
}
\seealso{
\code{\link[arm]{display}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merPlots.R
\name{plotREsim}
\alias{plotREsim}
\title{Plot the results of a simulation of the random effects}
\usage{
plotREsim(
  data,
  level = 0.95,
  stat = "median",
  sd = TRUE,
  sigmaScale = NULL,
  oddsRatio = FALSE,
  labs = FALSE,
  facet = TRUE
)
}
\arguments{
\item{data}{a data.frame generated by \code{\link{REsim}} with simulations of
the random effects of a \code{\link{merMod}}}

\item{level}{the width of the confidence interval}

\item{stat}{a character value indicating the variable name in data of the
midpoint of the estimated interval, e.g. "mean" or "median"}

\item{sd}{a logical indicating whether or not to plot error bars around
the estimates (default is TRUE). Calculates the width of the error bars
based on \code{level} and the variable named "sd" in \code{data}}

\item{sigmaScale}{a numeric value to divide the estimate and the standard
deviation by in the case of doing an effect size calculation}

\item{oddsRatio}{logical, should the parameters be converted to odds ratios
before plotting}

\item{labs}{logical, include the labels of the groups on the x-axis}

\item{facet}{Accepts either logical (\code{TRUE}) or \code{list} to specify which
random effects to plot. If \code{TRUE}, facets by both \code{groupFctr} and \code{term}.
If \code{list} selects the panel specified by the named elements of the list}
}
\value{
a ggplot2 plot of the coefficient effects
}
\description{
Plot the simulated random effects on a ggplot2 chart. Points that
are distinguishable from zero (i.e. the confidence band based on \code{level}
does not cross the red line) are highlighted. Currently, the plots are ordered
according to the grouping factor.
}
\examples{
\donttest{
 fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
 (p1 <- plotREsim(REsim(fm1)))
 #Plot just the random effects for the Days slope
 (p2 <- plotREsim(REsim(fm1), facet= list(groupFctr= "Subject", term= "Days")))
 }
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{RHSForm}
\alias{RHSForm}
\title{Parse merMod formulas}
\usage{
RHSForm(form, as.form = FALSE)
}
\description{
Parse merMod formulas
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{reOnly}
\alias{reOnly}
\title{Random Effects formula only}
\usage{
reOnly(f, response = FALSE)
}
\arguments{
\item{f}{a model formula}

\item{response}{logical, should the result include the response}
}
\value{
a formula
}
\description{
Random Effects formula only
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{fetch.merMod.msgs}
\alias{fetch.merMod.msgs}
\title{Extract all warning msgs from a merMod object}
\usage{
fetch.merMod.msgs(x)
}
\arguments{
\item{x}{a merMod object}
}
\description{
Extract all warning msgs from a merMod object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{buildModelMatrix}
\alias{buildModelMatrix}
\title{Build model matrix}
\source{
Taken from predict.merMod in lme4
}
\usage{
buildModelMatrix(model, newdata, which = "full")
}
\arguments{
\item{model}{a merMod object from lme4}

\item{newdata}{a data frame to construct the matrix from}

\item{which}{a character which matrix to return,default is full matrix with fixed and
random terms, other options are "fixed" and "random"}
}
\description{
a function to create a model matrix with all predictor terms in
both the group level and fixed effect level
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merList.R
\name{modelFixedEff}
\alias{modelFixedEff}
\title{Extract averaged fixed effect parameters across a list of merMod objects}
\usage{
modelFixedEff(modList, ...)
}
\arguments{
\item{modList}{an object of class merModList}

\item{...}{additional arguments to pass to \code{\link{tidy}}}
}
\value{
a data.frame of the averaged fixed effect parameters
}
\description{
Extract averaged fixed effect parameters across a list of merMod objects
}
\details{
The Rubin correction for combining estimates and standard errors from
Rubin (1987) is applied to adjust for the within and between imputation variances.
}
\examples{
\donttest{
sim_list <- replicate(n = 10,
        expr = sleepstudy[sample(row.names(sleepstudy), 180),],
        simplify=FALSE)
fml <- "Reaction ~ Days + (Days | Subject)"
mod <- lmerModList(fml, data = sim_list)
modelFixedEff(mod)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merData.R
\name{subsetList}
\alias{subsetList}
\title{Subset a data.frame using a list of conditions}
\usage{
subsetList(data, list)
}
\arguments{
\item{data}{a data.frame}

\item{list}{a named list of splitting conditions}
}
\value{
a data frame with values that match the conditions in the list
}
\description{
Split a data.frame by elements in a list
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merData.R
\name{shuffle}
\alias{shuffle}
\title{Randomly reorder a dataframe}
\usage{
shuffle(data)
}
\arguments{
\item{data}{a data frame}
}
\value{
a data frame of the same dimensions with the rows reordered
randomly
}
\description{
Randomly reorder a dataframe by row
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predictInterval_helpers.R
\name{combine_postvar_blocks}
\alias{combine_postvar_blocks}
\title{Combine a per-term list of postVar arrays into a block-diagonal array}
\usage{
combine_postvar_blocks(pv_list, reMeans)
}
\arguments{
\item{pv_list}{List of per-term postVar arrays.}

\item{reMeans}{The `as.matrix(ranef(m)[[j]])` matrix for the same
grouping factor; used to size and name the combined array.}
}
\value{
3-D array of shape `ncol(reMeans) × ncol(reMeans) × nrow(reMeans)`.
}
\description{
When a grouping factor has multiple random-effect term blocks — from the
double-bar syntax (`(x + y || g)`), explicit splits (`(1|g) + (0+x|g)`),
or mixed correlated + uncorrelated specs — `lme4::ranef(..., condVar =
TRUE)` returns `postVar` as a list of arrays, one per term block. Each
element has shape `d_m × d_m × n_levels`. This helper concatenates the
blocks into a single `d × d × n_levels` array with block-diagonal
structure per level (zeros off-diagonal because uncorrelated). The block
ordering and column ordering match `colnames(reMeans)` / lme4's `cnms`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merData.R
\name{findFormFuns}
\alias{findFormFuns}
\title{\code{findFormFuns} used by \link[merTools]{averageObs} to calculate proper
averages}
\usage{
findFormFuns(merMod, origData = NULL)
}
\arguments{
\item{merMod}{the merMod object from which to draw the average observation}

\item{origData}{(default=NULL) a data frame containing the original,
untransformed data used to call the model. This MUST be specified if
the original variables used in formula function calls are NOT present
as 'main effects'.}
}
\value{
a data frame with a single row for the average observation, but with full
factor levels. See details for more.
}
\description{
The purpose is to properly derive data for the average observation in the
data by being 'aware' of formulas that contain interactions and/or function
calls. For example, in the old behavior, if the formula contained a square
term specified as \code{I(x^2)}, we were returning the mean of \code{x(^2)} not the
square of mean(x).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subBoot.R
\name{thetaExtract}
\alias{thetaExtract}
\title{Extract theta parameters from a merMod model}
\usage{
thetaExtract(merMod)
}
\arguments{
\item{merMod}{a valid merMod object}
}
\value{
a vector of the covariance, theta, parameters from a \code{\link{merMod}}
}
\description{
A convenience function that returns the theta parameters for a
\code{\link{merMod}} object.
}
\examples{
(fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy))
thetaExtract(fm1) #(a numeric vector of the covariance parameters)
}
\seealso{
merMod
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merData.R
\name{REquantile}
\alias{REquantile}
\title{Identify group level associated with RE quantile}
\usage{
REquantile(merMod, quantile, groupFctr, term = "(Intercept)")
}
\arguments{
\item{merMod}{a merMod object with one or more random effect levels}

\item{quantile}{a numeric vector with values between 0 and 100 for quantiles}

\item{groupFctr}{a character of the name of the random effect grouping factor to extract
quantiles from}

\item{term}{a character of the random effect to extract for the grouping factor
specified. Default is the intercept.}
}
\value{
a vector of the level of the random effect grouping term that corresponds
to each quantile
}
\description{
For a user specified quantile (or quantiles) of the random effect
terms in a merMod object. This allows the user to easily identify the observation
associated with the nth percentile effect.
}
\examples{
\donttest{
fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
REquantile(fm1, quantile = 0.25, groupFctr = "Subject")
REquantile(fm1, quantile = 0.25, groupFctr = "Subject", term = "Days")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{ICC}
\alias{ICC}
\title{Calculate the intraclass correlation using mixed effect models}
\usage{
ICC(outcome, group, data, subset = NULL)
}
\arguments{
\item{outcome}{a character representing the variable of the outcome}

\item{group}{a character representing the name of the grouping term}

\item{data}{a data.frame}

\item{subset}{an optional subset}
}
\value{
a numeric for the intraclass correlation
}
\description{
Calculate the intraclass correlation using mixed effect models
}
\examples{
data(sleepstudy)
ICC(outcome = "Reaction", group = "Subject", data = sleepstudy)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parallel.R
\name{setup_parallel}
\alias{setup_parallel}
\title{Set up parallel environment}
\usage{
setup_parallel()
}
\value{
Nothing
}
\description{
Set up parallel environment
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predictInterval_helpers.R
\name{simulate_random_effects}
\alias{simulate_random_effects}
\title{Simulate random‑effect contributions for all grouping factors}
\usage{
simulate_random_effects(
  merMod,
  newdata,
  n.sims,
  .parallel = FALSE,
  seed = NULL,
  new.levels = c("zero", "draw")
)
}
\arguments{
\item{merMod}{A merMod object.}

\item{newdata}{Data frame containing the prediction covariates.}

\item{n.sims}{Number of simulations.}

\item{.parallel}{Logical flag to enable parallel execution via foreach.}

\item{seed}{Optional random seed for reproducibility.}

\item{new.levels}{Character; how to treat grouping levels present in
`newdata` but absent from the fitted model. `"zero"` drops the random
effect for such levels; `"draw"` samples each unobserved level's effect
from the estimated random-effect covariance (`VarCorr`).}
}
\value{
List of matrices (rows = observations, cols = simulations).
}
\description{
For each random effect term the function draws `n.sims` samples from the
conditional multivariate normal distribution (BLUPs + post‑var). The result is a
named list where each element is an array of dimensions `nrow(newdata)` ×
`n.sims`.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merData.R
\name{draw}
\alias{draw}
\alias{draw.merMod}
\title{Draw a single observation out of an object matching some criteria}
\usage{
draw(object, type = c("random", "average"), varList = NULL, seed = NULL, ...)

\method{draw}{merMod}(object, type = c("random", "average"), varList = NULL, seed = NULL, ...)
}
\arguments{
\item{object}{the object to draw from}

\item{type}{what kind of draw to make. Options include random or average}

\item{varList}{a list specifying filters to subset the data by when making the
draw}

\item{seed}{numeric, optional argument to set seed for simulations, ignored if type="average"}

\item{...}{additional arguments required by certain methods}
}
\value{
a data.frame with a single row representing the desired observation
}
\description{
Draw is used to select a single observation out of an R object.
Additional parameters allow the user to control how that observation is
chosen in order to manipulate that observation later. This is a generic
function with methods for a number of objects.
}
\details{
In cases of tie, ".", may be substituted for factors.
}
\examples{
fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
# Random case
draw(fm1, type = "random")
# Average
draw(fm1, type = "average")
# Subset
draw(fm1, type = "average", varList = list("Subject" = "308"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{sum.mm}
\alias{sum.mm}
\title{Title}
\usage{
\method{sum}{mm}(
  object,
  correlation = (p <= getOption("lme4.summary.cor.max")),
  use.hessian = NULL,
  ...
)
}
\arguments{
\item{object}{a merMod object}

\item{correlation}{optional p value}

\item{use.hessian}{logical}

\item{...}{additional arguments to pass through}
}
\value{
a summary of the object
}
\description{
Title
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merData.R
\name{stripAttributes}
\alias{stripAttributes}
\title{Remove attributes from a data.frame}
\usage{
stripAttributes(data)
}
\arguments{
\item{data}{a data.frame}
}
\value{
a data frame with variable names cleaned to remove all attributes except for
names, row.names, and class
}
\description{
Strips attributes off of a data frame that come with a merMod model.frame
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merExtract.R
\name{REsim}
\alias{REsim}
\title{Simulate random effects from merMod
\code{REsim} simulates random effects from merMod object posterior distributions}
\usage{
REsim(merMod, n.sims = 200, oddsRatio = FALSE, seed = NULL)
}
\arguments{
\item{merMod}{a merMod object from the lme4 package}

\item{n.sims}{number of simulations to use}

\item{oddsRatio}{logical, should parameters be converted to odds ratios?}

\item{seed}{numeric, optional argument to set seed for simulations}
}
\value{
a data frame with the following columns
\describe{
  \item{\code{groupFctr}}{Name of the grouping factor}
  \item{\code{groupID}}{Level of the grouping factor}
  \item{\code{term}}{Name of random term (intercept/coefficient)}
  \item{\code{mean}}{Mean of the simulations}
  \item{\code{median}}{Median of the simulations}
  \item{\code{sd}}{Standard deviation of the simulations, \code{NA} if \code{oddsRatio=TRUE}}
}
}
\description{
Simulate random effects from merMod
\code{REsim} simulates random effects from merMod object posterior distributions
}
\details{
Use the Gelman sim technique to build empirical Bayes estimates.
 Uses the sim function in the arm package
}
\examples{
require(lme4)
m2 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
re2 <- REsim(m2, 25)
head(re2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merData.R
\name{superFactor}
\alias{superFactor}
\title{Create a factor with unobserved levels}
\usage{
superFactor(x, fullLev)
}
\arguments{
\item{x}{a vector to be converted to a factor}

\item{fullLev}{a vector of factor levels to be assigned to x}
}
\value{
a factor variable with all observed levels of x and all levels
of x in fullLev
}
\description{
Create a factor variable and include unobserved levels
for compatibility with model prediction functions
}
\examples{
\donttest{
regularFactor <- c("A", "B", "C")
regularFactor <- factor(regularFactor)
levels(regularFactor)
# Now make it super
newLevs <- c("D", "E", "F")
regularFactor <- superFactor(regularFactor, fullLev = newLevs)
levels(regularFactor) # now super
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merPlots.R
\name{plotREimpact}
\alias{plotREimpact}
\title{Plot the impact of grouping-factor levels on predictions}
\usage{
plotREimpact(data, level = 0.95, facet = TRUE, point_size = 2.5)
}
\arguments{
\item{data}{either a single data.frame produced by \code{\link{REimpact}},
or a named list of such data.frames. When a named list is supplied the names
are used to colour and label the series.}

\item{level}{the width of the confidence interval (default 0.95).}

\item{facet}{logical, facet the plot by \code{case} (default \code{TRUE}).}

\item{point_size}{numeric size of the plotted central points.}
}
\value{
a ggplot2 object.
}
\description{
Plot the output of one or more \code{\link{REimpact}} calls on a
single chart. Each point is the weighted-average fitted value for a bin of
the expected-rank distribution of a grouping factor, with a confidence
interval derived from the weighted standard error. Supplying a named list of
\code{REimpact} results overlays them on the same axes -- for example to
compare the influence of two different grouping factors, or the same factor
across two models -- which previously required hand-assembling the data
frames (#84).
}
\examples{
\donttest{
 m1 <- lmer(Reaction ~ Days + (1 | Subject), sleepstudy)
 imp <- REimpact(m1, newdata = sleepstudy[1, ], groupFctr = "Subject",
                 breaks = 4)
 plotREimpact(imp)

 # Compare two grouping factors on the same chart
 g1 <- lmer(y ~ lectage + studage + (1 | d) + (1 | s), data = InstEval)
 d_eff <- REimpact(g1, newdata = InstEval[1, ], groupFctr = "d", breaks = 4)
 s_eff <- REimpact(g1, newdata = InstEval[1, ], groupFctr = "s", breaks = 4)
 plotREimpact(list("Instructor (d)" = d_eff, "Student (s)" = s_eff))
 }
}
\seealso{
\code{\link{REimpact}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merList.R
\name{ranef.merModList}
\alias{ranef.merModList}
\title{Extract random-effects estimates for a merModList}
\usage{
\method{ranef}{merModList}(object, ...)
}
\arguments{
\item{object}{an object of a class of fitted models with
    random effects, typically a
    \code{\linkS4class{merMod}} object.}

\item{...}{some methods for these generic functions
    require additional arguments.}
}
\value{
a named, numeric vector of random-effects estimates.
}
\description{
Extract random-effects estimates for a merModList
}
\details{
Extract the estimates of the random-effects parameters from a list of
fitted \code{merMod} models. Takes the mean of the individual \code{ranef}
objects for each of the component models in the \code{merModList}.
}
\examples{
\donttest{
sim_list <- replicate(n = 10,
        expr = sleepstudy[sample(row.names(sleepstudy), 180),],
        simplify=FALSE)
fml <- "Reaction ~ Days + (Days | Subject)"
mod <- lmerModList(fml, data = sim_list)
ranef(mod)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subBoot.R
\name{subBoot}
\alias{subBoot}
\title{Bootstrap a subset of an lme4 model}
\usage{
subBoot(merMod, n = NULL, FUN, R = 100, seed = NULL, warn = FALSE)
}
\arguments{
\item{merMod}{a valid merMod object}

\item{n}{the number of rows to sample from the original data
in the merMod object, by default will resample the entire model frame}

\item{FUN}{the function to apply to each bootstrapped model}

\item{R}{the number of bootstrap replicates, default is 100}

\item{seed}{numeric, optional argument to set seed for simulations}

\item{warn}{logical, if TRUE, warnings from lmer will be issued, otherwise they will be suppressed
default is FALSE}
}
\value{
a data.frame of parameters extracted from each of the R replications.
The original values are appended to the top of the matrix.
}
\description{
Bootstrap a subset of an lme4 model
}
\details{
This function allows users to estimate parameters of a
large merMod object using bootstraps on a subset of the data.
}
\examples{
\donttest{
(fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy))
resultMatrix <- subBoot(fm1, n = 160, FUN = thetaExtract, R = 20)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merList.R
\name{lmerModList}
\alias{lmerModList}
\alias{blmerModList}
\alias{glmerModList}
\alias{bglmerModList}
\title{Apply a multilevel model to a list of data frames}
\usage{
lmerModList(formula, data, parallel = FALSE, ...)

blmerModList(formula, data, parallel = FALSE, ...)

glmerModList(formula, data, parallel = FALSE, ...)

bglmerModList(formula, data, parallel = FALSE, ...)
}
\arguments{
\item{formula}{a formula to pass through compatible with merMod}

\item{data}{a list object with each element being a data.frame}

\item{parallel}{logical, should the models be run in parallel? Default FALSE. If so,
the `future_lapply` function from the `future.apply` package is used. See
details.}

\item{...}{additional arguments to pass to the estimating function}
}
\value{
a list of fitted merMod objects of class merModList

a merModList

a merModList

a merModList
}
\description{
Apply a multilevel model to a list of data frames

Apply a Bayesian multilevel model to a list of data frames

Apply a generalized linear multilevel model to a list of data frames

Apply a Bayesian generalized linear multilevel model to a list of data frames
}
\details{
Parallel computing is provided by the `futures` package, and its
extension the `future.apply` package to provide the `future_lapply` function
for easy parallel computations on lists. To use this package, simply register
a parallel backend using the `plan()` function from `futures` - an example
is to use `plan(multisession)`
}
\examples{
\donttest{
sim_list <- replicate(n = 10,
        expr = sleepstudy[sample(row.names(sleepstudy), 180),],
        simplify=FALSE)
fml <- "Reaction ~ Days + (Days | Subject)"
mod <- lmerModList(fml, data = sim_list)
summary(mod)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merList.R
\name{REcorrExtract}
\alias{REcorrExtract}
\title{Extract the correlations between the slopes and the intercepts from a model}
\usage{
REcorrExtract(model)
}
\arguments{
\item{model}{an object that inherits from class merMod}
}
\value{
a numeric vector of the correlations among the effects
}
\description{
Extract the correlations between the slopes and the intercepts from a model
}
\examples{
fm1 <- lmer(Reaction ~ Days + (Days | Subject), sleepstudy)
REcorrExtract(fm1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merList.R
\name{summary.merModList}
\alias{summary.merModList}
\title{Print the results of a merMod list}
\usage{
\method{summary}{merModList}(object, ...)
}
\arguments{
\item{object}{a modList of class merModList}

\item{...}{additional arguments}
}
\value{
summary content printed to console
}
\description{
Print the results of a merMod list
}
\examples{
sim_list <- replicate(n = 10,
        expr = sleepstudy[sample(row.names(sleepstudy), 180),],
        simplify=FALSE)
fml <- "Reaction ~ Days + (Days | Subject)"
mod <- lmerModList(fml, data = sim_list)
print(mod)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers.R
\name{reTermCount}
\alias{reTermCount}
\title{Count the number of random effect terms}
\source{
From lme4 package
}
\usage{
reTermCount(model)
}
\description{
Count the number of random effect terms
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predictInterval_helpers.R
\name{inverse_link}
\alias{inverse_link}
\title{Inverse link function for GLMM families}
\usage{
inverse_link(eta, family, link)
}
\arguments{
\item{eta}{Linear predictor (fixed + random effects).}

\item{family}{Character string specifying the distribution family.}

\item{link}{Character string specifying the link function.}
}
\value{
Vector of distribution parameters (prob, lambda, or rate).
}
\description{
Internal helper that applies the inverse link function to convert linear
predictors to distribution parameters.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predictInterval_helpers.R
\name{simulate_fixed_effects}
\alias{simulate_fixed_effects}
\title{Simulate fixed-effect predictions}
\usage{
simulate_fixed_effects(
  merMod,
  newdata,
  n.sims,
  ignore.fixed.terms = NULL,
  fix.intercept.variance = FALSE,
  .parallel = FALSE,
  seed = NULL
)
}
\arguments{
\item{merMod}{A merMod object.}

\item{newdata}{Data frame containing the prediction covariates.}

\item{n.sims}{Number of simulations.}

\item{ignore.fixed.terms}{Vector of fixed-effect names or indices to treat as known.}

\item{fix.intercept.variance}{Logical flag for intercept variance correction.}

\item{.parallel}{Logical flag to enable parallel execution via foreach.}
}
\value{
Matrix of dimensions `nrow(newdata)` x `n.sims` containing simulated
  fixed-effect predictions.
}
\description{
Draws `n.sims` samples from the multivariate normal distribution defined by
the fixed-effect estimates and their variance-covariance matrix, then
multiplies by the model matrix to produce predictions. It also respects
`ignore.fixed.terms` and optional intercept variance adjustment.
}
\keyword{internal}
